﻿local Root = BossEncounter3;

local Audio = Root:GetOrNewModule("Audio");

local Resources = Root:GetOrNewModule("Resources");

-- Will not interrupt musics initiated by other means except
-- if the foreign PlayMusic was issued while we were expecting
-- a BossEncounter3 music playing.

Audio.playing = false;
Audio.last = nil;
Audio.duration = 0;
Audio.start = 0;
Audio.count = { };

local choices = { };

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the name of the music to examine.                       *
-- ********************************************************************
-- * Get the number of times the specified music has been played.     *
-- ********************************************************************
function Audio:GetMusicCount(name)
    return self.count[name] or 0;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> type: of music to select.                                     *
-- ********************************************************************
-- * Randomly select a music of a given type.                         *
-- ********************************************************************
function Audio:SelectMusicType(type)
    Resources:ListMusicType(type);

    local leastCount = math.huge;
    local num = Resources:GetListSize();
    local i, name;

    for i=1, num do
        name = Resources:GetListEntry(i); -- Find out the minimum.
        leastCount = min(self:GetMusicCount(name), leastCount);
    end

    wipe(choices);
    for i=1, num do
        name = Resources:GetListEntry(i); -- Build the choice list.
        if self:GetMusicCount(name) == leastCount then
            tinsert(choices, name);
        end
    end

    if #choices > 0 then
        return choices[ math.random(1, #choices) ];
  else
        return nil;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the name of the music to start.                         *
-- * >> duration: the amount of time the music should be playing.     *
-- * Once the amount of time is elapsed, it will switch to a          *
-- * persistent silence till the Stop API is used.                    *
-- ********************************************************************
-- * Plays a music.                                                   *
-- * If it is not found or for some reason can't be played, false is  *
-- * returned; true elsewise is.                                      *
-- ********************************************************************
function Audio:PlayMusic(name, duration)
    local filename = select(2, Resources:GetMusicInfo(name));
    if ( filename ) then
        local success = PlayMusic(filename);
        if ( success == 1 ) then
            self.count[name] = self:GetMusicCount(name) + 1;
            self.playing = true;
            self.duration = duration or 0;
            self.start = GetTime();
            self.last = name;
            return true;
        end
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the name of the music to start.                         *
-- ********************************************************************
-- * Plays a music once.                                              *
-- ********************************************************************
function Audio:PlayMusicOnce(name)
    local duration = select(4, Resources:GetMusicInfo(name));
    return self:PlayMusic(name, duration - 1);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> type: of music to start. If there are multiple musics         *
-- * referenced for asked type, one is chosen randomly amongst them.  *
-- ********************************************************************
-- * Randomly plays a music of a given type.                          *
-- * The least played music is very likely to be selected.            *
-- ********************************************************************
function Audio:PlayMusicType(type)
    local music = self:SelectMusicType(type);
    if music then
        return self:PlayMusic(music);
  else
        return false;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> importance: the desired importance of the music.              *
-- ********************************************************************
-- * Play an appropriate battle music.                                *
-- ********************************************************************
function Audio:PlayBattleMusic(importance)
    Resources:ListMusicType("BATTLE");

    local num = Resources:GetListSize();
    local total = 0;
    local list = { };
    local i, name, myImportance, delta, score;
    for i=1, num do
        name = Resources:GetListEntry(i);
        myImportance = select(5, Resources:GetMusicInfo(name));

        delta = abs(importance - myImportance);
        score = floor(max(0, 20 - delta) ^ 1.5 + 0.5);

        if score > 0 then
            tinsert(list, {name = name, min = total, max = total+score-1});
            total = total + score;
        end
    end

    Root:Debug("Music play probability: (importance="..importance..")");
    for i=1, #list do
        local min, max = list[i].min, list[i].max;
        local chance = (max - min + 1) * 100 / total;
        Root:DebugFmt("%d. %s => %.2f%% [%d - %d]", i, list[i].name, chance, min, max);
    end

    local rand = math.random(0, max(0, total-1));
    Root:DebugFmt("Rolling [%d - %d], got %d.", 0, total-1, rand);
    for i=1, #list do
        if rand >= list[i].min and rand <= list[i].max then
            Root:DebugFmt("%s chosen!", list[i].name);
            return Audio:PlayMusic(list[i].name);
        end
    end
    return false;
end

-- ********************************************************************
-- * Stops the currently playing music.                               *
-- ********************************************************************
function Audio:StopMusic()
    if ( self.playing ) then
        self.playing = false;
        StopMusic();
    end
end

-- ********************************************************************
-- * Get the name of the last music successfully started.             *
-- ********************************************************************
function Audio:GetLastMusic()
    return self.last;
end

-- ********************************************************************
-- * Get the duration of specified music, if available.               *
-- ********************************************************************
function Audio:GetMusicDuration(name)
    return select(4, Resources:GetMusicInfo(name));
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the name of the sound to play.                          *
-- ********************************************************************
-- * Plays a music.                                                   *
-- * If it is not found or for some reason can't be played, false is  *
-- * returned; true elsewise is.                                      *
-- ********************************************************************
function Audio:PlaySound(name)
    local filename = select(2, Resources:GetSoundInfo(name));
    if ( filename ) then
        local success = PlaySoundFile(filename);
        if ( success == 1 ) then
            return true;
        end
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> type: of sound to start. If there are multiple sounds         *
-- * referenced for asked type, one is chosen randomly amongst them.  *
-- ********************************************************************
-- * Randomly plays a sound of a given type.                          *
-- ********************************************************************
function Audio:PlaySoundType(type)
    Resources:ListSoundType(type);
    local name = Resources:GetListRandomEntry();
    if ( name ) then
        return self:PlaySound(name);
  else
        return false;
    end
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function Audio:OnUpdate(elapsed)
    if self.playing and self.duration > 0 then
        if (GetTime() - self.start) > self.duration then
            -- Stop the music and prevent normal music from resuming by issuing the Silence music.
            if not self:PlayMusic("SILENCE") then
                -- In case Silence music is not available, we stop the music.
                self:StopMusic();
            end
            Root:InvokeHandler("OnMusicEnd", self.last);
        end
    end
end