﻿local Root = BossEncounter3;

local QuickModule = Root:GetOrNewModule("QuickModule");

local NPCInfo = Root:GetOrNewModule("NPCInfo");

local db = { };

local perName = { };

local emptyTable = { };

local function buildName(t) -- Build an unique name for each quick module (that will never conflict with normal modules either).
    if t.boss and t.boss > 0 then
        return tostring(t.boss);

elseif t.adds and t.adds[1] then
        return tostring(t.adds[1]);

elseif t.mooks and t.mooks[1] then
        return tostring(t.mooks[1]);
  else
        error("No way to name one of the quick modules (One of them has no boss, no adds, no mooks), please check.", 0);
    end
end

function QuickModule:OnBootComplete()
    local info, name, volatile, bypass, id;
    for _, info in ipairs(db) do
        name = buildName(info);
        volatile = info.volatile;
        bypass = info.bypass;

        perName[name] = info;
        info._name = name;
        info.volatile = nil; -- No longer need these fields.
        info.bypass = nil;

        id = info.boss or 0;
        if id > 0 then
            NPCInfo:Register(id, name, volatile, bypass);
        end
        for _, id in ipairs(info.adds or emptyTable) do
            NPCInfo:Register(id, name, volatile, bypass);
        end
        for _, id in ipairs(info.mooks or emptyTable) do
            NPCInfo:Register(id, name, volatile, bypass);
        end
    end
end

function QuickModule:GetNumDescriptionTables()
    return #db;
end

function QuickModule:GetDescriptionTable(query)
    if type(query) == "string" then
        return perName[query];
  else
        return db[query];
    end
end

function QuickModule:AddDescriptionTable(t)
    tinsert(db, t);
end

-- --------------------------------------------------------------------
-- **                  >> Quick modules start here <<                **
-- --------------------------------------------------------------------

-- One can design here basic bosses that do not need to have specific modules
-- for them (i.e: bosses the basic system can describe).

-- *** EXAMPLE *************************************************************************
-- Players interested by the creation of quick modules should read this one carefully.
-- It covers all possible features of quick modules and can be safely copy-pasted to be
-- used as a template. Optional lines you do not want to set explicitly can be deleted.
-- *************************************************************************************
QuickModule:AddDescriptionTable( {
    -- General informations
    boss = 12345,                         -- The NPC ID of the main boss. Use any respectable WoW database to get it. You can provide 0 when you consider there is no main boss in the encounter.
    adds = {12, 34, 56},                  -- Adds are IMPORTANT mobs accompanying the boss. They should never respawn and must be killed in order to win the encounter. Provide their NPC ID.
    mooks = { },                          -- Mooks are like adds, except they can respawn and do not need to die to win the encounter. Do NOT use mooks when there are a LOT of those enemies.
    range = 15,                           -- Specifies what the minimal distance between players should be. Set 0 or delete this line if there is no min distance.
    volatile = false,                     -- Specifies true if the module should be triggered as soon as the boss or one of its adds is selected by anyone.
    bypass = false,                       -- Specifies true if the system should skip hostility check on the boss (that means the module is enabled even if the mob is not hostile).
    game = "CATA",                        -- The game the encounter belongs to. Can be WOW, TBC, LK or CATA. Defaults to CATA. You can generally delete this line.
    environment = "PARTY",                -- The kind of environment the encounter belongs to. Can be SOLO, PARTY or RAID. Defaults to PARTY.

    -- /!\ Important: it's up to you to decide between "adds" and "mooks". You can't combine them. Both table cannot be filled at the same time.
    -- Also, do NOT provide "adds" or "mooks" that come in big numbers (e.g: whelps of Onyxia) because one row is created in the enemy list for EACH one of them.

    -- Behavior settings
    timeout = 60,                         -- The time amount BE3 will wait when it no longer finds the boss or adds before closing the module by itself. Remove this line to use the default timeout.
    berserk = 300,                        -- The berserk timer of the boss. Set 0 or delete this line for no berserk.
    safeEngage = false,                   -- If true, the boss (or its adds) won't be considered engaged unless it targets a group member.
    ignoreStandardClear = false,          -- If true, the winning condition of defeating the boss (and its adds if there are some) will no longer apply.
    ignoreLeaveCombat = false,            -- If true, the module won't fail if the boss and its adds go out of combat.
    ignoreWipe = false,                   -- If true, the module won't fail if all members of the group are killed.

    -- Benchmark settings
    greatTime = 30,                       -- Specify the time amount before which we'll consider the group has performed well.
    poorTime = 60,                        -- Specify the time amount after which we'll consider the group has performed poorly.
    timeCoeff = 1.0,                      -- This value will multiply the greatTime and poorTime values. Useful for bosses with adds, because DPS is spent on the adds. Defaults to 1.
    ignoreResults = false,                -- If true, there won't be any end sequence upon defeating the boss.
    ignoreDamageMeter = false,            -- If true, there won't be any damage graph after clearing the boss.
    ignoreScore = false,                  -- If true, there won't be any score *and* any result message after clearing the boss (no attempt/etc. card and no "X has been cleared in..." messages).
    importance = 0,                       -- The importance of the boss, on a scale from 0 to 100. This is used to determinate the probability for each music to be played. Defaults to 0.

    -- Note: setting "ignoreResults" to true is *strictly equivalent* to setting both "ignoreDamageMeter" and "ignoreScore" to true.

    -- ** Beginning proper boss description **

    yells = {                             -- Yells is a section allowing you to describe what important stuff the boss might yell during the combat and what those yells trigger.
        { trigger = "SomeTrigger",        -- This refers to the field of the "locale" table (see below) that will be read to determinate what the boss must yell for this entry to trigger.
          callback = "StartPhaseTwo" },   -- The name of the method that will be fired when this entry is triggered. See the "extra" section below.
    },

    thresholds = {                        -- Thresholds is a section allowing you to describe things that happen when the boss gets below a certain health percentage.
        { value = 0.50,                   -- Value is the health ratio value that will cause the threshold to trigger. Example: 0.30 will trigger the threshold once the boss gets below 30% HP.
          description = "MyThreshold",    -- This refers to the field of the "locale" table (see below) that will be used as a description of the threshold.
          callback = "StartPhaseTwo" },   -- The name of the method that will be fired when this entry is triggered. See the "extra" section below.
    },

    timers = {                            -- Timers is a section where you declare all the timers your module will use.
        { id = "MyTimer",                 -- The identifier of the timer. This is what the system will use to refer to it.
          category = "all/high",          -- The left part specifies "who" should pay attention to the timer. The right part specifies the importance of the timer.
                                            --   Possible "who": dps_melee, dps_ranged, healer, tank, dps, caster, melee, shield, offensive | Possible "importance": low, normal, high, critical
          spell = 12345,                  -- The spellID that will be examined to get the label of the timer. If it isn't provided, then BE3 will look into the locale table to get the label.
          default = 15,                   -- The time amount of the timer when it is run anytime except at the beginning of the encounter.
          first = 10,                     -- The time amount of the timer when the encounter begins. If it is different from 0, it will be automatically started at the beginning of the encounter.
          loop = true,                    -- If true, the timer will immediatly start again upon expiring (with "default" time amount).
          finish = true,                  -- If finish is true, the timer will finish as soon as the timer hits 0. Not recommanded if there is an action from the boss you can use to sync the timer to.
                                            --   If finish is false, you'll typically need to watch for some spells of the boss that cause this timer to finish. See "spells" section below.
          callback = "PrintHelloWorld",   -- The name of the method that will be fired when the timer finishes. See the "extra" section below.
        },
    },

    zones = {                             -- Zones is a section where you describe all the zones your module can display.
        { id = "MyZone",                  -- The identifier of the zone. This is what the system will use to refer to it.
          category = "all/high",          -- The left part specifies "who" should pay attention to the zone. The right part specifies the importance of the zone. See timer above for details.
          color = "green",                -- The color of the zone. Either "green" (good zone of goodness), "yellow" or "red" (evil zone of evilness).

          duration = 15,                  -- The duration of the zone (seconds).
          radius = 7,                     -- The radius of the zone (in yards).
          type = "fixed",                 -- How the zone is positionned. It can be "fixed" (requires x and y below in that case), "aimed" (will aim at an unit) or "anchor" (will follow an unit).
          x = 0.5,                        -- The X position of the zone if it is a fixed zone. Elsewise this line should be removed. It is relative to the map, 0.00 is left, 0.50 is middle, 1.00 is right.
          y = 0.5,                        -- The Y position. 0.00 is top, 0.50 is middle, 1.00 is bottom.
          multiple = false,               -- Allows more than one zone of this type to be displayed at the same type. Defaults to false.
        },
    },

    spells = {                            -- Spells is a section where you describe all the spells your module will watch for.
        { id = 12345,                     -- The SpellID to watch. Use a WoW database to get it. You can also use a list if there are several spells to list, like {12, 34, 56}.
          trigger = "effectgain",         -- How the spell is detected. Can be "cast" (boss starts a cast bar or the spell is instant), "damage" or "heal", "effectgain" or "effectfade" (buff/debuff).
          cooldown = 0,                   -- Cooldown allows you to specify that the spell cannot be detected several times in a quick succession. Use 0 or delete this line for no cooldown.
          onSelf = false,                 -- If true, actions will only occur if the spell is cast on you. Be careful, "cast" won't have the information of who is the target IF the spell is NOT instant.

          zone = "MyZone",                -- The name of the zone that will be shown (or hidden for "effectfade") once this spell is detected. Delete this line if there is no zone to show/hide.
          finishTimer = "MyTimer",        -- The name of the timer that will be finished once the spell is detected. Delete this line if the spell is not used to sync a timer (see timer infos above).
          startTimer = nil,               -- The name of the timer that will be started once the spell is detected. Delete this line if no timer is started. Do not use this for timers with "loop" set to true.
          callback = "DoSomething",       -- The name of the method called when the spell is detected. See the "extra" section below. Delete if not used. 
                                          -- Note: for "callback", GUID of the target will be passed as parameter, if available. If it's an effect gain/fade event, stack count will be passed as well too.
        },
    },

    -- ** Extra resources needed

    settings = {                          -- In this table you can define your own settings for the module. They can be used in your custom methods (see below).
        {
          id = "Foo",                                      -- The identifier the system (and you in your methods) will use to refer to the setting.
          type = "BOOLEAN",                                -- What kind of value the setting stores. BOOLEAN is generally the only thing you need for modules. DROPDOWN might be interesting sometimes too.
          default = true,                                  -- The default value of the setting.

          action = "SYMBOL",                               -- Either RAID_WARNING or SYMBOL in most cases. This tells what the setting will do. It is used to control permissions.
          exclusive = false,                               -- Set this if only one person should have control over the option. This is important for raid warnings (and somewhat symbols).

          label = {                                        -- The title of the option.
              ["enUS"] = "Some option",
              ["frFR"] = "Une option",
          },
          explain = {                                      -- The description of the option.
              ["enUS"] = "The description of the option.",
              ["frFR"] = "La description de l'option.",
          },
        },
    },

    locale = {                            -- In this table are stored the localizations needed by your module. Some of the fields are special, see below.
        ["enUS"] = {
            ["Title"] = "My quick module", -- "Title" field can be used to set a custom title to your module. By default it is the boss name. Highly recommanded when there is no main boss mob.

            ["EngageTrigger"] = "When I say this, the module is engaged!", -- This field can be used to specify a mob yell that will cause the module to trigger. Delete if not required.
            ["ClearTrigger"] = "When I say this, you win!",                -- This field can be used to specify a mob yell that will cause the encounter to be won. Delete if not required.
            ["FailTrigger"] = "When I say this, you lose!",                -- This field can be used to specify a mob yell that will cause the encounter to be lost. Delete if not required.

            ["SomeTrigger"] = "Blablablah",                                -- This field is used for our custom yell (see above).
            ["MyThreshold"] = "Phase 2 starts",                            -- This field is used for our threshold (see above).

            ["MyAlert"] = "Move your fucking ass out of here!",            -- This is a general purpose custom field that can be used by custom methods (see "extra" below).

        },
        ["frFR"] = {
            ["Title"] = "Mon module rapide",

            ["EngageTrigger"] = "Quand je dis ceci, le module est engagé !",
            ["ClearTrigger"] = "Quand je dis ceci, vous gagnez !",
            ["FailTrigger"] = "Quand je dis ceci, vous perdez !",

            ["SomeTrigger"] = "Blablabla",
            ["MyThreshold"] = "Début de phase 2",

            ["MyAlert"] = "Bouge tes fesses de là!",
        },
    },

    extra = {                             -- In this table you specify extra methods you module will need. This is required when using "callback" fields.
        PrintHelloWorld = function(self)
            if self:GetSetting("Foo") then -- Check the Foo setting is true.
                self:Message("Hello world!"); -- A sample custom function displaying "Hello world!" message.
            end
        end,

        PrintAlertMessage = function(self)
            self:Alert(self:Localize("MyAlert")); -- A sample custom function displaying my localized alert message in locale table.
        end,

        StartPhaseTwo = function(self)
            self:SetPhase(2);             -- A sample custom function that causes "Phase 2" to be displayed.
        end,

        DoSomething = function(self, guid)
                                          -- A sample custom function that, unlike what its name suggests, doesn't do anything! Also, since it's a spell event callback, the GUID of the target is passed as parameter.
        end,
    },
} );