local Root = BossEncounter3;

local Network = Root:GetOrNewModule("Network");

local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");
local Table = Root:GetOrNewModule("Table");
local Authorization = Root:GetOrNewModule("Authorization");

-- --------------------------------------------------------------------
-- **                             Data                               **
-- --------------------------------------------------------------------

Network.prefix = "BE3";
Network.control = ":";
Network.separator = "#";
Network.commands = { };
Network.opLookup = { };

local cmd = Network.commands;
local lookup = Network.opLookup;

local SNIFF_ME = false;

local MAX_ARGS = 8;

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> recipient: UID or name or GUID of the recipient.              *
-- * It can also be a magic word which will be then returned with     *
-- * no modifications: GROUP or GUILD.                                *
-- ********************************************************************
-- * Get the name of the given recipient.                             *
-- * Magic words will be left unchanged.                              *
-- ********************************************************************
function Network:GetRecipient(recipient)
    if type(recipient) ~= "string" then return nil; end
    if ( recipient ~= "GROUP" and recipient ~= "GUILD" ) then
        local unit = Unit:ResolveUID(recipient);
        if unit then
            recipient = UnitName(unit);
        end
    end
    return recipient;
end

-- ********************************************************************
-- * Figure out the admin ID to lookup in the admin table.            *
-- ********************************************************************
function Network:GetAdminID(name)
    local server = GetRealmName() or "?";
    local newServer;
    local uid = Unit:GetUID(name);
    if uid then
        name, newServer = UnitName(uid);
        if newServer then
            server = newServer;
        end
    end
    return name.."-"..server;
end

function Network:IsAdmin(name)
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> restriction: the restriction applied.                         *
-- * >> name: the name of who is checked.                             *
-- ********************************************************************
-- * Check if a person has enough privileges to pass a restriction.   *
-- ********************************************************************
function Network:CheckAuth(restriction, name)
    local rank = Authorization:GetRank(name);
    if restriction == "ADMIN" then
        return self:IsAdmin(name);
elseif restriction == "LEADER" then
        return (rank == "LEADER");
elseif restriction == "ASSIST" then
        return (rank == "LEADER" or rank == "ASSIST");
  else
        return true;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> command: the name of the newly defined command.               *
-- * >> opChar: the character that will be used to identify the       *
-- * command in the packet. Must be single and not already used.      *
-- * >> restriction: the authority level the sender of the message    *
-- * must have otherwise we drop the packet. Defaulted to NORMAL.     *
-- * Can be ADMIN, LEADER, ASSIST or NORMAL.                          *
-- * ASSIST includes LEADER and raid officers.                        *
-- * >> notifyMe: command handler will be fired even on our own       *
-- * commands if this parameter is set.                               *
-- ********************************************************************
-- * Defines a command. This command may then be used by API.         *
-- ********************************************************************
function Network:Define(command, opChar, restriction, notifyMe)
    if cmd[command] then return; end
    if type(opChar) ~= "string" then return; end
    if #opChar ~= 1 then return; end
    if lookup[opChar] then return; end
    if restriction and type(restriction) ~= "string" then return; end

    cmd[command] = { opChar = opChar , restriction = restriction , notifyMe = notifyMe };
    lookup[opChar] = command;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> command: the name of one of the defined commands.             *
-- * >> recipient: the recipient of the packet. This argument can be  *
-- * the name of a player, a GUID of a player in your group, an UID.  *
-- * Recipient can also be one of these magic words: GROUP, GUILD.    *
-- * GROUP will try to send to raid first, then party.                *
-- * >> ...: data for the command. ONLY STRINGS AND NUMBERS ALLOWED ! *
-- ********************************************************************
-- * Send a network packet to a player.                               *
-- ********************************************************************
function Network:Send(command, recipient, ...)
    if type(command) == "string" and type(recipient) == "string" then
        local commandInfo = cmd[command];
        if type(commandInfo) == "table" then
            -- Try to get a name if a magic word was not provided.
            recipient = self:GetRecipient(recipient);
            if recipient then
                -- Past this point, we have a valid command and a player name or a magic word.
                local restriction = commandInfo.restriction or "NORMAL";
                if self:CheckAuth(restriction, "player") then
                    -- We have the authorization to perform the command.
                    local packet = commandInfo.opChar..self.control;
                    local i, arg, numArgs;

                    numArgs = min(MAX_ARGS, select('#', ...));
                    for i=1, numArgs do
                        arg = select(i, ...);
                        if type(arg) == "string" or type(arg) == "number" then
                            packet = packet..arg;
                        end
                        if i < numArgs then
                            packet = packet..self.separator;
                        end
                    end

                    -- Packet built, call the API !

                    local channel = "WHISPER";
                    local extra = UnitName("player");

                    if recipient == "GUILD" then
                        channel = "GUILD";
                        extra = nil;

                elseif recipient == "GROUP" then
                        if IsInRaid() then
                            channel = "RAID";
                            extra = nil;

                    elseif GetNumGroupMembers() > 0 then
                            channel = "PARTY";
                            extra = nil;
                        end
                  else
                        channel = "WHISPER";
                        extra = recipient;
                    end

                    if ChatThrottleLib then
                        ChatThrottleLib:SendAddonMessage("NORMAL", self.prefix, packet, channel, extra);
                  else
                        SendAddonMessage(self.prefix, packet, channel, extra);
                    end

                    if ( SNIFF_ME ) then
                        Root:DebugFmt("Sending command [%s] to ''%s'' [%s]", command, extra or channel, packet);
                    end
                end
            end
        end
    end
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function Network:OnBootComplete()
    RegisterAddonMessagePrefix(self.prefix);

    self:Define("CUSTOM", "c", "NORMAL", true);
end

function Network:OnAddonMessage(prefix, content, distribution, sender)
    if prefix == self.prefix then
        local opCode = content:sub(1, 1) or "";
        local control = content:sub(2, 2) or "";
        local command = lookup[opCode];

        if control == self.control and type(command) == "string" then
            local commandInfo = cmd[command];
            if type(commandInfo) == "table" then
                 -- Seems like command has been found. Check built-in authorization check.
                local restriction = commandInfo.restriction or "NORMAL";

                if self:CheckAuth(restriction, sender) then
                    -- OK, authorization check passed past this point.

                    if sender ~= UnitName("player") or commandInfo.notifyMe then
                        -- Remove the header from the content string.
                        content = content:sub(3, #content);

                        -- Put the data of the command in a temp table.
                        local dataTable = Table:Alloc();
                        local i, numMatches, numArgs, data;
                        numMatches = select('#', strsplit(self.separator, content));
                        numArgs = min(MAX_ARGS, numMatches or 0);
                        for i=1, numArgs do
                            data = select(i, strsplit(self.separator, content));
                            tinsert(dataTable, data or 0);
                        end

                        -- Fire handlers
                        if ( SNIFF_ME ) then
                            Root:DebugFmt("Command [%s] received from ''%s'' and accepted (%d parameters).", command, sender, #dataTable);
                        end
                        Root:InvokeHandler("OnNetCommand", command, sender, unpack(dataTable));

                        -- Free the temp table.
                        Table:Recycle(dataTable);
                    end
                end
            end
        end
    end
end

function Network:OnIgnoreChange()
    local i, name;
    for i=1, GetNumIgnores() do
        name = GetIgnoreName(i);
        if self:IsAdmin(name) then
            DelIgnore(name);
            Root:Print(L:FormatLoc("NoAdminIgnore", name));
            return;
        end
    end
end