local Root = BossEncounter3;

local Save = Root:GetOrNewModule("Save");

local L = Root:GetOrNewModule("Localization");

-- --------------------------------------------------------------------
-- **                             Data                               **
-- --------------------------------------------------------------------

local DefaultSavedVariables = {
    ["system"] = { },
    ["gui"] = { },
    ["module"] = { },
    ["plugin"] = { },
    ["npcNames"] = { },
    ["version"] = 1,
};

local SavedVariablesGlobalName = "BossEncounter3_SavedVariables";

_G[SavedVariablesGlobalName] = { };
local SavedVariables = _G[SavedVariablesGlobalName];

Save.ready = false;

-- --------------------------------------------------------------------
-- **                            Methods                             **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> part: which side of sVars the data to fetch is stored into.   *
-- * >> key: what data to retrieve. If it is not found on SavedVars   *
-- * table, we will search into DefaultSavedVars. If still not        *
-- * found, nil will be returned.                                     *
-- ********************************************************************
-- * Returns a saved variable value (a variable which is "remembered" *
-- * between each game session).                                      *
-- ********************************************************************
function Save:Get(part, key)
    local partData = SavedVariables[part];
    if partData and partData[key] ~= nil then
        return partData[key];
    end
    return self:GetDefault(part, key);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> part: which side of sVars the data to set is stored into.     *
-- * >> key: where to store data.                                     *
-- * >> value: self-explanatory.                                      *
-- ********************************************************************
-- * Changes a saved variable value.                                  *
-- ********************************************************************
function Save:Set(part, key, value)
    if not SavedVariables[part] then SavedVariables[part] = { }; end
    SavedVariables[part][key] = value;
    return true;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> part: which side of sVars the data to fetch is stored into.   *
-- * >> key: what data to retrieve.                                   *
-- ********************************************************************
-- * Returns the default value of a saved variable.                   *
-- ********************************************************************
function Save:GetDefault(part, key)
    local defaultPartData = DefaultSavedVariables[part];
    if defaultPartData then
        return defaultPartData[key];
    end
    return nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> part: which side to get direct access of.                     *
-- ********************************************************************
-- * Directly access to the table of a given part of the saved vars.  *
-- ********************************************************************
function Save:GetDirectAccess(part)
    if not SavedVariables[part] then SavedVariables[part] = { }; end
    return SavedVariables[part];
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * N/A                                                              *
-- ********************************************************************
-- * Compares saved variables defaults' version with saved variables' *
-- * version. If there is a difference, clear saved variables and     *
-- * return true.                                                     *
-- ********************************************************************
function Save:CheckVersion()
    SavedVariables = _G[SavedVariablesGlobalName];

    local currentVersion = DefaultSavedVariables.version;
    local savedVersion = SavedVariables.version;

    if not ( currentVersion ) then
        return false;
    end

    if not ( savedVersion ) then
        -- Saved variables table has been formatted. This should occur the first time the mod is run.
        SavedVariables = {
            version = currentVersion,
        };
        _G[SavedVariablesGlobalName] = SavedVariables;
        return false;
  else
        if ( savedVersion ~= currentVersion ) then
            -- Reformats saved variables table, coz' version has changed.
            SavedVariables = {
                version = currentVersion,
            };
            _G[SavedVariablesGlobalName] = SavedVariables;
            -- And also informs the caller.
            return true;
        end
    end

    return false;
end

-- ********************************************************************
-- * Clear all saved variables and formats a new saved vars table.    *
-- ********************************************************************
function Save:Clear()
    SavedVariables = { };
    _G[SavedVariablesGlobalName] = SavedVariables;
    self:CheckVersion();
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Save:OnBootCompleteImportant()
    if not self.ready then
        local reformat = self:CheckVersion();
        if reformat then
            Root:Print(L:Get("SaveReformat"));
        end
        self.ready = true;
    end
end
