local Root = BossEncounter3;

local Unit = Root:GetOrNewModule("Unit");

local Table = Root:GetOrNewModule("Table");

-- --------------------------------------------------------------------
-- **                             Data                               **
-- --------------------------------------------------------------------

Unit.nextUpdate = 0;
Unit.nextDisconnectCheck = 0;
Unit.mouseoverGUID = "none";

local UIDLookup = { };
local UnitList = { };

local connectState = { };

local UPDATE_UNIT_LIST_RATE = 0.30;
local CHECK_DISCONNECT_RATE = 0.50;

-- Useful table for building the unit list.

local StaticPassList = { };
local TempPassList   = { };
local TargetPassList = { };

local AddTargetToTable = function(uid, target, nestLevels)
    local i;
    for i=1, nestLevels do
        tinsert(target, uid..strrep("target", i));
    end
end;

local me = {
    player = true,
    pet = true,
    vehicle = true,
}

local guidTypes = {
    [0] = "player",
    [1] = "object",
    [3] = "npc",
    [4] = "pet",
    [5] = "vehicle",
};

local affiliations = { };
local enumeration = { };

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of who is checked.                             *
-- ********************************************************************
-- * Determinates the unit type of the unit a GUID belongs to.        *
-- * Will either return "npc" (incl. pets) or "player".               *
-- * "unknown" may be returned in case of erroneous data.             *
-- ********************************************************************
function Unit:GetTypeFromGUID(guid)
    if guid then
        local digit = (guid):sub(5, 5);
        local value = tonumber(digit, 16);
        if value then
            value = bit.band(value, 0x07);
            if guidTypes[value] then
                return guidTypes[value];
            end
        end
    end
    return "unknown";
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID to extract the mob ID from.                    *
-- ********************************************************************
-- * Get the mob ID portion of a GUID (returns decimal number).       *
-- ********************************************************************
function Unit:GetMobID(guid)
    local npcType = self:GetTypeFromGUID(guid);
    if npcType == "npc" or npcType == "vehicle" or npcType == "object" then
        local id = (guid):sub(7, 10);
        return tonumber(id, 16);
    end
    return nil;
end

-- ********************************************************************
-- * Get the numbers of different units that are listed in the global *
-- * unit list. No unit apparears twice in this list, so if Bob can   *
-- * be accessed from "player" and "target" UIDs, the UID that will   *
-- * be kept to access Bob will be "player".                          *
-- ********************************************************************
function Unit:GetNumUID()
    return #UnitList;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid or name: the GUID or name to find an UID for.            *
-- *                             - OR -                               *
-- * >> index: the index to get info about in the global unit list.   *
-- ********************************************************************
-- * Try to get an UID pointing at the given GUID or name.            *
-- * This function has a high chance for failing.                     *
-- * In that case, nil will be returned.                              *
-- *                             - OR -                               *
-- * You can also explore with this function the global unit list by  *
-- * passing the index of the element of the list you want to get     *
-- * infos from. GUID, name and UID will then be returned.            *
-- ********************************************************************
function Unit:GetUID(value)
    if type(value) == "number" then
        -- Find the nth entry in the global unit list.
        local info = UnitList[value];
        if info then
            local guid, name = strsplit("|", info, 2);
            local uid = self:GetUID(guid);
            return guid, name, uid;
        end
        return nil, nil, nil;

elseif type(value) == "string" then
        -- Find a guid or name in the quick lookup table.
        local expectedUID = UIDLookup[value];
        if expectedUID then
            local name = UnitName(expectedUID);
            local guid = UnitGUID(expectedUID);
            if value == name or value == guid then -- Safety check in case UIDs have changed in-between.
                return expectedUID;
            end
        end
        return nil;
    end
end

-- ********************************************************************
-- * Calling this API will ask a complete reconstruction of           *
-- * accessable units by either their name or GUID.                   *
-- * This is necessary for GetUID function.                           *
-- * RebuildUnitList should be called whenever any UID changes or at  *
-- * a reasonnable periodic update rate.                              *
-- ********************************************************************
function Unit:RebuildList()
    -- Clean up

    wipe(UnitList);
    wipe(UIDLookup);

    wipe(StaticPassList);
    wipe(TempPassList);
    wipe(TargetPassList);

    -- Enumeration of stuff in the correct priority list.
    -- Static UIDs will get the first slots (the more stable ones), temporary UIDs will get not-so-stable slots and target UIDs will get the most unstable ones.

    local i;

    -- Static UIDs enumeration.
    tinsert(StaticPassList, "player"); -- Make sure player IS in the list, even if it appears twice.
    if IsInRaid() then
        for i=1, GetNumGroupMembers(), 1 do
            tinsert(StaticPassList, "raid"..i);
            tinsert(TempPassList, "raidpet"..i);
        end
  else
        for i=1, GetNumGroupMembers(), 1 do
            tinsert(StaticPassList, "party"..i);
            tinsert(TempPassList, "partypet"..i);
        end
    end
    for i=1, 4 do
        tinsert(StaticPassList, "boss"..i);
    end

    -- Temp UIDs enumeration.
    tinsert(TempPassList, "pet");
    tinsert(TempPassList, "target");
    tinsert(TempPassList, "focus");
    tinsert(TempPassList, "mouseover");
    tinsert(TempPassList, "vehicle");

    -- The target of target UIDs are automatically generated from Static and Temp UIDs.

    -- Do each pass
    self:AddToList(StaticPassList, AddTargetToTable, TargetPassList, 3);
    self:AddToList(TempPassList, AddTargetToTable, TargetPassList, 3);
    self:AddToList(TargetPassList);

    -- Schedule the next rebuild
    self.nextUpdate = GetTime() + UPDATE_UNIT_LIST_RATE;

    -- Fire unit list updated handler
    Root:InvokeHandler("OnUnitListUpdate");
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> UID: the unit to add to the unit list and lookup table.       *
-- *                          - OR -                                  *
-- * >> list: a list containing UIDs we should try to add.            *
-- * >> callback: callback to call each time an unit is added         *
-- * succesfully in the global unit list, passing UID.                *
-- * >> ...: arguments to pass to callback.                           *
-- ********************************************************************
-- * This will add the given unit's name and guid to the lookup table *
-- * and create if it doesn't exist yet an entry for the unit in the  *
-- * global unit list.                                                *
-- * Returns false if the unit has already been referenced (through   *
-- * another but equivalent UID for exemple) or if the unit doesn't   *
-- * exist.                                                           *
-- * Returns true if the unit has been added to the list correctly.   *
-- ********************************************************************
function Unit:AddToList(UID, callback, ...)
    if type(UID) == "table" then
        local k, v;
        for k, v in ipairs(UID) do
            self:AddToList(v, callback, ...);
        end

elseif type(UID) == "string" and UnitExists(UID) then
        local name = UnitName(UID);
        local guid = UnitGUID(UID);

        self:ResolveAffiliation(UID);

        if not UIDLookup[name] then
            UIDLookup[name] = UID;
        end
        if not UIDLookup[guid] then
            -- GUID still not in the lookup. First occurence of this entity in the list.
            UIDLookup[guid] = UID;
            tinsert(UnitList, guid.."|"..name);
            if type(callback) == "function" then
                callback(UID, ...);
            end
            return true;
        end
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: either a GUID, a name, or even an UID.                  *
-- ********************************************************************
-- * Try to transform the argument unit into a valid UID.             *
-- * Return an UID or nil.                                            *
-- ********************************************************************
function Unit:ResolveUID(unit)
    if unit and not UnitExists(unit) then
        -- Invalid UID, try to resolve.
        unit = self:GetUID(unit);
    end
    return unit;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mobID: the mobID we want to find an instance of.              *
-- * >> allowDead: if true, dead units will be valid candidates.      *
-- ********************************************************************
-- * Get the GUID of the first instance of the mobID that is          *
-- * encountered in the global unit list. Also return the UID.        *
-- ********************************************************************
function Unit:GetGUIDFromMobID(mobID, allowDead)
    if mobID then
        local i, guid, name, uid;
        for i=1, self:GetNumUID() do
            guid, name, uid = self:GetUID(i);
            if uid and self:GetMobID(guid) == mobID and name ~= UNKNOWN and (not UnitIsDeadOrGhost(uid) or allowDead) then
                return guid, uid;
            end
        end
    end
    return nil, nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> uid: the unit to check.                                       *
-- * >> isBuff: true if it's a buff, false if it's a debuff.          *
-- * >> effect: the name -or- spellID of the effect.                  *
-- * >> mineOnly: determinate if only effects put by the              *
-- *              player are valid results.                           *
-- ********************************************************************
-- * Check if a given effect is present on the given unit.            *
-- * Returns startTime, duration, stackCount if present.              *
-- *         nil, nil, nil if not.                                    *
-- ********************************************************************
function Unit:SearchEffect(uid, isBuff, effect, mineOnly)
    local filter;
    if isBuff then
        filter = "HELPFUL";
  else
        filter = "HARMFUL";
    end

    local i;
    for i=1, 40 do
        local name, rank, icon, count, debuffType, duration, expiration, caster, stealable, consolidate, spellId = UnitAura(uid, i, filter);
        if name == effect or spellId == effect then
            if (not mineOnly) or (mineOnly and me[caster]) then
                return expiration - duration, duration, count;
            end
        end
    end
    return nil, nil, nil;
end

-- ********************************************************************
-- * Enumerate all UIDs of raid members.                              *
-- * If you are not in a raid, party UIDs including yourself          *
-- * will be put instead. Pets are ignored.                           *
-- ********************************************************************
function Unit:EnumerateRaid()
    local i;
    wipe(enumeration);
    if IsInRaid() then
        for i=1, GetNumGroupMembers() do
            tinsert(enumeration, "raid"..i);
        end
  else
        for i=1, GetNumGroupMembers() do
            tinsert(enumeration, "party"..i);
        end
        tinsert(enumeration, "player");
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> role: the desired role.                                       *
-- ********************************************************************
-- * Enumerate all UIDs of raid members with the given role.          *
-- * If you are not in a raid, party UIDs including yourself          *
-- * will be put instead.                                             *
-- ********************************************************************
function Unit:EnumerateRaidByRole(role)
    local i;
    wipe(enumeration);
    if IsInRaid() then
        for i=1, GetNumGroupMembers() do
            if self:GetRole("raid"..i) == role then
                tinsert(enumeration, "raid"..i);
            end
        end
  else
        for i=1, GetNumGroupMembers() do
            if self:GetRole("party"..i) == role then
                tinsert(enumeration, "party"..i);
            end
        end
        if self:GetRole("player") == role then
            tinsert(enumeration, "player");
        end
    end
end

-- ********************************************************************
-- * Get the size of the last enumeration.                            *
-- ********************************************************************
function Unit:GetEnumerationSize()
    return #enumeration;
end

-- ********************************************************************
-- * Get the n-th UID in the last enumeration.                        *
-- ********************************************************************
function Unit:GetEnumerationUID(index)
    return enumeration[index];
end

-- ********************************************************************
-- * Calling this API will check if any raid member has gone from     *
-- * connected to disconnected status (or opposite).                  *
-- ********************************************************************
function Unit:CheckDisconnect()
    local baseUID = "party";
    if IsInRaid() then baseUID = "raid"; end
    local i, uid, guid, connected, name;
    for i=1, GetNumGroupMembers() do
        uid = baseUID..i;
        guid = UnitGUID(uid);
        if guid then
            connected = UnitIsConnected(uid) or 0;
            name = UnitName(uid);
            if connectState[guid] and connected ~= connectState[guid] then
                if connected == 0 then
                    -- This guy has (been) disconnected.
                    Root:InvokeHandler("OnConnectChange", guid, name, false);
                end
                if connected == 1 and UnitAffectingCombat("player") then
                    -- This guy has reconnected.
                    Root:InvokeHandler("OnConnectChange", guid, name, true);
                end
            end
            connectState[guid] = connected;
        end
    end
    -- Schedule the next check
    self.nextDisconnectCheck = GetTime() + CHECK_DISCONNECT_RATE;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> petUID: the UID of the pet.                                   *
-- ********************************************************************
-- * Determinate the UID of the master of a pet, if possible.         *
-- ********************************************************************
function Unit:GetPetMasterUID(petUID)
    if petUID then
        -- First obvious check
        if UnitIsUnit("pet", petUID) or UnitIsUnit("vehicle", petUID) then
            return "player";
        end

        -- Now check the raid/party
        local i, baseUID;
        if IsInRaid() then
            baseUID = "raid";
      else
            baseUID = "party";
        end
        for i=1, GetNumGroupMembers() do
            if UnitIsUnit(baseUID.."pet"..i, petUID) then
                return baseUID..i;
            end
        end
    end
    return nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> uid: the unit parsed.                                         *
-- ********************************************************************
-- * Resolve the affiliation of a given unit.                         *
-- ********************************************************************
function Unit:ResolveAffiliation(uid)
    local guid = UnitGUID(uid);
    local name = UnitName(uid);
    local isPlayer = UnitIsPlayer(uid);

    local _, class = UnitClass(uid);
    local info = affiliations[guid];

    -- If the unit is a player, we only proceed if we can get its class info and name, elsewise we will retry later.
    -- Do not try to resolve units that have already been classified as affiliated to us.

    if ((not info) or (not info.partner)) and ((not isPlayer) or class) and name ~= UNKNOWN then
        if not info then
            info = Table:Alloc();
        end
        info.name = name;
        info.partner = self:IsPartner(uid);

        if isPlayer then
            -- This unit is a player.
            info.type = "PLAYER";
            info.class = class;
            info.master = nil;
      else
            -- This unit could be a pet.
            local masterUID = self:GetPetMasterUID(uid);
            if masterUID then
                info.partner = info.partner or self:IsPartner(masterUID); -- The master may be affiliated with us, so we are affiliated with the pet.

                -- It is indeed a pet.
                info.type = "PET";
                info.class = "PET";
                info.master = UnitGUID(masterUID);
          else
                -- Likely to be a normal NPC, or a guardian.
                info.type = "NPC";
                info.class = "NPC";
                info.master = nil;
            end
        end



        affiliations[guid] = info;

    end

    -- Note: this function is bad for handling treants / spirit wolves.
    --       The "summon" combat event is good though for those.
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit queried.                           *
-- ********************************************************************
-- * Retrieve the affiliation of a given unit.                        *
-- * Return isPartner, type, name, class or master                    *
-- ********************************************************************
function Unit:GetAffiliation(guid)
    local info = affiliations[guid];
    if info then
        if info.type == "PLAYER" then
            return info.partner, info.type, info.name, info.class;
      else
            return info.partner, info.type, info.name, info.master;
        end
    end
    return false, "UNKNOWN", UNKNOWN, nil;
end


-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the pet resolved.                           *
-- * >> name: the name of the pet resolved.                           *
-- * >> masterGUID: the guid of the master of the pet.                *
-- ********************************************************************
-- * Allows you to flag an unit as a pet owned by another unit.       *
-- * Works great with the SUMMON combat event.                        *
-- ********************************************************************
function Unit:SetPetAffiliation(guid, name, masterGUID)
    local info = affiliations[guid];
    if not info then
        info = Table:Alloc();
    end

    local petUID = self:GetUID(guid);
    local masterUID = self:GetUID(masterGUID);

    info.name = name;
    info.partner = self:IsPartner(petUID) or self:IsPartner(masterUID);
    info.type = "PET";
    info.class = "PET";
    info.master = masterGUID;

    affiliations[guid] = info;

end


-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: GUID, name or UID of the unit queried.                  *
-- * This defaults to "player".                                       *
-- ********************************************************************
-- * Get a player's faction.                                          *
-- ********************************************************************
function Unit:GetFaction(unit)
    unit = self:ResolveUID(unit or "player");
    if unit then
        local faction = UnitFactionGroup(unit);
        if faction then
            return string.upper(faction);
        end
    end
    return "NONE";
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: GUID, name or UID of the unit queried.                  *
-- ********************************************************************
-- * Check if an unit is one of our partners.                         *
-- ********************************************************************
function Unit:IsPartner(unit)
    unit = self:ResolveUID(unit);
    if unit then
        return UnitIsUnit(unit, "player") or UnitIsUnit(unit, "pet") or UnitIsUnit(unit, "vehicle") or UnitInRaid(unit) or UnitInParty(unit);
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: GUID, name or UID of the unit resolved.                 *
-- ********************************************************************
-- * Try to find the unit that is tanking the specified unit.         *
-- * Return guid, name, uid.                                          *
-- ********************************************************************
function Unit:GetTankOf(unit)
    local mobUID = self:ResolveUID(unit);
    if mobUID then
        self:EnumerateRaid();
        local i, unit, isTanking;
        for i=1, self:GetEnumerationSize() do
            unit = self:GetEnumerationUID(i);
            isTanking = UnitDetailedThreatSituation(unit, mobUID);
            if isTanking then
                local guid = UnitGUID(unit);
                local name = UnitName(unit);
                return guid, name, unit;
            end
        end
    end
    return nil, nil, nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: GUID, name or UID of the unit queried.                  *
-- ********************************************************************
-- * Get the role of an unit (DAMAGER, HEALER, TANK).                 *
-- * Role will default to DAMAGER.                                    *
-- ********************************************************************
function Unit:GetRole(unit)
    unit = self:ResolveUID(unit);
    if unit then
        if GetPartyAssignment("MAINTANK", unit) then
            return "TANK";
    elseif GetPartyAssignment("MAINASSIST", unit) then
            return "HEALER";
      else
            local role = UnitGroupRolesAssigned(unit);
            if role ~= "NONE" then
                return role;
            end
        end
    end
    return "DAMAGER";
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function Unit:OnBootComplete()
    self:OnWorldEnter();
end

function Unit:OnWorldEnter()
    Table:Recycle(affiliations, true);
    affiliations = Table:Alloc();
end


function Unit:OnUpdate(elapsed)
    local mouseoverGUID = UnitGUID("mouseover") or "none";
    if self.mouseoverGUID ~= mouseoverGUID then
        self.mouseoverGUID = mouseoverGUID;
        self:RebuildList();
    end

    local t = GetTime();
    if t > self.nextUpdate then
        self:RebuildList();
    end
    if t > self.nextDisconnectCheck then
        self:CheckDisconnect();
    end
end