local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");

local AlertMessage = GUI:GetOrNewWidget("AlertMessage");

AlertMessage.TYPE = "Button";
AlertMessage.TEMPLATE = "BossEncounter3_AlertMessageTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Display an alert message. Only one message at a time.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local ALPHA_RATE = 4;

local BLUR_TOTAL_SCALE = 2.0;
local BLUR_TOTAL_TIMER = 1.0;
local BLUR_FADE_TIMER = 0.6;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> message: the message to be displayed.                         *
-- * >> duration: duration, can be omitted.                           *
-- ********************************************************************
-- * Display an alert (previous one will be replaced).                *
-- ********************************************************************
function AlertMessage:Display(message, duration)
    self.alpha = 0;
    self.blurTimer = 0;
    self.duration = duration or (1.5 + #message / 15);
    self.up = true;

    self.text:SetText(message);
    self.blur:SetText(message);

    self:SetWidth(self.text:GetWidth() + 64);
    self:SetHeight(self.text:GetHeight() + 16);

    self:Show();
    Audio:PlaySound("ALERT");
end

-- ********************************************************************
-- * Clear the alert.                                                 *
-- ********************************************************************
function AlertMessage:Clear()
    self.up = false;
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function AlertMessage:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);

    -- ** Children **
    self.text = GUI:RequestChild("MonoFont", nil, self, 1);
    self.text:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.text:Show();
    self.blur = GUI:RequestChild("MonoFont", nil, self, 2);
    self.blur:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.blur:Show();
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HS");
    self.band:SetFixedBorders(true);

    -- ** Properties **
    self.alpha = 0;
    self.blurTimer = 0;
    self.duration = 0;
    self.up = false;

    -- ** Initial config **
end

function AlertMessage:OnUpdate(elapsed)
    if self.up then
        if self.alpha < 1 then
            self.alpha = min(1, self.alpha + ALPHA_RATE * elapsed);
      else
            if self.duration > 0 then
                self.duration = max(0, self.duration - elapsed);
          else
                self:Clear();
            end
        end
  else
        if self.alpha > 0 then
            self.alpha = max(0, self.alpha - ALPHA_RATE * elapsed);
      else
            self:Hide();
        end
    end

    -- Blur handling

    self.blurTimer = self.blurTimer + elapsed;

    local blurScale = max(0.01, (self.blurTimer / BLUR_TOTAL_TIMER) ^ 0.5 * BLUR_TOTAL_SCALE);
    local blurAlpha;
    if self.blurTimer < BLUR_FADE_TIMER then
        blurAlpha = 1;
elseif self.blurTimer < BLUR_TOTAL_TIMER then
        blurAlpha = 1 - (self.blurTimer - BLUR_FADE_TIMER) / (BLUR_TOTAL_TIMER - BLUR_FADE_TIMER);
  else
        blurAlpha = 0;
    end
    if blurScale > 1 then
        self.text:SetAlpha(1);
  else
        self.text:SetAlpha(0);
    end
    self.blur:SetAlpha(0.2 * blurAlpha);
    self.blur:SetScale(blurScale);
    self:SetAlpha(self.alpha);
end