local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

local Graph = GUI:GetOrNewWidget("Graph");

Graph.TEMPLATE = "BossEncounter3_GraphTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- A DPS graph.

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local MAX_COLUMNS = 25;

local MAX_HEIGHT = 176;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: the time needed to scroll-in the graph.                 *
-- ********************************************************************
-- * Display the DPS graph.                                           *
-- ********************************************************************
function Graph:Open(time)
    if self.status == "STANDBY" or self.status == "CLOSING" then
        self.status = "OPENING";

        self:Reset();

        self:ScrollIn(time);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: the time to perform the scroll-out animation.           *
-- *          0 will cause the graph to be instantly removed.         *
-- ********************************************************************
-- * Stop displaying the DPS graph.                                   *
-- ********************************************************************
function Graph:Remove(time)
    if time == 0 or (not self:IsShown()) then
        self.status = "STANDBY";
        self:Hide();

elseif self.status == "OPENING" or self.status == "RUNNING" then
        self.status = "CLOSING";
        self:ScrollOut(time);
    end
end

-- ********************************************************************
-- * Hide all elements of the graph.                                  *
-- ********************************************************************
function Graph:Reset()
    -- Hide all columns
    local i;
    for i=1, MAX_COLUMNS do
        self.columns[i]:Hide();
    end

    -- Remove average line
    self.average = 0;
    self.averageHeight = 0;

    self.timer = 0;
    self.timerMax = 0;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> num: the number of columns that will get displayed.           *
-- ********************************************************************
-- * Set the number of columns that will get displayed in the graph.  *
-- ********************************************************************
function Graph:SetNumUsedColumns(num)
    self.usedColumns = max(5, num);

    local sizePerColumn = 480 / self.usedColumns;
    self.space = sizePerColumn;
    self.width = sizePerColumn * 0.8;
end

-- ********************************************************************
-- * Get the max number of columns displayable in the graph.          *
-- ********************************************************************
function Graph:GetNumColumns()
    return MAX_COLUMNS;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the index of the column to get.                        *
-- ********************************************************************
-- * Get the max number of columns that will get displayed in the     *
-- * graph and the number of columns that are set as used.            *
-- ********************************************************************
function Graph:GetColumn(index)
    return self.columns[index];
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> dpsTable: the DPS table to pull DPS info from.                *
-- * >> growTimer: the time for columns to grow.                      *
-- ********************************************************************
-- * Automatically set up the columns according to a DPS table.       *
-- * graph and the number of columns that are set as used.            *
-- ********************************************************************
function Graph:AutosetColumns(dpsTable, growTimer)
    local numEntries = min(self:GetNumColumns(), #dpsTable);
    local i, column, special, dpsEntry;

    local maximum = max(1, dpsTable.maximum);
    self.average = dpsTable.average;
    self.averageHeight = self.average * MAX_HEIGHT / maximum;

    self:SetNumUsedColumns(numEntries);

    for i=1, self:GetNumColumns() do
        column = self:GetColumn(i);
        if i <= numEntries then
            dpsEntry = dpsTable[i];

            -- Tooltip setup
            local pct = dpsEntry.dps * 100 / dpsTable.total;
            local wastedDps = max(0, dpsEntry.totalDps - dpsEntry.dps);
            local relevancePct = min(100, dpsEntry.dps * 100 / dpsEntry.totalDps);
            local petString = "";
            if dpsEntry.pet > 0 then
                petString = L:FormatLoc("DamageMeter-Pet", dpsEntry.pet);
            end
            local detail = L:FormatLoc("DamageGraph-Entry", i, dpsEntry.dps, petString, pct, wastedDps, relevancePct + 0.5);
            Utility:SetTooltip(column, dpsEntry.name, detail);

            -- Special handling
            local isLocal = (dpsTable[i].guid == UnitGUID("player"));
            local killingBlow = (dpsTable[i].guid == dpsTable.coupDeGrace);

            -- Display of the column
            column:Set(self.width, dpsEntry.dps * MAX_HEIGHT / maximum, dpsEntry.class, dpsEntry.dps, isLocal, killingBlow);
            column:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 16 + self.space * (i - 1 + 0.1), 32);
      else
            column:Hide();
        end
    end

    self.timer = 0;
    self.timerMax = growTimer;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function Graph:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **

    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetText(L:Get("DamageGraph-Title"));
    self.title:SetPoint("TOP", self, "TOP", 0, -8);
    self.closeButton = CreateFrame("Button", nil, self, "UIPanelCloseButton");
    self.closeButton:SetScript("OnClick", self.OnCloseButtonClick);
    self.closeButton:SetPoint("TOPRIGHT", self, "TOPRIGHT", -2, -1);

    self.columns = { };
    local i;
    for i=1, MAX_COLUMNS do
        self.columns[i] = GUI:RequestChild("GraphColumn", nil, self, 1);
    end

    self.averageBar = CreateFrame("Frame", nil, self);
    self.averageBar:SetWidth(480);
    self.averageBar:SetHeight(1);
    self.averageBar:SetFrameLevel(self:GetFrameLevel() + 2);
    self.averageBar.texture = self.averageBar:CreateTexture(nil, "BACKGROUND");
    self.averageBar.texture:SetAllPoints(true);
    self.averageBar.texture:SetTexture(1, 1, 1);
    self.averageBar.label = self:CreateFontString(self.averageBar, "ARTWORK", "GameFontHighlightSmall");
    self.averageBar.label:SetPoint("BOTTOMRIGHT", self.averageBar, "TOPRIGHT", 0, 0);

    -- ** Properties **
    self.status = "STANDBY";
    self.usedColumns = 5;
    self.width = 0;
    self.space = 0;
    self.average = 0;
    self.averageHeight = 0;
    self.timer = 0;
    self.timerMax = 0;

    -- ** Initial config **
    Utility:SetBackdrop(self, "OPTION");
    self:Reset();
end

function Graph:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);

    if self.status == "STANDBY" then
        self:Hide();
  else
        self.timer = min(self.timerMax, self.timer + elapsed);

        local progression = self.timer / max(1, self.timerMax);
        local gxProgression = sin(progression * 90)^2;

        -- Column handling
        local i, column;
        for i=1, self:GetNumColumns() do
            column = self:GetColumn(i);
            column:Update(progression, gxProgression);
        end

        -- Average handling
        self.averageBar:SetPoint("BOTTOM", self, "BOTTOM", 0, 31 + gxProgression * self.averageHeight);
        self.averageBar.label:SetAlpha(progression);
        self.averageBar.label:SetText(L:FormatLoc("DamageMeter-Average", gxProgression * self.average));
    end
end

function Graph:OnCloseButtonClick()
    self:GetParent():Notify("OnInterruptRequest");
end

function Graph:OnScrollOutEnd()
    self:Remove(0);
end