local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local L = Root:GetOrNewModule("Localization");
local Audio = Root:GetOrNewModule("Audio");
local Utility = Root:GetOrNewModule("Utility");

local System = Root:GetOrNewModule("System");
local Stats = System:GetOrNewModule("Stats");

local ResultAnim = GUI:GetOrNewWidget("ResultAnim");

ResultAnim.TEMPLATE = "BossEncounter3_ResultAnimTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Show results after clearing an encounter.

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local TEXT_SCALE = 0.75;

local MASK_FINAL_ALPHA = 0.5;

local RESULT_TEXT_YPOSITION = 0.80;
local TIME_TEXT_YPOSITION = 0.76;
local RECORD_TEXT_YPOSITION = 0.72;

-- In sec.
local FADE_DURATION = 0.20;
local SCROLL_DURATION = 0.30;

-- The following is only used for the full sequences.

local rankToMusic = {
    ["E"] = "VICTORY_FAIR",
    ["D"] = "VICTORY_GOOD",
    ["C"] = "VICTORY_GOOD",
    ["B"] = "VICTORY_GREAT",
    ["A"] = "VICTORY_GREAT",
    ["S"] = "VICTORY_EXCELLENT",
};

local musicTimings = {
    ["VICTORY_FAIR"] = {
        waitTimer =  0.5, -- Time before introducing the score frame after the result text has been shown.
        showTimer =  1.0, -- Time the score frame will need to take to show up.
        prePause  =  0.8, -- Time the score frame will wait until we play the score sequence. That is, growing bars and showing the rank.
        growTimer =  1.2, -- Time the bars in the score frame have to grow to their final value.
        postPause = 20.0, -- Time the score frame will wait after showing its final results before getting out of the screen.
        hideTimer =  2.0, -- Time the score frame will need to take to get out of screen.
    },
    ["VICTORY_GOOD"] = {
        waitTimer =  0.4,
        showTimer =  1.0,
        prePause  =  0.3,
        growTimer =  1.7,
        postPause = 20.0,
    },
    ["VICTORY_GREAT"] = {
        waitTimer =  0.1,
        showTimer =  1.0,
        prePause  =  0.1,
        growTimer =  1.5,
        postPause = 20.0,
    },
    ["VICTORY_EXCELLENT"] = {
        waitTimer =  0.5,
        showTimer =  1.0,
        prePause  =  1.0,
        growTimer =  1.0,
        postPause = 30.0,
    },
};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Starts playing the result sequence.                              *
-- ********************************************************************
function ResultAnim:Play()
    local name, attempts, attemptScore, speedScore, techniqueScore, clearTime, previousTime, previousScore, numKilled, sequence, dps = Stats:GetCurrentInfo();
    if self.status == "STANDBY" then
        self.status = "OPENING";
        self.timer = FADE_DURATION;
        self.previousScore = previousScore;
        self.sequence = sequence;
        self.dps = dps;

        if attemptScore and speedScore and techniqueScore then
            self.score = attemptScore + speedScore + techniqueScore;
            self.rank = Utility:PtsToRank(self.score);
            self.newRecord = self.score > (previousScore or 0);
            self.attempt, self.speed, self.technique = attemptScore, speedScore, techniqueScore;
      else
            self.score = nil;
            self.rank = nil;
            self.newRecord = false;
            self.attempt, self.speed, self.technique = 0, 0, 0;
        end

        -- Prepare the things that do not vary in the sequence.
        if self.score then
            self.resultText:SetText(L:FormatLoc("Defeated", name, attempts));
            if clearTime then
                self.timeText:SetText(L:FormatLoc("FightTime", Utility:FormatCountdown("%M'%S''%C", clearTime, false, false, false)));
                if previousTime then
                    self.recordText:SetText(L:FormatLoc("TimeRecord", Utility:FormatCountdown("%M'%S''%C", previousTime, false, false, false)));
              else
                    self.recordText:SetText("");
                end
          else
                self.timeText:SetText(L:Get("IncompleteFight"));
                self.recordText:SetText("");
            end
      else
            self.resultText:SetText("");
            self.timeText:SetText("");
            self.recordText:SetText("");
        end

        -- Start music and pick timing data.
        if sequence == "FULL" then
            local music, timings;
            if self.rank then
                music = rankToMusic[self.rank];
                Audio:PlayMusic(music);
                timings = musicTimings[music];
          else
                timings = musicTimings["VICTORY_GOOD"]; -- Pick one as default.
            end

            -- Change the timings of the score frame according to the music.
            self.scoreWaitTimer = timings.waitTimer;
            self.scoreShowTimer = timings.showTimer;
            self.scorePrePause  =  timings.prePause;
            self.scoreGrowTimer = timings.growTimer;
            self.scorePostPause = timings.postPause;
      else
            Audio:PlayMusicOnce("JINGLE");

            -- Set fixed timings
            self.scoreWaitTimer = 0.5;
            self.scoreShowTimer = 1.5;
            self.scorePrePause  = 0.0;
            self.scoreGrowTimer = 0.0;
            self.scorePostPause = 4.0;
        end

        self:Show();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> atOnce: if set, the result sequence will be hidden at once.   *
-- ********************************************************************
-- * Stops playing the result sequence.                               *
-- ********************************************************************
function ResultAnim:Stop(atOnce)
    if self.status ~= "STANDBY" and self.status ~= "CLOSING" then
        Audio:StopMusic();
    end
    if atOnce then
        self.status = "STANDBY";
        self:GetScoreAnim():Remove(0);
        self:GetDamageGraph():Remove(0);
        self:GetSummaryAnim():Remove(0);
        self:Hide();

elseif self.status ~= "STANDBY" and self.status ~= "CLOSING" then
        self.status = "CLOSING";
        self.timer = FADE_DURATION;
        self:GetScoreAnim():Remove(FADE_DURATION);
        self:GetDamageGraph():Remove(FADE_DURATION);
        self:GetSummaryAnim():Remove(FADE_DURATION);
    end
end

-- ********************************************************************
-- * Get the current internal status of the result anim.              *
-- ********************************************************************
function ResultAnim:GetStatus()
    return self.status;
end

-- ********************************************************************
-- * Get the score anim attached to the given result anim.            *
-- ********************************************************************
function ResultAnim:GetScoreAnim()
    return self.ScoreAnim;
end

-- ********************************************************************
-- * Get the damage graph attached to the given result anim.          *
-- ********************************************************************
function ResultAnim:GetDamageGraph()
    return self.DamageGraph;
end

-- ********************************************************************
-- * Get the summary anim attached to the given result anim.          *
-- ********************************************************************
function ResultAnim:GetSummaryAnim()
    return self.SummaryAnim;
end

-- ********************************************************************
-- * Set the appropriate positions for components of the result anim. *
-- ********************************************************************
function ResultAnim:SetComponentPositions()
    if self.sequence == "FULL" then
        if self.dps and self.score then
            self.ScoreAnim:SetDefaultPosition("CENTER", "CENTER", -256, 0);
            self.DamageGraph:SetDefaultPosition("CENTER", "CENTER", 256, 0);

    elseif self.dps then
            self.DamageGraph:SetDefaultPosition("CENTER", "CENTER", 0, 0);

    elseif self.score then
            self.ScoreAnim:SetDefaultPosition("CENTER", "CENTER", 0, 0);
        end
  else
        self.SummaryAnim:SetPoint("CENTER", self, "CENTER", 0, 32);
    end
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function ResultAnim:OnLoad()
    -- ** Children **
    self.mask = self:CreateTexture(nil, "BACKGROUND");
    self.mask:SetTexture(0, 0, 0, 1);
    self.mask:SetAllPoints(self);
    self.resultText = GUI:RequestChild("MonoFont", nil, self, 1);
    self.resultText:SetScale(TEXT_SCALE);
    self.timeText = GUI:RequestChild("MonoFont", nil, self, 1);
    self.timeText:SetScale(TEXT_SCALE);
    self.recordText = GUI:RequestChild("MonoFont", nil, self, 1);
    self.recordText:SetScale(TEXT_SCALE);
    self.ScoreAnim = GUI:RequestChild("ScoreAnim", nil, self, 1);
    self.DamageGraph = GUI:RequestChild("Graph", nil, self, 1);
    self.SummaryAnim = GUI:RequestChild("SummaryAnim", nil, self, 1);

    -- ** Properties **
    self.status = "STANDBY";
    self.timer = 0.00;
    self.previousScore = nil;
    self.sequence = nil;
    self.dps = nil;
    self.score = nil;
    self.rank = nil;
    self.newRecord = false;
    self.attempt, self.speed, self.technique = 0, 0, 0;
    self.scoreWaitTimer = 0;
    self.scoreShowTimer = 0;
    self.scorePrePause  = 0;
    self.scoreGrowTimer = 0;
    self.scorePostPause = 0;

    -- ** Initial config **
end

function ResultAnim:OnUpdate(elapsed)
    -- Status transition

    local progression = 0;
    local maskAlpha = MASK_FINAL_ALPHA;
    local ofsX = 0;
    local alpha = 1;

    self.timer = max(0, self.timer - elapsed);

    if self.status == "STANDBY" then
        self:Hide();

elseif self.status == "OPENING" then
        progression = 1 - self.timer / FADE_DURATION;
        maskAlpha = progression * MASK_FINAL_ALPHA;
        alpha = 0;

        if self.timer == 0 then
            self.status = "SCROLLING";
            self.timer = SCROLL_DURATION;
        end

elseif self.status == "SCROLLING" then
        progression = 1 - self.timer / SCROLL_DURATION;
        ofsX = 0.75 * (1 - progression);

        if self.timer == 0 then
            self.status = "SCORE_WAITING";
            self.timer = self.scoreWaitTimer;
        end

elseif self.status == "SCORE_WAITING" then
        if self.timer == 0 then
            self.status = "SCORE_SHOWING";

            self:SetComponentPositions();

            if self.sequence == "FULL" then
                if self.dps then
                    self:GetDamageGraph():Open(self.scoreShowTimer);
                end
                if self.score then
                    self:GetScoreAnim():Open(self.scoreShowTimer, self.previousScore);
                end
          else
                self:GetSummaryAnim():Play(self.scoreShowTimer, self.score, self.dps, self.newRecord);
            end
        end

elseif self.status == "SCORE_SHOWING" then
        if (not self:GetScoreAnim():Scrolling()) and (not self:GetDamageGraph():Scrolling()) and self:GetSummaryAnim():GetStatus() ~= "SHOW" then
            self.status = "SCORE_PREPAUSE";
            self.timer = self.scorePrePause;
        end

elseif self.status == "SCORE_PREPAUSE" then
        if self.timer == 0 then
            if self.sequence == "FULL" then
                if self.score then
                    self.status = "SCORE_RUNNING";
                    self:GetScoreAnim():Play(self.attempt, self.speed, self.technique, self.scoreGrowTimer, self.score > 300, self.newRecord);
              else
                    self.status = "SCORE_POSTPAUSE";
                    self.timer = self.scoreGrowTimer + self.scorePostPause;
                end
                if self.dps then
                    self:GetDamageGraph():AutosetColumns(self.dps, self.scoreGrowTimer);
                end
          else
                self.status = "SCORE_POSTPAUSE";
                self.timer = self.scoreGrowTimer + self.scorePostPause;
            end
        end

elseif self.status == "SCORE_RUNNING" then
        if self:GetScoreAnim():GetStatus() == "STANDBY" then
            self.status = "SCORE_POSTPAUSE";
            self.timer = self.scorePostPause;
        end

elseif self.status == "SCORE_POSTPAUSE" then
        if self.timer == 0 then
            self:Stop(false);
        end

elseif self.status == "CLOSING" then
        progression = 1 - self.timer / FADE_DURATION;
        maskAlpha = (1 - progression) * MASK_FINAL_ALPHA;
        alpha = 1 - progression;

        if self.timer == 0 then
            self:Stop(true);
        end
    end

    self.mask:SetAlpha(maskAlpha);

    self.resultText:SetAlpha(alpha);
    self.resultText:ClearAllPoints();
    self.resultText:SetPoint("CENTER", self, "CENTER", ofsX * self:GetWidth() / TEXT_SCALE, (RESULT_TEXT_YPOSITION - 0.5) * self:GetHeight() / TEXT_SCALE);
    self.timeText:SetAlpha(alpha);
    self.timeText:ClearAllPoints();
    self.timeText:SetPoint("CENTER", self, "CENTER", -ofsX * self:GetWidth() / TEXT_SCALE, (TIME_TEXT_YPOSITION - 0.5) * self:GetHeight() / TEXT_SCALE);
    self.recordText:SetAlpha(alpha);
    self.recordText:ClearAllPoints();
    self.recordText:SetPoint("CENTER", self, "CENTER", -ofsX * self:GetWidth() / TEXT_SCALE, (RECORD_TEXT_YPOSITION - 0.5) * self:GetHeight() / TEXT_SCALE);
end

function ResultAnim:OnInterruptRequest()
    self:Stop();
end