local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local L = Root:GetOrNewModule("Localization");
local Audio = Root:GetOrNewModule("Audio");
local Utility = Root:GetOrNewModule("Utility");
local Resources = Root:GetOrNewModule("Resources");

local ScoreAnim = GUI:GetOrNewWidget("ScoreAnim");

ScoreAnim.TEMPLATE = "BossEncounter3_ScoreAnimTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- A frame displaying evaluation of the encounter.

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local criterias = {
    [1] = {
        key = "attempt",
        color = {0.00, 0.00, 1.00},
    },
    [2] = {
        key = "speed",
        color = {1.00, 0.00, 0.00},
    },
    [3] = {
        key = "technique",
        color = {0.00, 1.00, 0.00},
    },
}

local LABEL_SCALE = 0.5;

local NICE_SCORE_TIMER  = 2.00;
local NICE_SCORE_CYCLES = 4;
local NEW_RECORD_TIMER  = 1.50;
local NEW_RECORD_CYCLES = 3;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: time for the scroll in.                                 *
-- * >> record: the previous record. Nil if there is none.            *
-- ********************************************************************
-- * Make the score anim appear in its initial state.                 *
-- ********************************************************************
function ScoreAnim:Open(time, record)
    self.status = "STANDBY";
    self.timer = 0.00;

    -- Reset bars

    local i;
    for i=1, #self.criterias do
        self.criterias[i]:Reset();
    end

    self.total:Reset();

    self.record:Reset();
    if record then
        self.record:SetValue(record);
    end

    self:ScrollIn(time);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> attempt, speed, technique: the value of each criteria (0~120) *
-- * >> growTimer: the time needed by the score anim to grow          *
-- *               the criteria bars to their final value.            *
-- * >> niceScore: if true, nice score message will be displayed.     *
-- * >> newRecord: if true, new record message will be displayed.     *
-- ********************************************************************
-- * Execute the score anim.                                          *
-- ********************************************************************
function ScoreAnim:Play(attempt, speed, technique, growTimer, niceScore, newRecord)
    if self.status == "STANDBY" then
        self.status = "GROWING";
        self.timer = growTimer;

        self.growTimer = growTimer;
        self.niceScore = niceScore;
        self.newRecord = newRecord;

        self.attempt = attempt;
        self.speed = speed;
        self.technique = technique;
        self.score = attempt + speed + technique;

        local i;
        for i=1, #self.criterias do
            self.criterias[i]:Grow(self[criterias[i].key], growTimer);
        end
        self.total:Grow(self.score, growTimer);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: time needed to be scrolled out (0 = at once).           *
-- ********************************************************************
-- * Stop displaying a score anim.                                    *
-- ********************************************************************
function ScoreAnim:Remove(time)
    if time == 0 or (not self:IsShown()) then
        self.status = "STANDBY";
        self:Hide();

 elseif not self:ScrollingOut() then
        self:ScrollOut(time);
    end
end

-- ********************************************************************
-- * Get the current internal status of the score anim.               *
-- * Can be either STANDBY, GROWING, NICESCORE or NEWRECORD.          *
-- ********************************************************************
function ScoreAnim:GetStatus()
    return self.status;
end

-- ********************************************************************
-- * Play the nice score jingle.                                      *
-- ********************************************************************
function ScoreAnim:PlayNiceScore()
    self.status = "NICESCORE";
    self.timer = NICE_SCORE_TIMER;
    Audio:PlaySound("GREATPERFORMANCE");

    self.total:ShowComment(L:Get("Score-Nice"), NICE_SCORE_TIMER, NICE_SCORE_CYCLES);
end

-- ********************************************************************
-- * Play the new record jingle.                                      *
-- ********************************************************************
function ScoreAnim:PlayNewRecord()
    self.status = "NEWRECORD";
    self.timer = NEW_RECORD_TIMER;
    Audio:PlaySound("NEWRECORD");

    self.record:Grow(self.score, NEW_RECORD_TIMER);
    self.record:ShowComment(L:Get("Score-New"), NEW_RECORD_TIMER, NEW_RECORD_CYCLES);
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function ScoreAnim:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetPoint("TOP", self, "TOP", -8, -8);
    self.title:SetText(L:Get("Score-Title"));
    self.closeButton = CreateFrame("Button", nil, self, "UIPanelCloseButton");
    self.closeButton:SetScript("OnClick", self.OnCloseButtonClick);
    self.closeButton:SetPoint("TOPRIGHT", self, "TOPRIGHT", -2, -1);
    self.total = GUI:RequestChild("ScoreTotal", nil, self, 1);
    self.total:Setup(L:Get("Score-Total"), 1, 1, 0);
    self.total:SetPoint("BOTTOM", self, "BOTTOM", 0, 40);
    self.record = GUI:RequestChild("ScoreTotal", nil, self, 1);
    self.record:Setup(L:Get("Score-Record"), 1, 0.5, 0);
    self.record:SetPoint("BOTTOM", self, "BOTTOM", 0, 8);

    self.criterias = { };
    local i, criteria, key;
    for i=1, #criterias do
        criteria = criterias[i];
        key = criteria.key;
        self.criterias[i] = GUI:RequestChild("ScoreCriteria", nil, self, 1);
        self.criterias[i]:SetCriteria(L:Get("CriteriaShort:"..key), L:Get("CriteriaFull:"..key), L:Get("CriteriaHelp:"..key), unpack(criteria.color));
        self.criterias[i]:SetPoint("LEFT", self, "TOPLEFT", 8, -16 - i*40);
    end

    -- ** Properties **
    self.status = "STANDBY";
    self.timer = 0.00;

    -- ** Initial config **
    Utility:SetBackdrop(self, "OPTION");
end

function ScoreAnim:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);

    self.timer = max(0, self.timer - elapsed);

    if self.status == "GROWING" then
        if self.timer == 0 then
            if self.niceScore then
                self:PlayNiceScore();
        elseif self.newRecord then
                self:PlayNewRecord();
          else
                self.status = "STANDBY";
            end
        end

elseif self.status == "NICESCORE" then
        if self.timer == 0 then
            if self.newRecord then
                self:PlayNewRecord();
          else
                self.status = "STANDBY";
            end
        end

elseif self.status == "NEWRECORD" then
        if self.timer == 0 then
            self.status = "STANDBY";
        end
    end
end

function ScoreAnim:OnCloseButtonClick()
    self:GetParent():Notify("OnInterruptRequest");
end

function ScoreAnim:OnScrollOutEnd()
    self:Remove(0);
end