local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local L = Root:GetOrNewModule("Localization");

local SummaryAnim = GUI:GetOrNewWidget("SummaryAnim");

SummaryAnim.TEMPLATE = "BossEncounter3_SummaryAnimTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Displays a quick evaluation of the encounter.

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local ELEMENT_ANIM_DURATION = 0.25; -- Normalised value between 0 and 1.

local animOrder = {
    "totalDps",
    "myDps",
    "score",
    "newRecord",
};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: total time amount available to appear.                  *
-- * >> score: the score amount (can be nil).                         *
-- * >> dps: DPS data of the combat (can be nil).                     *
-- * >> newRecord: if true, new record message will be displayed.     *
-- ********************************************************************
-- * Execute the summary anim.                                        *
-- ********************************************************************
function SummaryAnim:Play(time, score, dps, newRecord)
    if self.status == "STANDBY" then
        self.status = "SHOW";
        self.timer = time;
        self.timerMax = max(0.01, time);

        -- Set dps text
        local myDps = nil;
        if dps then
            local me = UnitGUID("player");
            local i;
            for i=1, #dps do
                if dps[i].guid == me then
                    myDps = dps[i].dps;
                end
            end
            self.totalDps:SetText(L:FormatLoc("Summary-TotalDps", dps.total));
      else
            self.totalDps:SetText("");
        end
        if myDps then
            self.myDps:SetText(L:FormatLoc("Summary-MyDps", myDps));
      else
            self.myDps:SetText("");
        end

        -- Set score text
        if score then
            self.score:SetText(L:FormatLoc("Summary-Score", score));
      else
            self.score:SetText("");
        end
        if newRecord then
            self.newRecord:SetText(L:Get("Summary-NewRecord"));
      else
            self.newRecord:SetText("");
        end

        self:Show();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: time needed to disappear.                               *
-- ********************************************************************
-- * Stop displaying a summary anim.                                  *
-- ********************************************************************
function SummaryAnim:Remove(time)
    if time == 0 or (not self:IsShown()) then
        self.status = "STANDBY";
        self:Hide();

 elseif self.status ~= "FADE" then
        self.status = "FADE";
        self.timer = time;
        self.timerMax = time;
    end
end

-- ********************************************************************
-- * Get the current internal status of the summary anim.             *
-- * Can be either STANDBY, SHOW, WAIT, FADE.                         *
-- ********************************************************************
function SummaryAnim:GetStatus()
    return self.status;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function SummaryAnim:OnLoad()
    -- ** Children **
    self.totalDps = GUI:RequestChild("MonoFont", nil, self, 1);
    self.totalDps:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0);
    self.myDps = GUI:RequestChild("MonoFont", nil, self, 1);
    self.myDps:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -40);
    self.score = GUI:RequestChild("MonoFont", nil, self, 1);
    self.score:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 40);
    self.newRecord = GUI:RequestChild("MonoFont", nil, self, 1);
    self.newRecord:SetPoint("BOTTOM", self, "BOTTOM", 0, 0);

    -- ** Properties **
    self.status = "STANDBY";
    self.timer = 0;
    self.timerMax = 0.01;

    -- ** Initial config **
end

function SummaryAnim:OnUpdate(elapsed)
    local alpha = 1;
    local progression = 1;

    self.timer = max(0, self.timer - elapsed);

    if self.status == "SHOW" then
        if self.timer == 0 then
            self.status = "WAIT";
        end
        progression = 1 - self.timer / self.timerMax;

elseif self.status == "FADE" then
        if self.timer == 0 then
            self.status = "STANDBY";
            self:Hide();
        end
        alpha = self.timer / self.timerMax;
    end

    local elementProgression;
    local beginAnim, endAnim;
    local i, element;
    for i=1, #animOrder do
        -- Determinate the progression, which is a bounded value between 0.00 and 1.00.
        beginAnim = (i-1) * ((1.0 - ELEMENT_ANIM_DURATION) / (#animOrder - 1));
        endAnim = beginAnim + ELEMENT_ANIM_DURATION;
        elementProgression = (progression - beginAnim) / (endAnim - beginAnim);
        elementProgression = max(0, min(elementProgression, 1));

        -- Apply the progression on the element
        element = self[animOrder[i]];
        element:SetAlpha(elementProgression);
    end

    self:SetAlpha(alpha);
end
