local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");

local BossBar = GUI:GetOrNewWidget("BossBar");

BossBar.TYPE = "Button";
BossBar.TEMPLATE = "BossEncounter3_BossBarTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Boss bar is a bar specialized to display big boss bars.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local MAX_TOKENS = 9;

local SWEEP_TIMER = 0.75;

local DEFAULT_HOLD = 2.50;
local VARIATION_SPEED = 0.20;
local VARIATION_HISPEED = 0.50;

local TEXT_SCALE = 0.75;

local BORDER_LEVEL = 1;
local INSIDE_LEVEL = 2;
local FILL_LEVEL = 3;
local FILLSPECIAL_LEVEL = 4;
local FILLVARIATION_LEVEL = 5;
local SWEEP_LEVEL = 6;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function DoSweep(self)
    self.sweepRequest = false;
    self.sweepTimer = SWEEP_TIMER;
end

local function SyncCurrentValue(self)
    self.currentValue = self.value;
    self.variationHold = 0;
    self.variationSync = true;
end

local function UpdateTokens(self, instantly)
    local numTokens = self.numTokens;
    local currentToken = self.currentToken;
    local i, token;
    for i=1, MAX_TOKENS do
        token = self.tokens[i];
        if i < currentToken and i <= numTokens then
            token:Enable();
            token:SetSpecial(self.special >= (i / numTokens));
      else
            if not instantly then
                token:HideUp();
          else
                token:Hide();
            end
        end
    end
end

local function GetCurrentToken(self, value)
    local numTokens = self.numTokens;
    local lifePerToken = self.max / numTokens;
    local currentToken = min(1 + math.floor(value / lifePerToken), numTokens);
    return currentToken;
end

local function CheckTokenChange(self, value)
    local oldToken = self.currentToken;
    local currentToken = GetCurrentToken(self, value);
    if currentToken ~= oldToken then
        self.currentToken = currentToken;
        self:Sweep();
        UpdateTokens(self, false);
        if currentToken > oldToken then
            SyncCurrentValue(self);
        end
    end
end

local function UpdateCurrentValue(self, elapsed)
    local currentValue = self.currentValue;
    if currentValue then
        local lifePerToken = self.max / self.numTokens;
        local value = self.value;
        local delta = GetCurrentToken(self, value) - self.currentToken;
        local speed;
        if delta == 0 then
            speed = VARIATION_SPEED;
    elseif abs(delta) < 5 then
            speed = VARIATION_HISPEED;
      else
            speed = math.huge;
        end
        if self.variationHold > 0 then
           self.variationHold = max(0, self.variationHold - elapsed);
      else
            if currentValue == value then
                self.variationSync = true;
          else
                if currentValue > value then
                    currentValue = max(value, currentValue - elapsed * speed * lifePerToken);
              else
                    currentValue = min(value, currentValue + elapsed * speed * lifePerToken);
                end
            end
        end
        CheckTokenChange(self, currentValue);
        self.currentValue = currentValue;
    end
end

local function GetFill(self)
    local numTokens = self.numTokens;
    local lifePerToken = self.max / numTokens;
    local tokenMin = (self.currentToken - 1) * lifePerToken;
    local tokenMax = tokenMin + lifePerToken;
    local fill = (self.value - tokenMin) / lifePerToken;
    local currentFill = ((self.currentValue or 0) - tokenMin) / lifePerToken;
    local specialFill = (self.special * self.max - tokenMin) / lifePerToken;
    return max(0, min(1, fill)), max(0, min(1, currentFill)), max(0, min(1, specialFill));
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> state: whether to do the blinking or not.                     *
-- ********************************************************************
-- * Make the boss bar blink (or stop the blinking).                  *
-- ********************************************************************
function BossBar:Blink(state)
    self.blink = state;
end

-- ********************************************************************
-- * Play the sweep animation on the bar.                             *
-- ********************************************************************
function BossBar:Sweep()
    if self.sweepTimer <= 0 then
        self.sweepRequest = true;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> max: self-explanatory.                                        *
-- ********************************************************************
-- * Set the maximum of the boss bar.                                 *
-- ********************************************************************
function BossBar:SetMax(max)
    max = math.max(1, max);
    if self.max ~= max then
        local reference = math.max(1000, UnitHealthMax("player") * 10);
        local numTokens = math.max(0, min(MAX_TOKENS, math.floor(max / reference))) + 1;

        self.max = max;
        self.value = min(self.value, max);
        self.currentValue = nil;
        self.numTokens = numTokens;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the desired new value.                                 *
-- ********************************************************************
-- * Change the value of the boss bar.                                *
-- ********************************************************************
function BossBar:SetValue(value)
    -- Cap the value
    local valueMax = self.max;
    value = max(0, min(valueMax, value));

    -- Value update
    self.value = value;

    -- Variation update
    if not self.currentValue then
        SyncCurrentValue(self);

elseif value > self.currentValue then
        self.variationHold = 0;
  else
        if value < self.currentValue and self.variationSync then
            self.variationHold = DEFAULT_HOLD;
            self.variationSync = false;
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> special: the position of the special value (0 ~ 1).           *
-- ********************************************************************
-- * Change the position of the special value of the boss bar.        *
-- ********************************************************************
function BossBar:SetSpecial(special)
    self.special = max(0, min(1, special));

    UpdateTokens(self, false);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> text: the text to display.                                    *
-- ********************************************************************
-- * Set the text of the boss bar.                                    *
-- ********************************************************************
function BossBar:SetText(text)
    self.text:SetText(text);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function BossBar:OnLoad()
    local _, borderFilename, borderWidth, borderHeight = Resources:GetTextureInfo("BOSSBAR_BORDER");
    local _, insideFilename, insideWidth, insideHeight = Resources:GetTextureInfo("BOSSBAR_INSIDE");
    local _, sweepFilename, sweepWidth, sweepHeight = Resources:GetTextureInfo("SWEEP");
    local _, fillFilename = Resources:GetTextureInfo("BOSSBAR_FILL");
    local _, fillVariationFilename = Resources:GetTextureInfo("BOSSBAR_FILLVARIATION");
    local _, fillSpecialFilename = Resources:GetTextureInfo("BOSSBAR_FILLSPECIAL");

    local width = self:GetWidth();
    local height = self:GetHeight();

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "VERTICAL", true);

    -- ** Children **
    self.text = GUI:RequestChild("MonoFont", nil, self, 1);
    self.text:SetPoint("BOTTOM", self, "TOP", 0, 0);
    self.text:SetScale(TEXT_SCALE);
    self.border = self:CreateTexture();
    self.border:SetPoint("CENTER", self, "CENTER", 0, 0);
    Utility:ShowTextureRegion(self.border, borderFilename, borderWidth, borderHeight, 0, 0, width, height);
    self.inside = self:CreateTexture();
    self.inside:SetPoint("CENTER", self, "CENTER", 0, 0);
    Utility:ShowTextureRegion(self.inside, insideFilename, insideWidth, insideHeight, 0, 0, width, height);
    self.percentage = GUI:RequestChild("MonoFont", nil, self, 1);
    self.percentage:SetPoint("TOP", self.inside, "TOP", 0, -7);
    self.percentage:SetScale(TEXT_SCALE);
    self.sweep = self:CreateTexture();
    self.fill = self:CreateTexture();
    self.fill:SetTexture(fillFilename);
    self.fill:SetPoint("TOPLEFT", self, "TOPLEFT", 4, -4);
    self.fillVariation = self:CreateTexture();
    self.fillVariation:SetTexture(fillVariationFilename);
    self.fillSpecial = self:CreateTexture();
    self.fillSpecial:SetTexture(fillSpecialFilename);
    self.fillSpecial:SetPoint("TOPLEFT", self, "TOPLEFT", 4, -4);

    -- Draw levels
    self.border:SetDrawLayer("ARTWORK", BORDER_LEVEL);
    self.inside:SetDrawLayer("ARTWORK", INSIDE_LEVEL);
    self.sweep:SetDrawLayer("ARTWORK", SWEEP_LEVEL);
    self.fill:SetDrawLayer("ARTWORK", FILL_LEVEL);
    self.fillVariation:SetDrawLayer("ARTWORK", FILLVARIATION_LEVEL);
    self.fillSpecial:SetDrawLayer("ARTWORK", FILLSPECIAL_LEVEL);

    self.tokens = { };
    local i, token;
    for i=1, MAX_TOKENS do
        token = GUI:RequestChild("BossBarToken", nil, self);
        token:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 8 + 42 * (i - 1), 2);
        token:Hide();
        tinsert(self.tokens, token);
    end

    -- ** Properties **
    self.max = 0;
    self.value = 0;
    self.currentValue = 0;
    self.special = 0;
    self.variationHold = 0;
    self.variationSync = false;
    self.blink = false;
    self.numTokens = 1;
    self.currentToken = 1;
    self.sweepRequest = false;
    self.sweepTimer = -SWEEP_TIMER;
    self.sweepFilename = sweepFilename;
    self.sweepWidth = sweepWidth;
    self.sweepHeight = sweepHeight;
end

function BossBar:OnUpdate(elapsed)
    local value = self.value;
    local currentValue = self.currentValue;
    local fill, varFill, specialFill = GetFill(self);
    local fillableWidth = self:GetWidth() - 8;
    local fillableHeight = 16;

    -- Start sweep anim if not scrolling.
    if self.sweepRequest then
        if not self:Scrolling() and elapsed > 0 then
            DoSweep(self);
      else
            self.sweepTimer = 2*SWEEP_TIMER;
        end
    end

    -- Update sweep anim.
    self.sweepTimer = self.sweepTimer - elapsed;

    local sweepProgression = 1 - self.sweepTimer / SWEEP_TIMER;
    local sweepStart = fillableWidth * sweepProgression - self.sweepWidth;
    local sweepEnd = sweepStart + self.sweepWidth;
    local position = nil;

    if sweepStart < 0 then
        position = sweepEnd;
        Utility:ShowTextureRegion(self.sweep, self.sweepFilename, self.sweepWidth, self.sweepHeight, self.sweepWidth - sweepEnd, 0, sweepEnd, fillableHeight);

elseif sweepEnd > fillableWidth then
        local sweepWidth = fillableWidth - sweepStart;
        if sweepWidth > 0 then
            position = fillableWidth;
            Utility:ShowTextureRegion(self.sweep, self.sweepFilename, self.sweepWidth, self.sweepHeight, 0, 0, sweepWidth, fillableHeight);
        end

elseif sweepStart < fillableWidth then
        position = sweepProgression * fillableWidth;
        Utility:ShowTextureRegion(self.sweep, self.sweepFilename, self.sweepWidth, self.sweepHeight, 0, 0, self.sweepWidth, fillableHeight);
    end
    if position then
        self.sweep:SetPoint("TOPRIGHT", self, "TOPLEFT", 4 + position, -4);
        self.sweep:Show();
  else
        self.sweep:Hide();
    end

    -- Update current value.
    UpdateCurrentValue(self, elapsed);

    -- Update percentage.
    local percentage = self.value / self.max;
    self.percentage:SetText(math.floor(percentage * 100 + 0.5));

    -- Update fill and fill variation textures.
    varFill = min(varFill, sweepProgression);
    specialFill = min(specialFill, varFill);
    fill = min(fill, varFill);
    if self.blink then
        local interval = max(0.10, min(1.00, 0.05 + percentage * 4));
        if math.fmod(GetTime(), interval) < 0.05 then
            varFill, specialFill, fill = 0, 0, 0;
        end
    end
    if varFill > fill then
        self.fillVariation:SetWidth((varFill - fill) * fillableWidth);
        self.fillVariation:SetHeight(fillableHeight);
        self.fillVariation:SetPoint("TOPLEFT", self, "TOPLEFT", 4 + fillableWidth * fill, -4);
        self.fillVariation:Show();
  else
        self.fillVariation:Hide();
    end
    local pulse = 0.9 + 0.1 * cos(GetTime() * 360);
    if fill > 0 then
        self.fill:SetWidth(fill * fillableWidth);
        self.fill:SetHeight(fillableHeight);
        self.fill:SetVertexColor(pulse, pulse, pulse);
        self.fill:Show();
  else
        self.fill:Hide();
    end
    if specialFill > 0 then
        self.fillSpecial:SetWidth(specialFill * fillableWidth);
        self.fillSpecial:SetHeight(fillableHeight);
        self.fillSpecial:SetVertexColor(pulse, pulse, pulse);
        self.fillSpecial:Show();
  else
        self.fillSpecial:Hide();
    end

    -- Update scrolling.
    self:UpdateScroll(elapsed);
end

function BossBar:OnShow()
    self:Blink(false);
    self:Sweep();
    UpdateTokens(self, true);
    SyncCurrentValue(self);
    self:OnUpdate(0);
end

function BossBar:OnScrollOutEnd()
    self:Hide();
end