local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local CoordinateSystem = GUI:GetOrNewWidget("CoordinateSystem");

local System = Root:GetOrNewModule("System");
local Distance = System:GetOrNewModule("Distance");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Give the inheriting frame methods to manipulate map coordinates
-- and transform yards distances into map coordinates.
-- The origin is always the player's current position.

-- Adds:
-- ----------------
-- SetOrigin()
-- GetCoordinates()
-- GetUnitCoordinates()
-- GetDistance()
-- GetAngle()

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mx, my: map coordinates of the origin.                        *
-- ********************************************************************
-- * Set the origin to the specified map point.                       *
-- * Generally, the chosen origin doesn't matter.                     *
-- * The origin shouldn't be modified once set.                       *
-- ********************************************************************
local function SetOrigin(self, mx, my)
    self.originX, self.originY = mx, my;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mx, my: map coordinates to transform.                         *
-- ********************************************************************
-- * Transform map coordinates into game world yards coordinates.     *
-- ********************************************************************
local function GetCoordinates(self, mx, my)
    local meterScale = Distance:GetMeterScale();
    if meterScale > 0 then
        mx, my = Distance:NormalizePosition(mx - self.originX, 1 - my - self.originY);
        return mx * meterScale, my * meterScale;
    end
    return nil, nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: unit to get coordinates of.                             *
-- ********************************************************************
-- * Transform unit's coordinates into game world yards coordinates.  *
-- ********************************************************************
local function GetUnitCoordinates(self, unit)
    local mx, my = GetPlayerMapPosition(unit);
    if mx > 0 and my > 0 then
        return self:GetCoordinates(mx, my);
    end
    return nil, nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x1, y1: first point coordinates.                              *
-- * >> x2, y2: second point coordinates.                             *
-- ********************************************************************
-- * Get the distance in yards between 2 points.                      *
-- ********************************************************************
local function GetDistance(self, x1, y1, x2, y2)
    return Distance:GetDistance(x1, y1, x2, y2);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x1, y1: first point coordinates.                              *
-- * >> x2, y2: second point coordinates.                             *
-- ********************************************************************
-- * Get the angle (in degrees) between 2 points.                     *
-- ********************************************************************
local function GetAngle(self, x1, y1, x2, y2)
    return Distance:Atanfull(x1, y1, x2, y2);
end

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

function CoordinateSystem.Install(self)
    -- Install properties
    self.originX = 0;
    self.originY = 0;

    -- Install methods
    self.SetOrigin = SetOrigin;
    self.GetCoordinates = GetCoordinates;
    self.GetUnitCoordinates = GetUnitCoordinates;
    self.GetDistance = GetDistance;
    self.GetAngle = GetAngle;

    return true;
end