local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

local PositionSystem = GUI:GetOrNewWidget("PositionSystem");

local System = Root:GetOrNewModule("System");
local View = System:GetOrNewModule("View");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- This system uses the Save and Menu systems.

-- Give the inheriting frame methods to memorize position of the frame,
-- drag the frame, set the anchor of the frame and its scale.

-- Adds:
-- ------------
-- MayMove()
-- MayDrag()
-- ResetPosition()
-- SetDefaultPosition()
-- RegisterPosition()
-- SetAnchor()
-- LockPosition()
-- ChangeScale()
-- UpdatePosition()
-- GetPosition()
-- SetPosition()
-- GetRealPosition()
-- SuspendDrag()
-- ResumeDrag()

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local SETANCHOR_DROPDOWN_FUNC = function(_, self, anchor) self:SetAnchor(anchor); end
local RESETDEFAULTS_DROPDOWN_FUNC = function(_, self) self:ResetPosition(); end
local LOCK_DROPDOWN_FUNC = function(_, self, state) self:LockPosition(state); end
local SIZE_DROPDOWN_FUNC = function(dropDown, self) View:StartSizeChange(self, dropDown); end

local verticalAnchor = {
    ["TOPLEFT"] = "TOP",
    ["TOP"] = "TOP",
    ["TOPRIGHT"] = "TOP",
    ["LEFT"] = "CENTER",
    ["CENTER"] = "CENTER",
    ["RIGHT"] = "CENTER",
    ["BOTTOMLEFT"] = "BOTTOM",
    ["BOTTOM"] = "BOTTOM",
    ["BOTTOMRIGHT"] = "BOTTOM",
};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Determinate if we can move the frame.                            *
-- ********************************************************************
local function MayMove(self)
    return not self:IsLocked();
end

-- ********************************************************************
-- * Determinate if the user should be able to drag the frame.        *
-- ********************************************************************
local function MayDrag(self)
    return (self:MayMove() and self.dragAllowed and (not self.position.locked));
end

-- ********************************************************************
-- * Apply the default position to the frame.                         *
-- ********************************************************************
local function ResetPosition(self)
    if self:MayMove() then
        local position = self.position;
        local default = self.defaultPosition;

        if default then
            position.scale = 1;
            self:SetScale(1);
            self:ClearAllPoints();
            self:SetPoint(default.point, GUI:GetBase(), default.relativePoint, default.x, default.y);
            self:SetAnchor(default.anchor);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> ...: the point info of the default position.                  *
-- * The default position is relative to UIParent in general.         *
-- ********************************************************************
-- * Register the default position of the frame.                      *
-- * Immediately apply the default position if the frame has none.    *
-- * Should be called after creating the frame.                       *
-- ********************************************************************
local function SetDefaultPosition(self, point, relativePoint, ofsX, ofsY)
    local isHidden = (not self:IsShown());
    self:Show(); -- The frame has to be shown so we can get its position on screen and save it.

    self.defaultPosition = {
        point = point,
        relativePoint = relativePoint,
        x = ofsX,
        y = ofsY,
        anchor = verticalAnchor[point] or "CENTER",
    };

    -- Now check if the frame needs to have this default position applied immediately.
    local position = self.position;
    if (not position.x) or (not position.y) or (not self.draggable) then
        -- No position data or disallowed to move the widget -> use default position
        self:ResetPosition();
  else
        -- Load a previously saved position. Apply the scale and anchor.
        self:ChangeScale(position.scale);
        self:SetAnchor(position.anchor);
    end

    if isHidden then
        self:Hide();
    end
end

-- ********************************************************************
-- * Register the new position of the frame.                          *
-- ********************************************************************
local function RegisterPosition(self)
    local position = self.position;
    position.x, position.y = self:GetPosition();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> anchor: the new vertical anchor.                              *
-- * Can be "TOP", "CENTER" or "BOTTOM".                              *
-- ********************************************************************
-- * Change the vertical anchor of the frame.                         *
-- ********************************************************************
local function SetAnchor(self, anchor)
    if anchor then
        self.position.anchor = anchor;
        self:RegisterPosition();
        self:UpdatePosition();
    end
end

-- ********************************************************************
-- * Get the main anchor of the frame.                                *
-- ********************************************************************
local function GetAnchor(self)
    return self.position.anchor;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> state: whether the position is locked or not.                 *
-- ********************************************************************
-- * Set whether a frame can be moved by the user or not.             *
-- ********************************************************************
local function LockPosition(self, state)
    if state == nil then state = true; end
    self.position.locked = state;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new scale value.                                   *
-- ********************************************************************
-- * Change the scale of a frame.                                     *
-- ********************************************************************
local function ChangeScale(self, value)
    self.position.scale = value;

    -- We need to update the position if the scaling is changed.
    self:UpdatePosition();
end

-- ********************************************************************
-- * Set the frame at the correct position.                           *
-- ********************************************************************
local function UpdatePosition(self)
    if self:MayMove() then
        local x, y = self:GetRealPosition();
        self:SetPosition(x, y);
    end
end

-- ********************************************************************
-- * Get coordinates of the frame (percentage-based).                 *
-- ********************************************************************
local function GetPosition(self)
    local position = self.position;
    local anchor = position.anchor;
    local scale = position.scale or 1;
    local x, y = self:GetCenter();

    if anchor == "TOP" then
        y = self:GetTop();
elseif anchor == "CENTER" then
        -- Nothing special to do.
elseif anchor == "BOTTOM" then
        y = self:GetBottom();
    end

    return GUI:DotToPctX(x or 0) * scale, GUI:DotToPctY(y or 0) * scale;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x, y: percentage-based coordinates.                           *
-- * (0.5, 0.5) is exactly the center, (0, 0) is the bottomleft.      *
-- * This is only true if CENTER anchor is used.                      *
-- ********************************************************************
-- * Set the position of the frame, respecting the desired anchor     *
-- * and working with percentage-based coordinates.                   *
-- ********************************************************************
local function SetPosition(self, x, y)
    local position = self.position;
    local anchor = position.anchor;
    local scale = position.scale or 1;

    self:ClearAllPoints();
    self:SetScale(scale);
    self:SetPoint(anchor, GUI:GetBase(), "BOTTOMLEFT", GUI:PctToDotX(x) / scale, GUI:PctToDotY(y) / scale);
end

-- ********************************************************************
-- * Get real coordinates of the frame (percentage-based).            *
-- ********************************************************************
local function GetRealPosition(self)
    local position = self.position;
    return position.x or 0, position.y or 0;
end

-- ********************************************************************
-- * Modify a percentage-based point according to the main anchor.    *
-- * No change if the anchor is CENTER.                               *
-- ********************************************************************
local function AdjustToAnchor(self, x, y)
    local h = self:GetHeight();
    local position = self.position;
    local anchor, scale = position.anchor, position.scale;
    if anchor == "TOP" then
        y = y + GUI:DotToPctY(h / 2);
elseif anchor == "CENTER" then
        -- Nothing special to do.
elseif anchor == "BOTTOM" then
        y = y - GUI:DotToPctY(h / 2);
    end
    return x, y;
end

-- ********************************************************************
-- * Suspend temporarily dragging feature if a subsystem can only     *
-- * work safely when denying the user the dragging feature.          *
-- ********************************************************************
local function SuspendDrag(self)
    if self.dragAllowed and self.draggable then
        self.dragAllowed = false;

        self:StopMovingOrSizing();
        self:SetClampedToScreen(false);
    end
end

-- ********************************************************************
-- * Reallow the user to use the dragging feature.                    *
-- ********************************************************************
local function ResumeDrag(self)
    if (not self.dragAllowed) and self.draggable then
        self.dragAllowed = true;

        self:SetClampedToScreen(true);
    end
end

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

function PositionSystem.Install(self, isDraggable)
    -- Install dependencies
    if isDraggable then
        GUI:InstallSystem("SaveSystem", self);
        self.position = self:ReadValue("position") or { };
        self:WriteValue("position", self.position);
  else
        self.position = { };
    end

    -- Install properties
    self.dragAllowed = isDraggable;

    -- Install methods
    self.MayDrag = MayDrag;
    self.MayMove = MayMove;
    self.ResetPosition = ResetPosition;
    self.SetDefaultPosition = SetDefaultPosition;
    self.RegisterPosition = RegisterPosition;
    self.SetAnchor = SetAnchor;
    self.GetAnchor = GetAnchor;
    self.LockPosition = LockPosition;
    self.ChangeScale = ChangeScale;
    self.UpdatePosition = UpdatePosition;
    self.GetPosition = GetPosition;
    self.SetPosition = SetPosition;
    self.GetRealPosition = GetRealPosition;
    self.AdjustToAnchor = AdjustToAnchor;
    self.SuspendDrag = SuspendDrag;
    self.ResumeDrag = ResumeDrag;

    -- Set constants
    self.draggable = isDraggable;

    -- Handle dragging.
    if isDraggable then
        Utility:MakeDraggable(self, self.MayDrag, nil, PositionSystem.OnDragStop);
        self:SetClampedToScreen(true);

        -- Set up menu commands
        GUI:InstallSystem("MenuSystem", self);
        self:RegisterCommandSet(PositionSystem.MenuCommandSet);
    end

    return true;
end

function PositionSystem.OnDragStop(self)
    self:StopMovingOrSizing();
    self:RegisterPosition();
    self:UpdatePosition();
end

function PositionSystem.MenuCommandSet(frame)
    local info;

    if UIDROPDOWNMENU_MENU_LEVEL == 2 then
        info         = UIDropDownMenu_CreateInfo();
        info.text    = L:Get("Top");
        info.func    = SETANCHOR_DROPDOWN_FUNC;
        info.arg1    = frame;
        info.arg2    = "TOP";
        info.checked = (frame.position.anchor == "TOP");
        UIDropDownMenu_AddButton(info, 2);

        info         = UIDropDownMenu_CreateInfo();
        info.text    = L:Get("Center");
        info.func    = SETANCHOR_DROPDOWN_FUNC;
        info.arg1    = frame;
        info.arg2    = "CENTER";
        info.checked = (frame.position.anchor == "CENTER");
        UIDropDownMenu_AddButton(info, 2);

        info         = UIDropDownMenu_CreateInfo();
        info.text    = L:Get("Bottom");
        info.func    = SETANCHOR_DROPDOWN_FUNC;
        info.arg1    = frame;
        info.arg2    = "BOTTOM";
        info.checked = (frame.position.anchor == "BOTTOM");
        UIDropDownMenu_AddButton(info, 2);

elseif UIDROPDOWNMENU_MENU_LEVEL == 1 then
        info                  = UIDropDownMenu_CreateInfo();
        info.text             = L:Get("Anchor");
        info.keepShownOnClick = true;
        info.hasArrow         = true;
        info.notCheckable     = true;
        UIDropDownMenu_AddButton(info);

        info              = UIDropDownMenu_CreateInfo();
        info.text         = L:Get("DefaultPosition");
        info.func         = RESETDEFAULTS_DROPDOWN_FUNC;
        info.arg1         = frame;
        info.notCheckable = true;
        UIDropDownMenu_AddButton(info);

        info         = UIDropDownMenu_CreateInfo();
        info.text    = L:Get("Lock");
        info.func    = LOCK_DROPDOWN_FUNC;
        info.arg1    = frame;
        info.arg2    = (not frame.position.locked);
        info.checked = frame.position.locked;
        UIDropDownMenu_AddButton(info);

        info         = UIDropDownMenu_CreateInfo();
        info.text    = L:Get("Size");
        info.func    = SIZE_DROPDOWN_FUNC;
        info.arg1    = frame;
        info.notCheckable = true;
        UIDropDownMenu_AddButton(info);
    end
end