local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local ScrollSystem = GUI:GetOrNewWidget("ScrollSystem");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- This system uses the Position system.

-- Give the inheriting frame methods to scroll in and out.

-- Adds:
-- ------------
-- ScrollIn()
-- ScrollOut()
-- Scrolling()
-- ScrollingIn()
-- ScrollingOut()
-- UpdateScroll()         -- needs to be called on each OnUpdate.

-- Your frame can also contain OnScrollInEnd and OnScrollOutEnd handlers to do something once scroll in/out is finished.
-- OnScrollInEnd and OnScrollOutEnd are guaranteed to be run when combat restriction are inactive (in the case of secure frames).

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

local DEFAULT_SCROLLIN_TIMER  = 0.30;
local DEFAULT_SCROLLOUT_TIMER = 0.30;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- Internal

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: the time amount allowed to do the scrolling.            *
-- * >> startX, startY: the normalized starting coordinates.          *
-- * >> endX, endY: the normalized ending coordinates.                *
-- ********************************************************************
-- * Starts scrolling the frame out.                                  *
-- ********************************************************************
local function Scroll(self, time, startX, startY, endX, endY)
    self.scrollTimer = time;
    self.scrollTimerMax = time;
    self.sX, self.sY = startX, startY;
    self.eX, self.eY = endX, endY;
    self:SuspendDrag();
    self:UpdateScroll(0);
end

-- ********************************************************************
-- * Called when whatever scrolling is finished.                      *
-- ********************************************************************
local function ScrollEnd(self)
    self:ResumeDrag();
    if self.scrollingIn then
        self.scrollingIn = false;
        self:UpdatePosition();
        if self.OnScrollInEnd then
            self:OnScrollInEnd();
        end
    end
    if self.scrollingOut then
        self.scrollingOut = false;
        if self.OnScrollOutEnd then
            self:OnScrollOutEnd();
        end
    end
end

-- ********************************************************************
-- * Return true if scroll animation should be completely avoided.    *
-- * If scroll animation is disabled, it is replaced by a fade anim.  *
-- ********************************************************************
local function AvoidScroll(self)
    return self:IsProtected();
end

-- API

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: the time amount allowed to scroll-in.                   *
-- ********************************************************************
-- * Starts scrolling the frame in.                                   *
-- * Asking a frame to scroll in will automatically show it.          *
-- * Return true if scroll-in command has been sent successfully.     *
-- ********************************************************************
local function ScrollIn(self, time)
    if not self.scrollingIn and (not self:IsShown() or self.scrollingOut) then
        if not self:IsShown() then
            GUI:Show(self);
        end
        local base = GUI:GetBase();
        local rX, rY = self:GetRealPosition();
        local sX, sY;
        if self.scrollStyle == "VERTICAL" then
            sX = rX;
            if rY < 0.5 then
                sY = GUI:DotToPctY(-self:GetHeight()/2);
          else
                sY = GUI:DotToPctY(base:GetHeight() + self:GetHeight()/2);
            end
            _, sY = self:AdjustToAnchor(sX, sY);

    elseif self.scrollStyle == "HORIZONTAL" then
            sY = rY;
            if rX < 0.5 then
                sX = GUI:DotToPctX(-self:GetWidth()/2);
          else
                sX = GUI:DotToPctX(base:GetWidth() + self:GetWidth()/2);
            end
            sX, _ = self:AdjustToAnchor(sX, sY);
        end
        self.scrollingIn = true;
        self.scrollingOut = false;
        Scroll(self, time or DEFAULT_SCROLLIN_TIMER, sX, sY, rX, rY);
        return true;
  else
        return false;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> time: the time amount allowed to scroll-out.                  *
-- ********************************************************************
-- * Starts scrolling the frame out.                                  *
-- * Return true if scroll-out command has been sent successfully.    *
-- ********************************************************************
local function ScrollOut(self, time)
    if not self.scrollingOut and (self:IsShown() or self.scrollingIn) then
        local base = GUI:GetBase();
        local rX, rY = self:GetRealPosition();
        local eX, eY;
        if self.scrollStyle == "VERTICAL" then
            eX = rX;
            if rY < 0.5 then
                eY = GUI:DotToPctY(-self:GetHeight()/2);
          else
                eY = GUI:DotToPctY(base:GetHeight() + self:GetHeight()/2);
            end
            _, eY = self:AdjustToAnchor(eX, eY);

    elseif self.scrollStyle == "HORIZONTAL" then
            eY = rY;
            if rX < 0.5 then
                eX = GUI:DotToPctX(-self:GetWidth()/2);
          else
                eX = GUI:DotToPctX(base:GetWidth() + self:GetWidth()/2);
            end
            eX, _ = self:AdjustToAnchor(eX, eY);
        end
        self.scrollingOut = true;
        self.scrollingIn = false;
        Scroll(self, time or DEFAULT_SCROLLOUT_TIMER, rX, rY, eX, eY);
        return true;
  else
        return false;
    end
end

-- ********************************************************************
-- * Check if the frame is scrolling or not.                          *
-- ********************************************************************
local function Scrolling(self)
    return self:ScrollingIn() or self:ScrollingOut();
end

-- ********************************************************************
-- * Check if the frame is scrolling in or not.                       *
-- ********************************************************************
local function ScrollingIn(self)
    return self.scrollingIn and self:IsShown();
end

-- ********************************************************************
-- * Check if the frame is scrolling out or not.                      *
-- ********************************************************************
local function ScrollingOut(self)
    return self.scrollingOut and self:IsShown();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> elapsed: the time elapsed since the last update.              *
-- ********************************************************************
-- * Update the inherited scroll system.                              *
-- ********************************************************************
local function UpdateScroll(self, elapsed)
    if self:Scrolling() then
        if not AvoidScroll(self) then
            if self:MayMove() and elapsed > 0 then
                self.scrollTimer = max(0, self.scrollTimer - elapsed);

                local progression = 1 - self.scrollTimer / self.scrollTimerMax;
                local curve;

                if self.scrollCurved then
                    if self.scrollingIn then
                        curve = 1 - exp(-5 * progression);
                        if progression > 0.80 then
                            curve = curve + (1 - curve) * (progression - 0.80) / 0.20;
                        end
                  else
                        curve = exp(-5 * (1-progression));
                        if progression < 0.20 then
                            curve = curve * progression / 0.20;
                        end
                    end
              else
                    curve = progression;
                end

                local pX = self.sX + (self.eX - self.sX) * curve;
                local pY = self.sY + (self.eY - self.sY) * curve;

                self:SetPosition(pX, pY);

                if self.scrollTimer == 0 then
                    ScrollEnd(self);
                end
            end
      else
            self.scrollTimer = max(0, self.scrollTimer - elapsed);

            local progression = 1 - self.scrollTimer / self.scrollTimerMax;
            local alpha = 1;

            if self.scrollingIn then
                alpha = progression;
          else
                alpha = 1 - progression;
            end

            if self:MayMove() then
                if self.scrollTimer == 0 then
                    ScrollEnd(self);
                end
          else
                alpha = max(0.2, alpha);
            end

            self:SetAlpha(alpha);
        end
    end
end

-- --------------------------------------------------------------------
-- **                              API                               **
-- --------------------------------------------------------------------

function ScrollSystem.Install(self, style, curved)
    -- Install dependencies
    GUI:InstallSystem("PositionSystem", self);

    -- Install properties
    self.scrollingIn = false;
    self.scrollingOut = false;
    self.scrollTimer = 0;
    self.scrollTimerMax = 0;
    self.sX, self.sY = 0, 0;
    self.eX, self.eY = 0, 0;

    -- Install methods
    self.ScrollIn = ScrollIn;
    self.ScrollOut = ScrollOut;
    self.Scrolling = Scrolling;
    self.ScrollingIn = ScrollingIn;
    self.ScrollingOut = ScrollingOut;
    self.UpdateScroll = UpdateScroll;

    -- Set constants
    self.scrollStyle = style or "HORIZONTAL";
    self.scrollCurved = curved or false;

    return true;
end
