local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");

local EnemyList = GUI:GetOrNewWidget("EnemyList");

EnemyList.TYPE = "Button";
EnemyList.TEMPLATE = "BossEncounter3_EnemyListTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- EnemyList is a final GUI element displaying a list of enemies.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> bossBar: the BossBar widget used by the EnemyList.            *
-- ********************************************************************
-- * Connect a boss bar to the enemy list.                            *
-- ********************************************************************
function EnemyList:ConnectBossBar(bossBar)
    self.BossBar = bossBar;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- * >> selectable: specify the unit should be selectable.            *
-- * >> class: class of the unit (not required).                      *
-- * >> role: role of the unit (not required).                        *
-- ********************************************************************
-- * Add a new unit in the enemy list.                                *
-- * If you provide a GUID, the row will display ???s until the exact *
-- * GUID of the unit is provided.                                    *
-- ********************************************************************
function EnemyList:AddUnit(id, selectable, class, role)
    if self:GetEnemyRow(id) ~= nil then
        Root:Debug("Attempted to add duplicate unit in EnemyList: "..id);
  else
        local E = self:GetFreeEnemyRow();
        E:Watch(id, selectable, class, role);
        E:SetExtraAlgorithm(self.algorithm);
        E:SetMainUnit(id == self.mainUnit);
        E:SetThresholdDelay(self.thresholdDelay);
        self.list:Insert(E);
        self:Update();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- * >> guid: the new GUID.                                           *
-- ********************************************************************
-- * Modify the GUID of a monitored unit.                             *
-- ********************************************************************
function EnemyList:SetUnit(id, guid)
    local E = self:GetEnemyRow(id);
    if E then
        E:Set(guid);
        self:Update();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- ********************************************************************
-- * Remove an unit from the enemy list.                              *
-- ********************************************************************
function EnemyList:RemoveUnit(id)
    local E = self:GetEnemyRow(id);
    if E then
        E:Stop();
        self.list:Remove(E);
        self:Update();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- * >> label: the label of the threshold.                            *
-- * >> threshold: the health ratio where the threshold is.           *
-- ********************************************************************
-- * Add an health threshold indicator to an enemy row.               *
-- ********************************************************************
function EnemyList:AddThreshold(id, label, threshold)
    local E = self:GetEnemyRow(id);
    if E then
        E:AddThreshold(label, threshold);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- * >> label: the label of the threshold.                            *
-- ********************************************************************
-- * Remove an health threshold indicator.                            *
-- ********************************************************************
function EnemyList:RemoveThreshold(id, label)
    local E = self:GetEnemyRow(id);
    if E then
        E:RemoveThreshold(label);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> algorithm: extra algorithm to use.                            *
-- * Either nil, a name, or a custom algorithm table.                 *
-- ********************************************************************
-- * Set the extra algorithm of the enemy list.                       *
-- ********************************************************************
function EnemyList:SetExtraAlgorithm(algorithm)
    self.algorithm = algorithm;

    local i, E;
    for i=1, #self.units do
        E = self.units[i];
        if E:IsUsed() then
            E:SetExtraAlgorithm(algorithm);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit. Nil to clear.                *
-- ********************************************************************
-- * Set the main unit of the unit list.                              *
-- * It will be the unit displayed in the boss health bar if there's  *
-- * one attached to this widget.                                     *
-- ********************************************************************
function EnemyList:SetMainUnit(id)
    self.mainUnit = id;

    local i, E;
    for i=1, #self.units do
        E = self.units[i];
        if E:IsUsed() then
            E:SetMainUnit( id and E:IsUnit(id) );
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> healthDelay: how far ahead the threshold is displayed.        *
-- ********************************************************************
-- * Set the delay to display an health threshold on enemy rows.      *
-- * If this is set to nil, the default delay is used (10%).          *
-- ********************************************************************
function EnemyList:SetThresholdDelay(healthDelay)
    self.thresholdDelay = healthDelay;

    local i, E;
    for i=1, #self.units do
        E = self.units[i];
        if E:IsUsed() then
            E:SetThresholdDelay(healthDelay);
        end
    end
end

-- ********************************************************************
-- * Stop all enemy rows from the enemy list.                         *
-- ********************************************************************
function EnemyList:Clear()
    local i;
    for i=1, #self.units do
        self.units[i]:Stop();
    end
    self.list:Clear();
    self:Update();
end

-- ********************************************************************
-- * Update the enemy list: check if we should make it appear,        *
-- * and if we should make it disappear.                              *
-- ********************************************************************
function EnemyList:Update()
    self:UpdateSize();

    if self.list:IsEmpty() then
        self:ScrollOut();
  else
        self:ScrollIn();
    end
end

-- ********************************************************************
-- * Get a free enemyRow widget (or create a new one).                *
-- ********************************************************************
function EnemyList:GetFreeEnemyRow()
    local i, E;
    for i=1, #self.units do
        E = self.units[i];
        if not E:IsUsed() then -- This row is available.
            return E;
        end
    end
    -- No existing row available, create a new one.
    E = GUI:RequestChild("EnemyRow", nil, self);
    tinsert(self.units, E);
    return E;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- ********************************************************************
-- * Search the row of specified event.                               *
-- ********************************************************************
function EnemyList:GetEnemyRow(id)
    local i, E;
    for i=1, #self.units do
        E = self.units[i];
        if E:IsUnit(id) then
            return E;
        end
    end
end

-- ********************************************************************
-- * Update the size of the event list.                               *
-- ********************************************************************
function EnemyList:UpdateSize()
    if self:IsLocked() then return; end

    self.list:Update();

    local w = max(self.title:GetWidth(), self.list:GetWidth());
    local h = self.title:GetHeight() + self.separator:GetHeight() + self.list:GetHeight();
    self:SetWidth(w);
    self:SetHeight(h);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function EnemyList:OnLoad()
    local _, titleFilename, titleWidth, titleHeight = Resources:GetTextureInfo("ENEMIES");
    local _, separatorFilename = Resources:GetTextureInfo("SEPARATOR");

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.title = self:CreateTexture(nil, "ARTWORK");
    self.title:SetPoint("TOP", self, "TOP", 0, 0);
    self.title:SetTexture(titleFilename);
    self.title:SetWidth(titleWidth);
    self.title:SetHeight(titleHeight);
    self.separator = self:CreateTexture(nil, "BACKGROUND");
    self.separator:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -32);
    self.separator:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", 0, -40);
    self.separator:SetTexture(separatorFilename);
    self.list = GUI:RequestChild("List", nil, self, 1);
    self.list:SetPoint("TOP", self, "TOP", 0, -40);
    self.units = { }; -- Will be dynamically filled for our needs.
    self.BossBar = nil;

    -- ** Properties **
    self.algorithm = nil;
    self.mainUnit = nil;
    self.mainUnitGUID = nil;
    self.thresholdDelay = nil;

    -- ** Initial config **
    self:UpdateSize();
end

function EnemyList:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);
end

function EnemyList:OnScrollOutEnd()
    self:Hide();
end

function EnemyList:OnMainUnitEnabled(guid, name)
    local BossBar = self.BossBar;
    if BossBar then
        self.mainUnitGUID = guid;
        BossBar:ScrollIn();
        BossBar:Blink(false);
        BossBar:SetSpecial(0);
        BossBar:SetText(name);
        BossBar:OnUpdate(0.01); -- Hack to remove a graphical glitch.
    end
end

function EnemyList:OnMainUnitHealthUpdated(health, healthMax)
    local BossBar = self.BossBar;
    if BossBar and self.mainUnitGUID then
        BossBar:SetMax(healthMax);
        BossBar:SetValue(health);
    end
end

function EnemyList:OnMainUnitThresholdShown(threshold)
    local BossBar = self.BossBar;
    if BossBar then
        BossBar:SetSpecial(threshold);
    end
end

function EnemyList:OnMainUnitThresholdHidden()
    local BossBar = self.BossBar;
    if BossBar then
        BossBar:SetSpecial(0);
    end
end

function EnemyList:OnMainUnitDisabled(guid)
    local BossBar = self.BossBar;
    if BossBar and self.mainUnitGUID == guid then
        self.mainUnitGUID = nil;
        BossBar:SetSpecial(0);
        BossBar:ScrollOut();
    end
end

function EnemyList:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end