local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");
local Table = Root:GetOrNewModule("Table");
local Utility = Root:GetOrNewModule("Utility");
local NPCName = Root:GetOrNewModule("NPCName");

local Module = Root:GetOrNewModule("Module");

local System = Root:GetOrNewModule("System");
local Combat = System:GetOrNewModule("Combat");

local EnemyRow = GUI:GetOrNewWidget("EnemyRow");

EnemyRow.TEMPLATE = "BossEncounter3_EnemyRowTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- EnemyRow is a GUI element allowing to display a bar, along
-- with a counter and a label. A second optionnal bar can be displayed.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local DEFAULT_RATE = 0.25;

local BAR_LABEL_SCALE = 0.75;
local ROW_LABEL_SCALE = 0.50;

local ROLE_COLOR = {
    ["NONE"] = {1, 1, 1},
    ["DPS"]  = {1, 0, 0},
    ["HEAL"] = {0, 1, 0},
};

local MAX_NAME_LENGTH = 20;

local THRESHOLD_VISIBILITY = 0.10;

local SHOW_THRESHOLDS_THRESHOLD = 4;
local SHOW_FULL_THRESHOLD = 2;

-- --------------------------------------------------------------------
-- **                             Locals                             **
-- --------------------------------------------------------------------

-- This table contains the basic algorithms that can be used freely by the modules.
-- Simply pass the algorithm name to the AssignAlgorithm method and provide the parameters
-- asked by the algorithm to start it.

local basicAlgorithms = {
    -- Algorithms have access to the running boss module, in the "module" field.
    -- -------------------------------------------------------------------------

    -- MP algorithm *********************************
    -- Simply displays the mana points of the target.

    ["mana"] = {
        label = {
            ["enUS"] = "MP",
            ["frFR"] = "PM",
        },
        color = {0, 0, 1},   -- Color of the bar.
        allowDead = false,   -- If dead, zero is used as value.
        rate = DEFAULT_RATE, -- Variation per second (in %).
        digits = 5,          -- Number of digits allowed on the display.
        useMultiples = true, -- Allow to use K or M if the amount is too high.

        GetValue = function(self, uid)
            return UnitPower(uid, SPELL_POWER_MANA), UnitPowerMax(uid, SPELL_POWER_MANA);
        end,
    },

    -- EP algorithm ***********************************
    -- Simply displays the energy points of the target.

    ["power"] = {
        label = {
            ["enUS"] = "EP",
            ["frFR"] = "PE",
        },
        color = {1, 0.5, 0},
        allowDead = false,
        rate = DEFAULT_RATE,
        digits = 5,
        useMultiples = false,

        GetValue = function(self, uid)
            return UnitPower(uid), UnitPowerMax(uid);
        end,
    },

    -- Revive algorithm ***********************************************
    -- Monitors time left before an add will revive after being killed.

    ["revive"] = {
        label = {
            ["enUS"] = "Regen", -- Warning: no accents or special chars because they won't be displayed by the monofont.
            ["frFR"] = "Regen",
        },
        color = {1, 0.5, 0},
        allowDead = true, -- Allow the extra bar to display a value even when the unit is dead.
        rate = nil,       -- No smoothing
        digits = 2,
        useMultiples = false,

        GetValue = function(self, uid)
            local guid = UnitGUID(uid);
            return self:GetValueMasked(guid);
        end,

        GetValueMasked = function(self, guid)
            return self.module:GetReviveTimer(guid);
        end,
    },
};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function CheckMainUnit(self)
    if self.main and self.found then
        self:Notify("OnMainUnitEnabled", self.guid, self.displayName);
        self:Notify("OnMainUnitUpdated", self.lastHP, self.lastHPM);
    end
end

-- ********************************************************************
-- * Check if the enemy row is being used.                            *
-- ********************************************************************
function EnemyRow:IsUsed()
    return (self.id > 0 or self.guid ~= "");
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the compared unit.                     *
-- ********************************************************************
-- * Check if the enemy row is watching the specified unit.           *
-- ********************************************************************
function EnemyRow:IsUnit(id)
    return (self.id == id or self.guid == id);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: mobID -or- GUID of the unit.                              *
-- * >> selectable: set whether the unit can be selected or not.      *
-- * >> class: class of the unit (not required).                      *
-- * >> role: role of the unit (not required).                        *
-- ********************************************************************
-- * Watch an unit.                                                   *
-- ********************************************************************
function EnemyRow:Watch(id, selectable, class, role)
    if type(id) == "string" then
        self.id = Unit:GetMobID(id) or 0;
        self.guid = id;
  else
        self.id = id;
        self.guid = "";
    end

    self:SetMainUnit(false);

    self.selectable = selectable;
    self.class = class; -- TODO: unused for now.
    self.role = role or "NONE";

    self:Set(self.guid);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID being watched.                                 *
-- ********************************************************************
-- * Watch an unit once its GUID is known.                            *
-- ********************************************************************
function EnemyRow:Set(guid)
    self.found = false;
    self.guid = guid;
    self.name = "";
    self.displayName = "";

    self.dead = false;
    self.symbol = 0;
    self.exact = false;
    self.lastHP = 0;
    self.lastHPM = 1;
    self.damageHP = 0;

    self:UpdateSelect();
    self:UpdateLayout();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: the UID of the unit.                                    *
-- ********************************************************************
-- * Set the enemy as the specified unit.                             *
-- ********************************************************************
function EnemyRow:Discover(unit)
    self.found = true;
    self.guid = UnitGUID(unit);
    self.name = UnitName(unit);
    self.displayName = Utility:Abbreviate(Utility:ToFlatString(self.name), MAX_NAME_LENGTH);

    if self.selectable then
        NPCName:Register(self.id, self.name);
    end

    self:UpdateSelect();
    self:UpdateLayout();

    CheckMainUnit(self);
end

-- ********************************************************************
-- * Stop watching the unit.                                          *
-- ********************************************************************
function EnemyRow:Stop()
    self:SetMainUnit(false);

    self.id = 0;
    self.guid = "";
    self.found = false;
    self.main = false;
    self:SetExtraAlgorithm(nil);
    self:ClearThresholds();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> algorithm: the table defining the algorithm to use.           *
-- * You can also input a string to use one of predefined algorithms. *
-- * See above for informations about algorithms and predefined ones. *
-- ********************************************************************
-- * Assign the algorithm used to set the value of the extra bar.     *
-- * It can be predefined algorithm or a custom one.                  *
-- * Passing nil will clear the algorithm.                            *
-- ********************************************************************
function EnemyRow:SetExtraAlgorithm(algorithm)
    if not algorithm then
        if self.algorithm then
            self.algorithm.module = nil; -- Break the reference to module so it isn't "eaten" by the recursive recycle.
            self.algorithm = Table:Recycle(self.algorithm, true);
        end
        self:DisableSecondaryBar();

elseif type(algorithm) == "string" then
        local predefinedAlgorithm = basicAlgorithms[algorithm];
        self:SetExtraAlgorithm(predefinedAlgorithm);

elseif type(algorithm) == "table" then
        self.algorithm = Table:Copy(algorithm);
        self.algorithm.module = Module:GetActive();

        local name = L:ReadTable(algorithm.label);
        local r, g, b = unpack(algorithm.color);
        self:EnableSecondaryBar(name, r, g, b, algorithm.rate, algorithm.digits, algorithm.useMultiples);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> flag: the main unit flag. True to enable.                     *
-- ********************************************************************
-- * Specify if the enemy is the main unit or not.                    *
-- * If true, the enemy row will notify the enemy list when the main  *
-- * unit health changes.                                             *
-- ********************************************************************
function EnemyRow:SetMainUnit(flag)
    if self.main and not flag then
        self.main = false;
        if self.found then
            self:Notify("OnMainUnitDisabled", self.guid);
        end

elseif not self.main and flag then
        self.main = true;
        CheckMainUnit(self);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> label: the label of the threshold.                            *
-- * >> threshold: the health ratio where the threshold is.           *
-- ********************************************************************
-- * Add an health threshold indicator to the enemy row health bar.   *
-- ********************************************************************
function EnemyRow:AddThreshold(label, threshold)
    local newThreshold = Table:Alloc();
    newThreshold.label = label;
    newThreshold.value = threshold;
    tinsert(self.thresholds, newThreshold);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> label: the label of the threshold.                            *
-- ********************************************************************
-- * Remove an health threshold indicator.                            *
-- ********************************************************************
function EnemyRow:RemoveThreshold(label)
    local i, threshold;
    for i=#self.thresholds, 1, -1 do
        threshold = self.thresholds[i];
        if threshold.label == label then
            Table:Recycle(threshold);
            tremove(self.thresholds, i);
        end
    end
    if self.currentThreshold == label then
        self:HideThreshold();
    end
end

-- ********************************************************************
-- * Clear health thresholds indicators.                              *
-- ********************************************************************
function EnemyRow:ClearThresholds()
    wipe(self.thresholds);
    self:HideThreshold();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> healthDelay: how far ahead the threshold is displayed.        *
-- ********************************************************************
-- * Set the delay to display an health threshold on the bar.         *
-- * If this is set to nil, the default delay is used (10%).          *
-- ********************************************************************
function EnemyRow:SetThresholdDelay(healthDelay)
    self.thresholdDelay = healthDelay;
    self:HideThreshold();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width: the new width.                                         *
-- ********************************************************************
-- * Resize the bars of the enemy row.                                *
-- ********************************************************************
function EnemyRow:ResizeBar(width)
    self.health:ResizeBar(width);
    self.extra:ResizeBar(width);

    self:UpdateSize();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> label: the label to display.                                  *
-- * >> symbol: symbol to display alongside the label.                *
-- * >> r, g, b: color of the label (not required).                   *                          
-- ********************************************************************
-- * Set the enemy row label.                                         *
-- ********************************************************************
function EnemyRow:SetLabel(label, symbol, r, g, b)
    if r and g and b then
        self.label:SetColor(r, g, b);
  else
        self.label:SetColor(1, 1, 1);
    end
    self.label:SetText(label);
    Utility:SetIcon(self.symbolTexture, symbol);
    self:UpdateSize();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the secondary bar.                              *
-- * >> r, g, b: color of the secondary bar.                          *
-- * >> rate: % of change per second.                                 *
-- * >> digits: max digits allowed on display.                        *
-- * >> useMultiples: allow the use of kilo/million when too high.    *
-- ********************************************************************
-- * Enable the enemy row secondary bar.                              *
-- ********************************************************************
function EnemyRow:EnableSecondaryBar(name, r, g, b, rate, digits, useMultiples)
    self.extraLabel:SetText(name);
    self.extraBar:SetColor(r, g, b);
    if rate then
        self.extra:SetRate(rate);
  else
        self.extra:SetInfiniteRate();
    end
    self.extra:SetDimensions(digits or 5, useMultiples or false);

    self.extraLabel:Show();
    self.extra:Show();
    self.extraLabel:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 0);
    self.healthLabel:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 20);

    self:UpdateSize();
end

-- ********************************************************************
-- * Disable the enemy row secondary bar.                             *
-- ********************************************************************
function EnemyRow:DisableSecondaryBar()
    self.extraLabel:Hide();
    self.extra:Hide();
    self.healthLabel:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 0);

    self:UpdateSize();
end

-- ********************************************************************
-- * Set the health bar value.                                        *
-- ********************************************************************
function EnemyRow:SetHealth(health, healthMax)
    self.health:SetValue(health, healthMax);
end

-- ********************************************************************
-- * Set the secondary bar value.                                     *
-- ********************************************************************
function EnemyRow:SetExtra(value, valueMax)
    if self.found and value then
        self.extra:SetValue(value, valueMax);
        self.extra:SetHidden(false);
  else
        self.extra:SetHidden(true);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> state: whether the unit is dead or alive.                     *
-- ********************************************************************
-- * Flag the unit as dead or alive.                                  *
-- ********************************************************************
function EnemyRow:SetDead(state)
    self.dead = state;

    self:UpdateWithoutUnit();
end

-- ********************************************************************
-- * Make the row selectable, if possible.                            *
-- ********************************************************************
function EnemyRow:UpdateSelect()
    local name;
    if self.selectable then
        if self.found then
            name = self.name;
      else
            name = NPCName:Get(self.id);
        end
  else
        name = nil;
    end
    if name then
        -- Make the row selectable.
        self:SetTargetAction(name);
  else
        -- Make the row not selectable.
        self:ClearAction();
    end
end

-- ********************************************************************
-- * Set the appropriate layout for the enemy row.                    *
-- ********************************************************************
function EnemyRow:UpdateLayout()
    if self.found then
        local r, g, b = unpack(ROLE_COLOR[self.role]);
        self:SetLabel(self.displayName, self.symbol, r, g, b);
  else
        self:SetLabel("???", 0);
    end
    self.health:SetHidden(not self.found);
    self.extra:SetHidden(not self.found);
    self:UpdateSize();
end

-- ********************************************************************
-- * Update the enemy row according to the state of the unit.         *
-- ********************************************************************
function EnemyRow:UpdateUnit(unit)
    local algorithm = self.algorithm;
    local extra, extraMax = 0, nil;
    local health, healthMax = 0, self.lastHPM;

    -- We have exact infos about the unit for now.
    self.exact = true;

    -- Death transition check
    if self.dead and (not UnitIsDeadOrGhost(unit)) then
        self:SetDead(false);
elseif (not self.dead) and UnitIsDeadOrGhost(unit) then
        self:SetDead(true);
    end

    -- Health value update
    if not self.dead then
        health, healthMax = UnitHealth(unit), UnitHealthMax(unit);
        self.lastHP = health;
        self.lastHPM = healthMax;
        self.damageHP = 0;
    end

    -- Extra value update
    if algorithm and (not self.dead or algorithm.allowDead) then
        extra, extraMax = algorithm:GetValue(unit);
    end

    -- Symbol change check
    local currentSymbol = (not self.dead and GetRaidTargetIndex(unit)) or 0;
    if currentSymbol ~= self.symbol then
        self.symbol = currentSymbol;
        self:UpdateLayout();
    end

    -- Update bars
    self:SetHealth(health, healthMax);
    self:SetExtra(extra, extraMax);
    self:CheckThreshold(health / max(1, healthMax));

    -- Main unit health notification
    if self.main then
        self:Notify("OnMainUnitHealthUpdated", health, healthMax);
    end
end

-- ********************************************************************
-- * Update the enemy row without knowing the state of the unit.      *
-- ********************************************************************
function EnemyRow:UpdateWithoutUnit()
    local algorithm = self.algorithm;
    local extra, extraMax = 0, nil;
    local health, healthMax = (self.lastHP - self.damageHP), self.lastHPM;

    -- We have inexact infos about the unit.
    self.exact = false;

    -- Health value update
    if self.dead then
        health = 0;
    end

    -- Extra value update
    if algorithm and (not self.dead or algorithm.allowDead) then
        if algorithm.GetValueMasked then
            extra, extraMax = algorithm:GetValueMasked(self.guid);
      else
            extra, extraMax = nil, nil; -- Will be hidden instead.
        end
    end

    -- Update bars
    self:SetHealth(health, healthMax);
    self:SetExtra(extra, extraMax);
    self:CheckThreshold(0);

    -- Main unit health notification
    if self.main then
        self:Notify("OnMainUnitHealthUpdated", health, healthMax);
    end
end

-- ********************************************************************
-- * Show the health threshold indicator.                             *
-- ********************************************************************
function EnemyRow:ShowThreshold(label, threshold)
    if self.currentThreshold ~= label and self.showThresholds then
        self.currentThreshold = label;
        self.threshold:Enable(label, threshold, self.showFull);

        if self.main then
            self:Notify("OnMainUnitThresholdShown", threshold);
        end
    end
end

-- ********************************************************************
-- * Hide the health threshold indicator.                             *
-- ********************************************************************
function EnemyRow:HideThreshold()
    if self.currentThreshold then
        self.currentThreshold = nil;
        self.threshold:Disable();

        if self.main then
            self:Notify("OnMainUnitThresholdHidden");
        end
    end
end

-- ********************************************************************
-- * Check the health threshold indicator.                            *
-- ********************************************************************
function EnemyRow:CheckThreshold(current)
    local closest = 0;
    local closestValue = 1;
    local i, threshold, distance;
    for i=1, #self.thresholds do
        threshold = self.thresholds[i];
        distance = current - threshold.value;
        if distance > 0 and distance < closestValue then
            closestValue = distance;
            closest = i;
        end
    end
    if closest > 0 and closestValue <= (self.thresholdDelay or THRESHOLD_VISIBILITY) then
        threshold = self.thresholds[closest];
        self:ShowThreshold(threshold.label, threshold.value);
  else
        self:HideThreshold();
    end
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole enemy row.        *
-- ********************************************************************
function EnemyRow:UpdateSize()
    if self:IsLocked() then return; end

    local cw = self.healthLabel:GetWidth() * BAR_LABEL_SCALE + self.health:GetWidth();
    local lh = self.label:GetHeight() * ROW_LABEL_SCALE;
    local ch = self.health:GetHeight();
    if self.extra:IsShown() then
        cw = max(cw, self.extraLabel:GetWidth() * BAR_LABEL_SCALE + self.extra:GetWidth());
        ch = ch + self.extra:GetHeight();
    end

    local width = cw;
    local height = lh + ch;

    self:SetWidth(width);
    self:SetHeight(height);

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function EnemyRow:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);
    GUI:InstallSystem("ActionSystem", self);

    -- ** Children **

    self.symbolTexture = self:CreateTexture(nil, "ARTWORK");
    self.symbolTexture:SetWidth(12);
    self.symbolTexture:SetHeight(12);
    self.symbolTexture:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0);
    self.label = GUI:RequestChild("MonoFont", nil, self, 1);
    self.label:SetPoint("TOP", self, "TOP", 0, 0);
    self.label:SetScale(ROW_LABEL_SCALE);
    self.health = GUI:RequestChild("Counter", nil, self, 1);
    self.healthLabel = GUI:RequestChild("MonoFont", nil, self, 1);
    self.healthLabel:SetText(L:Get("HP"));
    self.healthLabel:SetScale(BAR_LABEL_SCALE);
    self.healthBar = self.health:GetBar();
    self.health:SetRate(DEFAULT_RATE);
    self.health:SetDimensions(5, true);
    self.health:TogglePercentage(true);
    self.health:SetPoint("LEFT", self.healthLabel, "RIGHT", 0, 0);
    self.extra = GUI:RequestChild("Counter", nil, self, 1);
    self.extraLabel = GUI:RequestChild("MonoFont", nil, self, 1);
    self.extraLabel:SetScale(BAR_LABEL_SCALE);
    self.extraBar = self.extra:GetBar();
    self.extraBar:SetStyle("SIMPLE");
    self.extra:SetPoint("LEFT", self.extraLabel, "RIGHT", 0, 0);
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HN");
    self.threshold = GUI:RequestChild("Threshold", nil, self.healthBar, 2);
    self.threshold:Bind(self.healthBar, self.healthBar:GetInsets());

    -- ** Properties **
    self.id = 0;
    self.guid = "";
    self.name = "";
    self.displayName = "";
    self.found = false;
    self.class = nil;
    self.role = nil;
    self.selectable = false;
    self.main = false;
    self.dead = false;
    self.symbol = 0;
    self.exact = false;
    self.lastHP = 0;
    self.lastHPM = 1;
    self.damageHP = 0;
    self.algorithm = nil;
    self.showThresholds = false;
    self.showFull = false;
    self.thresholds = { };
    self.currentThreshold = nil;
    self.thresholdDelay = nil;

    -- ** Initial config **
    Combat:RegisterCallback(self.OnCombatEvent, self);
    self:SetExtraAlgorithm(nil);
    self:ResizeBar(160);
    self:ClearAction();
end

function EnemyRow:OnUpdate(elapsed)
    local uid = Unit:GetUID(self.guid);
    if (not self.found) and uid then
        self:Discover(uid);
    end
    if self.found then
        if uid then
            self:UpdateUnit(uid);
      else
            self:UpdateWithoutUnit();
        end
    end
    self:UpdateAction();
end

function EnemyRow:OnDamage(value, zero)
    if not self.exact then
        if zero and value >= 0 then
            self.damageHP = 0;
            self.lastHP = 0;

    elseif zero and value < 0 then
            self.damageHP = 0;
            self.lastHP = self.lastHPM;
      else
            self.damageHP = self.damageHP + value;
        end
    end
end

EnemyRow.OnCombatEvent = {
    ["DAMAGE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed)
        if self:IsUnit(targetGUID) then
            self:OnDamage(amount, (overkill or 0) > 0);
        end
    end,
    ["HEAL"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed)
        if self:IsUnit(targetGUID) then
            self:OnDamage(-amount, (overheal or 0) > 0);
        end
    end,
    ["DEATH"] = function(self, guid, name)
        if self:IsUnit(guid) then
            self:SetDead(true);
        end
    end,
};

function EnemyRow:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end

function EnemyRow:OnShow()
    self:UpdateLayout();
    self:OnUpdate(0);
end

function EnemyRow:OnDifficultyUpdate(difficulty)
    self.showThresholds = (difficulty <= SHOW_THRESHOLDS_THRESHOLD);
    self.showFull       = (difficulty <= SHOW_FULL_THRESHOLD);
end