local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");

local EventList = GUI:GetOrNewWidget("EventList");

EventList.TYPE = "Button";
EventList.TEMPLATE = "BossEncounter3_EventListTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- EventList is a final GUI element displaying a list of timers.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local function sortFunc(eventRow1, eventRow2)
    return eventRow1:GetTimer() < eventRow2:GetTimer();
end

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- * >> display: the displayed name of the event (can be nil).        *
-- * >> timer: the time amount of the event.                          *
-- * >> alert: do something when the timer gets low or not?           *
-- * >> callback: callback upon event expiration.                     *
-- * >> ...: arguments for callback.                                  *
-- ********************************************************************
-- * Start a new event.                                               *
-- * If display is nil, the event will be invisible.                  *
-- * Alert can be 0 (none), 1 (important), 2 (critical).              *
-- ********************************************************************
function EventList:StartEvent(name, display, timer, alert, callback, ...)
    if self:GetEventRow(name) ~= nil then
        Root:Debug("Attempted to run duplicate event in EventList: "..name);
  else
        local E = self:GetFreeEventRow();
        E:Start(name, display, timer, alert, callback, ...);
        if display then
            self.list:Insert(E);
            self:Update();
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- ********************************************************************
-- * Stop an event.                                                   *
-- ********************************************************************
function EventList:StopEvent(name)
    local E = self:GetEventRow(name);
    if E then
        E:Stop();
        self.list:Remove(E);
        self:Update();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- ********************************************************************
-- * Hide the timer of an event.                                      *
-- ********************************************************************
function EventList:HideEventTimer(name)
    local E = self:GetEventRow(name);
    if E then
        E:HideTimer();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- * >> icon: the icon to display. Nil to clear.                      *
-- ********************************************************************
-- * Set the icon for an event.                                       *
-- ********************************************************************
function EventList:SetIcon(name, icon)
    local E = self:GetEventRow(name);
    if E then
        E:SetIcon(icon);
    end
end

-- ********************************************************************
-- * Stop all event rows from the event list.                         *
-- ********************************************************************
function EventList:Clear()
    local i;
    for i=1, #self.events do
        self.events[i]:Stop();
    end
    self.list:Clear();
    self:Update();
end

-- ********************************************************************
-- * Update the event list: check if we should make it appear,        *
-- * if we should make it disappear, and if we should sort it.        *
-- ********************************************************************
function EventList:Update()
    self:Sort();
    if self.list:IsEmpty() then
        self:ScrollOut();
  else
        self:ScrollIn();
    end
end

-- ********************************************************************
-- * Sort the events in the event list.                               *
-- ********************************************************************
function EventList:Sort()
    self.list:Sort(sortFunc);
    self:UpdateSize();
end

-- ********************************************************************
-- * Get a free eventRow widget (or create a new one).                *
-- ********************************************************************
function EventList:GetFreeEventRow()
    local i, eventRow;
    for i=1, #self.events do
        eventRow = self.events[i];
        if not eventRow:GetEventName() then -- This row is available.
            return eventRow;
        end
    end
    -- No existing row available, create a new one.
    eventRow = GUI:RequestChild("EventRow", nil, self);
    tinsert(self.events, eventRow);
    return eventRow;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- ********************************************************************
-- * Search the row of specified event.                               *
-- ********************************************************************
function EventList:GetEventRow(name)
    local i, eventRow;
    for i=1, #self.events do
        eventRow = self.events[i];
        if eventRow:GetEventName() == name then
            return eventRow;
        end
    end
end

-- ********************************************************************
-- * Update the size of the event list.                               *
-- ********************************************************************
function EventList:UpdateSize()
    self.list:UpdateSize();

    local w = max(self.title:GetWidth(), self.list:GetWidth());
    local h = self.title:GetHeight() + self.separator:GetHeight() + self.list:GetHeight();
    self:SetWidth(w);
    self:SetHeight(h);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function EventList:OnLoad()
    local _, titleFilename, titleWidth, titleHeight = Resources:GetTextureInfo("EVENTS");
    local _, separatorFilename = Resources:GetTextureInfo("SEPARATOR");

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.title = self:CreateTexture(nil, "ARTWORK");
    self.title:SetPoint("TOP", self, "TOP", 0, 0);
    self.title:SetTexture(titleFilename);
    self.title:SetWidth(titleWidth);
    self.title:SetHeight(titleHeight);
    self.separator = self:CreateTexture(nil, "BACKGROUND");
    self.separator:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -32);
    self.separator:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", 0, -40);
    self.separator:SetTexture(separatorFilename);
    self.list = GUI:RequestChild("List", nil, self, 1);
    self.list:SetPoint("TOP", self, "TOP", 0, -40);
    self.events = { }; -- Will be dynamically filled for our needs.

    -- ** Properties **

    -- ** Initial config **
    self:UpdateSize();
end

function EventList:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);
end

function EventList:OnEventTrigger(name)
    self:StopEvent(name);
end

function EventList:OnScrollOutEnd()
    self:Hide();
end

function EventList:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end