local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");
local Utility = Root:GetOrNewModule("Utility");

local EventRow = GUI:GetOrNewWidget("EventRow");

EventRow.TEMPLATE = "BossEncounter3_EventRowTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- EventRow is a GUI element allowing to display a bar, along
-- with a timer and a label, on a nice black band.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local LABEL_SCALE = 0.5;

local ALERT_THRESHOLD = 5;

local BLINK_DURATION = 2.00;
local BLINK_CYCLE = 0.50;

local MAX_NAME_LENGTH = 19;

local ALERT_COLOR = { [0] = {0, 0, 0}, [1] = {1, 1, 0}, [2] = {1, 0, 0} };

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width: the new width.                                         *
-- ********************************************************************
-- * Resize the bar of the event row.                                 *
-- ********************************************************************
function EventRow:ResizeBar(width)
    self.bar:Resize(width);
    self:UpdateSize();
end

-- ********************************************************************
-- * Get the name of the event being handled by the event row.        *
-- * Return nil if no event is running on this frame.                 *
-- ********************************************************************
function EventRow:GetEventName()
    return self.name;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the event.                         *
-- * >> display: the displayed name of the event.                     *
-- * >> timer: the time amount of the event.                          *
-- * >> alert: do something when the timer gets low or not?           *
-- * >> callback: callback upon event expiration.                     *
-- * >> ...: arguments for callback.                                  *
-- ********************************************************************
-- * Start a new event row.                                           *
-- ********************************************************************
function EventRow:Start(name, display, timer, alert, callback, ...)
    self.name = name;
    self.timerMax = 0;
    self.triggered = false;
    self.color = alert or 0;
    self.alerted = ((not alert) or (alert == 0));
    self.callback = callback;
    self.args = {...};

    self.label:SetText(Utility:Abbreviate(display or name, MAX_NAME_LENGTH));
    self.timer:SetColor(1, 1, 1);
    self:SetTimer(timer);
    self:Resume();
    self:ShowTimer();

    self:Update();
end

-- ********************************************************************
-- * Stop the event row, nullifying the timer.                        *
-- ********************************************************************
function EventRow:Stop()
    self.name = nil;
    self.triggered = true;
    self.blink = 0;
    self:SetIcon(nil);
end

-- ********************************************************************
-- * Pause the timer of the event row.                                *
-- ********************************************************************
function EventRow:Pause()
    self.timer:Stop();
end

-- ********************************************************************
-- * Resume the timer of the event row.                               *
-- ********************************************************************
function EventRow:Resume()
    self.timer:SetDelta(-1);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> timer: the new timer.                                         *
-- ********************************************************************
-- * Change the timer of the event row.                               *
-- ********************************************************************
function EventRow:SetTimer(timer)
    self.timerMax = max(self.timerMax, timer);
    self.bar:SetMinMax(0, self.timerMax);

    self.timer:SetValue(timer);

    self:Update();
end

-- ********************************************************************
-- * Get the timer value of the event row.                            *
-- ********************************************************************
function EventRow:GetTimer()
    return self.timer:GetValue();
end

-- ********************************************************************
-- * Hide the timer of the event row.                                 *
-- ********************************************************************
function EventRow:HideTimer()
    self.timer:SetHidden(true);

    self:UpdateSize();
end

-- ********************************************************************
-- * Show the timer of the event row.                                 *
-- ********************************************************************
function EventRow:ShowTimer()
    self.timer:SetHidden(false);

    self:UpdateSize();
end

-- ********************************************************************
-- * Trigger the callback of the event row.                           *
-- ********************************************************************
function EventRow:Trigger()
    if not self.triggered then
        self.triggered = true;

        self:Notify("OnEventTrigger", self.name);

        if self.callback then
            self.callback(unpack(self.args));
        end
    end
end
-- ********************************************************************
-- * Arguments:                                                       *
-- * >> alert: the type of alert.                                     *
-- * 0 is none, 1 is important, 2 is critical.                        *
-- ********************************************************************
-- * Make the event row blink, putting some emphasis on it.           *
-- ********************************************************************
function EventRow:Blink(alert)
    if alert > 0 then
        self.blink = BLINK_DURATION;
        if alert == 1 then
            Audio:PlaySound("BLINK");
    elseif alert > 1 then
            Audio:PlaySound("BLINKCRITICAL");
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> icon: the icon to display. Nil to clear.                      *
-- ********************************************************************
-- * Set an icon on the event row.                                    *
-- ********************************************************************
function EventRow:SetIcon(icon)
    if icon then
        self.icon:SetTexture(icon);
        self.icon:Show();
        self.label:SetPoint("TOP", self, "TOP", 12, 0);
        self.bar:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 24, 0);
        self:ResizeBar(120);
  else
        self.icon:Hide();
        self.label:SetPoint("TOP", self, "TOP", 0, 0);
        self.bar:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 0);
        self:ResizeBar(144);
    end
end

-- ********************************************************************
-- * Update the event row.                                            *
-- ********************************************************************
function EventRow:Update()
    if self.timer:IsDisplayHidden() then
        self.bar:SetValue(0);
  else
        local timeLeft = self.timer:GetValue();
        local progression = 1 - timeLeft / self.timerMax
        local r, g = progression^2, (1 - progression)^2;
        self.bar:SetColor(r, g, 0);
        self.bar:SetValue(timeLeft);

        if self.timer:GetDelta() < 0 and (not self.triggered) then
            if (not self.alerted) and timeLeft < ALERT_THRESHOLD then
                self.alerted = true;
                if not self.timer:IsDisplayHidden() then
                    self.timer:SetColor( unpack(ALERT_COLOR[self.color]) );
                    self:Blink(self.color);
                end
            end
        end
    end
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole event row.        *
-- ********************************************************************
function EventRow:UpdateSize()
    local width = self.label:GetWidth() * LABEL_SCALE;
    local height = self.label:GetHeight() * LABEL_SCALE + self.timer:GetHeight();

    if self.icon:IsShown() then
        width = max(width, self.icon:GetWidth() + self.bar:GetWidth() + self.timer:GetWidth());
  else
        width = max(width, self.bar:GetWidth() + self.timer:GetWidth());
    end

    self:SetWidth(width);
    self:SetHeight(height);

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function EventRow:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.label = GUI:RequestChild("MonoFont", nil, self, 1);
    self.label:SetScale(LABEL_SCALE);
    self.timer = GUI:RequestChild("Timer", nil, self, 1);
    self.timer:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0);
    self.bar = GUI:RequestChild("Bar", nil, self, 1);
    self.icon = self:CreateTexture(nil, "ARTWORK");
    self.icon:SetWidth(24);
    self.icon:SetHeight(24);
    self.icon:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 2);
    self.icon:Hide();
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HN");

    -- ** Properties **
    self.name = nil;
    self.timerMax = 0;
    self.triggered = false;
    self.color = 0;
    self.alerted = false;
    self.callback = nil;
    self.args = {};
    self.blink = 0;

    -- ** Initial config **
    self:Pause();
    self:SetIcon(nil);
end

function EventRow:OnUpdate(elapsed)
    self:Update();
    self.blink = max(0, self.blink - elapsed);

    local a = 1 - abs(cos(self.blink * 180 / BLINK_CYCLE));
    local r, g, b = unpack(ALERT_COLOR[self.color]);
    self.band:SetColor(r * a, g * a, b * a);
end

function EventRow:OnTimerExpire(timer)
    -- Called when the child timer has expired.
    self:Trigger();
end

function EventRow:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end