local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");

local FloatingTimer = GUI:GetOrNewWidget("FloatingTimer");

FloatingTimer.TEMPLATE = "BossEncounter3_FloatingTimerTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Floating timer is a simple, general timer.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local TIMER_SCALE = 2.5;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function getCurrentSecond(self)
    return math.floor(self.timer:GetValue());
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mode: < 0 is countdown mode.                                  *
-- *          > 0 is watch mode.                                      *
-- ********************************************************************
-- * Set the mode of a floating timer.                                *
-- ********************************************************************
function FloatingTimer:SetMode(mode)
    self.mode = mode;
    self:Clear();
end

-- ********************************************************************
-- * Get the value of a floating timer.                               *
-- ********************************************************************
function FloatingTimer:GetValue()
    return self.timer:GetValue();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new value.                                         *
-- ********************************************************************
-- * Change the value of a floating timer.                            *
-- ********************************************************************
function FloatingTimer:SetValue(value)
    self.timer:SetValue(value);
end

-- ********************************************************************
-- * Start the floating timer.                                        *
-- ********************************************************************
function FloatingTimer:Start()
    self.removeDelay = 0;
    self.lastSecond = getCurrentSecond(self);
    Audio:PlaySound("WOOSH");
    self:Resume();
    self:ScrollIn();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> removeDelay: optional delay before removing the timer.        *
-- ********************************************************************
-- * Stop the floating timer.                                         *
-- ********************************************************************
function FloatingTimer:Stop(removeDelay)
    self.timer:Stop();
    if removeDelay and removeDelay > 0 then
        self.removeDelay = removeDelay;
        Audio:PlaySound("POPUP");
  else
        self:ScrollOut();
    end
end

-- ********************************************************************
-- * Resume the floating timer.                                       *
-- ********************************************************************
function FloatingTimer:Resume()
    self.timer:SetDelta(self.mode);
end

-- ********************************************************************
-- * Clear the floating timer.                                        *
-- ********************************************************************
function FloatingTimer:Clear()
    self:SetValue(0);
    self:Stop();
end

-- ********************************************************************
-- * Update the floating timer size.                                  *
-- ********************************************************************
function FloatingTimer:UpdateSize()
    self:SetWidth(self.timer:GetWidth() * TIMER_SCALE);
    self:SetHeight(self.timer:GetHeight() * TIMER_SCALE);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function FloatingTimer:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, false);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.timer = GUI:RequestChild("Timer", nil, self, 0);
    self.timer:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.timer:SetStyle("FULL");
    self.timer:SetScale(TIMER_SCALE);

    -- ** Properties **
    self.mode = 0;
    self.removeDelay = 0;
    self.lastSecond = 0;

    -- ** Initial config **
    self:UpdateSize();
end

function FloatingTimer:OnUpdate(elapsed)
    local timerAlpha = 1;

    self:UpdateScroll(elapsed);

    -- Handle time tick playback
    if self.timer:GetDelta() ~= 0 then
        local second = getCurrentSecond(self);
        if second ~= self.lastSecond then
            Audio:PlaySound("TIMETICK");
            self.lastSecond = second;
        end
    end

    -- Handle remove delay
    if self.removeDelay > 0 then
        self.removeDelay = max(0, self.removeDelay - elapsed);

        -- Handle blinking
        if math.fmod(self.removeDelay, 0.20) < 0.10 then
            timerAlpha = 0;
        end

        if self.removeDelay == 0 then
            self:ScrollOut();
        end
    end

    self.timer:SetAlpha(timerAlpha);
end

function FloatingTimer:OnScrollOutEnd()
    self:Hide();
end

function FloatingTimer:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end