﻿local Root = BossEncounter3;

local GUI = Root:GetOrNewModule("GUI");

local Table = Root:GetOrNewModule("Table");

local System = Root:GetOrNewModule("System");
local Settings = System:GetOrNewModule("Settings");

-- --------------------------------------------------------------------
-- **                           Properties                           **
-- --------------------------------------------------------------------

GUI._show = { };
GUI._hide = { };

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Get the base reference object for parent and position.           *
-- * It is UIParent by default. All objects made with Request will    *
-- * have the object referenced by this function as initial parent.   *
-- ********************************************************************
function GUI:GetBase()
    return _G["UIParent"];
end

-- ********************************************************************
-- * Convert a pixel value to a % value.                              *
-- ********************************************************************
function GUI:DotToPctX(x)
    local base = self:GetBase();
    return x / base:GetWidth();
end
function GUI:DotToPctY(y)
    local base = self:GetBase();
    return y / base:GetHeight();
end

-- ********************************************************************
-- * Convert a % value to a pixel value.                              *
-- ********************************************************************
function GUI:PctToDotX(x)
    local base = self:GetBase();
    return x * base:GetWidth();
end
function GUI:PctToDotY(y)
    local base = self:GetBase();
    return y * base:GetHeight();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> widgetName: the name of the widget.                           *
-- ********************************************************************
-- * Get an existing widget class or create a new one.                *
-- ********************************************************************
function GUI:GetOrNewWidget(widgetName)
    if type(self[widgetName]) == "table" then
        return self[widgetName];
  else
        self[widgetName] = {
            TYPE = "Frame",
            TEMPLATE = nil,
            WATCHSETTINGS = { },
        };
        return self[widgetName];
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> widgetName: the name of the widget.                           *
-- * >> instanceName: how the new instance will be named.             *
-- * instanceName can be nil (for unnamed frames) or 'AUTO'.          *
-- * It can also be a number.                                         *
-- ********************************************************************
-- * Request an instance of the specified widget class.               *
-- ********************************************************************
function GUI:Request(widgetName, instanceName)
    if type(instanceName) == "number" then
        instanceName = Root:GetName().."_"..widgetName..instanceName;
elseif instanceName == 'AUTO' then
        instanceName = Root:GetName().."_"..widgetName;
    end

    local widget = self[widgetName];
    if not widget then
        error("Unknown widget class: "..widgetName, 0);
    end

    local instance = CreateFrame(widget.TYPE, instanceName, self:GetBase(), widget.TEMPLATE);
    instance._loaded = false;
    instance._class = widgetName;
    instance._systems = { };
    self:InstallMethods(widget, instance);

    if instance.OnLoad then
        instance:OnLoad();
        instance._loaded = true;
  else
        error("Missing OnLoad for "..widgetName.." widget class!", 0);
    end

    tinsert(self, instance);

    -- Notify the new element of the current difficulty and settings it has registered for.
    local id;
    self:FireWidgetHandler(instance, "OnDifficultyUpdate", Settings:GetDifficulty());
    for id in pairs(widget.WATCHSETTINGS) do
        if Settings:Exists(id) then
            self:FireWidgetHandler(instance, "OnSettingUpdate", id, Settings:ReadReal(id));
      else
            error("Widget class "..widgetName.." watches unknown setting: "..id, 0);
        end
    end

    return instance;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> widgetName: the name of the widget.                           *
-- * >> instanceName: how the new instance will be named.             *
-- * >> parent: the parent of the widget.                             *
-- * >> priority: frame levels above the parent of the child.         *
-- ********************************************************************
-- * Request an instance of the specified widget class with a parent. *
-- ********************************************************************
function GUI:RequestChild(widgetName, instanceName, parent, priority)
    local instance = self:Request(widgetName, instanceName);
    instance:SetParent(parent);
    if priority then
        instance:SetFrameLevel(parent:GetFrameLevel() + priority);
    end
    return instance;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> widget: the widget object.                                    *
-- * >> instance: the instance of that widget.                        *
-- ********************************************************************
-- * Populate an instance object's table with methods of its class.   *
-- ********************************************************************
function GUI:InstallMethods(widget, instance)
    local k, v;
    for k, v in pairs(widget) do
        if type(v) == "function" then
            instance[k] = v;

    elseif type(v) == "table" then
            instance[k] = Table:Copy(v);
        end
    end
    -- Install extra getters.
    instance.GetClass = function(self) return self._class; end
    instance.IsLoaded = function(self) return self._loaded; end
    instance.IsLocked = function(self) return self:IsProtected() and InCombatLockdown(); end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> systemName: the system name.                                  *
-- * >> instance: the instance of that widget.                        *
-- * >> ...: arguments passed to the install method.                  *
-- ********************************************************************
-- * Install a sub-system component into a widget's instance.         *
-- ********************************************************************
function GUI:InstallSystem(systemName, instance, ...)
    if not instance._systems[systemName] then
        local system = self:GetOrNewWidget(systemName);
        if system and system.Install then
            if not system.Install(instance, ...) then
                error("System "..systemName.." reported an install error!", 0);
          else
                instance._systems[systemName] = true;
            end
      else
            error("System "..systemName.." missing!", 0);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> element: the element whose widget is fired.                   *
-- * >> handlerName: the name of the handler.                         *
-- * >> ...: arguments passed to the handler.                         *
-- ********************************************************************
-- * Fire an handler on a GUI element of BE3.                         *
-- ********************************************************************
function GUI:FireWidgetHandler(element, handlerName, ...)
    local handler = element[handlerName];
    if handler then
        handler(element, ...);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> handlerName: the name of the handler.                         *
-- * >> ...: arguments passed to the handler.                         *
-- ********************************************************************
-- * Fire an handler on ALL GUI elements of BE3.                      *
-- * Don't abuse this method, it should be used for specific cases.   *
-- ********************************************************************
function GUI:FireHandler(handlerName, ...)
    local i;
    for i=1, #self do
        self:FireWidgetHandler(self[i], handlerName, ...);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the setting being monitored.                              *
-- ********************************************************************
-- * Register a widget class to be informed of the state of a setting *
-- * at initialization time and when widgets are created.             *
-- ********************************************************************
function GUI:WatchSettingUpdate(widgetName, id)
    local widget = self[widgetName];
    if widget then
        widget.WATCHSETTINGS[id] = true;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> object: the object to show.                                   *
-- ********************************************************************
-- * Ask the GUI to show an object in a secure way.                   *
-- * GUI:Show is exactly the same as calling Show method directly for *
-- * non-secure widgets. However, for secure widgets, GUI:Show will   *
-- * wait until you leave combat to show the widget.                  *
-- ********************************************************************
function GUI:Show(object)
    local i;
    for i=#self._hide, 1, -1 do
        if self._hide[i] == object then
            tremove(self._hide, i);
        end
    end
    if object:IsLocked() then
        tinsert(self._show, object);
  else
        object:Show();
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> object: the object to hide.                                   *
-- ********************************************************************
-- * Ask the GUI to hide an object in a secure way.                   *
-- * GUI:Hide is exactly the same as calling Hide method directly for *
-- * non-secure widgets. However, for secure widgets, GUI:Hide will   *
-- * wait until you leave combat to hide the widget.                  *
-- ********************************************************************
function GUI:Hide(object)
    local i;
    for i=#self._show, 1, -1 do
        if self._show[i] == object then
            tremove(self._show, i);
        end
    end
    if object:IsLocked() then
        tinsert(self._hide, object);
  else
        object:Hide();
    end
end

-- ********************************************************************
-- * Execute pending show/hide secure commands.                       *
-- ********************************************************************
function GUI:ExecuteSecureCommands()
    local i;
    for i=1, #self._hide do
        self._hide[i]:Hide();
    end
    for i=1, #self._show do
        self._show[i]:Show();
    end
    wipe(self._hide);
    wipe(self._show);
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function GUI:OnUpdate(elapsed)
    self:FireHandler("OnLogicUpdate", elapsed);
end

function GUI:OnCombatLeave()
    self:ExecuteSecureCommands();
end

function GUI:OnRankChange(rank)
    self:FireHandler("OnRankChange", rank);
end

function GUI:OnSettingsRefresh()
    self:FireHandler("OnSettingsRefresh");
end

function GUI:OnDifficultyUpdate(difficulty)
    self:FireHandler("OnDifficultyUpdate", difficulty);
end

function GUI:OnSettingUpdate(id, value)
    local i, element, widget;
    for i=1, #self do
        element = self[i];
        widget = self[element:GetClass()];
        if widget and widget.WATCHSETTINGS[id] then
            self:FireWidgetHandler(element, "OnSettingUpdate", id, value);
        end
    end
end