local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Audio = Root:GetOrNewModule("Audio");

local ImportantTimer = GUI:GetOrNewWidget("ImportantTimer");

ImportantTimer.TEMPLATE = "BossEncounter3_ImportantTimerTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Important timer is a timer with a bar and a icon graphic.
-- It grows below a given threshold.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local ZOOM_SCALE = 2;
local ZOOM_TIMER = 0.50;
local ZOOM_SWAY_TIMER = 2.50;

local LAST_SECONDS_THRESHOLD = 11;

local BLINK_RATE = 1.00;

local colors = {
    {1.0, 0.0, 0.0}, -- Level 1 - Red: little time remaining + zoom
    {1.0, 1.0, 0.0}, -- Level 2 - Yellow: some time remaining
    {0.0, 1.0, 0.0}, -- Level 3 - Green: generous time
    {1.0, 1.0, 1.0}, -- Level 4 - White: no countdown
}

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> yellowThreshold: the threshold below which the timer will be  *
-- *                     yellow. 0 will always disable this.          *
-- * >> redThreshold: the threshold below which the timer will be     *
-- *                  scaled up and red. 0 will always disable this.  *
-- * >> useLastSeconds: specifies if the last seconds sound should    *
-- *                    be used when below 10 seconds.                *
-- * >> swayOnStart: if true, the timer will have a sway effect once  *
-- *                 it is started. A sound will also be played.      *
-- ********************************************************************
-- * Set the behavior of the important timer.                         *
-- ********************************************************************
function ImportantTimer:SetBehavior(yellowThreshold, redThreshold, useLastSeconds, swayOnStart)
    self.yellowThreshold = yellowThreshold;
    self.redThreshold = redThreshold;
    self.useLastSeconds = useLastSeconds;
    self.swayOnStart = swayOnStart;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mode: < 0 is countdown mode.                                  *
-- *          > 0 is watch mode.                                      *
-- ********************************************************************
-- * Change the mode of an important timer.                           *
-- ********************************************************************
function ImportantTimer:SetMode(mode)
    self.mode = mode;
    self:Clear();
end

-- ********************************************************************
-- * Set the icon of the important timer.                             *
-- ********************************************************************
function ImportantTimer:SetIcon(textureName, scale)
    local _, filename, width, height = Resources:GetTextureInfo(textureName);
    self.icon:SetTexture(filename);
    self.icon:SetWidth(width * (scale or 1));
    self.icon:SetHeight(height * (scale or 1));
    self:UpdateSize();
end

-- ********************************************************************
-- * Get the value of an important timer.                             *
-- ********************************************************************
function ImportantTimer:GetValue()
    return self.timer:GetValue();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new value.                                         *
-- * >> max: max value (optionnal, will default to "value").          *
-- ********************************************************************
-- * Change the value of an important timer.                          *
-- ********************************************************************
function ImportantTimer:SetValue(value, max)
    self.bar:SetMinMax(0, max or value or 0);
    if value then
        self.timer:SetValue(value);
    end
    self:UpdateSize();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> max: max value (optionnal).                                   *
-- ********************************************************************
-- * Start the important timer.                                       *
-- ********************************************************************
function ImportantTimer:Start(max)
    if self.mode < 0 then
        self.lastSeconds = false;
        if self.swayOnStart then
            self:SwayTimer();
            Audio:PlaySound("STARTTIMER");
        end
    end
    if max then
        self:SetValue(nil, max);
    end
    self:Resume();
end

-- ********************************************************************
-- * Stop the important timer.                                        *
-- ********************************************************************
function ImportantTimer:Stop()
    self.timer:Stop();
end

-- ********************************************************************
-- * Resume the important timer.                                      *
-- ********************************************************************
function ImportantTimer:Resume()
    self.timer:SetDelta(self.mode);
end

-- ********************************************************************
-- * Clear the important timer.                                       *
-- ********************************************************************
function ImportantTimer:Clear()
    self:SetValue(0);
    self:Stop();
end

-- ********************************************************************
-- * Set the color scheme of the important timer.                     *
-- ********************************************************************
function ImportantTimer:SetColor(level)
    if self.color ~= level then
        self.color = level;

        local r, g, b = unpack(colors[level]);
        self.bar:SetColor(r, g, b);
        self.timer:SetColor(r, g, b);

        if level == 1 then
            self:GrowTimer();
      else
            self:ShrinkTimer();
        end
    end
end

-- ********************************************************************
-- * Make the timer normal size.                                      *
-- ********************************************************************
function ImportantTimer:ShrinkTimer()
    self.zoomDt = -1;
end

-- ********************************************************************
-- * Make the timer increased size.                                   *
-- ********************************************************************
function ImportantTimer:GrowTimer()
    self.zoomDt = 1;
end

-- ********************************************************************
-- * Make the timer sway between normal and increased size.           *
-- ********************************************************************
function ImportantTimer:SwayTimer()
    self.zoomSway = ZOOM_SWAY_TIMER;
end

-- ********************************************************************
-- * Hide the timer.                                                  *
-- ********************************************************************
function ImportantTimer:HideTimer()
    self.timer:Hide();
end

-- ********************************************************************
-- * Show the timer.                                                  *
-- ********************************************************************
function ImportantTimer:ShowTimer()
    self.timer:Show();
end

-- ********************************************************************
-- * Update the important timer width.                                *
-- ********************************************************************
function ImportantTimer:UpdateSize()
    local tw = self.timer:GetWidth() * ZOOM_SCALE;
    self.bar:Resize(tw);

    local width = tw + self.icon:GetWidth();
    local height = self.bar:GetHeight() + self.timer:GetHeight() * ZOOM_SCALE;

    self:SetWidth(width);
    self:SetHeight(height);

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function ImportantTimer:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.icon = self:CreateTexture(nil, "ARTWORK");
    self.icon:ClearAllPoints();
    self.icon:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 0);
    self.bar = GUI:RequestChild("Bar", nil, self, 0);
    self.bar:SetPoint("BOTTOMLEFT", self.icon, "BOTTOMRIGHT", 0, 0);
    self.bar:SetStyle("SIMPLE");
    self.timer = GUI:RequestChild("Timer", nil, self, 0);
    self.timer:SetPoint("BOTTOM", self.bar, "TOP", 0, 0);
    self.timer:SetStyle("FULL");

    -- ** Properties **
    self.mode = 0;
    self.color = 0;
    self.zoom = 0;
    self.zoomDt = 0;
    self.zoomSway = 0;
    self.lastSeconds = false;
    self.yellowThreshold = 0;
    self.redThreshold = 0;
    self.useLastSeconds = false;
    self.swayOnStart = false;
    self.blinkTimer = 0;

    -- ** Initial config **
    self:SetIcon("CLOCK", 0.5);
    self:SetColor(#colors);
    self:SetBehavior(180, 60, true, true);
    self:UpdateSize();
end

function ImportantTimer:OnShow()
    self.zoom = 0;
    self.zoomSway = 0;
    self.icon:SetAlpha(1);
    self:OnUpdate(0);
end

function ImportantTimer:OnUpdate(elapsed)
    local timeRemaining = self.timer:GetValue();
    local blinking = false;

    -- Handle coloring and last seconds sound.
    local color = #colors;
    if self.mode < 0 then
        if timeRemaining == 0 then
            blinking = true; -- + default color!
    elseif timeRemaining < self.redThreshold then
            color = 1;
    elseif timeRemaining < self.yellowThreshold then
            color = 2;
      else
            color = 3;
        end
        if timeRemaining <= LAST_SECONDS_THRESHOLD and not self.lastSeconds and self.timer:GetDelta() < 0 then
            self.lastSeconds = true;

            if self.useLastSeconds then
                Audio:PlaySound("LASTSECONDS");
            end
        end
    end
    self:SetColor(color);

    -- Handle bar
    self.bar:SetValue(timeRemaining);

    -- Handle growth
    if self.zoomSway > 0 then
        self.zoomSway = max(0, self.zoomSway - elapsed);

        local progress = math.abs(sin(self.zoomSway / ZOOM_SWAY_TIMER * 720));
        self.zoom = progress * ZOOM_TIMER * 0.5;
        self.icon:SetAlpha(1 - progress);
  else
        self.zoom = max(0, min(ZOOM_TIMER, self.zoom + self.zoomDt * elapsed));
    end
    self.timer:SetScale(1 + self.zoom / ZOOM_TIMER * (ZOOM_SCALE - 1));

    -- Handle blinking
    local timerAlpha = 1;
    if blinking then
        self.blinkTimer = self.blinkTimer + elapsed;
        timerAlpha = abs(cos(self.blinkTimer / BLINK_RATE * 180));
  else
        self.blinkTimer = 0;
    end
    self.timer:SetAlpha(timerAlpha);
end

function ImportantTimer:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end

function ImportantTimer:OnTimerExpire()
    self:Notify("OnImportantTimerExpire", self);
end