local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");

local MirrorBar = GUI:GetOrNewWidget("MirrorBar");

MirrorBar.TEMPLATE = "BossEncounter3_MirrorBarTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Mirror bar is a long emphasized bar which shows a mirrored value.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local ZOOM_TIME = 0.5;

local LABEL_SCALE = 0.5;

local MAX_DIGITS = 5;

local BAR_WIDTH = 384;
local ALTBAR_WIDTH = 376;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> label: the label of the bar.                                  *
-- * >> object: the object that has the method.                       *
-- * >> method: the method to call to get the current/max value.      *
-- * >> methodAlt: same as method, but for the alternate bar.         *
-- ********************************************************************
-- * Start the mirror bar.                                            *
-- ********************************************************************
function MirrorBar:Start(label, object, method, methodAlt)
    self.object = object;
    self.method = method;
    self.methodAlt = methodAlt;
    self.active = true;
    
    local current, max = self:GetValue();
    self.counter:SetDimensions(min(MAX_DIGITS, floor(log10(max)) + 1), true); -- Set the appropriate number of digits.
    self.counter:TogglePercentage(max > 100);

    self.label:SetText(label);

    Audio:PlaySound("HIT");
    self:Show();
end

-- ********************************************************************
-- * Stop the mirror bar.                                             *
-- ********************************************************************
function MirrorBar:Stop()
    self.active = false;
end

-- ********************************************************************
-- * Instantly get rid of the mirror bar.                             *
-- ********************************************************************
function MirrorBar:Clear()
    self.active = false;
    self.object = nil;
    self.method = nil;
    self.methodAlt = nil;
    self:Hide();
end

-- ********************************************************************
-- * Get the current/max value of the mirror bar.                     *
-- ********************************************************************
function MirrorBar:GetValue()
    if self.method then
        return self.method(self.object);
  else
        return 0, 1;
    end
end

-- ********************************************************************
-- * Get the alternate current/max value of the mirror bar.           *
-- ********************************************************************
function MirrorBar:GetAltValue()
    if self.methodAlt then
        return self.methodAlt(self.object);
  else
        return 0, 1;
    end
end

-- ********************************************************************
-- * Update the mirror bar width.                                     *
-- ********************************************************************
function MirrorBar:UpdateSize()
    local width = max(self.label:GetWidth() * LABEL_SCALE, self.counter:GetWidth());
    local height = self.label:GetHeight() * LABEL_SCALE + self.counter:GetHeight();

    self:SetWidth(width + 32);
    self:SetHeight(height);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function MirrorBar:OnLoad()
    -- ** Children **
    self.label = GUI:RequestChild("MonoFont", nil, self, 1);
    self.label:SetPoint("TOP", self, "TOP", 0, 0);
    self.label:SetScale(LABEL_SCALE);
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HS");
    self.band:SetFixedBorders(true);
    self.counter = GUI:RequestChild("Counter", nil, self, 1);
    self.counter:SetPoint("BOTTOM", self, "BOTTOM", 0, 0);
    self.counter:GetBar():SetStyle("SIMPLE");
    self.counter:ResizeBar(BAR_WIDTH);
    self.counter:SetBarColor(0.25, 0.5, 1.00);

    local bar = self.counter:GetBar();
    self.altBar = bar:CreateTexture(nil, "OVERLAY");
    self.altBar:SetTexture(1, 0, 0);
    self.altBar:SetWidth(ALTBAR_WIDTH);
    self.altBar:SetHeight(2);
    self.altBar:SetPoint("TOPLEFT", bar, "TOPLEFT", 3, -4);

    -- ** Properties **
    self.zoom = 0;

    -- ** Initial config **
    self:Clear();
    self:UpdateSize();
end

function MirrorBar:OnShow()
    self.zoom = ZOOM_TIME;
    self:OnUpdate(0);
end

function MirrorBar:OnUpdate(elapsed)
    -- Update the bar.
    local currentV, maxV = self:GetValue();
    self.counter:SetValue(currentV, maxV);

    -- Update the alternate bar.
    local altBar = self.altBar;
    local currentAltV, maxAltV = self:GetAltValue();
    local fraction = currentAltV / max(1, maxAltV);
    if fraction > 0 then
        altBar:SetWidth(ALTBAR_WIDTH * fraction);
        altBar:Show();
  else
        altBar:Hide();
    end

    -- Update the zoom effect.
    if self.active then
        self.zoom = max(0, self.zoom - elapsed);
  else
        self.zoom = min(ZOOM_TIME, self.zoom + elapsed);
        if self.zoom == ZOOM_TIME then
            self:Hide();
        end
    end
    local zoomProgression = self.zoom / ZOOM_TIME;
    local zoomScale = 1 + 2 * exp(-5 * (1 - zoomProgression));
    local zoomAlpha = 1 - exp(-5 * (1 - zoomProgression));
    local parent = self:GetParent();

    self:SetScale(zoomScale);
    self:SetAlpha(zoomAlpha);
    self:SetPoint("BOTTOM", parent, "BOTTOM", 0, parent:GetHeight() * 0.35);
end

function MirrorBar:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end