﻿local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

local MusicManager = GUI:GetOrNewWidget("MusicManager");

MusicManager.TEMPLATE = "BossEncounter3_MusicManagerTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- A music manager allows to check and toggle music plugins.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local MAX_MUSICS = 25;

local NAME_SCALE = 0.75;
local AUTHOR_SCALE = 0.50;
local PAGE_SCALE = 0.50;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function SetAll(self, state)
    local _, _, numMusics = Resources:GetPluginInfo(self.page);
    local i;
    for i=1, numMusics do
        Resources:SetPluginMusicState(self.page, i, state);
    end
    self:Refresh();
end

-- ********************************************************************
-- * Open the music manager.                                          *
-- ********************************************************************
function MusicManager:Open()
    if self:ScrollIn() then
        self.numPages = Resources:GetNumPlugins();
        self:MovePage(0);
    end
end

-- ********************************************************************
-- * Close the music manager.                                         *
-- ********************************************************************
function MusicManager:Close()
    self:ScrollOut();
end

-- ********************************************************************
-- * Move between pages.                                              *
-- ********************************************************************
function MusicManager:MovePage(delta)
    self.page = min(self.numPages, max(1, self.page + delta));

    -- Update page buttons
    if self.page > 1 then
        self.prevPageButton:Enable();
  else
        self.prevPageButton:Disable();
    end
    if self.page < self.numPages then
        self.nextPageButton:Enable();
  else
        self.nextPageButton:Disable();
    end

    -- Page counter
    self.pageText:SetText(L:FormatLoc("Plugin-Page", self.page, self.numPages));

    -- And now refresh!
    self:Refresh();
end

-- ********************************************************************
-- * Adjust the current page.                                         *
-- ********************************************************************
function MusicManager:Refresh()
    local page = self.page;
    local name, author, musics = Resources:GetPluginInfo(page);

    -- Set name and author texts
    if name and author then
        self.name:SetText(name);
        self.author:SetText(L:FormatLoc("Plugin-Author", author));
        self.enableButton:Enable();
        self.disableButton:Enable();
  else
        self.name:SetText("");
        self.author:SetText("");
        self.enableButton:Disable();
        self.disableButton:Disable();
    end

    -- Recalculate the list
    local list = self.list;
    local i, row;
    list:Clear();
    if musics > 0 then
        row = self.firstRow;
        row:Setup(0, 0);
        list:Insert(row);
    end
    for i=1, MAX_MUSICS do
        row = self.musicRows[i];
        if i <= musics then
            row:Setup(page, i);
            list:Insert(row);
      else
            row:Hide();
        end
    end

    -- Update the remainder
    local w = max(list:GetWidth(), 512);
    local h = self.header:GetHeight() + self.separator:GetHeight() + list:GetHeight() + 128;
    if list:IsEmpty() then
        h = h + self.noneText:GetHeight();
        self.noneText:Show();
  else
        self.noneText:Hide();  
    end
    self:SetWidth(w);
    self:SetHeight(h);
end

-- ********************************************************************
-- * Get the offsets for each header.                                 *
-- * Return titleOfs, durationOfs, playOfs, useOfs                    *
-- ********************************************************************
function MusicManager:GetOffsets()
    return 8, 368, 444, 508;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function MusicManager:OnLoad()
    local _, separatorFilename = Resources:GetTextureInfo("SEPARATOR");

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, false);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.header = GUI:RequestChild("MonoFont", nil, self, 1);
    self.header:SetPoint("TOP", self, "TOP", 0, -12);
    self.header:SetText(L:Get("Plugin-Header"));
    self.header:Show();
    self.separator = self:CreateTexture(nil, "BACKGROUND");
    self.separator:SetPoint("TOPLEFT", self, "TOPLEFT", 8, -32);
    self.separator:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", -8, -40);
    self.separator:SetTexture(separatorFilename);
    self.name = GUI:RequestChild("MonoFont", nil, self, 1);
    self.name:SetPoint("TOP", self.separator, "BOTTOM", 0, -16);
    self.name:SetColor(1, 1, 0.2);
    self.name:SetScale(NAME_SCALE);
    self.name:Show();
    self.author = GUI:RequestChild("MonoFont", nil, self, 1);
    self.author:SetPoint("TOP", self.name, "BOTTOM", 0, -24);
    self.author:SetColor(0.25, 1, 0.25);
    self.author:SetScale(AUTHOR_SCALE);
    self.author:Show();
    self.list = GUI:RequestChild("List", nil, self, 1);
    self.list:SetPoint("TOPLEFT", self.separator, "BOTTOMLEFT", 0, -64);
    self.list.spacing = 0;
    self.closeButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.closeButton:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -8, 8);
    self.closeButton:SetWidth(80);
    self.closeButton:SetHeight(24);
    self.closeButton:SetText(CLOSE);
    self.closeButton:SetScript("OnClick", self.OnCloseButtonClick);
    self.closeButton:Show();
    self.enableButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.enableButton:SetPoint("RIGHT", self.closeButton, "LEFT", -8, 0);
    self.enableButton:SetWidth(80);
    self.enableButton:SetHeight(24);
    self.enableButton:SetText(L:Get("Plugin-Enable"));
    self.enableButton:SetScript("OnClick", self.OnEnableButtonClick);
    self.enableButton:Show();
    self.disableButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.disableButton:SetPoint("RIGHT", self.enableButton, "LEFT", -8, 0);
    self.disableButton:SetWidth(80);
    self.disableButton:SetHeight(24);
    self.disableButton:SetText(L:Get("Plugin-Disable"));
    self.disableButton:SetScript("OnClick", self.OnDisableButtonClick);
    self.disableButton:Show();
    self.prevPageButton = CreateFrame("Button", nil, self, "BossEncounter3_PrevPageButtonTemplate");
    self.prevPageButton:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 6, 4);
    self.prevPageButton:SetScript("OnClick", self.OnPrevPageButtonClick);
    self.prevPageButton:Show();
    self.pageText = GUI:RequestChild("MonoFont", nil, self, 1);
    self.pageText:SetPoint("LEFT", self.prevPageButton, "RIGHT", 8, 0);
    self.pageText:SetScale(PAGE_SCALE);
    self.pageText:Show();
    self.nextPageButton = CreateFrame("Button", nil, self, "BossEncounter3_NextPageButtonTemplate");
    self.nextPageButton:SetPoint("LEFT", self.pageText, "RIGHT", 8, 0);
    self.nextPageButton:SetScript("OnClick", self.OnNextPageButtonClick);
    self.nextPageButton:Show();
    self.noneText = _G[self:GetName().."_NoneText"];
    self.noneText:SetText(L:Get("Plugin-None"));

    -- Prepare some music rows.
    self.musicRows = { };
    self.firstRow = GUI:RequestChild("MusicRow", self:GetName().."_Row0", self);
    local i;
    for i=1, MAX_MUSICS do
        self.musicRows[i] = GUI:RequestChild("MusicRow", self:GetName().."_Row"..i, self);
    end

    -- ** Properties **
    self.page = 0;
    self.numPages = 0;

    -- ** Initial config **
    Utility:SetBackdrop(self, "OPTION");
end

function MusicManager:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);
end

function MusicManager:OnScrollOutEnd()
    self:Hide();
end

function MusicManager.OnCloseButtonClick(button)
    button:GetParent():Close();
end

function MusicManager.OnEnableButtonClick(button)
    SetAll(button:GetParent(), true);
end

function MusicManager.OnDisableButtonClick(button)
    SetAll(button:GetParent(), false);
end

function MusicManager.OnPrevPageButtonClick(button)
    button:GetParent():MovePage(-1);
end

function MusicManager.OnNextPageButtonClick(button)
    button:GetParent():MovePage(1);
end