local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");

local NormalMessage = GUI:GetOrNewWidget("NormalMessage");

NormalMessage.TYPE = "Button";
NormalMessage.TEMPLATE = "BossEncounter3_NormalMessageTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Display a normal message. Messages can be queued.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local RESIZE_RATE = 1024; -- Pixel per second.
local ALPHA_RATE = 4;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> message: the message to be displayed.                         *
-- * >> duration: duration, can be omitted.                           *
-- ********************************************************************
-- * Add a normal message to the queue.                               *
-- ********************************************************************
function NormalMessage:Add(message, duration)
    duration = duration or (1.5 + #message / 15);
    tinsert(self.queue, {message = message, duration = duration});

    if self:IsReady() then
        self:Pull();
    end
end

-- ********************************************************************
-- * Get the size of the queue.                                       *
-- ********************************************************************
function NormalMessage:GetQueueSize()
    return #self.queue;
end

-- ********************************************************************
-- * Get a message from the queue.                                    *
-- ********************************************************************
function NormalMessage:GetQueueMessage(index)
    return self.queue[index];
end

-- ********************************************************************
-- * Clear current message and queue.                                 *
-- ********************************************************************
function NormalMessage:Clear()
    wipe(self.queue);
    if self.up then
        self:OnMessageExpired();
    end
end

-- ********************************************************************
-- * Check if the container is ready to display another message.      *
-- ********************************************************************
function NormalMessage:IsReady()
    return (not (self:Scrolling() or self.up or self.down));
end

-- ********************************************************************
-- * Pull the latest message in the queue and show it.                *
-- ********************************************************************
function NormalMessage:Pull()
    if self:GetQueueSize() == 0 then
        -- No message, hide the container.
        self:ScrollOut();
  else
        local info = self:GetQueueMessage(1);
        tremove(self.queue, 1);

        self.text:SetText(info.message);
        self:SetHeight(self.text:GetHeight() + 16);

        self.messageWidth = self.text:GetWidth() + 64;
        self.messageDuration = info.duration;
        self.up = true;

        if self:ScrollIn() then
            self.width = self.messageWidth;
        end
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function NormalMessage:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.text = GUI:RequestChild("MonoFont", nil, self, 1);
    self.text:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.text:Show();
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HS");
    self.band:SetFixedBorders(true);

    -- ** Properties **
    self.queue = { };
    self.up = false;
    self.down = false;
    self.alpha = 0;
    self.width = 128;
    self.messageDuration = 0;
    self.messageWidth = 0;

    -- ** Initial config **
end

function NormalMessage:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);

    if not self:Scrolling() then
        if self.width ~= self.messageWidth then
            if self.width < self.messageWidth then
                self.width = min(self.messageWidth, self.width + RESIZE_RATE * elapsed);
          else
                self.width = max(self.messageWidth, self.width - RESIZE_RATE * elapsed);
            end
      else
            if self.up then
                if self.alpha < 1 then
                    self.alpha = min(1, self.alpha + ALPHA_RATE * elapsed);
              else
                    if self.messageDuration > 0 then
                        self.messageDuration = max(0, self.messageDuration - elapsed);
                  else
                        self:OnMessageExpired();
                    end
                end

        elseif self.down then
                if self.alpha > 0 then
                    self.alpha = max(0, self.alpha - ALPHA_RATE * elapsed);
              else
                    self:OnMessageCleaned();
                end
            end
        end
    end

    self.text:SetAlpha(self.alpha);
    self:SetWidth(self.width);
end

function NormalMessage:OnMessageExpired() -- Called when the current message should be disposed of.
    self.up = false;
    self.down = true;
end

function NormalMessage:OnMessageCleaned() -- Called when the current message has been disposed of.
    self.down = false;
    self:Pull(); -- Take the next pending message.
end

function NormalMessage:OnScrollOutEnd()
    self:Hide();
end