local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");

local Band = GUI:GetOrNewWidget("Band");

Band.TEMPLATE = "BossEncounter3_BandTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Band is a GUI primitive element that allows one to display horizontal
-- or vertical bands of black fading to transparent.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local FIXED_BORDER_WIDTH = 32;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> char: the char being drawn.                                   *
-- ********************************************************************
-- * Grab the appropriate draw region for the specified char.         *
-- ********************************************************************
function Band:SetGraphic(name, r, g, b)
    local _, filename, width, height = Resources:GetTextureInfo(name);
    self.textureFilename = filename;
    self.textureWidth = width;
    self.textureHeight = height;
    self.texture:SetTexture(filename);
    self.border1:SetTexture(filename);
    self.border2:SetTexture(filename);

    -- Default color
    self:SetColor(r or 0, g or 0, b or 0);

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: color value.                                         *
-- ********************************************************************
-- * Change the color of the band.                                    *
-- ********************************************************************
function Band:SetColor(r, g, b)
    self.texture:SetVertexColor(r, g, b);
    self.border1:SetVertexColor(r, g, b);
    self.border2:SetVertexColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> flag: whether to enable fixed borders or not.                 *
-- ********************************************************************
-- * Make the borders of the band fixed width.                        *
-- * This is important if you resize skewed bands.                    *
-- ********************************************************************
function Band:SetFixedBorders(flag)
    if (not flag) then
        self.texture:SetAllPoints(true);
        self.border1:Hide();
        self.border2:Hide();
  else
        self.texture:ClearAllPoints();
        self.texture:SetPoint("TOPLEFT", self.border1, "TOPRIGHT");
        self.texture:SetPoint("BOTTOMRIGHT", self.border2, "BOTTOMLEFT");
        self.border1:Show();
        self.border2:Show();
    end
    self.fixedBorders = flag;
    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> flag: whether to enable fixed borders or not.                 *
-- ********************************************************************
-- * Make the borders of the band fixed width.                        *
-- ********************************************************************
function Band:Update()
    if self.fixedBorders then
        local beginX = FIXED_BORDER_WIDTH / self.textureWidth;
        local endX = 1 - beginX;
        self.border1:SetTexCoord(0, beginX, 0, 1);
        self.border2:SetTexCoord(endX, 1, 0, 1);
        self.texture:SetTexCoord(beginX, endX, 0, 1);
  else
        self.texture:SetTexCoord(0, 1, 0, 1);
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Band:OnLoad()
    -- ** Children **
    self.texture = self:CreateTexture(nil, "BACKGROUND");
    self.border1 = self:CreateTexture(nil, "BACKGROUND");
    self.border1:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0);
    self.border1:SetPoint("BOTTOMRIGHT", self, "BOTTOMLEFT", FIXED_BORDER_WIDTH, 0);
    self.border2 = self:CreateTexture(nil, "BACKGROUND");
    self.border2:SetPoint("TOPLEFT", self, "TOPRIGHT", -FIXED_BORDER_WIDTH, 0);
    self.border2:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0);

    -- ** Properties **
    self.textureFilename = "";
    self.textureWidth = width;
    self.textureHeight = height;
    self.fixedBorders = false;

    -- ** Initial config **
    self:SetFixedBorders(false);
    self:SetGraphic("BAND_HS");
    self:Show();
end

function Band:OnUpdate(elapsed)

end