local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");

local Bar = GUI:GetOrNewWidget("Bar");

Bar.TEMPLATE = "BossEncounter3_BarTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Bar is a GUI primitive element that allows to display... bars.

-- It supports several styles, resizing and recoloring.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local styles = {
    ["STANDARD"] = {
        border = "BAR_BORDER",
        inside = "BAR_INSIDE",
        fill = "BAR_FILL",
        stretch = false,
        edgeWidth = 16,
        insets = {2, 4},
    },
    ["SIMPLE"] = {
        inside = "SIMPLEBAR_INSIDE",
        fill = "SIMPLEBAR_FILL",
        stretch = false,
        edgeWidth = 16,
        insets = {2, 2},
    },
};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> style: the bar style name. See above.                         *
-- ********************************************************************
-- * Set the bar style. This will resize the bar.                     *
-- ********************************************************************
function Bar:SetStyle(style)
    local info = styles[style];
    if not info then return; end

    if info.border then
        local _, borderFilename, borderWidth, borderHeight = Resources:GetTextureInfo(info.border);
        self.borderTexture = borderFilename;
        self.borderWidth = borderWidth;
        self.borderHeight = borderHeight;
  else
        self.borderTexture = nil;
        self.borderWidth = 0;
        self.borderHeight = 0;
    end

    local _, insideFilename, insideWidth, insideHeight = Resources:GetTextureInfo(info.inside);
    self.insideTexture = insideFilename;
    self.insideWidth = insideWidth;
    self.insideHeight = insideHeight;

    local _, fillFilename, fillWidth, fillHeight = Resources:GetTextureInfo(info.fill);
    self.fill1:SetTexture(fillFilename);
    self.fill2:SetTexture(fillFilename);
    self.fill1:SetHeight(fillHeight);
    self.fill2:SetHeight(fillHeight);

    self.stretch = info.stretch or false;
    self.edgeWidth = info.edgeWidth or 0;

    self:SetInsets(unpack(info.insets));
    self:SetHeight(fillHeight);
    self:RedrawBorder();
    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width: the new width of the bar.                              *
-- ********************************************************************
-- * Change the width of the bar.                                     *
-- ********************************************************************
function Bar:Resize(width)
    if self.width ~= width then
        self.width = width;
        self:SetWidth(width);
        self:Notify("OnChildResized");
        self:RedrawBorder();
        self:Update();
    end
end

-- ********************************************************************
-- * Redraw the border of the bar.                                    *
-- ********************************************************************
function Bar:RedrawBorder()
    local textureFilename, texture, width, height, layer;
    local edgeWidth = self.edgeWidth;

    if self.borderTexture then
        textureFilename = self.borderTexture;
        texture = self.border;
        width = self.borderWidth;
        height = self.borderHeight;
        layer = "ARTWORK";

        self.border:Show();
        self.inside:SetTexture(self.insideTexture);
        self.inside:SetWidth(self:GetWidth());
        self.inside:SetHeight(self:GetHeight());
  else
        textureFilename = self.insideTexture;
        texture = self.inside;
        width = self.insideWidth;
        height = self.insideHeight;
        layer = "BACKGROUND";

        self.border:Hide();
    end

    if edgeWidth > 0 then
        self.edge1:Show();
        self.edge2:Show();
        self.edge1:SetDrawLayer(layer);
        self.edge2:SetDrawLayer(layer);
        Utility:ShowTextureRegion(self.edge1, textureFilename, width, height,                 0, 0,             edgeWidth, height);
        Utility:ShowTextureRegion(self.edge2, textureFilename, width, height, width - edgeWidth, 0,             edgeWidth, height);
        Utility:ShowTextureRegion(   texture, textureFilename, width, height,         edgeWidth, 0, width - 2 * edgeWidth, height);
  else
        self.edge1:Hide();
        self.edge2:Hide();
        texture:SetTexture(textureFilename);
        texture:SetTexCoord(0, 1, 0, 1);
    end

    texture:SetWidth(self:GetWidth() - 2 * edgeWidth);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> min, max: self-explanatory.                                   *
-- ********************************************************************
-- * Change limits of the bar.                                        *
-- ********************************************************************
function Bar:SetMinMax(min, max)
    self.min = min;
    self.max = max;

    self:SetValue(self.value1, self.value2);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value1/2: the desired new value.                              *
-- * Value2 is optional and grows from the opposite direction.        *
-- ********************************************************************
-- * Change the value of a bar.                                       *
-- ********************************************************************
function Bar:SetValue(value1, value2)
    self.value1 = max(self.min, min(self.max, value1));
    self.value2 = max(self.min, min(self.max, value2 or 0));

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r1, g1, b1: the colors for the first part.                    *
-- * >> r2, g2, b2: the colors for the second part.                   *
-- ********************************************************************
-- * Change the color of one or both parts of the bar.                *
-- ********************************************************************
function Bar:SetColor(r1, g1, b1, r2, g2, b2)
    if ( r1 and g1 and b1 ) then
        self.fill1:SetVertexColor(r1, g1, b1);
    end
    if ( r2 and g2 and b2 ) then
        self.fill2:SetVertexColor(r2, g2, b2);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> left, right: the left/right inset.                            *
-- ********************************************************************
-- * Set the insets of the fill portion of the bar.                   *
-- * Useful if the border hides a significant portion of the bar.     *
-- ********************************************************************
function Bar:SetInsets(left, right)
    self.leftInset = max(0, left);
    self.rightInset = max(0, right);

    self:UpdateDisposition();
end

-- ********************************************************************
-- * Get the insets of the bar.                                       *
-- ********************************************************************
function Bar:GetInsets()
    return self.leftInset, self.rightInset;
end

-- ********************************************************************
-- * Get the min/max values of the bar.                               *
-- ********************************************************************
function Bar:GetMinMax()
    return self.min, self.max;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> disposition: the disposition mode desired.                    *
-- * Can be "EDGE" or "CENTER".                                       *
-- ********************************************************************
-- * Set the disposition mode of the bar.                             *
-- * EDGE: grows from the sides to reach the center.                  *
-- * CENTER: grows from the center to reach the sides.                *
-- ********************************************************************
function Bar:SetDisposition(disposition)
    self.disposition = disposition;

    self:UpdateDisposition();
end

-- ********************************************************************
-- * Update the disposition of the bar, if it has changed.            *
-- ********************************************************************
function Bar:UpdateDisposition()
    self.fill1:ClearAllPoints();
    self.fill2:ClearAllPoints();

    if self.disposition == "EDGE" then
        self.fill1:SetPoint("LEFT",  self,  "LEFT",   self.leftInset, 0);
        self.fill2:SetPoint("RIGHT", self, "RIGHT", -self.rightInset, 0);
  else
        local offset = -(self.leftInset - self.rightInset)/2;
        self.fill1:SetPoint("RIGHT", self, "CENTER", offset, 0);
        self.fill2:SetPoint("LEFT",  self, "CENTER", offset, 0);
    end
end

-- ********************************************************************
-- * Update the bar.                                                  *
-- ********************************************************************
function Bar:Update()
    -- Figure out fill percentages
    local pct1, pct2 = 0, 0;
    local d = self.max - self.min;
    if d > 0 then
        pct1 = min(1, max(0, (self.value1 - self.min) / d));
        pct2 = min(1, max(0, (self.value2 - self.min) / d));
    end

    -- Update fill textures
    if self.disposition == "EDGE" then
        self:UpdateTexture(self.fill1, pct1, "RIGHT");
        self:UpdateTexture(self.fill2, pct2, "LEFT");
  else
        self:UpdateTexture(self.fill1, pct1, "LEFT");
        self:UpdateTexture(self.fill2, pct2, "RIGHT");
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> texture: the texture updated.                                 *
-- * >> fraction: the fraction of the texture shown.                  *
-- * >> expend: the direction toward which the texture expends.       *
-- ********************************************************************
-- * Update a texture bar.                                            *
-- ********************************************************************
function Bar:UpdateTexture(texture, fraction, expend)
    local usefulWidth = self.width - self.leftInset - self.rightInset;
    if self.disposition == "CENTER" then usefulWidth = usefulWidth / 2; end

    if fraction == 0 or usefulWidth <= 0 then
        texture:Hide();
  else
        local leftOffset = self.leftInset / self.width;
        local rightOffset = self.rightInset / self.width;
        local left = leftOffset;
        local right = 1-rightOffset;

        if self.stretch then
            texture:SetTexCoord(left, right, 0, 1);
      else
            if self.disposition == "CENTER" then
                if expend == "LEFT" then
                    right = right - 0.5;
            elseif expend == "RIGHT" then
                    left = left + 0.5;
                end
            end

            if expend == "RIGHT" then
                texture:SetTexCoord(left, left + (right - left) * fraction, 0, 1);
        elseif expend == "LEFT" then
                texture:SetTexCoord(right + (left - right) * fraction, right, 0, 1);
            end
        end

        texture:SetWidth(usefulWidth * fraction);
        texture:Show();
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Bar:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.border = self:CreateTexture(nil, "ARTWORK");
    self.border:ClearAllPoints();
    self.border:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.edge1 = self:CreateTexture(nil, "ARTWORK");
    self.edge1:ClearAllPoints();
    self.edge1:SetPoint("LEFT", self, "LEFT", 0, 0);
    self.edge2 = self:CreateTexture(nil, "ARTWORK");
    self.edge2:ClearAllPoints();
    self.edge2:SetPoint("RIGHT", self, "RIGHT", 0, 0);
    self.inside = self:CreateTexture(nil, "BACKGROUND");
    self.inside:ClearAllPoints();
    self.inside:SetPoint("CENTER", self, "CENTER", 0, 0);
    self.fill1 = self:CreateTexture(nil, "BORDER");
    self.fill2 = self:CreateTexture(nil, "BORDER");

    -- ** Properties **
    self.width = 256;
    self.min = 0;
    self.max = 1;
    self.value1 = 0;
    self.value2 = 0;
    self.leftInset = 0;
    self.rightInset = 0;
    self.disposition = "EDGE";
    self.stretch = false;
    self.edgeWidth = 0;
    self.insideTexture = nil;
    self.insideWidth = 0;
    self.insideHeight = 0;
    self.borderTexture = nil;
    self.borderWidth = 0;
    self.borderHeight = 0;

    -- ** Initial config **
    self:SetStyle("STANDARD");
    self:SetColor(1, 0, 0);
    self:SetDisposition("EDGE");
    self:SetMinMax(0, 1);
    self:Show();
end