local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local Counter = GUI:GetOrNewWidget("Counter");

Counter.TEMPLATE = "BossEncounter3_CounterTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Counter is a GUI primitive element based on a Bar and a Number.
-- It allows to display a value with a Number display of the value on
-- its side.

-- It comes with a value change smoothing effect and a percentage display.
-- It can also use kilo and million when the value is too high.
-- You can access the Bar object to customize it further.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local NUMBER_SCALE = 0.75;
local PERCENTAGE_SCALE = 0.50;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Get the bar object.                                              *
-- ********************************************************************
function Counter:GetBar()
    return self.bar;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width: the new width of the bar.                              *
-- ********************************************************************
-- * Change the width of the bar within the counter.                  *
-- ********************************************************************
function Counter:ResizeBar(width)
    self.bar:Resize(width);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> current: self-explanatory                                     *
-- * >> max: self-explanatory. Optionnal.                             *
-- ********************************************************************
-- * Change the current value / max of the counter.                   *
-- ********************************************************************
function Counter:SetValue(current, max)
    if max then
        self.max = max;
        self.current = min(self.current, max);
        self.bar:SetMinMax(0, max);
    end
    self.target = math.max(0, min(self.max, current));
end

-- ********************************************************************
-- * Get the current value.                                           *
-- ********************************************************************
function Counter:GetValue()
    return self.target;
end

-- ********************************************************************
-- * Get the current displayed value.                                 *
-- ********************************************************************
function Counter:GetDisplayValue()
    return self.current;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> rate: the new rate.                                           *
-- ********************************************************************
-- * Change the rate % of the counter.                                *
-- ********************************************************************
function Counter:SetRate(rate)
    self.rate = rate;
end

-- ********************************************************************
-- * This will make the counter updates its value instantly.          *
-- ********************************************************************
function Counter:SetInfiniteRate()
    self.rate = math.huge;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> flag: to hide or not to hide?                                 *
-- ********************************************************************
-- * Hide or show the counter.                                        *
-- ********************************************************************
function Counter:SetHidden(flag)
    self.hidden = flag;
    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: the new color.                                       *
-- ********************************************************************
-- * Set the color of the text of the counter.                        *
-- ********************************************************************
function Counter:SetTextColor(r, g, b)
    self.number:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: the new color.                                       *
-- ********************************************************************
-- * Set the color of the bar of the counter.                         *
-- ********************************************************************
function Counter:SetBarColor(r, g, b)
    self.bar:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> maxDigits: max allowed digits displayable.                    *
-- * >> useMultiples: if we go above the max value displayable, this  *
-- *                  allows use to use kilo and million multiples.   *
-- ********************************************************************
-- * Set the max values displayable with the number display.          *
-- ********************************************************************
function Counter:SetDimensions(maxDigits, useMultiples)
    self.maxDigits = maxDigits;
    self.useMultiples = useMultiples;
    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> state: true if percentage should be displayed on the bar.     *
-- ********************************************************************
-- * Set if the percentage should be shown on the bar.                *
-- ********************************************************************
function Counter:TogglePercentage(state)
    if state then
        self.showPercentage = true;
        self.percentage:Show();
  else
        self.showPercentage = false;
        self.percentage:Hide();
    end
    self:Update();
end

-- ********************************************************************
-- * Update the counter.                                              *
-- ********************************************************************
function Counter:Update()
    local maxDigits = self.maxDigits;
    local max = 10 ^ maxDigits;
    local value = self.current;
    local overflow = false;
    local suffix = "";

    if value >= max and self.useMultiples then
        suffix = "K";
        value = floor(value / 1000 + 0.5);
        maxDigits = maxDigits - 1;
        max = 10 ^ maxDigits;
    end
    if value >= max and self.useMultiples then
        suffix = "M";
        value = floor(value / 1000 + 0.5);
    end
    if value >= max then
        overflow = true;
    end

    -- Update number display
    local n = self.number;
    n:SetSuffix(suffix);
    n:SetPadding(maxDigits);
    n:SetHidden(self.hidden or overflow);
    n:SetValue(value);

    -- Update bar display
    if not self.hidden then
        self.bar:SetValue(self.current);
  else
        self.bar:SetValue(0);
    end

    -- Update percentage display
    local p = self.percentage;
    if self.showPercentage then
        if self.hidden then
            p:SetPadding(1);
            p:SetHidden(true);
      else
            p:SetValue(floor(self.current * 100 / math.max(self.max, 1) + 0.5));
            p:SetPadding(3);
            p:SetHidden(false);
        end
    end
end

-- ********************************************************************
-- * Update the size of the counter.                                  *
-- ********************************************************************
function Counter:UpdateSize()
    local w = self.bar:GetWidth() + self.number:GetWidth() * NUMBER_SCALE;
    local h = max(self.number:GetHeight() * NUMBER_SCALE, self.bar:GetHeight());
    self:SetWidth(w);
    self:SetHeight(h);

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Counter:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.number = GUI:RequestChild("Number", nil, self, 1);
    self.number:SetPoint("RIGHT", self, "RIGHT", 0, 0);
    self.number:SetScale(NUMBER_SCALE);
    self.bar = GUI:RequestChild("Bar", nil, self, 1);
    self.bar:SetPoint("LEFT", self, "LEFT", 0, 0);
    self.percentage = GUI:RequestChild("Number", nil, self, 2);
    self.percentage:SetPoint("CENTER", self.bar, "CENTER", 0, 0);
    self.percentage:SetScale(PERCENTAGE_SCALE);

    -- ** Properties **
    self.max = 1;
    self.current = 0;
    self.target = 0;
    self.rate = 0;
    self.hidden = false;
    self.maxDigits = 0;
    self.useMultiples = false;
    self.showPercentage = false;

    -- ** Initial config **
    self:SetInfiniteRate();
    self:ResizeBar(192);
    self:SetDimensions(5, false);
    self:TogglePercentage(false);
    self:SetValue(0, 1);
    self:UpdateSize();
    self:Show();
end

function Counter:OnUpdate(elapsed)
    if self.current < self.target then
        self.current = min(self.target, self.current + self.rate * self.max * elapsed);

elseif self.current > self.target then
        self.current = max(self.target, self.current - self.rate * self.max * elapsed);
    end

    self:Update();
end

function Counter:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end