local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");

local MonoFont = GUI:GetOrNewWidget("MonoFont");

MonoFont.TEMPLATE = nil;

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- MonoFont is a GUI primitive element that allows one to display messages
-- through a graphical mono-spaced font.
-- It is not suited for long and rich block of texts.
-- It should be used for simple messages and headers.
-- The text is displayed on a single line.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> fontName: font internal name.                                 *
-- * >> space: space, if desired. Default to 0.                       *
-- ********************************************************************
-- * Set the font that will be used for the mono font.                *
-- ********************************************************************
function MonoFont:SetFont(fontName, space)
    local name, filename, isGraphic, imgWidth, imgHeight, charWidth, charHeight, charPerLine, charPerColumn, charMap = Resources:GetFontInfo(fontName);
    if not isGraphic or not filename then return; end

    self.changed = true;
    self.texture = filename;
    self.charWidth = charWidth;
    self.charHeight = charHeight;
    self.charPerLine = charPerLine;
    self.charPerColumn = charPerColumn;
    self.usefulHorizontal = charWidth * charPerLine / imgWidth;
    self.usefulVertical = charHeight * charPerColumn / imgHeight;
    self.drawWidth = charWidth;
    self.drawHeight = charHeight;
    self.drawSpace = space or 0;
    self.pixelX = 1 / imgWidth;
    self.pixelY = 1 / imgHeight;
    self.matrix = charMap;

    local i;
    for i=1, #self do
        self[i]:SetTexture(self.texture);
    end

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width, height: the new desired dimensions.                    *
-- * >> space: space, if desired. Default to 0.                       *
-- ********************************************************************
-- * Resize the texture render region. Useful to scale down.          *
-- ********************************************************************
function MonoFont:Resize(width, height, space)
    self.changed = true;
    self.drawWidth = width;
    self.drawHeight = height;
    self.drawSpace = space or 0;

    local i, t;
    for i=1, #self do
        t = self[i];
        t:SetWidth(self.drawWidth);
        t:SetHeight(self.drawHeight);
    end

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> char: the char being drawn.                                   *
-- ********************************************************************
-- * Grab the appropriate draw region for the specified char.         *
-- ********************************************************************
function MonoFont:GetTextureCoord(char)
    local m = self.matrix;
    if m then
        local t = m[char];
        if t then
            local x, y = unpack(t);
            local mx, my = self.charPerLine, self.charPerColumn;
            local ux, uy = self.usefulHorizontal, self.usefulVertical;

            local left, right = x / mx * ux, (x + 1) / mx * ux;
            local top, bottom = y / my * uy, (y + 1) / my * uy;
            local correctionX = self.pixelX / 2;
            local correctionY = self.pixelY / 2;

            return left + correctionX, right - correctionX, top + correctionY, bottom - correctionY;
        end
    end
    return nil, nil, nil, nil;
end

-- ********************************************************************
-- * Get length of the text.                                          *
-- ********************************************************************
function MonoFont:GetLength()
    return #self.text;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> text: the text shown.                                         *
-- ********************************************************************
-- * Change the text of a mono font.                                  *
-- ********************************************************************
function MonoFont:SetText(text)
    if not text then
        text = '';
  else
        text = tostring(text);
        text = Utility:ToFlatString(text);
        text = Utility:ToCleanString(text);
    end

    self.changed = true;
    self.text = text;

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: color value.                                         *
-- ********************************************************************
-- * Change the color of a mono font.                                 *
-- ********************************************************************
function MonoFont:SetColor(r, g, b)
    self.r, self.g, self.b = r, g, b;

    local i;
    for i=1, #self do
        self[i]:SetVertexColor(self.r, self.g, self.b);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the n-th char we want to get.                          *
-- ********************************************************************
-- * Grab a specific char texture within the mono font.               *
-- ********************************************************************
function MonoFont:GetCharTexture(index)
    local t = self[index];
    if not t then
        t = self:CreateTexture(nil, "BACKGROUND");
        t:SetTexture(self.texture);
        t:SetVertexColor(self.r, self.g, self.b);
        t:SetWidth(self.drawWidth);
        t:SetHeight(self.drawHeight);
        t:ClearAllPoints();
        self[index] = t;
    end
    return t;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> texture: the texture being changed.                           *
-- * >> char: the char to display on the texture.                     *
-- ********************************************************************
-- * Grab a specific char texture within the mono font.               *
-- ********************************************************************
function MonoFont:SetCharTexture(index, char, offset)
    local t = self:GetCharTexture(index);
    local left, right, top, bottom = self:GetTextureCoord(char);
    if left then
        t:SetTexCoord(left, right, top, bottom);
        t:SetPoint("LEFT", self, "LEFT", offset, 0);
        t:Show();
        return true;
   else
        t:Hide();
        return false;
    end
end

-- ********************************************************************
-- * Update the mono font frame.                                      *
-- ********************************************************************
function MonoFont:Update()
    if not self.changed then return; end
    self.changed = false;

    local numChars = #self.text;
    local numTextures = #self;
    local x = 0;

    local i, j, k, char;
    j = 0;

    for i=1, max(numChars, numTextures) do
        char = (self.text):sub(i, i);
        if char == ' ' then
            x = x + self.drawWidth;
      else
            -- Char can be nil or a valid char at this point.
            j = j + 1;

            if self:SetCharTexture(j, char, x) then
                x = x + self.drawWidth;
            end
        end
        if i < numChars then
            x = x + self.drawSpace;
        end
    end
    for k=(j+1), numTextures do
        self:SetCharTexture(k, nil, 0);
    end

    self:SetWidth(x);
    self:SetHeight(self.drawHeight);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function MonoFont:OnLoad()
    -- ** Properties **
    self.changed = true;
    self.text = '';
    self.texture = nil;
    self.charWidth = 0;
    self.charHeight = 0;
    self.charPerLine = 0;
    self.charPerColumn = 0;
    self.usefulHorizontal = 0;
    self.usefulVertical = 0;
    self.drawWidth = 0;
    self.drawHeight = 0;
    self.drawSpace = 0;
    self.pixelX = 0;
    self.pixelY = 0;
    self.matrix = nil;
    self.r, self.g, self.b = 1, 1, 1;

    -- ** Initial config **
    self:SetFont("CYBERSPACE");
    self:Show();
end