local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local Number = GUI:GetOrNewWidget("Number");

Number.TEMPLATE = "BossEncounter3_NumberTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Number is a specialized GUI primitive element expanding on MonoFont.
-- It allows to display counters, countdowns or simple digits using
-- a graphical mono-spaced font.

-- It can be padded with zeros.
-- Extra zeros can have a different alpha value.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> fontName: font internal name.                                 *
-- * >> space: space, if desired. Default to 0.                       *
-- ********************************************************************
-- * Set the font that will be used for the number frame.             *
-- ********************************************************************
function Number:SetFont(fontName, space)
    self.font:SetFont(fontName, space);

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> width, height: the new desired dimensions.                    *
-- * >> space: space, if desired. Default to 0.                       *
-- ********************************************************************
-- * Resize the texture render region. Useful to scale down.          *
-- ********************************************************************
function Number:Resize(width, height, space)
    self.font:Resize(width, height, space);

    self:Update();
end

-- ********************************************************************
-- * Get the value of a number frame.                                 *
-- ********************************************************************
function Number:GetValue()
    return self.value;
end

-- ********************************************************************
-- * Get the displayed value of a number frame.                       *
-- ********************************************************************
function Number:GetDisplayValue()
    return self.display;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new value.                                         *
-- ********************************************************************
-- * Change the value of a number frame.                              *
-- ********************************************************************
function Number:SetValue(value)
    value = tonumber(value);
    if not value then value = 0; end

    self.value = value;

    if self.dv == 0 then
        -- No smoothing: instant update.
        self:SetDisplayValue(value);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> text: the suffix text.                                        *
-- ********************************************************************
-- * Set the suffix of a number frame.                                *
-- ********************************************************************
function Number:SetSuffix(text)
    self.suffix = text;

    self:SetDisplayValue(self.display);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the value shown.                                       *
-- ********************************************************************
-- * Change the display value of a number frame.                      *
-- ********************************************************************
function Number:SetDisplayValue(value)
    value = tonumber(value);
    if not value then value = 0; end

    self.display = value;

    local numDisplayDigits = self:GetDisplayNumDigits();
    local numDigits = self:GetNumDigits();
    local padThreshold = numDisplayDigits - numDigits;

    local str = '';
    local i, digit;
    for i=numDisplayDigits, 1, -1 do
        if not self:IsHidden() then
            digit = self:GetDigitValue(i);
            if digit then
                str = str..digit;
            end
      else
            str = str..'?';
        end
    end
    str = str..self.suffix;
    self.font:SetText(str);

    local i, texture;
    for i=1, numDisplayDigits do
        texture = self.font:GetCharTexture(i);
        if i > padThreshold and (not self:IsHidden()) then
            -- Normal digit.
            texture:SetAlpha(1);
      else
            -- Padded digit.
            texture:SetAlpha(self.padAlpha);
        end
    end

    self:Update();
end

-- ********************************************************************
-- * Check if the number is hidden or not.                            *
-- ********************************************************************
function Number:IsHidden(flag)
    return self.hidden;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> flag: whether the number is hidden or not.                    *
-- ********************************************************************
-- * Flag the number as hidden or not.                                *
-- ********************************************************************
function Number:SetHidden(flag)
    self.hidden = flag;

    self:SetDisplayValue(self.display);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: color value.                                         *
-- ********************************************************************
-- * Change the color of a mono font.                                 *
-- ********************************************************************
function Number:SetColor(r, g, b)
    self.font:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> digits: the number of digits always shown.                    *
-- ********************************************************************
-- * Set the number of digits of padding when value is too small.     *
-- ********************************************************************
function Number:SetPadding(digits, alpha)
    self.pad = max(1, digits);
    self.padAlpha = alpha or 0.5;

    self:SetDisplayValue(self.display);
end

-- ********************************************************************
-- * Get the number of digits of padding.                             *
-- ********************************************************************
function Number:GetPadding()
    return self.pad;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> dv: change of value per second.                               *
-- ********************************************************************
-- * Allow to smoothen transitions between 2 values.                  *
-- ********************************************************************
function Number:EnableSmooth(dv)
    self.dv = dv;
end

-- ********************************************************************
-- * Disable smooth transitions between 2 values.                     *
-- ********************************************************************
function Number:DisableSmooth()
    self.dv = 0;

    self:SetDisplayValue(self.value);
end

-- ********************************************************************
-- * Get the number of digits in the number.                          *
-- ********************************************************************
function Number:GetNumDigits()
    if self:IsHidden() then
        return self.pad;
  else
        local digits = 1;
        local value = self.display;
        while (value >= 10) do
            digits = digits + 1;
            value = value / 10;
        end
        return digits;
    end
end

-- ********************************************************************
-- * Get the number of digits in the number at display.               *
-- ********************************************************************
function Number:GetDisplayNumDigits()
    return max(self:GetNumDigits(), self.pad);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the n-th digit we want to get the value of.            *
-- ********************************************************************
-- * Get the value a specified digit should display.                  *
-- * Nil if the digit should be hidden.                               *
-- ********************************************************************
function Number:GetDigitValue(index)
    local modulo = 10 ^ (index - 1);
    local value = self.display;
    if value < modulo then
        if index <= self.pad then
            return 0;
      else
            return nil;
        end
  else
        value = floor(value / modulo);
        return math.fmod(value, 10);
    end
end

-- ********************************************************************
-- * Update the number frame.                                         *
-- ********************************************************************
function Number:Update()
    local f = self.font;

    f:Update();

    if self:GetWidth() ~= f:GetWidth() or self:GetHeight() ~= f:GetHeight() then
        self:SetWidth(f:GetWidth());
        self:SetHeight(f:GetHeight());

        self:Notify("OnChildResized");
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Number:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.font = GUI:RequestChild("MonoFont", nil, self, 0);
    self.font:SetPoint("LEFT", self, "LEFT", 0, 0);

    -- ** Properties **
    self.value = 0;
    self.display = 0;
    self.dv = 0;
    self.pad = 1;
    self.padAlpha = 1;
    self.hidden = false;
    self.suffix = "";

    -- ** Initial config **
    self:SetPadding(1);
    self:Show();
end

function Number:OnUpdate(elapsed)
    if self.display < self.value then
        self:SetDisplayValue( min(self.value, self.display + self.dv * elapsed) );

elseif self.display > self.value then
        self:SetDisplayValue( max(self.value, self.display - self.dv * elapsed) );
    end
end