local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

local Threshold = GUI:GetOrNewWidget("Threshold");

Threshold.TEMPLATE = "BossEncounter3_ThresholdTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Threshold is a GUI primitive element that allows to display
-- threshold values on a bar.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local BOUNCE_CYCLE = 1.00;

local ALPHA_RATE = 2.00;
local VALUE_RATE = 0.50;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> frame: the bound frame. Generally a bar.                      *
-- * >> leftInset, rightInset: the insets (default is 0).             *
-- ********************************************************************
-- * Set the frame on which depends the threshold object.             *
-- ********************************************************************
function Threshold:Bind(frame, leftInset, rightInset)
    self.frame = frame;
    self.leftInset = leftInset or 0;
    self.rightInset = rightInset or 0;

    -- Set the size of the tag.
    local height = frame:GetHeight()* 1.25;
    local width = height / 2;
    self.tag:SetWidth(width);
    self.tag:SetHeight(height);
    self.tag:Hide();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> label, value: infos about the threshold.                      *
-- * >> showValue: set if the % should be explicitly shown.           *
-- ********************************************************************
-- * Enable the threshold.                                            *
-- ********************************************************************
function Threshold:Enable(label, value, showValue)
    if showValue then
        self.text:SetText(L:FormatLoc("ThresholdFull", label, 100 * value));
        self.tag:Show();
  else
        self.text:SetText(L:FormatLoc("ThresholdShort", label));
        self.tag:Hide();
    end

    self.value = value;
    self:SetWidth(self.text:GetStringWidth() + 16);

    if not self:IsShown() then
        self:Show();
  else
        self.shown = true;
    end
end

-- ********************************************************************
-- * Disable the threshold.                                           *
-- ********************************************************************
function Threshold:Disable()
    self.shown = false;
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Threshold:OnLoad()
    local _, thresholdFilename = Resources:GetTextureInfo("THRESHOLD");

    -- ** Children **
    self.tag = self:CreateTexture(nil, "ARTWORK");
    self.tag:ClearAllPoints();
    self.tag:SetTexture(thresholdFilename);
    self.text = self:CreateFontString(self, "ARTWORK", "GameFontNormalSmall");
    self.text:SetPoint("CENTER", self, "CENTER", 0, 0);

    -- ** Properties **
    self.frame = nil;
    self.leftInset = 0;
    self.rightInset = 0;
    self.shown = false;
    self.alpha = 0;
    self.value = 0;
    self.position = 0;

    -- ** Initial config **
    self:SetHeight(16);
    Utility:SetBackdrop(self, "BUBBLE");
    self:Show();
end

function Threshold:OnUpdate(elapsed)
    local sway = abs(sin(GetTime() * 180 / BOUNCE_CYCLE));

    -- Alpha update
    if self.shown then
        self.alpha = min(1, self.alpha + elapsed * ALPHA_RATE);
  else
        self.alpha = max(0, self.alpha - elapsed * ALPHA_RATE);
        if self.alpha == 0 then
            self:Hide();
        end
    end
    self:SetAlpha(self.alpha);

    -- Position update
    if self.frame then
        self:SetPoint("BOTTOM", self.frame, "TOP", 0, 8 * sway);

        -- Tag update
        if self.tag:IsShown() then
            if self.position < self.value then
                self.position = min(self.value, self.position + elapsed * VALUE_RATE);

        elseif self.position > self.value then
                self.position = max(self.value, self.position - elapsed * VALUE_RATE);
            end
            self.tag:SetPoint("CENTER", self.frame, "LEFT", self.leftInset + (self.frame:GetWidth() - self.leftInset - self.rightInset) * self.position, 0);
        end
    end
end

function Threshold:OnShow()
    self.shown = true;
    self.alpha = 0;
    self:OnUpdate(0);
end