local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");

local Timer = GUI:GetOrNewWidget("Timer");

Timer.TEMPLATE = "BossEncounter3_TimerTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Timer is a very specialized GUI primitive element expanding on Number.
-- It allows to display timers (such as countdowns) using graphical
-- mono-spaced numbers.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local SUB_SCALE = 0.75;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> fontName: font internal name.                                 *
-- * >> space: space, if desired. Default to 0.                       *
-- ********************************************************************
-- * Set the font that will be used for the timer frame.              *
-- ********************************************************************
function Timer:SetFont(fontName, space)
    self.minNumber:SetFont(fontName, space);
    self.secNumber:SetFont(fontName, space);
    self.subNumber:SetFont(fontName, space);
    self.separator:SetFont(fontName, space);
end

-- ********************************************************************
-- * Get the value of a timer frame.                                  *
-- ********************************************************************
function Timer:GetValue()
    return self.value;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new value.                                         *
-- ********************************************************************
-- * Change the value of a timer frame.                               *
-- ********************************************************************
function Timer:SetValue(value)
    self.value = value;

    if self.style == "SHORT" then
        local sec = floor(value);
        self.secNumber:SetValue(sec);
  else
        local min = floor(value / 60);
        local sec = floor(value - min * 60);
        self.minNumber:SetValue(min);
        self.secNumber:SetValue(sec);
    end

    local cen = floor(math.fmod(value * 100, 100));
    self.subNumber:SetValue(cen);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> value: the new value.                                         *
-- ********************************************************************
-- * Set how the timer should vary per second.                        *
-- ********************************************************************
function Timer:SetDelta(dv)
    self.dv = dv;
end

-- ********************************************************************
-- * Get how the timer vary per second.                               *
-- ********************************************************************
function Timer:GetDelta()
    return self.dv;
end

-- ********************************************************************
-- * Make the timer still.                                            *
-- ********************************************************************
function Timer:Stop()
    self:SetDelta(0);
end

-- ********************************************************************
-- * Clear the timer.                                                 *
-- ********************************************************************
function Timer:Clear()
    self:SetValue(0);
    self:Stop();
end

-- ********************************************************************
-- * Check if the timer display is hidden.                            *
-- ********************************************************************
function Timer:IsDisplayHidden()
    return self.hidden;
end

-- ********************************************************************
-- * Make the timer display ??s instead.                              *
-- ********************************************************************
function Timer:SetHidden(flag)
    self.hidden = flag;

    self.minNumber:SetHidden(flag);
    self.secNumber:SetHidden(flag);

    if flag then
        self.subNumber:Hide();
  else
        self.subNumber:Show();
    end

    self:UpdateSize();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: color value.                                         *
-- ********************************************************************
-- * Change the color of a timer.                                     *
-- ********************************************************************
function Timer:SetColor(r, g, b)
    self.minNumber:SetColor(r, g, b);
    self.secNumber:SetColor(r, g, b);
    self.subNumber:SetColor(r, g, b);
    self.separator:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> style: either SHORT or FULL.                                  *
-- ********************************************************************
-- * Change the style of a timer.                                     *
-- ********************************************************************
function Timer:SetStyle(style)
    self.style = style;

    if style == "SHORT" then
        self.minNumber:Hide();
        self.separator:Hide();
        self.secNumber:SetPoint("LEFT", self, "LEFT", 0, 0);
        self.secNumber:SetPadding(3);

elseif style == "FULL" then
        self.minNumber:Show();
        self.minNumber:SetPoint("LEFT", self, "LEFT", 0, 0);
        self.minNumber:SetPadding(2);
        self.separator:Show();
        self.separator:SetPoint("LEFT", self.minNumber, "RIGHT", 0, 0);
        self.secNumber:SetPoint("LEFT", self.separator, "RIGHT", 0, 0);
        self.secNumber:SetPadding(2, 1);
    end

    self:SetValue(self.value);

    self:UpdateSize();
end

-- ********************************************************************
-- * Update the timer width.                                          *
-- ********************************************************************
function Timer:UpdateSize()
    local mn = self.minNumber;
    local sn = self.secNumber;
    local cn = self.subNumber;

    local width = sn:GetWidth();
    if self.style == "FULL" then
        width = width + mn:GetWidth() + self.separator:GetWidth();
    end
    if not self.hidden then
        width = width + cn:GetWidth() * SUB_SCALE;
    end
    self:SetWidth(width);
    self:SetHeight(mn:GetHeight());

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Timer:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);

    -- ** Children **
    self.minNumber = GUI:RequestChild("Number", nil, self, 0);
    self.minNumber:SetPoint("LEFT", self, "LEFT", 0, 0);
    self.minNumber:SetPadding(3);
    self.secNumber = GUI:RequestChild("Number", nil, self, 0);
    self.secNumber:SetPoint("LEFT", self, "LEFT", 0, 0);
    self.secNumber:SetPadding(3);
    self.subNumber = GUI:RequestChild("Number", nil, self, 0);
    self.subNumber:SetPoint("TOPRIGHT", self, "TOPRIGHT", 0, 0);
    self.subNumber:SetPadding(2, 1);
    self.subNumber:SetScale(SUB_SCALE);
    self.separator = GUI:RequestChild("MonoFont", nil, self, 0);
    self.separator:SetText(":");

    -- ** Properties **
    self.value = 0;
    self.dv = 0;
    self.hidden = false;
    self.style = "";

    -- ** Initial config **
    self:SetStyle("SHORT");
    self:Show();
end

function Timer:OnUpdate(elapsed)
    -- Do nothing
end

function Timer:OnLogicUpdate(elapsed)
    local newValue = max(0, self.value + self.dv * elapsed);
    local expired = (newValue == 0 and self.value > 0);
    self:SetValue(newValue);
    if expired then
        self:Notify("OnTimerExpire", self);
    end
end

function Timer:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end