local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Unit = Root:GetOrNewModule("Unit");
local Table = Root:GetOrNewModule("Table");
local Utility = Root:GetOrNewModule("Utility");

local Radar = GUI:GetOrNewWidget("Radar");

Radar.TYPE = "Button";
Radar.TEMPLATE = "BossEncounter3_RadarTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Radar is a final GUI element allowing to display pins on a map
-- and highlight important areas.
-- Can also provide a fullscreen gradient effect.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local colors = {
    ["none"]   = {1, 1, 1},
    ["green"]  = {0, 1, 0},
    ["yellow"] = {1, 1, 0},
    ["red"]    = {1, 0, 0},
};

local colorPriority = {
    ["none"]   = 0,
    ["green"]  = 1,
    ["yellow"] = 2,
    ["red"]    = 3,
};

local DEFAULT_VISIBILITY = 50;

local DISTANCE_FLASH_CYCLE = 0.50;
local DISTANCE_FLASH_HALF  = 0.25;

local TARGET_SWAY_CYCLE  = 1.00;

local BLINK_CYCLE  = 0.50;
local BLINK_UPTIME = 0.25;

local WARNING_FADE_TIMER = 0.4;
local WARNING_SCROLL_CYCLE = 2.5;
local WARNING_ALPHA = 0.6;

local ZONE_LEVEL         = 1;
local SYMBOL_SELF_LEVEL  = 2;
local DOT_SELF_LEVEL     = 3;
local DOT_LEVEL          = 4;
local SYMBOL_LEVEL       = 5;
local ARROW_LEVEL        = 6;

local DISTANCE_THRESHOLD = 4; -- Difficulty at or below which radar will display distance.
local COLOR_THRESHOLD    = 3; -- Difficulty at or below which radar will use colors for text and inside texture.
local WARNING_THRESHOLD  = 2; -- Difficulty at or below which radar will display fullscreen warnings.
local ARROW_THRESHOLD    = 1; -- Difficulty at or below which radar will display arrows.

local BLACK_COLOR = {r=0, g=0, b=0};

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> color: one of the allowed colors. See above.                  *
-- * >> radius: radius in yards of the zone.                          *
-- * >> x, y: game world coordinates of the zone.                     *
-- * >> duration: duration of the zone.                               *
-- ********************************************************************
-- * Add a fixed zone on the radar.                                   *
-- ********************************************************************
function Radar:AddDirectZone(name, color, radius, x, y, duration)
    local newZone = Table:Alloc();
    newZone.name         = name;
    newZone.color        = color;
    newZone.radius       = radius;
    newZone.x, newZone.y = x, y;
    newZone.duration     = duration;
    newZone.self         = false;
    tinsert(self.zones, newZone);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> color: one of the allowed colors. See above.                  *
-- * >> radius: radius in yards of the zone.                          *
-- * >> mx, my: map coordinates of the zone.                          *
-- * >> duration: duration of the zone.                               *
-- ********************************************************************
-- * Add a fixed zone on the radar.                                   *
-- ********************************************************************
function Radar:AddFixedZone(name, color, radius, mx, my, duration)
    local x, y = self:GetCoordinates(mx, my);
    if x and y then
        self:AddDirectZone(name, color, radius, x, y, duration);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> color: one of the allowed colors. See above.                  *
-- * >> radius: radius in yards of the zone.                          *
-- * >> unit: who the zone is anchored to.                            *
-- * >> duration: duration of the zone.                               *
-- ********************************************************************
-- * Add an anchored zone on the radar.                               *
-- ********************************************************************
function Radar:AddAnchoredZone(name, color, radius, unit, duration)
    unit = Unit:ResolveUID(unit);
    if unit then
        local guid = UnitGUID(unit);
        local newZone = Table:Alloc();
        newZone.name     = name;
        newZone.color    = color;
        newZone.radius   = radius;
        newZone.unit     = guid;
        newZone.duration = duration;
        newZone.self     = (UnitGUID("player") == guid);
        tinsert(self.zones, newZone);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> add: additive component to the zone.                          *
-- * >> factor: multiplicative component to the zone (default: 1).    *
-- ********************************************************************
-- * Change the size of an existing zone on the radar.                *
-- * Generally you'll only want to add A or multiply by B. Use the    *
-- * appropriate values. The addition is done before the product.     *
-- ********************************************************************
function Radar:ModifyZoneSize(name, add, factor)
    local i, zone;
    for i=1, #self.zones do
        zone = self.zones[i];
        if zone.name == name then
            zone.radius = (zone.radius + add) * (factor or 1);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> radius: the new radius.                                       *
-- ********************************************************************
-- * Set the size of an existing zone on the radar.                   *
-- ********************************************************************
function Radar:SetZoneSize(name, radius)
    local i, zone;
    for i=1, #self.zones do
        zone = self.zones[i];
        if zone.name == name then
            zone.radius = radius;
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: name of the zone.                                       *
-- * >> unit: allows to find the correct zone for anchored zones.     *
-- ********************************************************************
-- * Remove a zone from the radar.                                    *
-- ********************************************************************
function Radar:RemoveZone(name, unit)
    local guid = nil;
    unit = Unit:ResolveUID(unit);
    if unit then
        guid = UnitGUID(unit);
    end

    local i, zone;
    for i=#self.zones, 1, -1 do
        zone = self.zones[i];
        if zone.name == name and ((not guid) or (not zone.unit) or zone.unit == guid) then
            tremove(self.zones, i);
            Table:Recycle(zone);
        end
    end
end

-- ********************************************************************
-- * Clear all zones on the radar.                                    *
-- ********************************************************************
function Radar:Clear()
    Table:Recycle(self.zones, true);
    self.zones = Table:Alloc();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> range: the min range.                                         *
-- ********************************************************************
-- * Set the range below which the indicator will be red.             *
-- ********************************************************************
function Radar:SetMinRange(range)
    self.minRange = range;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> visibility: max visibility in yards.                          *
-- ********************************************************************
-- * Set the max visibility of the radar.                             *
-- * Passing nil will restore the default visibility.                 *
-- ********************************************************************
function Radar:SetVisibility(visibility)
    self.visibility = visibility or DEFAULT_VISIBILITY;
end

-- ********************************************************************
-- * Recalculate logic elements of the radar.                         *
-- ********************************************************************
function Radar:Update(elapsed)
    local i, j;

    -- Reset stuff
    Table:Recycle(self.pins, true);
    self.pins = Table:Alloc();
    Table:Recycle(self.arrows, true);
    self.arrows = Table:Alloc();

    Unit:EnumerateRaid();
    wipe(self.units);
    for i=1, Unit:GetEnumerationSize() do
        local uid = Unit:GetEnumerationUID(i);
        if UnitIsUnit(uid, "target") or not UnitIsDeadOrGhost(uid) then -- We don't care about dead players unless we target them.
            tinsert(self.units, uid);
        end
    end

    self.minDistance = nil;
    self.color = "none";
    self.useFullscreenColor = false;

    -- First get position of the player.
    self.px, self.py = self:GetUnitCoordinates("player");

    if self.px and self.py then
        -- Then place pins.
        for i=1, #self.units do
            local unit = self.units[i];
            local newPin = Table:Alloc();
            newPin.name = UnitName(unit);
            newPin.guid = UnitGUID(unit);
            newPin.class = select(2, UnitClass(unit));
            newPin.symbol = GetRaidTargetIndex(unit) or 0;
            newPin.self = (newPin.guid == UnitGUID("player"));
            newPin.target = (newPin.guid == UnitGUID("target"));
            newPin.blink = false;
            newPin.x, newPin.y = self:GetUnitCoordinates(unit);
            if newPin.x and newPin.y then
                local distance = self:GetDistance(self.px, self.py, newPin.x, newPin.y);
                newPin.distance = distance;
                if (not newPin.self) and ((not self.minDistance) or distance < self.minDistance) then
                    self.minDistance = distance;
                end
                tinsert(self.pins, newPin);
            end
        end

        -- Then place zones and arrows.
        for i=1, #self.zones do
            local zone = self.zones[i];
            if zone.unit then
                local unit = Unit:ResolveUID(zone.unit);
                if unit then
                    zone.x, zone.y = self:GetUnitCoordinates(unit);
                end
            end
            if zone.x and zone.y then
                if not zone.self then
                    -- The zone is not anchored on us.
                    local distance = self:GetDistance(self.px, self.py, zone.x, zone.y);

                    if distance < zone.radius then
                        if self.hasColor and colorPriority[zone.color] > colorPriority[self.color] then
                            self.color = zone.color;
                            if not zone.unit then
                                -- Zone is static, should move out of it.
                                self.useFullscreenColor = true;
                            end
                        end
                        if self.hasArrows and zone.color == "red" then
                            self:AddArrow(self:GetAngle(zone.x, zone.y, self.px, self.py), "red");
                        end
                  else
                        if self.hasArrows and zone.color == "green" then
                            self:AddArrow(self:GetAngle(self.px, self.py, zone.x, zone.y), "green");
                        end
                    end
              else
                    -- The zone is anchored on us.
                    for j=1, #self.pins do
                        local pin = self.pins[j];
                        local distance = self:GetDistance(pin.x, pin.y, zone.x, zone.y);

                        if (not pin.self) and distance < zone.radius then
                            if self.hasColor and colorPriority[zone.color] > colorPriority[self.color] then
                                self.color = zone.color;
                                self.useFullscreenColor = true;
                            end
                            if self.hasArrows and zone.color == "red" then
                                pin.blink = true;
                            end
                        end
                    end
                end
            end
        end
    end

    -- Handle expiration of zones.
    for i=#self.zones, 1, -1 do
        local zone = self.zones[i];
        zone.duration = max(0, zone.duration - elapsed);

        if zone.duration == 0 then
            tremove(self.zones, i);
        end
    end
end

-- ********************************************************************
-- * Recalculate graphical elements on the radar.                     *
-- ********************************************************************
function Radar:Render()
    local v = self.visibility;
    local w = self.radarWidth;
    local h = self.radarHeight;
    local r = GetPlayerFacing();
    local blinkShown = math.fmod(GetTime(), BLINK_CYCLE) <= BLINK_UPTIME;

    -- Rotate stuff!
    self.rotation = r;
    -- self.inside:SetRotation(r); -- TODO: buggy & wrong direction

    -- Radar color!
    local color = colors[self.color];
    self.inside:SetVertexColor(unpack(color));

    -- Update distance counter.
    if self.hasDistance and self.minDistance and self.minDistance < 1000 then
        self.range:SetHidden(false);
        self.range:SetValue(self.minDistance);
        if self.hasColor and self.minDistance < self.minRange and UnitAffectingCombat("player") and math.fmod(GetTime(), DISTANCE_FLASH_CYCLE) < DISTANCE_FLASH_HALF then
            self.range:SetColor(1, 0, 0);
      else
            self.range:SetColor(1, 1, 1);
        end
  else
        self.range:SetHidden(true);
        self.range:SetColor(1, 1, 1);
    end

    -- Render pins
    local i, pin, pinTexture;
    for i=1, max(#self.pinTextures, #self.pins) do
        pin = self.pins[i];
        pinTexture = self:GetPinTexture(i);

        if pin and (not pin.blink or blinkShown) then
            local scale, r, g, b;
            if pin.symbol > 0 then
                r, g, b = 1, 1, 1;
                Utility:SetIcon(pinTexture, pin.symbol);
                pinTexture:SetVertexColor(1, 1, 1);
                if pin.self then
                    pinTexture:SetDrawLayer("ARTWORK", SYMBOL_SELF_LEVEL);
                    scale = 2.5;
              else
                    pinTexture:SetDrawLayer("ARTWORK", SYMBOL_LEVEL);
                    scale = 1.75;
                end
          else
                local color = RAID_CLASS_COLORS[pin.class] or BLACK_COLOR;
                r, g, b = color.r, color.g, color.b;
                pinTexture:SetTexture(self.pinFilename);
                pinTexture:SetTexCoord(0, 1, 0, 1);
                if pin.self then
                    pinTexture:SetDrawLayer("ARTWORK", DOT_SELF_LEVEL);
                    scale = 1.5;
              else
                    pinTexture:SetDrawLayer("ARTWORK", DOT_LEVEL);
                    scale = 1;
                end
            end
            if pin.target then
                local sway = abs(sin(GetTime() * 180 / TARGET_SWAY_CYCLE));
                r, g, b = sway * r, sway * g, sway * b;
                scale = scale * (2 - sway);
            end
            pinTexture:SetWidth(6 * scale);
            pinTexture:SetHeight(6 * scale);
            pinTexture:SetVertexColor(r, g, b);
            self:SetElementPosition(pinTexture, pin.x, pin.y);
      else
            pinTexture:Hide();
        end
    end

    -- Render zones
    local i, zone, zoneTexture;
    for i=1, max(#self.zoneTextures, #self.zones) do
        zone = self.zones[i];
        zoneTexture = self:GetZoneTexture(i);

        if zone and zone.x and zone.y then
            local color = colors[zone.color];
            local sizeX = zone.radius / v * w;
            local sizeY = zone.radius / v * h; -- Generally equal to sizeX (radar is a circle).
            zoneTexture:SetVertexColor(unpack(color));
            zoneTexture:SetWidth(sizeX);
            zoneTexture:SetHeight(sizeY);
            self:SetElementPosition(zoneTexture, zone.x, zone.y, 0.5);
      else
            zoneTexture:Hide();
        end
    end

    -- Render arrows
    local i, arrow, arrowTexture;
    for i=1, max(#self.arrowTextures, #self.arrows) do
        arrow = self.arrows[i];
        arrowTexture = self:GetArrowTexture(i);

        if arrow and self.px and self.py then
            local color = colors[arrow.color];
            local pi = math.pi;
            local a = arrow.angle / 180 * pi;
            local x = self.px + math.cos(a) * 1/5 * v;
            local y = self.py + math.sin(a) * 1/5 * v;
            arrowTexture:SetVertexColor(unpack(color));
            arrowTexture:SetRotation(a - r - pi/2);
            self:SetElementPosition(arrowTexture, x, y);
      else
            arrowTexture:Hide();
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> element: the element being placed.                            *
-- * >> x, y: coordinates of the point to position the element at.    *
-- * >> baseAlpha: base alpha value of the element. Default is 1.     *
-- ********************************************************************
-- * Set the position of an element.                                  *
-- ********************************************************************
function Radar:SetElementPosition(element, x, y, baseAlpha)
    if self.px and self.py then
        local v = self.visibility;
        local a = self.rotation;
        local c = math.cos(a);
        local s = math.sin(a);
        local ox = x - self.px;
        local oy = y - self.py;

        x = ox * c + oy * s;
        y = oy * c - ox * s;
        baseAlpha = baseAlpha or 1;

        local d = (x^2 + y^2)^0.5;
        if d <= v then
            local dx = x/(v*2) * self.radarWidth;
            local dy = y/(v*2) * self.radarHeight;
            local t = 4/5 * v;
            local alpha;
            if d > t then
                alpha = 1 - min(1, (d - t) / (v - t));
          else
                alpha = 1;
            end

            element:SetPoint("CENTER", self.inside, "CENTER", dx, dy);
            element:SetAlpha(alpha * baseAlpha);
            element:Show();
            return true;
        end
    end

    element:Hide();
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> angle: direction the arrow points at. 0 is top.               *
-- * >> color: the color of the arrow.                                *
-- ********************************************************************
-- * Add an arrow on the radar.                                       *
-- ********************************************************************
function Radar:AddArrow(angle, color)
    local newArrow = Table:Alloc();
    newArrow.angle = angle;
    newArrow.color = color;
    tinsert(self.arrows, newArrow);
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole radar.            *
-- ********************************************************************
function Radar:UpdateSize()
    self.inside:SetWidth(self.radarWidth);
    self.inside:SetHeight(self.radarHeight);
    self.border:SetWidth(self.radarWidth * 1.08);
    self.border:SetHeight(self.radarHeight * 1.08);

    self:SetWidth(self.radarWidth);
    self:SetHeight(self.radarHeight + self.range:GetHeight());
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the n-th pin we want to get.                           *
-- ********************************************************************
-- * Grab a specific pin texture.                                     *
-- ********************************************************************
function Radar:GetPinTexture(index)
    local t = self.pinTextures[index];
    if not t then
        t = self:CreateTexture(nil, "ARTWORK");
        t:ClearAllPoints();
        self.pinTextures[index] = t;
    end
    return t;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the n-th zone we want to get.                          *
-- ********************************************************************
-- * Grab a specific zone texture.                                    *
-- ********************************************************************
function Radar:GetZoneTexture(index)
    local t = self.zoneTextures[index];
    if not t then
        t = self:CreateTexture(nil, "ARTWORK", nil, ZONE_LEVEL);
        t:SetTexture(self.zoneFilename);
        t:ClearAllPoints();
        self.zoneTextures[index] = t;
    end
    return t;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> index: the n-th arrow we want to get.                         *
-- ********************************************************************
-- * Grab a specific arrow texture.                                   *
-- ********************************************************************
function Radar:GetArrowTexture(index)
    local t = self.arrowTextures[index];
    if not t then
        t = self:CreateTexture(nil, "ARTWORK", nil, ARROW_LEVEL);
        t:SetTexture(self.arrowFilename);
        t:SetWidth(self.arrowWidth);
        t:SetHeight(self.arrowHeight);
        t:ClearAllPoints();
        self.arrowTextures[index] = t;
    end
    return t;
end

-- ********************************************************************
-- * Arguments:                                       - EASTER EGG -  *
-- * >> color: color of the line.                                     *
-- ********************************************************************
-- * Build a "line" from the current position of the player.          *
-- * Will last 20 seconds.                                            *
-- ********************************************************************
local TRIANGULATE_STEP = 10;
function Radar:Triangulate(color)
    local distance = 100;
    if color == "yellow" then
        distance = 50;
elseif color == "green" then
        distance = 20;
    end
    local steps = max(1, floor(distance / TRIANGULATE_STEP + 0.5));
    local r = GetPlayerFacing() + math.pi/2;
    local c, s = math.cos(r), math.sin(r);
    local i, x, y;
    for i=1, steps do
        x = self.px + c * TRIANGULATE_STEP * (i-1);
        y = self.py + s * TRIANGULATE_STEP * (i-1);

        local newZone = Table:Alloc();
        newZone.name         = "";
        newZone.color        = color or "red";
        newZone.radius       = TRIANGULATE_STEP/2;
        newZone.x, newZone.y = x, y;
        newZone.duration     = 20;
        newZone.self         = false;
        if newZone.x and newZone.y then
            tinsert(self.zones, newZone);
        end
    end
    self:ScrollIn();
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function Radar:OnLoad()
    local _, borderFilename = Resources:GetTextureInfo("RADAR_BORDER");
    local _, warningFilename = Resources:GetTextureInfo("WARNING_BACKGROUND");
    local _, insideFilename, width, height = Resources:GetTextureInfo("RADAR_INSIDE");
    _, self.pinFilename = Resources:GetTextureInfo("RADAR_PIN");
    _, self.zoneFilename = Resources:GetTextureInfo("RADAR_ZONE");
    _, self.arrowFilename, self.arrowWidth, self.arrowHeight = Resources:GetTextureInfo("RADAR_ARROW");

    -- ** Systems **
    GUI:InstallSystem("CoordinateSystem", self);
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "VERTICAL");

    -- ** Children **
    self.inside = self:CreateTexture(nil, "BORDER", nil, 1);
    self.inside:SetTexture(insideFilename);
    self.inside:SetPoint("TOP", self, "TOP", 0, 0);
    self.border = self:CreateTexture(nil, "BORDER", nil, 2);
    self.border:SetTexture(borderFilename);
    self.border:SetPoint("CENTER", self.inside, "CENTER", 0, 0);
    self.range = GUI:RequestChild("Number", nil, self, 1);
    self.range:SetPoint("BOTTOM", self, "BOTTOM", 0, 0);
    self.range:SetSuffix("M");
    self.range:SetPadding(3);
    self.range:EnableSmooth(50);
    self.warning = self:CreateTexture(nil, "BACKGROUND");
    self.warning:SetTexture(warningFilename);
    self.warning:SetAllPoints(UIParent);
    self.warning:SetBlendMode("ADD");
    self.warning:Hide();
    self.pinTextures = { };
    self.zoneTextures = { };
    self.arrowTextures = { };

    -- ** Properties **
    self.radarWidth = width / 2;
    self.radarHeight = height / 2;
    self.minRange = 0;
    self.visibility = DEFAULT_VISIBILITY;
    self.pins = { };
    self.zones = { };
    self.arrows = { };
    self.units = { };
    self.px = 0;
    self.py = 0;
    self.rotation = 0;
    self.minDistance = 0;
    self.color = "none";
    self.warningTimer = 0;
    self.useFullscreenColor = false;
    self.hasDistance = false;
    self.hasColor = false;
    self.hasWarning = false;
    self.hasArrows = false;

    -- ** Initial config **
    self:SetOrigin(0.5, 0.5);
    self:UpdateSize();
end

function Radar:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);

    self:Update(elapsed);
    self:Render();

    -- Update fullscreen warning texture
    if self.hasWarning and self.useFullscreenColor and self.color == "red" then
        self.warningTimer = min(self.warningTimer + elapsed, WARNING_FADE_TIMER);
  else
        self.warningTimer = max(self.warningTimer - elapsed, 0);
    end
    local alpha = self.warningTimer / WARNING_FADE_TIMER;
    if alpha > 0 then
        local phaseProgression = math.fmod(GetTime(), WARNING_SCROLL_CYCLE) / WARNING_SCROLL_CYCLE;
        local startX = 256 * phaseProgression;
        local endX   = 256 + startX;
        self.warning:SetTexCoord(0, 1, startX / 512, endX / 512);
        self.warning:SetAlpha(alpha * WARNING_ALPHA);
        self.warning:Show();
  else
        self.warning:Hide();
    end
end

function Radar:OnShow()
    self:OnUpdate(0);
end

function Radar:OnScrollOutEnd()
    self:Hide();
end

function Radar:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end

function Radar:OnDifficultyUpdate(difficulty)
    self.hasDistance = (difficulty <= DISTANCE_THRESHOLD);
    self.hasColor    = (difficulty <= COLOR_THRESHOLD);
    self.hasWarning  = (difficulty <= WARNING_THRESHOLD);
    self.hasArrows   = (difficulty <= ARROW_THRESHOLD);
end