local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

local System = Root:GetOrNewModule("System");
local Settings = System:GetOrNewModule("Settings");

local SettingFrame = GUI:GetOrNewWidget("SettingFrame");

SettingFrame.TEMPLATE = "BossEncounter3_SettingFrameTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- A setting frame allows to interpret graphically a setting table
-- and change the settings values obviously.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local MAX_CHECKBUTTONS = 10;
local MAX_SLIDERS = 2;
local MAX_BIT_TABLES = 1;
local NUM_BITS_PER_BIT_TABLE = 8;
local MAX_DROPDOWNS = 2;
local MAX_EDITBOXES = 1;

local initializeDropDown = function(dropDown)
    local i;
    for i=1, #dropDown.values do
        info = UIDropDownMenu_CreateInfo();
        info.text = dropDown.values[i];
        info.func = function(self, dropDown, value) SettingFrame.OnDropDownSelect(dropDown, value); end;
        info.arg1 = dropDown;
        info.arg2 = i;
        UIDropDownMenu_AddButton(info);
    end
end

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> global: true to open global options page.                     *
-- *            false to open module options page.                    *
-- ********************************************************************
-- * Open the setting frame.                                          *
-- ********************************************************************
function SettingFrame:Open(global)
    if self:ScrollIn() then
        self:SetPage(global);
    end
end

-- ********************************************************************
-- * Close the setting frame.                                         *
-- ********************************************************************
function SettingFrame:Close()
    self:ScrollOut();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> global: true to open global options page.                     *
-- *            false to open module options page.                    *
-- ********************************************************************
-- * Set the page of the setting frame.                               *
-- * Either global or module settings.                                *
-- ********************************************************************
function SettingFrame:SetPage(global)
    self.global = global;
    self:Refresh();
end

-- ********************************************************************
-- * Toggle the page of the setting frame.                            *
-- ********************************************************************
function SettingFrame:TogglePage()
    self:SetPage(not self.global);
end

-- ********************************************************************
-- * Use the default settings (for the current page only).            *
-- ********************************************************************
function SettingFrame:SetDefaults()
    local i, id, default, numSettings;
    if self.global then
        numSettings = Settings:GetNumGlobal();
  else
        numSettings = Settings:GetNumSpecific();
    end
    for i=1, numSettings do
        if self.global then
            id = Settings:GetGlobal(i);
      else
            id = Settings:GetSpecific(i);
        end
        _, default = Settings:GetInfo(id);
        Settings:Write(id, default);
    end
    self:Refresh();
end

-- ********************************************************************
-- * Refresh the setting frame: recalculate the necessary widgets,    *
-- * put them in place and set their values.                          *
-- ********************************************************************
function SettingFrame:Refresh()
    local numSettings = 0;
    local currentCheckButton = 1;
    local currentSlider = 1;
    local currentBitTable = 1;
    local currentDropDown = 1;
    local currentEditBox = 1;
    local checkButton, slider, bitTable, dropDown, editBox;
    local i, id, type, extra, default, perCharacter, action, exclusive, label, explanation;
    local minV, maxV, step, bits, valueKey;
    local realValue, locked, authColor;
    local lowText, highText;
    local j, button;

    -- Handle the page button and the title.
    if self.global then
        numSettings = Settings:GetNumGlobal();
        self.title:SetText(L:Get("Setting-GlobalTitle"));
        self.pageButton:SetText(L:Get("Setting-ModulePage"));
        if Settings:GetNumSpecific() > 0 then
            self.pageButton:Enable();
      else
            self.pageButton:Disable();
        end
  else
        numSettings = Settings:GetNumSpecific();
        if numSettings > 0 then
            self.title:SetText(L:Get("Setting-ModuleTitle"));
            self.pageButton:SetText(L:Get("Setting-GlobalPage"));
            self.pageButton:Enable();
      else
            -- No module settings loaded. Immediately switch to global settings page.
            self:SetPage(true);
            return;
        end
    end

    -- Start position data
    local width = self.title:GetWidth() + 32;
    local height = 80;
    local y = -48;

    -- Now set up the layout and widgets.
    for i=1, numSettings do
        if self.global then
            id = Settings:GetGlobal(i);
      else
            id = Settings:GetSpecific(i);
        end
        type, default, perCharacter, action, exclusive, label, explanation = Settings:GetInfo(id);
        minV, maxV, step, bits, valueKey = Settings:GetExtraInfo(id);
        realValue = Settings:ReadReal(id);
        locked = not Settings:HasPermission(id);

        if locked then
            authColor = "|cffa08080";
      else
            authColor = "|cffffffff";
        end

        if type == "BOOLEAN" then
            checkButton = self.checkButtons[currentCheckButton];
            currentCheckButton = currentCheckButton + 1;

            checkButton.id = id;

            checkButton:SetChecked(realValue);
            checkButton:SetText(authColor..label.."|r");
            self:CreateTooltip(checkButton, label, explanation, locked, exclusive, perCharacter);

            local hoverFrame = checkButton:GetHoverFrame();
            local spellId = Utility:ExtractSpellID(label);
            if spellId then
                -- A spell ID is contained within the label. Allows to get informations about the spell by hovering on the text.
                Utility:SetTooltip(hoverFrame, spellId);
          else
                Utility:SetTooltip(hoverFrame, 'CLEAR');
            end

            checkButton:Show();
            checkButton:SetPoint("TOPLEFT", self, "TOPLEFT", 8, y);

            y = y - 32;
            height = height + 32;
            width = max(width, checkButton:GetTextObject():GetStringWidth() + 52);

     elseif type == "NUMBER" then
            slider = self.sliders[currentSlider];
            currentSlider = currentSlider + 1;

            slider.id = id;

            if step < 0.1 then
                lowText, highText = "%.2f", "%.2f";
                slider.valueFormat = "%.2f";

        elseif step < 1 then
                lowText, highText = "%.1f", "%.1f";
                slider.valueFormat = "%.1f";
          else
                lowText, highText = "%d", "%d";
                slider.valueFormat = "%d";
            end

            slider:SetMinMaxValues(minV, maxV);
            slider:SetValueStep(step);
            slider:SetLowText(lowText:format(minV));
            slider:SetHighText(highText:format(maxV));

            slider:SetValue(realValue);
            slider:SetText(authColor..label.."|r");
            self:CreateTooltip(slider, label, explanation, locked, exclusive, perCharacter);

            local hoverFrame = slider:GetHoverFrame();
            local spellId = Utility:ExtractSpellID(label);
            if spellId then
                -- A spell ID is contained within the label. Allows to get informations about the spell by hovering on the text.
                Utility:SetTooltip(hoverFrame, spellId);
          else
                Utility:SetTooltip(hoverFrame, 'CLEAR');
            end

            slider:Show();
            slider:SetPoint("TOPLEFT", self, "TOPLEFT", 16, y-6);

            y = y - 40;
            height = height + 40;
            width = max(width, slider:GetTextObject():GetStringWidth() + 172);

     elseif type == "BIT_TABLE" then
            bitTable = self.bitTables[currentBitTable];
            currentBitTable = currentBitTable + 1;

            -- Prepare each bit

            local bitTitle, bitExplanation, bitValue;
            local totalWidth = 0;

            for j=1, NUM_BITS_PER_BIT_TABLE do
                button = bitTable.bits[j];

                if j <= bits then
                    bitTitle = L:Get(valueKey..":"..j);
                    bitExplanation = explanation.."\n\n|cffffffff"..bitTitle.."|r";
                    bitValue = bit.band(realValue, 2^(j-1));

                    button:SetChecked(bitValue ~= 0);
                    button:SetID(j);

                    button:SetText(bitTitle); -- First letter only.
                    self:CreateTooltip(button, label, bitExplanation, locked, exclusive, perCharacter);

                    button:SetPoint("BOTTOMLEFT", bitTable, "BOTTOMLEFT", 8+(j-1)*96, 4);
                    button:Show();

                    totalWidth = totalWidth + 96;
               else
                    button:Hide(); -- Unused bit.
                end
            end

            -- Finalize the bit table

            bitTable.id = id;

            bitTable:SetText(label);
            bitTable:SetWidth(totalWidth + 16);
            bitTable:SetHeight(56);

            bitTable:SetPoint("TOP", self, "TOP", 0, y);
            bitTable:Show();

            y = y - 64;
            height = height + 64;
            width = max(width, totalWidth + 32);

     elseif type == "DROPDOWN" then
            dropDown = self.dropDowns[currentDropDown];
            currentDropDown = currentDropDown + 1;

            wipe(dropDown.values);
            for j=minV, maxV do
                dropDown.values[j] = L:Get(valueKey..":"..j);
            end

            dropDown.id = id;

            dropDown:SetText(dropDown.values[realValue] or "?");
            dropDown:SetLabel(authColor..label.."|r");
            self:CreateTooltip(dropDown, label, explanation, locked, exclusive, perCharacter);

            UIDropDownMenu_Initialize(dropDown, initializeDropDown);
            UIDropDownMenu_SetWidth(dropDown, 256);

            dropDown:SetPoint("TOP", self, "TOP", 0, y - 16);
            dropDown:Show();

            y = y - 56;
            height = height + 56;
            width = max(width, 288);

     elseif type == "STRING" then
            editBox = self.editBoxes[currentEditBox];
            currentEditBox = currentEditBox + 1;

            editBox.id = id;

            editBox:SetMaxLetters(maxV);

            editBox:SetText(tostring(realValue) or "");
            editBox.border:SetText(authColor..label.."|r");
            self:CreateTooltip(editBox, label, explanation, locked, exclusive, perCharacter);

            editBox:Show();
            editBox:SetPoint("TOPLEFT", self, "TOPLEFT", 16, y-12);

            y = y - 40;
            height = height + 40;
            width = max(width, editBox:GetWidth() + 32);
        end
    end

    -- Hide unused remaining widgets.
    for i=currentCheckButton, MAX_CHECKBUTTONS do
        self.checkButtons[i]:Hide();
    end
    for i=currentSlider, MAX_SLIDERS do
        self.sliders[i]:Hide();
    end
    for i=currentBitTable, MAX_BIT_TABLES do
        self.bitTables[i]:Hide();
    end
    for i=currentDropDown, MAX_DROPDOWNS do
        self.dropDowns[i]:Hide();
    end
    for i=currentEditBox, MAX_EDITBOXES do
        self.editBoxes[i]:Hide();
    end

    -- Finalize width.
    self:SetWidth(width);
    self:SetHeight(height);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> object: the object that receives the tooltip.                 *
-- * >> title: the title of the tooltip (can be nil).                 *
-- * >> description: the description of the tooltip (can be nil).     *
-- * >> isLocked: true if you can't modify the setting.               *
-- * >> isExclusive: true if the setting is exclusive.                *
-- * >> isPerCharacter: true if the setting is saved per character.   *
-- ********************************************************************
-- * Build the tooltip for one of the objects of the setting frame.   *
-- ********************************************************************
function SettingFrame:CreateTooltip(object, title, description, isLocked, isExclusive, isPerCharacter)
    if isLocked then
        description = (description or "")..L:Get("Setting-Lock");
  else
        if isExclusive then
            description = (description or "")..L:Get("Setting-Exclusive");
        end
        if isPerCharacter then
            description = (description or "")..L:Get("Setting-PerCharacter");
        end
    end
    Utility:SetTooltip(object, title, description);
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function SettingFrame:OnLoad()
    local _, separatorFilename = Resources:GetTextureInfo("SEPARATOR");
    local i;

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, false);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetPoint("TOP", self, "TOP", 0, -12);
    self.title:Show();
    self.separator = self:CreateTexture(nil, "BACKGROUND");
    self.separator:SetPoint("TOPLEFT", self, "TOPLEFT", 8, -32);
    self.separator:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", -8, -40);
    self.separator:SetTexture(separatorFilename);
    self.closeButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.closeButton:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -8, 8);
    self.closeButton:SetWidth(80);
    self.closeButton:SetHeight(24);
    self.closeButton:SetText(CLOSE);
    self.closeButton:SetScript("OnClick", self.OnCloseButtonClick);
    self.closeButton:Show();
    self.defaultButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.defaultButton:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 8, 8);
    self.defaultButton:SetWidth(80);
    self.defaultButton:SetHeight(24);
    self.defaultButton:SetText(DEFAULT);
    self.defaultButton:SetScript("OnClick", self.OnDefaultButtonClick);
    self.defaultButton:Show();
    self.pageButton = CreateFrame("Button", nil, self, "OptionsButtonTemplate");
    self.pageButton:SetPoint("BOTTOM", self, "BOTTOM", 0, 8);
    self.pageButton:SetWidth(160);
    self.pageButton:SetHeight(24);
    self.pageButton:SetScript("OnClick", self.OnPageButtonClick);
    self.pageButton:Show();

    -- Prepare some checkbuttons.
    self.checkButtons = { };
    for i=1, MAX_CHECKBUTTONS do
        self.checkButtons[i] = CreateFrame("CheckButton", self:GetName().."_CheckButton"..i, self, "BossEncounter3_CheckButtonTemplate");
        self.checkButtons[i]:SetScript("OnClick", self.OnCheckButtonClick);
    end

    -- Prepare some sliders.
    self.sliders = { };
    for i=1, MAX_SLIDERS do
        self.sliders[i] = CreateFrame("Slider", self:GetName().."_Slider"..i, self, "BossEncounter3_SliderTemplate");
        self.sliders[i]:SetScript("OnValueChanged", self.OnSliderValueChanged);
    end

    -- Prepare some bit tables.
    self.bitTables = { };
    for i=1, MAX_BIT_TABLES do
        self.bitTables[i] = CreateFrame("Frame", self:GetName().."_BitTable"..i, self, "BossEncounter3_BitTableTemplate");
        self:OnBitTableLoad(self.bitTables[i]);
    end

    -- Prepare some dropdowns.
    self.dropDowns = { };
    for i=1, MAX_DROPDOWNS do
        self.dropDowns[i] = CreateFrame("Frame", self:GetName().."_DropDown"..i, self, "BossEncounter3_DropDownTemplate");
        self.dropDowns[i].values = { };
    end

    -- Prepare some editboxes.
    self.editBoxes = { };
    for i=1, MAX_EDITBOXES do
        self.editBoxes[i] = CreateFrame("EditBox", self:GetName().."_EditBox"..i, self, "BossEncounter3_EditBoxTemplate");
        self.editBoxes[i]:SetScript("OnTextChanged", self.OnEditBoxTextChanged);
        Utility:SetBackdrop(self.editBoxes[i].border, "OPTION");
    end

    -- ** Properties **
    self.global = true;

    -- ** Initial config **
    Utility:SetBackdrop(self, "OPTION");
end

function SettingFrame:OnBitTableLoad(bitTable)
    bitTable.bits = { };

    local i;
    for i=1, NUM_BITS_PER_BIT_TABLE do
        bitTable.bits[i] = CreateFrame("CheckButton", self:GetName().."_Bit"..i, bitTable, "BossEncounter3_CheckButtonTemplate");
        bitTable.bits[i]:SetScript("OnClick", self.OnBitTableButtonClick);
    end

    Utility:SetBackdrop(bitTable, "OPTION");
end

function SettingFrame:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);
end

function SettingFrame.OnCloseButtonClick(button)
    button:GetParent():Close();
end

function SettingFrame.OnDefaultButtonClick(button)
    button:GetParent():SetDefaults();
end

function SettingFrame.OnPageButtonClick(button)
    button:GetParent():TogglePage();
end

function SettingFrame.OnCheckButtonClick(button)
    if button:GetChecked() then
        PlaySound("igMainMenuOptionCheckBoxOn");
        Settings:Write(button.id, true);
  else
        PlaySound("igMainMenuOptionCheckBoxOff");
        Settings:Write(button.id, false);
    end
end

function SettingFrame.OnBitTableButtonClick(button)
    local bitTable = button:GetParent();
    local bitRank = button:GetID()-1;
    local bitMask = (2^NUM_BITS_PER_BIT_TABLE-1) - 2^bitRank;
    local bitCurrent = Settings:ReadReal(bitTable.id);
    local bitValue;

    if button:GetChecked() then
        PlaySound("igMainMenuOptionCheckBoxOn");
        bitValue = 2^bitRank;
  else
        PlaySound("igMainMenuOptionCheckBoxOff");
        bitValue = 0;
    end

    local value = bit.bor(bitValue, bit.band(bitMask, bitCurrent));
    Settings:Write(bitTable.id, value);
end

function SettingFrame.OnSliderValueChanged(slider)
    local value = slider:GetValue();

    local valueText = (slider.valueFormat):format(value);
    slider:SetValueText(valueText);

    Settings:Write(slider.id, value);
end

function SettingFrame.OnDropDownSelect(dropDown, value)
    dropDown:SetText(dropDown.values[value] or "?");
    Settings:Write(dropDown.id, value);
end

function SettingFrame.OnEditBoxTextChanged(editBox)
    Settings:Write(editBox.id, editBox:GetText());
end

function SettingFrame:OnScrollOutEnd()
    self:Hide();
end

function SettingFrame:OnDifficultyUpdate()
    if self:IsShown() then
        self:Refresh();
    end
end

function SettingFrame:OnRankChange()
    if self:IsShown() then
        self:Refresh();
    end
end

function SettingFrame:OnSettingsRefresh()
    if self:IsShown() then
        self:Refresh();
    end
end