local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local L = Root:GetOrNewModule("Localization");
local Audio = Root:GetOrNewModule("Audio");
local Resources = Root:GetOrNewModule("Resources");

local StatusWindow = GUI:GetOrNewWidget("StatusWindow");

local System = Root:GetOrNewModule("System");
local View = System:GetOrNewModule("View");

StatusWindow.TYPE = "Button";
StatusWindow.TEMPLATE = "BossEncounter3_StatusWindowTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- StatusWindow is a final GUI element allowing to display a title and
-- an important timer. It also advertises BE3!

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local TITLE_SCALE = 0.5;
local AD_DURATION = 4.0;

local WRENCH_FLASH = 2.5;
local WRENCH_CYCLE = WRENCH_FLASH / 4;
local WRENCH_FADE  = 1.0;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Get the important timer of the status window.                    *
-- ********************************************************************
function StatusWindow:GetImportantTimer()
    return self.importantTimer;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> title: the desired title.                                     *
-- ********************************************************************
-- * Set the title of the status window.                              *
-- ********************************************************************
function StatusWindow:SetTitle(title)
    self.titleText = title;
    if self.ad == 0 then
        self:DisplayTitle(title);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> title: the desired title.                                     *
-- ********************************************************************
-- * Set the displayed title of the status window.                    *
-- ********************************************************************
function StatusWindow:DisplayTitle(title)
    self.title:SetText(title);
    self:UpdateSize();
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole status window.    *
-- ********************************************************************
function StatusWindow:UpdateSize()
    local lh = self.title:GetHeight() * TITLE_SCALE;
    local th = self.importantTimer:GetHeight();

    local width = max(self.title:GetWidth() * TITLE_SCALE, self.importantTimer:GetWidth()) + 64;
    local height = lh + th;

    self:SetWidth(width);
    self:SetHeight(height);
end

-- ********************************************************************
-- * Enable the wrench button.                                        *
-- ********************************************************************
function StatusWindow:AllowOptions()
    self.wrenchState = true;
end

-- ********************************************************************
-- * Disable the wrench button.                                       *
-- ********************************************************************
function StatusWindow:DisableOptions(instantly)
    if self.wrenchShown and not instantly then
        self.wrenchTimer = WRENCH_FADE;
  else
        self.wrenchTimer = 0;
    end

    self.wrenchState = false;
    self.wrenchShown = false;
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function StatusWindow:OnLoad()
    local _, wrenchFilename = Resources:GetTextureInfo("WRENCH");

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "VERTICAL");

    -- ** Children **
    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetPoint("TOP", self, "TOP", 0, 0);
    self.title:SetScale(TITLE_SCALE);
    self.importantTimer = GUI:RequestChild("ImportantTimer", nil, self, 1);
    self.importantTimer:SetPoint("BOTTOM", self, "BOTTOM", 0, 0);
    self.importantTimer:Show();
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HS");
    self.wrenchButton = CreateFrame("Button", nil, self);
    self.wrenchButtonIcon = self.wrenchButton:CreateTexture(nil, "ARTWORK");
    self.wrenchButtonIcon:SetTexture(wrenchFilename);
    self.wrenchButtonIcon:SetAllPoints(self.wrenchButton);
    self.wrenchButton:SetScript("OnClick", self.OnWrenchButtonClick);
    self.wrenchButton:SetWidth(32);
    self.wrenchButton:SetHeight(32);
    self.wrenchButton:SetPoint("TOPRIGHT", self, "TOPRIGHT", -32, -12);
    self.wrenchButton:Hide();

    -- ** Properties **
    self.ad = 0;
    self.titleText = "";
    self.wrenchState = false;
    self.wrenchShown = false;
    self.wrenchTimer = 0;

    -- ** Initial config **
    self:SetTitle("");
    self:UpdateSize();
end

function StatusWindow:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);

    if self.ad > 0 then
        self.ad = max(0, self.ad - elapsed);
        if self.ad == 0 then
            self:DisplayTitle(self.titleText);
        end

  elseif self.wrenchState and not self.wrenchShown then
        self.wrenchShown = true;
        self.wrenchTimer = WRENCH_FLASH;
        Audio:PlaySound("NOTIFY");
    end

    if self.wrenchTimer == 0 and not self.wrenchShown then
        self.wrenchButton:Hide();
  else
        self.wrenchTimer = max(0, self.wrenchTimer - elapsed);
        if self.wrenchShown then
            self.wrenchButton:SetAlpha(0.5 * (1 + cos(self.wrenchTimer * 360 / WRENCH_CYCLE)));
      else
            self.wrenchButton:SetAlpha(self.wrenchTimer / WRENCH_FADE);
        end
        self.wrenchButton:Show();
    end
end

function StatusWindow:OnShow()
    self.ad = AD_DURATION;
    self:DisplayTitle(L:Get("Ad"));
    self:OnUpdate(0);
end

function StatusWindow:OnScrollOutEnd()
    self:Hide();
end

function StatusWindow:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end

function StatusWindow:OnWrenchButtonClick()
    View:OpenOptions();
end