local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Audio = Root:GetOrNewModule("Audio");

local SuperAlert = GUI:GetOrNewWidget("SuperAlert");

SuperAlert.TYPE = "Frame";
SuperAlert.TEMPLATE = "BossEncounter3_SuperAlertTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Display a super alert. Only one alert at a time.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local TITLE_SCALE = 1.00;
local SUBTITLE_SCALE = 0.75;

local UI_PARENT_ALPHA = 0.50;
local UI_PARENT_ALPHA_SIMPLE = 0.25;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> title: the title of the alert.                                *
-- * >> subTitle: the sub title of the alert.                         *
-- * >> duration: duration of the alert (required).                   *
-- * >> focus: true to put emphasis on the alert.                     *
-- ********************************************************************
-- * Display a super alert (previous one will be replaced).           *
-- ********************************************************************
function SuperAlert:Display(title, subTitle, duration, focus)
    self.focus = focus;
    self.simple = (title == "" and subTitle == "");

    if self.simple then
        self.band:Hide();
        self.importantTimer:HideTimer();
  else
        self.band:Show();
        self.importantTimer:ShowTimer();
    end

    self.title:SetText(title);
    self.subTitle:SetText(subTitle);
    self.importantTimer:Clear();
    self.importantTimer:SetValue(duration);
    self.importantTimer:Start();
    self:UpdateSize();

    self:Show();
    Audio:PlaySound("HIT");
end

-- ********************************************************************
-- * Clear the alert.                                                 *
-- ********************************************************************
function SuperAlert:Clear()
    self:Hide();
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole super alert.      *
-- ********************************************************************
function SuperAlert:UpdateSize()
    local width = max(self.title:GetWidth() * TITLE_SCALE, self.subTitle:GetWidth() * SUBTITLE_SCALE, self.importantTimer:GetWidth()) + 64;
    local height = self.title:GetHeight() * TITLE_SCALE + self.subTitle:GetHeight() * SUBTITLE_SCALE + self.importantTimer:GetHeight() + 16;

    self:SetWidth(width);
    self:SetHeight(height);
end

-- ********************************************************************
-- * Update the UI alpha.                                             *
-- ********************************************************************
function SuperAlert:UpdateUIAlpha()
    if self.focus then
        if self.simple then
            UIParent:SetAlpha(UI_PARENT_ALPHA_SIMPLE);
      else
            UIParent:SetAlpha(UI_PARENT_ALPHA);
        end
  else
        UIParent:SetAlpha(1);
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function SuperAlert:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, false);

    -- ** Children **
    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetPoint("TOP", self, "TOP", 0, 0);
    self.title:SetScale(TITLE_SCALE);
    self.subTitle = GUI:RequestChild("MonoFont", nil, self, 1);
    self.subTitle:SetPoint("TOP", self, "TOP", 0, -28);
    self.subTitle:SetScale(SUBTITLE_SCALE);
    self.subTitle:SetColor(1, 1, 0);
    self.importantTimer = GUI:RequestChild("ImportantTimer", nil, self, 1);
    self.importantTimer:SetPoint("BOTTOM", self, "BOTTOM", 0, 0);
    self.importantTimer:SetIcon("WARNING");
    self.importantTimer:SetBehavior(60, 15, false, false);
    self.importantTimer:SetMode(-1);
    self.importantTimer:Show();
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HN");

    -- ** Properties **
    self.focus = false;
    self.simple = false;
end

function SuperAlert:OnShow()
    self:UpdateUIAlpha();
end

function SuperAlert:OnHide()
    self.focus = false;
    self:UpdateUIAlpha();
end

function SuperAlert:OnImportantTimerExpire(importantTimer)
    self:Hide();
end

function SuperAlert:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end