local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Resources = Root:GetOrNewModule("Resources");
local Unit = Root:GetOrNewModule("Unit");

local UnitList = GUI:GetOrNewWidget("UnitList");

UnitList.TYPE = "Button";
UnitList.TEMPLATE = "BossEncounter3_UnitListTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- UnitList is a final GUI element displaying a list of raid units.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> title: set the title of the unit list.                        *
-- ********************************************************************
-- * Set the title of the unit list.                                  *
-- ********************************************************************
function UnitList:SetTitle(title)
    self.title:SetText(title);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> playerList: the list of players. Can be the magic word "ALL". *
-- ********************************************************************
-- * Set the players displayed in the unit list.                      *
-- ********************************************************************
function UnitList:SetPlayers(playerList)
    self:Clear();

    local units = self.units;
    local i, uid;

    if playerList == "ALL" then
        Unit:EnumerateRaid();
        for i=1, Unit:GetEnumerationSize() do
            tinsert(units, Unit:GetEnumerationUID(i));
        end
  else
        for i=1, #playerList do
            tinsert(units, playerList[i]);
        end
    end

    for i=1, #units do
        local E = self:GetFreeUnitRow();
        uid = Unit:ResolveUID(units[i]);
        if uid then
            E:SetUnit(uid);
            E:SetUpdateFunc(self.updateFunc);
            self.list:Insert(E);
        end
    end

    self:Update();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> updateFunc: the update function.                              *
-- ********************************************************************
-- * Set the update func that will update rows.                       *
-- ********************************************************************
function UnitList:SetUpdateFunc(updateFunc)
    self.updateFunc = updateFunc;

    local i;
    for i=1, #self.rows do
        self.rows[i]:SetUpdateFunc(updateFunc);
    end
end

-- ********************************************************************
-- * Clear the unit list.                                             *
-- ********************************************************************
function UnitList:Clear()
    wipe(self.units);

    local i;
    for i=1, #self.rows do
        self.rows[i]:Kill();
    end
    self.list:Clear();

    self:Update();
end

-- ********************************************************************
-- * Update the unit list.                                            *
-- ********************************************************************
function UnitList:Update()
    self:UpdateSize();

    if self.list:IsEmpty() then
        self:ScrollOut();
  else
        self:ScrollIn();
    end
end

-- ********************************************************************
-- * Get a free eventRow widget (or create a new one).                *
-- ********************************************************************
function UnitList:GetFreeUnitRow()
    local i, unitRow;
    for i=1, #self.rows do
        unitRow = self.rows[i];
        if not unitRow:GetGUID() then
            return unitRow;
        end
    end
    -- No existing row available, create a new one.
    unitRow = GUI:RequestChild("UnitRow", nil, self);
    tinsert(self.rows, unitRow);
    return unitRow;
end

-- ********************************************************************
-- * Update the size of the unit list.                                *
-- ********************************************************************
function UnitList:UpdateSize()
    if self:IsLocked() then return; end

    self.list:UpdateSize();

    local w = max(self.title:GetWidth(), self.list:GetWidth());
    local h = self.title:GetHeight() + self.separator:GetHeight() + self.list:GetHeight();
    self:SetWidth(w);
    self:SetHeight(h);
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function UnitList:OnLoad()
    local _, separatorFilename = Resources:GetTextureInfo("SEPARATOR");

    -- ** Systems **
    GUI:InstallSystem("PositionSystem", self, true);
    GUI:InstallSystem("ScrollSystem", self, "HORIZONTAL", true);

    -- ** Children **
    self.title = GUI:RequestChild("MonoFont", nil, self, 1);
    self.title:SetPoint("TOP", self, "TOP", 0, 0);
    self.separator = self:CreateTexture(nil, "BACKGROUND");
    self.separator:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -16);
    self.separator:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", 0, -24);
    self.separator:SetTexture(separatorFilename);
    self.list = GUI:RequestChild("List", nil, self, 1);
    self.list:SetPoint("TOP", self, "TOP", 0, -24);
    self.rows = { }; -- Will be dynamically filled for our needs.

    -- ** Properties **
    self.units = { };
    self.updateFunc = nil;

    -- ** Initial config **
    self:SetTitle("Units");
    self:UpdateSize();
end

function UnitList:OnUpdate(elapsed)
    self:UpdateScroll(elapsed);
end

function UnitList:OnScrollOutEnd()
    self:Hide();
end

function UnitList:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end