local Root = BossEncounter3;
local GUI = Root:GetOrNewModule("GUI");
local Utility = Root:GetOrNewModule("Utility");
local Unit = Root:GetOrNewModule("Unit");

local Module = Root:GetOrNewModule("Module");

local UnitRow = GUI:GetOrNewWidget("UnitRow");

UnitRow.TEMPLATE = "BossEncounter3_UnitRowTemplate";

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- UnitRow is a GUI element allowing to display a clickable player name
-- along with a label and a cross or a tick icon.

-- --------------------------------------------------------------------
-- **                            Constants                           **
-- --------------------------------------------------------------------

local LABEL_SCALE = 0.5;

local MAX_NAME_LENGTH = 12;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Get the GUID of the unit monitored by the unit row.              *
-- ********************************************************************
function UnitRow:GetGUID()
    return self.guid;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: unitId of the unit.                                     *
-- ********************************************************************
-- * Set the unit monitored by the specified unit row.                *
-- ********************************************************************
function UnitRow:SetUnit(unit)
    self.guid = UnitGUID(unit);
    self.symbol = 0;

    local name = UnitName(unit);
    self.label:SetText(Utility:Abbreviate(name, MAX_NAME_LENGTH));
    if UnitIsUnit(unit, "player") then
        self.label:SetColor(1, 1, 0);
  else
        self.label:SetColor(1, 1, 1);
    end
    self:SetTargetAction(name);

    self:SetBarColor(1, 1, 1);
    self:SetBarFill(0);
    self:SetTextColor(1, 1, 1);
    self:SetText("");

    self:UpdateLayout();
end

-- ********************************************************************
-- * Kill the unit row.                                               *
-- ********************************************************************
function UnitRow:Kill()
    self.guid = nil;
    self.symbol = 0;
    self.updateFunc = nil;

    self:UpdateLayout();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> updateFunc: the update function.                              *
-- ********************************************************************
-- * Set the function used to update the unit row.                    *
-- ********************************************************************
function UnitRow:SetUpdateFunc(updateFunc)
    self.updateFunc = updateFunc;
    self.module = Module:GetActive();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: color of the bar.                                    *
-- ********************************************************************
-- * Set the color of the bar.                                        *
-- ********************************************************************
function UnitRow:SetBarColor(r, g, b)
    self.bar:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> fill: fill ratio of the bar between 0 and 1.                  *
-- ********************************************************************
-- * Set the fill ratio of the bar.                                   *
-- ********************************************************************
function UnitRow:SetBarFill(fill)
    self.bar:SetValue(fill);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> r, g, b: colors of the text.                                  *
-- ********************************************************************
-- * Set the color of the text.                                       *
-- ********************************************************************
function UnitRow:SetTextColor(r, g, b)
    self.text:SetColor(r, g, b);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> text: the text to display.                                    *
-- ********************************************************************
-- * Set the text of the row.                                         *
-- ********************************************************************
function UnitRow:SetText(text)
    self.text:SetText(text);
end

-- ********************************************************************
-- * Update the layout of the unit row.                               *
-- ********************************************************************
function UnitRow:UpdateLayout()
    Utility:SetIcon(self.symbolTexture, self.symbol);

    self:UpdateSize();
end

-- ********************************************************************
-- * Recalculate the appropriate size for the whole event row.        *
-- ********************************************************************
function UnitRow:UpdateSize()
    local width = self.bar:GetWidth();
    local height = self.bar:GetHeight();

    self:SetWidth(width);
    self:SetHeight(height);

    self:Notify("OnChildResized");
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function UnitRow:OnLoad()
    -- ** Systems **
    GUI:InstallSystem("NotifySystem", self);
    GUI:InstallSystem("ActionSystem", self);

    -- ** Children **
    self.label = GUI:RequestChild("MonoFont", nil, self, 2);
    self.label:SetPoint("LEFT", self, "LEFT", 16 / LABEL_SCALE, 0);
    self.label:SetScale(LABEL_SCALE);
    self.symbolTexture = self.label:CreateTexture(nil, "ARTWORK");
    self.symbolTexture:SetWidth(20);
    self.symbolTexture:SetHeight(20);
    self.symbolTexture:SetPoint("RIGHT", self.label, "LEFT", 0, 0);
    self.text = GUI:RequestChild("MonoFont", nil, self, 2);
    self.text:SetPoint("RIGHT", self, "RIGHT", -16 / LABEL_SCALE, 0);
    self.text:SetScale(LABEL_SCALE);
    self.bar = GUI:RequestChild("Bar", nil, self, 1);
    self.bar:SetPoint("LEFT", self, "LEFT", 0, 0);
    self.bar:SetStyle("SIMPLE");
    self.bar:SetHeight(8);
    self.bar:SetAlpha(0.8);
    self.band = GUI:RequestChild("Band", nil, self, 0);
    self.band:SetAllPoints(self);
    self.band:SetGraphic("BAND_HN");

    -- ** Properties **
    self.guid = nil;
    self.symbol = 0;
    self.updateFunc = nil;
    self.module = nil;

    -- ** Initial config **
    self:ClearAction();
end

function UnitRow:OnUpdate(elapsed)
    local noUpdate = true;
    local unit = Unit:GetUID(self.guid);
    if unit then
        if self.updateFunc then
            self:updateFunc(unit);
            noUpdate = false;
        end

        -- Symbol change check
        local currentSymbol = GetRaidTargetIndex(unit) or 0;
        if currentSymbol ~= self.symbol then
            self.symbol = currentSymbol;
            self:UpdateLayout();
        end
    end
    if noUpdate then
        self:SetBarFill(0);
        self:SetText("");
    end
    self:UpdateAction();
end

function UnitRow:OnChildResized()
    if self:IsLoaded() then
        self:UpdateSize();
    end
end

function UnitRow:OnShow()
    self:UpdateLayout();
    self:OnUpdate(0);
end