﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local BOSS_ID = 46753;
local ADD_ID = 47175;

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "WindBurstAlert",
        type = "BOOLEAN",
        default = false,
        label = {
            ["enUS"] = "Super alert: Wind Burst",
            ["frFR"] = "Super alerte : Jaillissement de vent",
        },
        explain = {
            ["enUS"] = "Enable this if you need an aggressive alert when Wind Burst is cast.",
            ["frFR"] = "Cette option permet d'afficher une alerte agressive quand Jaillissement de vent est incanté.",
        },
    },
    {
        id = "LightningRod",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 89668,
                 symbol = 7 },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["PhaseThreeTrigger"] = "I WILL NO LONGER",
        ["WindBurstTitle"] = "Wind Burst",
        ["WindBurstExplain"] = "Come closer!",
        ["FeedbackBar"] = "Life of the lowest Stormling",
    },
    ["frFR"] = {
        ["PhaseThreeTrigger"] = "Je ne serai",
        ["WindBurstTitle"] = "Jaillissement",
        ["WindBurstExplain"] = "Allez au càc!",
        ["FeedbackBar"] = "Vie du plus bas Tourmentin",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("AlAkir", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(75);
        self:SetScoreBenchmarks(400, 600);

        self:RegisterRespawnableAdd(ADD_ID);
        self:IgnoreAdds();

        self:RegisterAnyYell(self:Localize("PhaseThreeTrigger"), "EyeOfTheStorm");        
    end,

    OnBossFound = function(self, uid)
        self:RegisterHealthThreshold(0.80, self:Localize("PhaseTwo"));
        self:RegisterHealthThreshold(0.25, self:Localize("PhaseThree"));
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();
    end,

    -- *******
    -- Methods
    -- *******

    GetMinAddHealth = function(self)
        local minHealth = nil;
        local maxHealth = nil;
        local num = Unit:GetNumUID();
        local i, guid, name, unit, id, health;
        for i=1, num do
            guid, name, unit = Unit:GetUID(i);
            id = Unit:GetMobID(guid);
            if id == ADD_ID and unit and not UnitIsDead(unit) then
                health = UnitHealth(unit);
                if (not minHealth) or (health < minHealth) then
                    minHealth = health;
                    maxHealth = UnitHealthMax(unit);
                end
            end
        end
        if minHealth then
            return minHealth, maxHealth;
      else
            return 0, 1000000; -- Set a default max value.
        end
    end,

    GetFeedbackTimer = function(self)
        local unit = Unit:ResolveUID(self:GetBossGUID());
        if unit then
            local start, duration = Unit:SearchEffect(unit, true, 87904);
            if start and duration then
                local timeLeft = max(0, start + duration - GetTime());
                return timeLeft, duration;
            end
        end
        return 0, 30;
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
        self:SetMainUnit(nil);
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:InterruptTimer("WindBurst1");
            self:SetMainUnit(BOSS_ID);

            self:StartMirrorValue(self:Localize("FeedbackBar"), self.GetMinAddHealth, self.GetFeedbackTimer);
        end
    end,

    OnPhaseThree = function(self)
        if self:SetIncreasingPhase(3) then
            self:StartTimer("WindBurst2", 27.5);
            self:StartTimer("LightningRod", 20);

            self:StopMirrorValue();
        end
    end,

    OnWindBurst = function(self)
        if self:GetSetting("WindBurstAlert") then
            local castTime = 5;
            if self:IsHeroic() then castTime = 4; end
            self:SuperAlert(self:Localize("WindBurstTitle"), self:Localize("WindBurstExplain"), castTime, true);
      else
            self:AlertIncomingSpell("all/high", 87770);
        end
    end,

    OnLightningRod = function(self, guid)
        if self:GetSetting("LightningRod") then
            self:AnnounceUnitSpell(89668, guid, 7, 10);
        end
        if self:IsMe(guid) then
            self:AlertSpell("all/high", 89668, guid);
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(BOSS_ID, false, false); -- ID, volatile, bypass
M:DefineBerserk(600);
M:DefineMinRange(0); -- Force the radar to be shown

-- Events

M:RegisterCastEvent(87770, "WindBurst1"); 
M:RegisterDamageEvent(88858, "WindBurst2");
M:RegisterEffectGainEvent(88301, "AcidRain");
M:RegisterEffectGainEvent(87904, "Feedback");
M:RegisterEffectGainEvent(89668, "LightningRod");

-- Timers

M:DefineVoidTimer("Feedback/87904", "dps/high", 20);
M:DefineRepeatingVoidTimer("LightningRod/89668", "all/high", 20);
M:DefineRepeatingVoidTimer("WindBurst2/88858", "all/high", 20);
M:DefineInitialRepeatingVoidTimer("WindBurst1/87770", "all/high", 27.5, 20);

-- Actions

M:DeclareAction("AcidRain", 600);
M:AddCallbackToAction("AcidRain", "OnPhaseTwo");

M:DeclareAction("EyeOfTheStorm", 600);
M:AddCallbackToAction("EyeOfTheStorm", "OnPhaseThree");

M:DeclareAction("WindBurst1", 5);
M:AddTriggerTimerToAction("WindBurst1", "WindBurst1");
M:AddCallbackToAction("WindBurst1", "OnWindBurst");

M:DeclareAction("WindBurst2", 5);
M:AddTriggerTimerToAction("WindBurst2", "WindBurst2");

M:DeclareAction("Feedback");
M:AddTriggerTimerToAction("Feedback", "Feedback");
M:AddStartTimerToAction("Feedback", "Feedback");

M:DeclareAction("LightningRod");
M:AddTriggerTimerToAction("LightningRod", "LightningRod");
M:AddCallbackToAction("LightningRod", "OnLightningRod", "targetguid");

-- Errors

M:DefineAvoidEffectError(89104); -- Relentless Storm
M:DefineAvoidDamageError(91020, 6); -- Ice Storm
M:DefineAvoidDamageError(87856, 4); -- Squall Line
M:DefineAvoidDamageError(89588); -- Lightning Cloud
M:DefineAvoidBombError(89667); -- Lightning Rod