﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local function healthFunc(self, uid)
    local health, healthMax = UnitHealth(uid), UnitHealthMax(uid);
    local ratio = min(1, health / healthMax);
    if health < 10000 then
        self:SetBarColor(1, 0, 0);
        self:SetTextColor(1, 0, 0);
  else
        self:SetBarColor(0, 1, 0);
        self:SetTextColor(1, 1, 1);
    end
    self:SetBarFill(ratio);
    self:SetText(health);
end

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "BreakThreshold",
        type = "NUMBER",
        min = 1,
        max = 4,
        step = 1,
        default = 3,
        meta = { spell = 82881,
                 threshold = true },
    },
    {
        id = "ShowHealth",
        type = "BOOLEAN",
        default = false,
        label = {
            ["enUS"] = "Show health",
            ["frFR"] = "Montrer l'état des PV",
        },
        explain = {
            ["enUS"] = "Display a clickable player health list.",
            ["frFR"] = "Affiche une liste cliquable des PV des joueurs.",
        },
    },
    {
        id = "SlimeWarning",
        type = "BOOLEAN",
        action = "RAID_WARNING",
        default = false,
        exclusive = true,
        meta = { spell = 82935,
                 multiple = true },
    },
    {
        id = "SlimeSymbol",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 82935,
                 symbol = 1,
                 multiple = true,
                 silent = true },
    },

    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Mortality"] = "Mortality",
        ["FeudEnd"] = "Feud: end",
        ["Break"] = "You have Break %d times!",
        ["SpreadOut"] = "Spread out!",
        ["Regroup"] = "Regroup!",
        ["NotStacked"] = "has taken undivided damage; the raid wasn't stacked.",
        ["AttemptedHealing"] = "tried to heal despite the 99% healing reduction.",
        ["HealthUI"] = "Health",
    },
    ["frFR"] = {
        ["Mortality"] = "Mortalité",
        ["FeudEnd"] = "Fin de querelle",
        ["Break"] = "Vous avez %d fois brèche !",
        ["SpreadOut"] = "Espacez-vous !",
        ["Regroup"] = "Regroupez-vous !",
        ["NotStacked"] = "a subi des dégâts non divisés; le raid n'était pas groupé.",
        ["AttemptedHealing"] = "a essayé de soigner en dépit de la réduction de 99% des soins.",
        ["HealthUI"] = "Etat : PV",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Chimaeron", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    protection = false,
    causticSlime = { },

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(60);
        self:SetScoreBenchmarks(400, 600);

        self:CheckHealthDisplay();
    end,

    OnBossFound = function(self, uid)
        self:RegisterHealthThreshold(0.205, self:Localize("Mortality"));
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();
        self:OnProtectionChange("player");
    end,

    OnSettingUpdate = function(self, id, value)
        if id == "ShowHealth" then
            self:CheckHealthDisplay();
        end
    end,

    OnHeal = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, school, absorbed)
        if amount >= 250 and self:GetPhase() == 2 and self:IsUnitAffiliated(actorGUID) then
            self:TriggerError("AttemptedHealing", actorName);
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckHealthDisplay = function(self)
        if self:GetSetting("ShowHealth") then
            self:StartUnitMonitor(self:Localize("HealthUI"), healthFunc);
      else
            self:StopUnitMonitor();
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:InterruptTimer("Massacre");
            self:InterruptTimer("Feud");
            self:InterruptTimer("CausticSlime");
        end
    end,

    OnProtectionChange = function(self, unit) -- Unit is a GUID or unitId.
        if self:IsMe(unit) then
            timeLeft = self:GetEffectInfo("player", 82705);
            if timeLeft then
                self.protection = true;
          else
                self.protection = false;
            end

            if self:GetState() == "ENGAGED" and self:IsUnitAlive("player") then
                if self.protection then
                    self:ChangeMinRange(6);
                    self:AlertCategorized("all/normal", self:Localize("SpreadOut"));
              else
                    self:ChangeMinRange(0);
                    self:AlertCategorized("all/high", self:Localize("Regroup"));
                end
            end
        end
    end,

    OnBreak = function(self, guid, stack)
        if self:IsMe(guid) then
            self:AlertThreshold("BreakThreshold", stack, self:Localize("Break"));
        end
    end,

    OnDoubleAttack = function(self, guid)
        self:AlertSpell("shield/high", 88826, guid);
    end,

    OnCausticSlimeTarget = function(self, guid, name, amount)
        tinsert(self.causticSlime, guid);

        if amount >= 50000 and not self.protection then
            self:TriggerError("NotStacked", name);
        end
    end,

    OnCausticSlime = function(self)
        local warn = self:GetSetting("SlimeWarning");
        local symbol = self:GetSetting("SlimeSymbol");
        if warn then
            self:AnnounceUnitListSpell(self.causticSlime, 82935, symbol, 5);
    elseif symbol then
            self:PlaceSymbolList(self.causticSlime, 5);
        end
        wipe(self.causticSlime);
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(43296, false, false); -- ID, volatile, bypass
M:DefineBerserk(450);
M:DefineMinRange(0); -- Force the radar to be shown before the engage.

-- Events

M:RegisterEffectGainEvent(82881, "Break");
M:RegisterEffectGainEvent(88826, "DoubleAttack");
M:RegisterEffectGainEvent(82705, "ProtectionChange");
M:RegisterEffectFadeEvent(82705, "ProtectionChange");
M:RegisterCastEvent(82934, "Mortality");
M:RegisterCastEvent(88872, "Feud");
M:RegisterCastEvent(82848, "Massacre");
M:RegisterDamageEvent(82935, "CausticSlimeTarget");

-- Timers

M:DefineInitialRepeatingVoidTimer("Massacre/82848", "all/critical", 30, 30);
M:DefineInitialRepeatingVoidTimer("Feud/88872", "all/high", 90, 90);
M:DefineVoidTimer("FeudEnd", "all/high", 30);
M:DefineVoidTimer("CausticSlime/82935", "healer/critical", 15);

-- Actions

M:DeclareAction("Mortality");
M:AddCallbackToAction("Mortality", "OnPhaseTwo");

M:DeclareAction("Break");
M:AddCallbackToAction("Break", "OnBreak", "targetguid", "stack");

M:DeclareAction("DoubleAttack");
M:AddCallbackToAction("DoubleAttack", "OnDoubleAttack", "actorguid");

M:DeclareAction("Feud");
M:AddTriggerTimerToAction("Feud", "Feud");
M:AddStartTimerToAction("Feud", "FeudEnd");
M:AddStartTimerToAction("Feud", "CausticSlime");

M:DeclareAction("Massacre");
M:AddTriggerTimerToAction("Massacre", "Massacre");

M:DeclareAction("ProtectionChange");
M:AddCallbackToAction("ProtectionChange", "OnProtectionChange", "targetguid");

M:DeclareAction("CausticSlimeTarget");
M:AddCallbackToAction("CausticSlimeTarget", "OnCausticSlimeTarget", "targetguid", "targetname", "amount");
M:AddExternalActionToAction("CausticSlimeTarget", "CausticSlime");

M:DeclareAction("CausticSlime", 2.5, 0.8);
M:AddCallbackToAction("CausticSlime", "OnCausticSlime");

-- Errors

M:DefineBossAggroError();
M:DefineCustomError("NotStacked");
M:DefineCustomError("AttemptedHealing");