﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local CHOGALL_ID = 43324;
local ADHERENT_ID = 43622;

local function corruptionFunc(self, uid)
    local corruption, corruptionMax = UnitPower(uid, ALTERNATE_POWER_INDEX), UnitPowerMax(uid, ALTERNATE_POWER_INDEX);
    local ratio = corruption / max(1, corruptionMax);
    if corruption >= 75 then
        self:SetBarColor(1, 0, 0);
elseif corruption >= 50 then
        self:SetBarColor(1, 0.5, 0);
elseif corruption >= 25 then
        self:SetBarColor(1, 1, 0);
  else
        self:SetBarColor(0, 1, 0);
    end
    if corruption == 100 then
        self:SetTextColor(1, 0, 0);
  else
        self:SetTextColor(1, 1, 1);
    end
    self:SetBarFill(ratio);
    self:SetText(corruption);
end

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "ShowCorruption",
        type = "BOOLEAN",
        default = false,
        label = {
            ["enUS"] = "Show corruption",
            ["frFR"] = "Montrer l'état de la corruption",
        },
        explain = {
            ["enUS"] = "Display everyone's corruption in a list.",
            ["frFR"] = "Affiche la corruption de tout le monde dans une liste.",
        },
    },
    {
        id = "Conversion",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 91317,
                 symbol = 1,
                 multiple = true },
    },
    {
        id = "Crash",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 81685,
                 symbol = 7,
                 silent = true },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["FirstFury"] = "First Fury",
        ["Fire"] = "You're burning!",
        ["Sickness"] = "You're about to vomit!",
        ["CorruptionUI"] = "Corruption",
    },
    ["frFR"] = {
        ["FirstFury"] = "Première fureur",
        ["Fire"] = "Vous brûlez !",
        ["Sickness"] = "Vous allez vomir !",
        ["CorruptionUI"] = "Corruption",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Chogall", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    lastSickness = { },
    conversion = { },
    initialAdherent = false,
    adherentLeft = 0,
    crashGUID = "",
    crashAttempts = 0,

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(80);
        self:SetScoreBenchmarks(400, 600);
        self:DefineNearVictory(0.200, 0.80, 60);

        self:RegisterRespawnableAdd(ADHERENT_ID);
        self:IgnoreAdds();

        self:CheckCorruptionDisplay();
    end,

    OnBossFound = function(self, uid)
        self:RegisterHealthThreshold(0.85, self:Localize("FirstFury"));
        self:RegisterHealthThreshold(0.25, self:Localize("PhaseTwo"));
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();

        self:CheckSickness();
    end,

    OnSettingUpdate = function(self, id, value)
        if id == "ShowCorruption" then
            self:CheckCorruptionDisplay();
        end
    end,

    OnDeath = function(self, guid, name)
        local mobID = self:GetMobID(guid);
        if mobID == ADHERENT_ID then
            self.adherentLeft = max(0, self.adherentLeft - 1);
            if self.adherentLeft == 0 then
                self:InterruptTimer("FesteringBlood");
            end
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckCorruptionDisplay = function(self)
        if self:GetSetting("ShowCorruption") then
            self:StartUnitMonitor(self:Localize("CorruptionUI"), corruptionFunc);
      else
            self:StopUnitMonitor();
        end
    end,

    CheckSickness = function(self)
        self:EnumerateRaid();
        local i, uid, guid, timeLeft;
        for i=1, self:GetEnumerationSize() do
            uid = self:GetEnumerationUID(i);
            guid = UnitGUID(uid);
            timeLeft = self:GetEffectInfo(uid, 82235);
            if timeLeft then
                local elapsed = GetTime() - (self.lastSickness[guid] or 0);
                if elapsed > 8 then
                    self:OnSickness(guid, timeLeft);
                end
            end
        end

        self:StartTimer("CheckSickness");
    end,

    CheckCrash = function(self)
        local tGUID, tName, tUID = self:GetTargetInfo(self.crashGUID);
        local aGUID, aName, aUID = self:GetTankUnit(self.crashGUID);
        if tGUID and tGUID ~= aGUID and tUID then
            if self:IsMe(tUID) or self:GetDistance("player", tUID) < 10 then
                self:AlertSpell("all/high", 81685, tUID);
            end
            self:ShowZone("Crash", tUID);
            if self:GetSetting("Crash") then
                self:PlaceSymbol(tUID, 7, 4);
            end
      else
            self.crashAttempts = self.crashAttempts - 1;
            if self.crashAttempts > 0 then
                self:StartTimer("CheckCrash");
            end
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self.adherentLeft = 0;
            self:InterruptTimer("FesteringBlood");
            self:InterruptTimer("Adherent");
            self:InterruptTimer("Conversion");
        end
    end,

    OnCrash = function(self, guid)
        self.crashGUID = guid;
        self.crashAttempts = 10;

        self:InterruptTimer("CheckCrash");
        self:CheckCrash();
    end,

    OnFury = function(self)
        self:AlertIncomingSpell("shield/high", 82524);
    end,

    OnAdherent = function(self)
        local numSpawn = 1;
        if self:IsHeroic() then
            numSpawn = 2;
        end
        self.initialAdherent = true;
        self.adherentLeft = self.adherentLeft + numSpawn;
        self:StartTimer("FesteringBlood");
    end,

    OnConversionTarget = function(self, guid)
        tinsert(self.conversion, guid);
    end,

    OnConversion = function(self)
        if self:GetSetting("Conversion") then
            self:AnnounceUnitListSpell(self.conversion, 91317, true, 8);
        end
        wipe(self.conversion);

        local nextTimer = nil;
        if not self.initialAdherent then
            nextTimer = 22.5;
        end
        self:StartTimer("Conversion", nextTimer);
    end,

    OnFire = function(self, guid)
        if self:IsMe(guid) and self:CheckCooldown("Fire", 5) then
            self:AlertCategorized("all/normal", self:Localize("Fire"));
        end
    end,

    OnSickness = function(self, guid, timeLeft)
        self.lastSickness[guid] = GetTime();

        self:ShowZone("Sickness", guid, timeLeft);

        if self:IsMe(guid) and timeLeft > 2.5 then
            self:AlertCategorized("all/high", self:Localize("Sickness"));
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(CHOGALL_ID, false, false); -- ID, volatile, bypass
M:DefineBerserk(600);

-- Events

M:RegisterCastEvent(82524, "Fury");
M:RegisterCastEvent(81628, "Adherent");
M:RegisterCastEvent(82630, "ConsumeBlood");
M:RegisterCastEvent(82414, "Creations");
M:RegisterCastEvent(81685, "Crash");
M:RegisterEffectGainEvent(91317, "ConversionTarget");
M:RegisterDamageEvent(81538, "Fire");

-- Timers

M:DefineTimer("CheckCrash", "invisible", 0.05, nil, "CheckCrash", true);
M:DefineTimer("CheckSickness", "invisible", 0.5, nil, "CheckSickness", true);
M:DefineRepeatingVoidTimer("Creations/82414", "dps/high", 30);
M:DefineRepeatingVoidTimer("Adherent/81628", "shield/high", 90);
M:DefineRepeatingVoidTimer("Fury/82524", "shield/high", 47.5);
M:DefineInitialVoidTimer("Conversion/91317", "all/high", 35, 10);
M:DefineVoidTimer("FesteringBlood/82299", "offensive/high", 40);

-- Zones

M:DefineAnchoredZone("Sickness",  "all/high", "yellow", 5, 5, true);
M:DefineAimedZone(   "Crash", "all/critical",    "red", 8, 5, true);

-- Actions

M:DeclareAction("Fury");
M:AddTriggerTimerToAction("Fury", "Fury");
M:AddCallbackToAction("Fury", "OnFury");

M:DeclareAction("Adherent", 10);
M:AddTriggerTimerToAction("Adherent", "Adherent");
M:AddCallbackToAction("Adherent", "OnAdherent");

M:DeclareAction("ConsumeBlood");
M:AddCallbackToAction("ConsumeBlood", "OnPhaseTwo");

M:DeclareAction("Creations");
M:AddTriggerTimerToAction("Creations", "Creations");

M:DeclareAction("Crash");
M:AddCallbackToAction("Crash", "OnCrash", "actorguid");

M:DeclareAction("CheckCrash");
M:AddCallbackToAction("CheckCrash", "CheckCrash");

M:DeclareAction("CheckSickness");
M:AddCallbackToAction("CheckSickness", "CheckSickness");

M:DeclareAction("ConversionTarget");
M:AddCallbackToAction("ConversionTarget", "OnConversionTarget", "targetguid");
M:AddExternalActionToAction("ConversionTarget", "Conversion");

M:DeclareAction("Conversion", 10, 0.5);
M:AddTriggerTimerToAction("Conversion", "Conversion");
M:AddCallbackToAction("Conversion", "OnConversion");

M:DeclareAction("Fire");
M:AddCallbackToAction("Fire", "OnFire", "targetguid");

-- Errors

M:DefineBossAggroError();
M:DefineAvoidEffectError(91331, 2); -- Twisted Devotion
M:DefineAvoidDamageError(81538, 1); -- Fire
M:DefineAvoidDamageError(82919); -- Sprayed Corruption
M:DefineAvoidDamageError(81689); -- Corrupting Crash
M:DefineAvoidDamageError(81713); -- Depravity