﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local VALIONA_ID   = 45992;
local THERALION_ID = 45993;

local MAX_GUIDES = 5;

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "Blackout",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 86788,
                 symbol = 7 },
    },
    {
        id = "EngulfingMagic",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 86622,
                 symbol = 1,
                 multiple = true },
    },
    {
        id = "Guides",
        type = "STRING",
        max = 100,
        default = "Player1 ; Player2",
        exclusive = true,
        action = "SYMBOL",
        label = {
            ["enUS"] = "Guides",
            ["frFR"] = "Guides",
        },
        explain = {
            ["enUS"] = "Enter here the names of players which should be followed during Theralion's phase.\n\nSymbols will be put on them, starting from diamond.",
            ["frFR"] = "Entrez ici les noms des joueurs qui doivent être suivis durant la phase de Theralion.\n\nDes symboles seront placés sur eux, en commençant par le losange.",
        },
    },

    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Title"] = "Dragon Twins",
        ["Breath"] = "Breath incoming!",
        ["DamageUndivided"] = "has taken undivided damage.",
    },
    ["frFR"] = {
        ["Title"] = "Les dragons jumeaux",
        ["Breath"] = "Souffle imminent !",
        ["DamageUndivided"] = "a subi des dégâts non divisés.",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("DragonTwins", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    guides = { },
    putSymbols = false,
    engulfingMagic = { },

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(60);
        self:SetScoreBenchmarks(400, 600);

        self:RegisterSpellYell(86059, "DeepBreath");
    end,

    OnEngaged = function(self)
        local list, decider = self:GetSetting("Guides");
        self:FillPlayerList(self.guides, list);
        self.putSymbols = self:IsMe(decider);

        self:OnPhaseOne();
    end,

    OnDamage = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed)
        local nextBoss = self:GetInactiveGUID()
        if (not spellId) and (nextBoss and actorGUID == nextBoss) then
            self:TogglePhase();
        end
    end,

    -- *******
    -- Methods
    -- *******

    GetInactiveGUID = function(self)
        if self:GetPhase() == 1 then
            return self:GetMobGUID(THERALION_ID);
      else
            return self:GetMobGUID(VALIONA_ID);
        end
    end,

    TogglePhase = function(self)
        if self:GetPhase() == 1 then
            self:OnPhaseTwo();
      else
            self:OnPhaseOne();
        end
    end,

    EnableGuides = function(self)
        local i, guid;
        for i=1, min(MAX_GUIDES, #self.guides) do
            guid = self.guides[i];
            self:ShowZone("SafeZone", guid);

            if self.putSymbols then
                self:PlaceSymbol(guid, 2 + i);
            end
        end
    end,

    DisableGuides = function(self)
        local i, guid;
        for i=1, min(MAX_GUIDES, #self.guides) do
            guid = self.guides[i];
            self:InterruptZone("SafeZone", guid);

            if self.putSymbols then
                self:ClearSymbol(guid, 2 + i);
            end
        end
    end,

    ScanMeteorite = function(self)
        -- Maybe don't care because there's already plenty of hints for phase 2 positionning.
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
        self:DisableGuides();
        self:ChangeMinRange(8);

        self:InterruptTimer("EngulfingMagic");

        self:SetMainUnit(VALIONA_ID);
    end,

    OnPhaseTwo = function(self)
        self:SetPhase(2);
        self:EnableGuides();
        self:ChangeMinRange(0);

        self:InterruptTimer("Blackout");
        self:InterruptTimer("DevouringFlames");

        self:SetMainUnit(THERALION_ID);
    end,

    OnBlackoutStart = function(self, guid)
        self:TriggerTimer("Blackout");
        self:ShowZone("Blackout", guid);
        if self:GetSetting("Blackout") then
            self:AnnounceUnitSpell(86788, guid, 7);
        end
        self:AlertSpell("all/high", 86788, guid);
    end,

    OnBlackoutEnd = function(self, guid)
        self:InterruptZone("Blackout", guid);
        if self:GetSetting("Blackout") then
            self:ClearSymbol(guid);
        end
    end,

    OnDeepBreath = function(self)
        self:AlertCategorized("all/high", self:Localize("Breath"));
    end,

    OnDevouringFlames = function(self)
        self:AlertCategorized("all/high", self:Localize("Breath"));
    end,

    OnEngulfingMagicTarget = function(self, guid)
        if self:IsMe(guid) then
            self:Bomb(guid);
        end
        tinsert(self.engulfingMagic, guid);
    end,

    OnEngulfingMagic = function(self, guid)
        if self:GetSetting("EngulfingMagic") then
            self:AnnounceUnitListSpell(self.engulfingMagic, 86622, true, 20);
        end
        wipe(self.engulfingMagic);
    end,

    OnMeteorDamage = function(self, name, amount)
        if amount > 75000 then
            self:TriggerError("DamageUndivided", name);
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss({VALIONA_ID, THERALION_ID}, false, false); -- ID, volatile, bypass
M:DefineBerserk(600);

-- Events

M:RegisterDamageEvent(86014, "MeteorDamage");
M:RegisterCastEvent({86840, 90950}, "DevouringFlames");
M:RegisterEffectGainEvent(86622, "EngulfingMagicTarget");
M:RegisterEffectGainEvent(86788, "BlackoutStart");
M:RegisterEffectFadeEvent(86622, "EngulfingMagicFade");
M:RegisterEffectFadeEvent(86788, "BlackoutEnd");

-- Timers

M:DefineRepeatingVoidTimer("EngulfingMagic/86622",       "all/normal", 37.5);
M:DefineInitialRepeatingVoidTimer("Blackout/86788",        "all/high",   45,   10);
M:DefineInitialRepeatingVoidTimer("DevouringFlames/86840", "all/high",   40, 27.5);

-- Zones

M:DefineAnchoredZone("EngulfingMagic", "all/high",    "red", 10,  20, true);
M:DefineAnchoredZone(      "Blackout", "all/high", "yellow",  8,  15, true);
M:DefineAnchoredZone(      "SafeZone", "all/high",  "green",  8, 180, true);

-- Actions

M:DeclareAction("DevouringFlames");
M:AddTriggerTimerToAction("DevouringFlames", "DevouringFlames");
M:AddCallbackToAction("DevouringFlames", "OnDevouringFlames");

M:DeclareAction("DeepBreath");
M:AddCallbackToAction("DeepBreath", "OnDeepBreath");

M:DeclareAction("BlackoutStart");
M:AddCallbackToAction("BlackoutStart", "OnBlackoutStart", "targetguid");

M:DeclareAction("BlackoutEnd");
M:AddCallbackToAction("BlackoutEnd", "OnBlackoutEnd", "targetguid");

M:DeclareAction("EngulfingMagicTarget");
M:AddShowZoneToAction("EngulfingMagicTarget", "EngulfingMagic");
M:AddCallbackToAction("EngulfingMagicTarget", "OnEngulfingMagicTarget", "targetguid");
M:AddExternalActionToAction("EngulfingMagicTarget", "EngulfingMagic");

M:DeclareAction("EngulfingMagic", 5, 0.5);
M:AddTriggerTimerToAction("EngulfingMagic", "EngulfingMagic");
M:AddCallbackToAction("EngulfingMagic", "OnEngulfingMagic");

M:DeclareAction("EngulfingMagicFade");
M:AddHideZoneToAction("EngulfingMagicFade", "EngulfingMagic");

M:DeclareAction("MeteorDamage");
M:AddCallbackToAction("MeteorDamage", "OnMeteorDamage", "targetname", "amount");

-- Errors

M:DefineCustomError("DamageUndivided");
M:DefineAvoidEffectError(86202); -- Twilight Shift
M:DefineAvoidDamageError({86199, 86228}); -- Deep Breath
M:DefineAvoidDamageError({86840, 90950}, 2); -- Devouring Flames
M:DefineAvoidDamageError(86505, 3); -- Fabulous Flames
M:DefineAvoidBombError(86622, 4); -- Engulfing Magic