﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local ARCANOTRON_ID = 42166;
local ELECTRON_ID   = 42179;
local MAGMATRON_ID  = 42178;
local TOXITRON_ID   = 42180;

local powerAlgorithm = {
    label = {
        ["enUS"] = "EP",
        ["frFR"] = "PE",
    },
    color = {1, 1, 0},
    allowDead = false,
    digits = 3,
    useMultiples = false,

    GetValue = function(self, uid)
        return UnitPower(uid), UnitPowerMax(uid);
    end,
};

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "PoisonThreshold",
        type = "NUMBER",
        min = 1,
        max = 15,
        step = 1,
        default = 5,
        meta = { spell = 80011,
                 threshold = true },
    },
    {
        id = "Fixate",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 80094,
                 symbol = 1,
                 multiple = true },
    },
    {
        id = "InterruptChain",
        type = "STRING",
        max = 100,
        default = "Player1 ; Player2",
        exclusive = true,
        action = "RAID_WARNING",
        label = {
            ["enUS"] = "Players assigned to interrupt",
            ["frFR"] = "Joueurs assignés à l'interruption",
        },
        explain = {
            ["enUS"] = "Enter here the names of players who are supposed to interrupt Arcanotron, separated by a semicolon.\n\nLeave the field blank if you don't want to use this feature.",
            ["frFR"] = "Entrez ici les noms des joueurs assignés à l'interruption d'Arcanotron, séparés par un point-virgule.\n\nLaissez le champ vide si vous ne souhaitez pas utiliser cet outil.",
        },
    },

    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Title"] = "Tron Council",
        ["Poison"] = "You have the Poison %d times!",
        ["FixateAlert"] = "%s is the target of a Bomb!",
        ["FixateSelf"] = "You are the target of a Bomb!",
        ["Stop"] = "Stop attacking %s!",
        ["BossPowerUp"] = "Take the boss out of the powerup!",
        ["LockTargetSelf"] = "Flamethrower on you!",
        ["LockTargetOther"] = "Flamethrower on %s!",
        ["AvoidFlamethrower"] = "was aligned with the flamethrower's target.",
    },
    ["frFR"] = {
        ["Title"] = "Conseil des Tron",
        ["Poison"] = "Vous avez %d fois le poison !",
        ["FixateAlert"] = "%s est la cible d'une bombe !",
        ["FixateSelf"] = "Vous êtes la cible d'une bombe !",
        ["Stop"] = "Arrêtez d'attaquer %s !",
        ["BossPowerUp"] = "Sortez le boss de la zone de puissance !",
        ["LockTargetSelf"] = "Lance-flamme sur vous !",
        ["LockTargetOther"] = "Lance-flamme sur %s !",
        ["AvoidFlamethrower"] = "était aligné avec la cible du lance-flamme.",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("TronCouncil", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    nextSymbol = 1,
    lastFlamethrowerTarget = "",

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(55);
        self:SetScoreBenchmarks(400, 600);

        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:SetExtraAlgorithm(powerAlgorithm);
        end
    end,

    OnEngaged = function(self)
        self:StartInterruptChain("InterruptChain");
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnActive = function(self, guid)
        local id = self:GetMobID(guid);
        if id == ARCANOTRON_ID then
            self:StartTimer("PowerUp", 10);

    elseif id == ELECTRON_ID then
            self:StartTimer("Conductor", 10);

    elseif id == MAGMATRON_ID then
            self:StartTimer("LockTarget", 20);

    elseif id == TOXITRON_ID then
            self:StartTimer("PoisonProtocol");
        end
    end,

    OnInactive = function(self, guid)
        local id = self:GetMobID(guid);
        if id == ARCANOTRON_ID then
            self:InterruptTimer("PowerUp");

    elseif id == ELECTRON_ID then
            self:InterruptTimer("Conductor");

    elseif id == MAGMATRON_ID then
            self:InterruptTimer("LockTarget");

    elseif id == TOXITRON_ID then
            self:InterruptTimer("PoisonProtocol");
        end
    end,

    OnPoison = function(self, guid, stack)
        if self:IsMe(guid) then
            self:AlertThreshold("PoisonThreshold", stack, self:Localize("Poison"));
        end
    end,

    OnFixate = function(self, guid)
        if self:GetSetting("Fixate") then
            self:AnnounceUnit(self:Localize("FixateAlert"), guid, self.nextSymbol, 15);
            self.nextSymbol = math.fmod(self.nextSymbol, 5) + 1;
        end
        if self:IsMe(guid) then
            self:AlertCategorized("all/high", self:Localize("FixateSelf"));
        end
    end,

    OnStop = function(self, guid, name)
        if UnitGUID("target") == guid then
            self:AlertCategorized("dps/critical", self:FormatLoc("Stop", name));
        end
    end,

    OnArcanotronSpell = function(self)
        if not self:IsInterruptChainRunning() then
            self:InterruptMob(ARCANOTRON_ID, "normal");
        end
    end,

    OnPowerUp = function(self, guid)
        if self:IsUnitAffiliated(guid) then
            if self:CheckCooldown("PowerUp", 30) then
                self:ShowZone("PowerUp", guid);
            end
      else
            local tankGUID = self:GetTankUnit(guid);
            if self:IsMe(tankGUID) then
                self:AlertCategorized("all/high", self:Localize("BossPowerUp"));
            end
        end
    end,

    OnBomb = function(self, guid)
        self:Bomb(guid);
    end,

    OnLockTarget = function(self, guid)
        self.lastFlamethrowerTarget = guid;

        if self:IsMe(guid) then
            self:AlertCategorized("all/high", self:Localize("LockTargetSelf"));
      else
            self:AlertUnitCategorized("healer/high", self:Localize("LockTargetOther"), guid);
        end
    end,

    OnFlamethrowerDamage = function(self, guid, name)
        if guid ~= self.lastFlamethrowerTarget then
            self:TriggerError("AvoidFlamethrower", name);
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss({ARCANOTRON_ID, ELECTRON_ID, MAGMATRON_ID, TOXITRON_ID}, false, false); -- ID, volatile, bypass
M:DefineBerserk(nil, 600);

-- Events

M:RegisterEffectGainEvent(92048, "ShadowInfusion");
M:RegisterEffectGainEvent(80094, "Fixate");
M:RegisterEffectGainEvent(78726, "Inactive");
M:RegisterEffectGainEvent(79888, "Conductor");
M:RegisterEffectGainEvent(79501, "LockTarget");
M:RegisterEffectGainEvent({79628, 79629}, "PowerUpGain");
M:RegisterEffectGainEvent(80011, "Poison");
M:RegisterEffectGainEvent(78740, "Active");
M:RegisterCastEvent(79624, "PowerUpCast");
M:RegisterCastEvent(80053, "PoisonProtocol");
M:RegisterCastEvent(72729, "Stop"); -- Power Conversion
M:RegisterCastEvent(79582, "Stop"); -- Barrier
M:RegisterCastEvent(79710, "ArcanotronSpell");
M:RegisterDamageEvent(79504, "Flamethrower");

-- Timers

M:DefineRepeatingVoidTimer(    "LockTarget/79501", "all/high", 40);
M:DefineRepeatingVoidTimer(     "Conductor/79888", "all/normal", 30);
M:DefineRepeatingVoidTimer("ShadowInfusion/92048", "all/high", 30);
M:DefineRepeatingVoidTimer(       "PowerUp/79624", "caster/normal", 30);
M:DefineRepeatingVoidTimer("PoisonProtocol/80053", "all/high", 45);
M:DefineInitialRepeatingVoidTimer( "Active/78740", "all/high", 45);

-- Zones

M:DefineAnchoredZone("Conductor",      "all/high",    "red",  8, 10, true);
M:DefineAnchoredZone("ShadowInfusion", "all/high",    "red", 16,  5, true);
M:DefineAnchoredZone("LockTarget", "all/critical", "yellow",  6,  8);
M:DefineAimedZone(   "PowerUp",        "all/high",  "green",  5, 25);

-- Actions

M:DeclareAction("ShadowInfusion");
M:AddTriggerTimerToAction("ShadowInfusion", "ShadowInfusion");
M:AddShowZoneToAction("ShadowInfusion", "ShadowInfusion");
M:AddCallbackToAction("ShadowInfusion", "OnBomb", "targetguid");

M:DeclareAction("Conductor");
M:AddTriggerTimerToAction("Conductor", "Conductor");
M:AddShowZoneToAction("Conductor", "Conductor");
M:AddCallbackToAction("Conductor", "OnBomb", "targetguid");

M:DeclareAction("LockTarget");
M:AddTriggerTimerToAction("LockTarget", "LockTarget");
M:AddShowZoneToAction("LockTarget", "LockTarget");
M:AddCallbackToAction("LockTarget", "OnLockTarget", "targetguid");

M:DeclareAction("Poison");
M:AddCallbackToAction("Poison", "OnPoison", "targetguid", "stack");

M:DeclareAction("Fixate", 1);
M:AddCallbackToAction("Fixate", "OnFixate", "targetguid");

M:DeclareAction("Stop");
M:AddCallbackToAction("Stop", "OnStop", "actorguid", "actorname");

M:DeclareAction("ArcanotronSpell");
M:AddCallbackToAction("ArcanotronSpell", "OnArcanotronSpell");

M:DeclareAction("PowerUpGain");
M:AddCallbackToAction("PowerUpGain", "OnPowerUp", "targetguid");

M:DeclareAction("Active");
M:AddTriggerTimerToAction("Active", "Active");
M:AddCallbackToAction("Active", "OnActive", "targetguid");

M:DeclareAction("Inactive");
M:AddCallbackToAction("Inactive", "OnInactive", "targetguid");

M:DeclareAction("Flamethrower");
M:AddCallbackToAction("Flamethrower", "OnFlamethrowerDamage", "targetguid", "targetname");

M:DeclareAction("PowerUpCast");
M:AddTriggerTimerToAction("PowerUpCast", "PowerUp");

M:DeclareAction("PoisonProtocol");
M:AddTriggerTimerToAction("PoisonProtocol", "PoisonProtocol");

-- Errors

M:DefineBossAggroError(100); -- 100 sec grace period.
M:DefineCustomError("AvoidFlamethrower", 0);
M:DefineAvoidDamageError(80092); -- Poison Bomb
M:DefineAvoidDamageError(79617); -- Magmatron shield backlash (Backdraft)
M:DefineAvoidDamageError(79912); -- Electron shield backlash (Static Shock)
M:DefineAvoidBombError(79889, 1); -- Conductor

-- Special

M:DefineInterruptChain(79710, 1, 15); -- Arcanic Annihilator interruptions