﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local BOSS_ID = 52530;

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "OptimizeSettings",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Optimize settings",
            ["frFR"] = "Optimiser les réglages",
        },
        explain = {
            ["enUS"] = "Remove visual filler for phase 2: reduce view range and particle density.",
            ["frFR"] = "Réduit le remplissage visuel pendant la phase 2 : réduit la distance de vue et la densité des particules.",
        },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Hatch"] = "Hatch",
        ["BlazingPower"] = "Blazing Power",
        ["FieryVortexTrigger"] = "These skies",
    },
    ["frFR"] = {
        ["Hatch"] = "Eclosion",
        ["BlazingPower"] = "Puissance flamboyante",
        ["FieryVortexTrigger"] = "Ce ciel",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Alysrazor", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(65);
        self:SetScoreBenchmarks(540, 780);

        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:SetExtraAlgorithm("power");
        end
    end,

    OnBossFound = function(self, uid)
        self:RegisterAnyYell(self:Localize("FieryVortexTrigger"), "FieryVortex");
    end,

    OnEngaged = function(self)
        self:OnPhaseOne(false);
    end,

    -- *******
    -- Methods
    -- *******

    IsIncomplete = function(self) -- Override!
        return false;
    end,

    GetBlazingPowerCount = function(self)
        local _, stackCount = self:GetEffectInfo("player", 99461);
        if stackCount then
            return stackCount, 25;
      else
            return 0, 25;
        end
    end,

    GetWingTimer = function(self)
        local timeLeft = self:GetEffectInfo("player", 98619);
        if timeLeft then
            return timeLeft, 30;
      else
            return 0, 30;
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self, restart)
        self:SetPhase(1);

        self:SetMainUnit(nil);

        if restart then
            self:StartTimer("Hatch");
            if self:IsHeroic() then
                self:StartTimer("FieryVortex", 225);
          else
                self:StartTimer("FieryVortex");
            end
    elseif self:IsHeroic() then
            self:InterruptTimer("FieryVortex");
            self:StartTimer("FieryVortex", 243);
        end
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:InterruptTimer("Molting");

            if self:GetSetting("OptimizeSettings") then
                self:OverrideCVar("particleDensity", 15);
                self:OverrideCVar("farclip", 125);
                self:OverrideCVar("shadowMode", 0);
                self:OverrideCVar("reflectionMode", 0);
                self:OverrideCVar("rippleDetail", 0);
                self:OverrideCVar("waterDetail", 0);
            end

            self:ShowZone("Vortex");
            self:ShowZone("Start");
        end
    end,

    OnPhaseThree = function(self)
        if self:SetIncreasingPhase(3) then
            self:SetMainUnit(BOSS_ID);

            self:RestoreCVars();

            self:InterruptZone("Vortex");
            self:InterruptZone("Start");
        end
    end,

    OnPhaseFour = function(self)
        self:SetIncreasingPhase(4);
    end,

    OnFieroblast = function(self, guid)
        self:InterruptMob(guid);
    end,

    OnTantrum = function(self, guid)
        local aGUID = self:GetTankUnit(guid);
        if self:IsMe(aGUID) then
            self:AlertSpell("tank/high", 99362, guid);
        end
    end,

    OnGushingWound = function(self, guid)
        self:AlertSpell("shield/high", 99308, guid);
    end,

    OnIgnition = function(self, guid)
        self:InterruptMob(guid);
    end,

    -- Abilities used for phase transitions

    OnFieryVortex = function(self)
        self:AlertIncomingSpell("all/critical", 99793);
        self:InterruptTimer("FieryVortex");
        self:OnPhaseTwo();
    end,

    OnFullPower = function(self)
        self:OnPhaseOne(true); -- Go back to phase 1.
    end,

    OnBurnoutGain = function(self)
        self:OnPhaseThree();
    end,

    OnBurnoutFade = function(self)
        self:OnPhaseFour();
    end,

    -- Mirror bar toggle

    OnWingGain = function(self, guid)
        if self:IsMe(guid) then
            self:StartMirrorValue(self:Localize("BlazingPower"), self.GetBlazingPowerCount, self.GetWingTimer);
        end
    end,

    OnWingFade = function(self, guid)
        if self:IsMe(guid) then
            self:StopMirrorValue();
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(BOSS_ID, true, false); -- ID, volatile, bypass
M:DefineFailCooldown(60);

-- Events

M:RegisterCastEvent({99464, 100836}, "Molting");
M:RegisterCastEvent(101223, "Fieroblast");
M:RegisterDamageEvent(99925, "FullPower");
M:RegisterEffectGainEvent(99432, "BurnoutGain");
M:RegisterEffectFadeEvent(99432, "BurnoutFade");
M:RegisterEffectGainEvent(99362, "Tantrum");
M:RegisterEffectGainEvent(99308, "GushingWound");
M:RegisterEffectGainEvent(99919, "Ignition");
M:RegisterEffectGainEvent(98619, "WingGain");
M:RegisterEffectFadeEvent(98619, "WingFade");

-- Timers

M:DefineRepeatingVoidTimer("Molting/99464", "dps/high", 60);
M:DefineInitialVoidTimer("Hatch", "all/high", 27.5, 45);
M:DefineInitialVoidTimer("FieryVortex/99794", "all/critical", 170, 190);

-- Zones

M:DefineFixedZone("Vortex", "all/critical", "red", 15, 999, 0.62693, 0.44203);
M:DefineFixedZone("Start", "all/high", "green", 5, 10, 0.61442, 0.43172);

-- Actions

M:DeclareAction("FieryVortex");
M:AddCallbackToAction("FieryVortex", "OnFieryVortex");

M:DeclareAction("Molting");
M:AddTriggerTimerToAction("Molting", "Molting");

M:DeclareAction("Fieroblast");
M:AddCallbackToAction("Fieroblast", "OnFieroblast", "actorguid");

M:DeclareAction("FullPower", 30);
M:AddCallbackToAction("FullPower", "OnFullPower");

M:DeclareAction("BurnoutGain");
M:AddCallbackToAction("BurnoutGain", "OnBurnoutGain");

M:DeclareAction("BurnoutFade");
M:AddCallbackToAction("BurnoutFade", "OnBurnoutFade");

M:DeclareAction("Tantrum");
M:AddCallbackToAction("Tantrum", "OnTantrum", "targetguid");

M:DeclareAction("GushingWound", 1.5);
M:AddCallbackToAction("GushingWound", "OnGushingWound", "targetguid");

M:DeclareAction("Ignition");
M:AddCallbackToAction("Ignition", "OnIgnition", "actorguid");

M:DeclareAction("WingGain");
M:AddCallbackToAction("WingGain", "OnWingGain", "targetguid");

M:DeclareAction("WingFade");
M:AddCallbackToAction("WingFade", "OnWingFade", "targetguid");

-- Errors

M:DefineAvoidDamageError(99794); -- Fiery Vortex