﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "BurningWoundThreshold",
        type = "NUMBER",
        min = 1,
        max = 10,
        step = 1,
        default = 4,
        meta = { spell = 99399,
                 threshold = true },
    },
    {
        id = "OptimizeSettings",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Optimize settings",
            ["frFR"] = "Optimiser les réglages",
        },
        explain = {
            ["enUS"] = "Remove as much visual filler as possible in this fight: reduce shadows, water details, particle density and view range.",
            ["frFR"] = "Retire le plus de remplissage visuel possible durant ce combat : réduit les ombres, les détails des liquides et la distance de vue.",
        },
    },
    {
        id = "SmashAlert",
        type = "BOOLEAN",
        default = false,
        label = {
            ["enUS"] = "Super alert: Sulfuras Smash",
            ["frFR"] = "Super alerte : Frappe de Sulfuras",
        },
        explain = {
            ["enUS"] = "Enable this if you need an aggressive alert when Sulfuras Smash is cast.",
            ["frFR"] = "Cette option permet d'afficher une alerte agressive quand Frappe de Sulfuras se produit.",
        },
    },
    {
        id = "Trap",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 98164,
                 symbol = 7,
                 silent = true },
    },
    {
        id = "BlazingHeat",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 100460,
                 symbol = 1,
                 multiple = true },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Transition"] = "Transition",
        ["Victory"] = "Victory",
        ["BurningWound"] = "You have Burning Wound %d times!",
        ["Boom"] = "Boom!",
        ["SmashTitle"] = "Sulfuras",
        ["SmashExplain"] = "Avoid the waves!",
    },
    ["frFR"] = {
        ["Transition"] = "Transition",
        ["Victory"] = "Victoire",
        ["BurningWound"] = "Vous avez %d fois blessure brûlante !",
        ["Boom"] = "Boum !",
        ["SmashTitle"] = "Sulfuras",
        ["SmashExplain"] = "Evitez les vagues !",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Ragnaros", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    sonsDead = 0,
    trapAttempts = 0,
    blazingHeat = 0,
    lastMeteor = { },

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(80);
        self:SetScoreBenchmarks(360, 660);
    end,

    OnBossFound = function(self, uid)
        self:RegisterHealthThreshold(0.705, self:Localize("PhaseTwo"));
        self:RegisterHealthThreshold(0.405, self:Localize("PhaseThree"));
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();

        if self:IsHeroic() then
            self:RegisterHealthThreshold(0.105, self:Localize("PhaseFour"), self.OnPhaseFour, self);
            self:DefineNearVictory(0.100, 0.90, 120);
      else
            self:RegisterHealthThreshold(0.105, self:Localize("Victory"), self.OnCleared, self);
            self:DefineNearVictory(0.250, 0.90, 180);
        end

        -- Tune options if desired
        if self:GetSetting("OptimizeSettings") then
            -- Scene details
            self:OverrideCVar("farclip", 200);
            self:OverrideCVar("particleDensity", 15);

            -- Totally useless features
            self:OverrideCVar("shadowMode", 0);
            self:OverrideCVar("reflectionMode", 0);
            self:OverrideCVar("rippleDetail", 0);
            self:OverrideCVar("waterDetail", 0);
        end
    end,

    OnDeath = function(self, guid, name)
        if self:GetMobID(guid) == 53140 then
            self.sonsDead = self.sonsDead + 1;
            if self.sonsDead == 8 then
                self:OnTransitionEnded();
            end
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckMeteor = function(self)
        self:EnumerateRaid();
        local i, uid, guid, timeLeft;
        for i=1, self:GetEnumerationSize() do
            uid = self:GetEnumerationUID(i);
            guid = UnitGUID(uid);
            timeLeft = self:GetEffectInfo(uid, 99849);
            if timeLeft then
                local elapsed = GetTime() - (self.lastMeteor[guid] or 0);
                if elapsed > 4 then
                    self:OnMeteor(guid, elapsed);
                end
            end
        end
        self:StartTimer("CheckMeteor");
    end,

    CheckTrap = function(self)
        local tGUID, tName, tUID = self:GetTargetInfo();
        local aGUID, aName, aUID = self:GetTankUnit();
        if tGUID and tGUID ~= aGUID and tUID then
            self:ShowZone("Trap", tUID);
            if self:IsMe(tUID) then
                self:AlertSpell("all/high", 98164, tUID);
            end
            if self:GetSetting("Trap") then
                self:PlaceSymbol(tUID, 7, 4);
            end
      else
            self.trapAttempts = self.trapAttempts - 1;
            if self.trapAttempts > 0 then
                self:StartTimer("CheckTrap");
            end
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:StartTimer("Smash", 19);
            self:StartTimer("MoltenSeeds", 25);
            self:StartTimer("Flames", 40);
        end
    end,

    OnPhaseThree = function(self)
        if self:SetIncreasingPhase(3) then
            self:StartTimer("LivingMeteor");
            self:StartTimer("Flames");
            self:StartTimer("CheckMeteor");
            self:InterruptTimer("MoltenSeeds");
        end
    end,

    OnPhaseFour = function(self) -- Heroic only
        if self:SetIncreasingPhase(4) then
            -- TODO later (heroic)
        end
    end,

    OnTransitionStarted = function(self, spellId)
        self.sonsDead = 0;
        if self:IsHeroic() then
            self:StartTimer("Transition", 60);
      else
            self:StartTimer("Transition");
        end
        self:InterruptTimer("Wrath");
        self:InterruptTimer("Hand");
        self:InterruptTimer("Smash");
        self:InterruptTimer("MagmaTrap");
        self:InterruptTimer("Flames");
        if spellId == 98952 then
            self:ShowZoneFixed("Sulfuras", 0.50513, 0.20861); -- Middle

    elseif spellId == 98953 then
            self:ShowZoneFixed("Sulfuras", 0.52717, 0.19665); -- East

    elseif spellId == 98951 then
            self:ShowZoneFixed("Sulfuras", 0.48346, 0.19646); -- West
        end
    end,

    OnTransitionEnded = function(self)
        self:InterruptTimer("Transition");
        if self:GetPhase() == 1 then
            self:OnPhaseTwo();
    elseif self:GetPhase() == 2 then
            self:OnPhaseThree();
        end
    end,

    OnWorldOfFlames = function(self, spellId)
        -- TODO later (heroic)
    end,

    OnMeteor = function(self, guid, elapsed)
        self.lastMeteor[guid] = GetTime();

        self:InterruptZone("Meteor", guid);
        self:ShowZone("Meteor", guid);

        if self:IsMe(guid) and elapsed > 8 then
            self:AlertSpell("all/high", 99849, guid);
        end
    end,

    OnBlazingHeat = function(self, guid)
        local count = self.blazingHeat + 1;
        self.blazingHeat = count;

        if self:GetSetting("BlazingHeat") then
            self:AnnounceUnitSpell(100460, guid, math.fmod(count, 2) + 1, 8);
        end
        if self:IsMe(guid) then
            self:AlertSpell("all/high", 100460, guid);
        end
        self:ShowZone("BlazingHeat", guid);
    end,

    OnSmash = function(self)
        self:TriggerTimer("Smash");
        if self:GetPhase() == 1 then
            self:StartTimer("Smash");
            self:StartTimer("Wrath");
      else
            self:StartTimer("Smash", 40);
        end
        if self:HasRole("CASTER") or self:IsTimerRunning("SectionMeleeFlames") then
            if self:GetSetting("SmashAlert") then
                self:SuperAlert(self:Localize("SmashTitle"), self:Localize("SmashExplain"), 4, true);
          else
                self:AlertIncomingSpell("caster/high", 98710);
            end
        end
    end,

    OnHand = function(self, guid)
        if guid == self:GetBossGUID() then
            self:TriggerTimer("Hand");
        end
    end,

    OnLivingMeteor = function(self)
        self:AlertIncomingSpell("all/high", 99268);
    end,

    OnMoltenSeeds = function(self)
        -- TODO: zones
        self:StartTimer("SectionSeeds");
    end,

    OnFlames = function(self, spellId)
        if self:GetPhase() == 2 then
            self:StartTimer("Flames", 40);
      else
            self:StartTimer("Flames");
        end
        if spellId == 99172 then
            -- North
            self:StartTimer("SectionMeleeFlames");
            if self:HasRole("MELEE") or self:IsTimerRunning("SectionSeeds") then
                -- All players are probably in melee range.
                self:SuperAlert("", "", 3, true);
            end

    elseif spellId == 99235 then
            -- TODO: middle

    elseif spellId == 99236 then
            -- TODO: south
        end
    end,

    OnMagmaTrap = function(self)
        self.trapAttempts = 10;
        self:InterruptTimer("CheckTrap");
        self:CheckTrap();
    end,

    OnBurningWound = function(self, guid, stack)
        if self:IsMe(guid) then
            self:AlertThreshold("BurningWoundThreshold", stack, self:Localize("BurningWound"));
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(52409, false, false); -- ID, volatile, bypass
M:DefineBerserk(1080);
M:DefineMinRange(6);

-- Events

M:RegisterCastEvent({98951, 98952, 98953}, "TransitionStart");
M:RegisterCastEvent({99172, 99235, 99236}, "WorldOfFlames");
M:RegisterCastEvent(100460, "BlazingHeat");
M:RegisterCastEvent(98710, "Smash");
M:RegisterCastEvent(98237, "Hand");
M:RegisterCastEvent(98263, "Wrath");
M:RegisterCastEvent({99172, 99235, 99236}, "Flames");
M:RegisterCastEvent(99268, "LivingMeteor");
M:RegisterCastEvent(98164, "MagmaTrap");
M:RegisterEffectGainEvent(100594, "Superheated");
M:RegisterEffectGainEvent(99399, "BurningWound");
M:RegisterDamageEvent(98498, "MoltenSeeds");
M:RegisterDamageEvent(98518, "MoltenSeedsDetonation");

-- Timers

M:DefineTimer("CheckMeteor", "invisible", 1, nil, "CheckMeteor", true);
M:DefineTimer("CheckTrap", "invisible", 0.05, nil, "CheckTrap", true);
M:DefineTimer("Transition", "all/high", 45, nil, "TransitionEnd", true);
M:DefineVoidTimer("SectionSeeds", "invisible", 25, nil);
M:DefineVoidTimer("SectionMeleeFlames", "invisible", 12, nil);
M:DefineVoidTimer("Boom", "all/critical", 10);
M:DefineVoidTimer("Wrath/98263", "caster/high", 12);
M:DefineVoidTimer("Flames/99171", "all/critical", 30);
M:DefineRepeatingVoidTimer("Hand/98237", "healer/high", 25);
M:DefineRepeatingVoidTimer("LivingMeteor/99268", "all/high", 45);
M:DefineRepeatingVoidTimer("MoltenSeeds/98520", "all/critical", 60);
M:DefineInitialVoidTimer("Smash/98710", "caster/critical", 30);
M:DefineInitialRepeatingVoidTimer("MagmaTrap/98164", "all/high", 25, 16);

-- Zones

M:DefineAimedZone("Trap", "all/high", "red", 8, 12.5, true);
M:DefineAnchoredZone("Meteor", "all/critical", "red", 8, 5, true);
M:DefineAnchoredZone("BlazingHeat", "all/high", "red", 6, 8, true);
M:DefineFixedZone("Sulfuras", "all/high", "red", 8, 50, true); -- During transitions.

-- Actions

M:DeclareAction("TransitionStart");
M:AddCallbackToAction("TransitionStart", "OnTransitionStarted", "spellid");

M:DeclareAction("TransitionEnd");
M:AddCallbackToAction("TransitionEnd", "OnTransitionEnded");

M:DeclareAction("WorldOfFlames");
M:AddCallbackToAction("WorldOfFlames", "OnWorldOfFlames", "spellid");

M:DeclareAction("CheckMeteor");
M:AddCallbackToAction("CheckMeteor", "CheckMeteor");

M:DeclareAction("CheckTrap");
M:AddCallbackToAction("CheckTrap", "CheckTrap");

M:DeclareAction("BlazingHeat");
M:AddCallbackToAction("BlazingHeat", "OnBlazingHeat", "targetguid");

M:DeclareAction("Smash");
M:AddCallbackToAction("Smash", "OnSmash");

M:DeclareAction("Hand");
M:AddCallbackToAction("Hand", "OnHand", "actorguid");

M:DeclareAction("Wrath", 5);
M:AddTriggerTimerToAction("Wrath", "Wrath");

M:DeclareAction("LivingMeteor", 5);
M:AddTriggerTimerToAction("LivingMeteor", "LivingMeteor");
M:AddCallbackToAction("LivingMeteor", "OnLivingMeteor");

M:DeclareAction("MoltenSeeds", 15);
M:AddTriggerTimerToAction("MoltenSeeds", "MoltenSeeds");
M:AddStartTimerToAction("MoltenSeeds", "Boom");
M:AddCallbackToAction("MoltenSeeds", "OnMoltenSeeds");

M:DeclareAction("MoltenSeedsDetonation", 15);
M:AddTriggerTimerToAction("MoltenSeedsDetonation", "Boom");

M:DeclareAction("Flames");
M:AddTriggerTimerToAction("Flames", "Flames");
M:AddCallbackToAction("Flames", "OnFlames", "spellid");

M:DeclareAction("MagmaTrap");
M:AddTriggerTimerToAction("MagmaTrap", "MagmaTrap");
M:AddCallbackToAction("MagmaTrap", "OnMagmaTrap");

M:DeclareAction("Superheated");
M:AddCallbackToAction("Superheated", "OnPhaseFour");

M:DeclareAction("BurningWound");
M:AddCallbackToAction("BurningWound", "OnBurningWound", "targetguid", "stack");

-- Errors

M:DefineAvoidEffectError(98928); -- Lava Wave
M:DefineAvoidDamageError(98708); -- Sulfuras Smash
M:DefineAvoidDamageError(99224); -- Engulfing Flame
M:DefineAvoidDamageError(99287); -- Meteor Impact