﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local MAX_GUIDES = 5;

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "OptimizeSettings",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Optimize settings",
            ["frFR"] = "Optimiser les réglages",
        },
        explain = {
            ["enUS"] = "Remove as much visual filler as possible in this fight: reduce shadows, water details, particle density, environment details, view range and disable projected textures.",
            ["frFR"] = "Retire le plus de remplissage visuel possible durant ce combat : réduit les ombres, les détails des liquides, les détails de l'environnement, la distance de vue et désactive les textures projetées.",
        },
    },
    {
        id = "Guides",
        type = "STRING",
        max = 100,
        default = "Player1 ; Player2",
        exclusive = true,
        action = "SYMBOL",
        label = {
            ["enUS"] = "Guide(s)",
            ["frFR"] = "Guide(s)",
        },
        explain = {
            ["enUS"] = "Enter here the names of players which should be followed during the encounter.\n\nSymbols will be put on them, starting from square.",
            ["frFR"] = "Entrez ici les noms des joueurs qui doivent être suivis pendant le combat.\n\nDes symboles seront placés sur eux, en commençant par le carré.",
        },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Title"] = "Driving Test",
        ["Elementals"] = "Elementals",
        ["MagmaFlow"] = "Magma flow coming!",
        ["ArmorDestroyed"] = "Armor destroyed!",
    },
    ["frFR"] = {
        ["Title"] = "Test de conduite",
        ["Elementals"] = "Elémentaires",
        ["MagmaFlow"] = "Coulée active !",
        ["ArmorDestroyed"] = "Armure détruite !",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Rhyolith", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    guides = { },
    putSymbols = false,

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(65);
        self:SetScoreBenchmarks(270, 420);

        self:RegisterAdd(53087);
        self:RegisterAdd(52577);
        self:IgnoreAdds();

        self:DisableCombatCheck();
    end,

    OnBossFound = function(self, uid)
        self:RegisterHealthThreshold(0.25, self:Localize("PhaseTwo"), self.OnPhaseTwo, self);
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();

        -- Exempt tanks from the immolate error
        self:EnumerateRaid();
        local i, uid;
        for i=1, self:GetEnumerationSize() do
            uid = self:GetEnumerationUID(i);
            if self:GetUnitRole(uid) == "TANK" then
                self:ExemptError("d:98598", uid);
                self:ExemptError("d:99845", uid);
            end
        end

        -- Get the guides list
        local list, decider = self:GetSetting("Guides");
        self:FillPlayerList(self.guides, list);
        self.putSymbols = self:IsMe(decider);

        -- Apply symbols on heal
        for i=1, min(MAX_GUIDES, #self.guides) do
            local guid = self.guides[i];
            self:ShowZone("SafeZone", guid);

            if self.putSymbols then
                self:PlaceSymbol(guid, 7 - i);
            end
        end

        -- Tune options if desired
        if self:GetSetting("OptimizeSettings") then
            if CombatText_UpdateDisplayedMessages then
                SHOW_COMBAT_TEXT = "0";
                CombatText_UpdateDisplayedMessages();
            end

            -- Scene details
            self:OverrideCVar("farclip", 500);
            self:OverrideCVar("groundEffectDist", 32);
            self:OverrideCVar("groundEffectDensity", 16);
            self:OverrideCVar("particleDensity", 50); -- Not standard
            self:OverrideCVar("showfootprintparticles", 0); -- Hidden to the typical user

            -- Totally useless features
            self:OverrideCVar("shadowMode", 0);
            self:OverrideCVar("reflectionMode", 0);
            self:OverrideCVar("rippleDetail", 0);
            self:OverrideCVar("waterDetail", 0);
            self:OverrideCVar("projectedTextures", 0);
        end
    end,

    OnUpdate = function(self)
        if UnitExists("boss1") and self:GetState() == "STANDBY" then
            self:OnEngaged();
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckSparkTank = function(self)
        self:InterruptZone("Immolate");

        local sparkGUID = self:GetMobGUID(53211);
        if sparkGUID then
            local guid = self:GetTankUnit(sparkGUID);
            if guid then
                self:ShowZone("Immolate", guid);
            end
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
        self:TriggerTimer("CheckSparkTank");
        self:SetMainUnit(nil);
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:InterruptTimer("CheckSparkTank");

            self:InterruptTimer("Elementals");
            self:InterruptTimer("HeatedVolcano");
            self:InterruptTimer("FlameStomp");
        end
    end,

    OnFlameStomp = function(self)
        if self:GetPhase() == 2 then
            self:StartTimer("FlameStomp", 13);
      else
            self:StartTimer("FlameStomp");
        end
    end,

    OnMagmaFlow = function(self)
        self:AlertCategorized("all/high", self:Localize("MagmaFlow"));
    end,

    OnArmorFade = function(self)
        self:AlertCategorized("all/high", self:Localize("ArmorDestroyed"));

        self:SetMainUnit(52558);
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(52558, true, false); -- ID, volatile, bypass
M:DeclareNPC(53087, false, false);
M:DeclareNPC(52577, false, false);
M:DefineBerserk(360, 300);

-- Events

M:RegisterCastEvent(97225, "MagmaFlow");
M:RegisterCastEvent(98493, "HeatedVolcano");
M:RegisterCastEvent(97282, "FlameStomp");
M:RegisterCastEvent({98136, 98552, 98045}, "Elementals");
M:RegisterEffectFadeEvent(98632, "ArmorFade", "LAST");

-- Timers

M:DefineRepeatingTimer("CheckSparkTank", "invisible", 2.5, nil, "CheckSparkTank", true);
M:DefineInitialRepeatingVoidTimer("Elementals", "offensive/high", 22.5, 21.5);
M:DefineInitialRepeatingVoidTimer("HeatedVolcano/98493", "offensive/high", 40, 55);
M:DefineInitialVoidTimer("FlameStomp/97282", "all/high", 30, 27.5);

-- Zones

M:DefineAnchoredZone("Immolate", "all/high",   "red", 12,   5);
M:DefineAnchoredZone("SafeZone", "all/high", "green",  8, 600, true);

-- Actions

M:DeclareAction("CheckSparkTank");
M:AddCallbackToAction("CheckSparkTank", "CheckSparkTank");

M:DeclareAction("Elementals", 5);
M:AddTriggerTimerToAction("Elementals", "Elementals");

M:DeclareAction("MagmaFlow");
M:AddCallbackToAction("MagmaFlow", "OnMagmaFlow");

M:DeclareAction("HeatedVolcano");
M:AddTriggerTimerToAction("HeatedVolcano", "HeatedVolcano");

M:DeclareAction("FlameStomp");
M:AddTriggerTimerToAction("FlameStomp", "FlameStomp");
M:AddCallbackToAction("FlameStomp", "OnFlameStomp");

M:DeclareAction("ArmorFade", 2);
M:AddCallbackToAction("ArmorFade", "OnArmorFade");

-- Errors

M:DefineAvoidDamageError(98034); -- Drink Magma
M:DefineAvoidDamageError(98649); -- Fusion
M:DefineAvoidDamageError({98598, 99845}, 2); -- Immolate
M:DefineAvoidDamageError({97230, 97234}); -- Magma Flow