﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local SHANNOX_ID = 53691;
local RIPLIMB_ID = 53694;
local RAGEFACE_ID = 53695;

local FREEZE_TRAP = 99836;
local BURN_TRAP = 99839;

local WARY_NAME = GetSpellInfo(100167);

local waryAlgorithm = {
    label = {
        ["enUS"] = "Avoid",
        ["frFR"] = "Evite",
    },
    color = {0, 1, 1},
    allowDead = false,
    digits = 2,
    useMultiples = false,

    GetValue = function(self, uid)
        start, duration = Unit:SearchEffect(uid, true, WARY_NAME);
        if start and duration then
            local timeLeft = max(0, start + duration - GetTime());
            return timeLeft, duration;
        end
        return 0, 30;
    end,
};

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "JaggedTearThreshold",
        type = "NUMBER",
        min = 1,
        max = 15,
        step = 1,
        default = 5,
        meta = { spell = 99937,
                 threshold = true },
    },
    {
        id = "FaceRage",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 99945,
                 symbol = 8,
                 silent = true },
    },
    {
        id = "Trap",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = FREEZE_TRAP,
                 symbol = 6,
                 silent = true },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["FrenziedDevotion"] = "P2 - Frenzied Devotion",
        ["JaggedTear"] = "You have %d Jagged Tear stacks!",
        ["Separated"] = "The boss is too far from his pets!",
        ["Separation"] = "enraged because he was too far away from his pets.",
        ["RangeTip"] = "Stay at max range!",
    },
    ["frFR"] = {
        ["FrenziedDevotion"] = "P2 - Dévotion frénétique",
        ["JaggedTear"] = "Vous avez %d fois Déchiqueture !",
        ["Separated"] = "Le boss est trop séparé de ses familiers !",
        ["Separation"] = "a enragé car il se trouvait trop loin de ses familiers.",
        ["RangeTip"] = "Restez à portée max !",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Shannox", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    lastTrap = 0,
    trapAttempts = 0,

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(60);
        self:SetScoreBenchmarks(400, 600);

        self:RegisterAdd(RIPLIMB_ID);
        self:RegisterAdd(RAGEFACE_ID);
        self:IgnoreAdds();

        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:SetExtraAlgorithm(waryAlgorithm);
        end
    end,

    OnEngaged = function(self)
        if not self:IsHeroic() then
            self:SetPhase(1);

            self:RegisterHealthThreshold(0.305, self:Localize("FrenziedDevotion"), self.OnBurnPhase, self);
        end
    end,

    OnAddDeath = function(self, guid, id, name)
        if id == RIPLIMB_ID then
            self:InterruptTimer("Spear");

    elseif id == RAGEFACE_ID then
            self:InterruptTimer("FaceRage");
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckTrap = function(self)
        local tGUID, tName, tUID = self:GetTargetInfo();
        local aGUID, aName, aUID = self:GetTankUnit();
        if tGUID and (tGUID ~= aGUID or self.trapAttempts == 1) and tUID then
            if self.lastTrap == FREEZE_TRAP then
                self:ShowZone("TrapFreeze", tUID);
                if self:IsMe(tUID) then
                    self:AlertSpell("all/critical", FREEZE_TRAP, tUID);
                end
          else
                self:ShowZone("TrapBurn", tUID);
                if self:IsMe(tUID) then
                    self:AlertSpell("all/high", BURN_TRAP, tUID);
                end
            end
            if self:GetSetting("Trap") then
                self:PlaceSymbol(tUID, 6, 4);
            end
      else
            self.trapAttempts = self.trapAttempts - 1;
            if self.trapAttempts > 0 then
                self:StartTimer("CheckTrap");
            end
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnRupture = function(self)
        self:AlertIncomingSpell("all/high", 99840);
    end,

    OnSpear = function(self)
        self:AlertIncomingSpell("all/high", 100002);
    end,

    OnAggro = function(self, guid)
        if self:IsMe(guid) then
            self:AlertSpell("all/normal", 100415, guid);
        end
    end,

    OnJaggedTear = function(self, guid, stack)
        if self:IsMe(guid) then
            self:AlertThreshold("JaggedTearThreshold", stack, self:Localize("JaggedTear"));
        end
    end,

    OnFaceRageGain = function(self, guid)
        self:AlertSpell("all/high", 99945, guid);

        if self:GetSetting("FaceRage") then
            self:PlaceSymbol(guid, 8);
        end
    end,

    OnFaceRageFade = function(self, guid)
        if self:GetSetting("FaceRage") then
            self:ClearSymbol(guid);
        end
    end,

    OnTrap = function(self, spellId)
        self.lastTrap = spellId;
        self.trapAttempts = 15;

        self:InterruptTimer("CheckTrap");
        self:CheckTrap();
    end,

    OnSeparated = function(self, name)
        self:AlertCategorized("all/high", self:Localize("Separated"));

        self:TriggerError("Separation", name);
    end,

    OnBurnPhase = function(self)
        if self:SetIncreasingPhase(2) then
            if self:HasRole("CASTER") then
                self:Alert(self:Localize("RangeTip"));
            end
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(SHANNOX_ID, true, false); -- ID, volatile, bypass
M:DefineBerserk(600);

-- Events

M:RegisterCastEvent(99840, "Rupture");
M:RegisterCastEvent({100002, 100031}, "Spear");
M:RegisterCastEvent({FREEZE_TRAP, BURN_TRAP}, "Trap");
M:RegisterCastEvent({99945, 99947}, "FaceRageGain");
M:RegisterEffectGainEvent(100415, "Aggro");
M:RegisterEffectGainEvent(99937, "JaggedTear");
M:RegisterEffectGainEvent({99945, 99947}, "FaceRageGain");
M:RegisterEffectFadeEvent({99945, 99947}, "FaceRageFade");
M:RegisterEffectFadeEvent(99835, "Separated");

-- Timers

M:DefineTimer("CheckTrap", "invisible", 0.05, nil, "CheckTrap", true);
M:DefineRepeatingVoidTimer("FaceRage/99948", "caster/critical", 28.5);
M:DefineInitialRepeatingVoidTimer("Spear/100002", "all/high", 42, 21);

-- Zones

M:DefineAimedZone("TrapFreeze", "all/critical", "red", 3, 60, true);
M:DefineAimedZone("TrapBurn", "all/high", "yellow", 3, 60, true);

-- Actions

M:DeclareAction("Rupture");
M:AddCallbackToAction("Rupture", "OnRupture");

M:DeclareAction("Spear", 3);
M:AddTriggerTimerToAction("Spear", "Spear");
M:AddCallbackToAction("Spear", "OnSpear");

M:DeclareAction("Aggro");
M:AddCallbackToAction("Aggro", "OnAggro", "targetguid");

M:DeclareAction("JaggedTear");
M:AddCallbackToAction("JaggedTear", "OnJaggedTear", "targetguid", "stack");

M:DeclareAction("FaceRageGain", 3);
M:AddTriggerTimerToAction("FaceRageGain", "FaceRage");
M:AddCallbackToAction("FaceRageGain", "OnFaceRageGain", "targetguid");

M:DeclareAction("FaceRageFade");
M:AddCallbackToAction("FaceRageFade", "OnFaceRageFade", "targetguid");

M:DeclareAction("Trap", 3);
M:AddCallbackToAction("Trap", "OnTrap", "spellid");

M:DeclareAction("CheckTrap");
M:AddCallbackToAction("CheckTrap", "CheckTrap");

M:DeclareAction("Separated", 5);
M:AddCallbackToAction("Separated", "OnSeparated", "targetname");

-- Errors

M:DefineBossAggroError();
M:DefineCustomError("Separation");
M:DefineAvoidEffectError(99837); -- Crystal Prison Trap
M:DefineAvoidEffectError(99838); -- Immolation Trap
M:DefineAvoidDamageError({99947, 99948}, 6); -- Face Rage