﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local BLACKHORN_ID = 56427;
local GORIONA_ID   = 56781;
local BOAT_ID      = 56598;

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "SunderThreshold",
        type = "NUMBER",
        min = 1,
        max = 5,
        step = 1,
        default = 1,
        meta = { spell = 108043,
                 threshold = true },
    },
    {
        id = "ShockwaveAlert",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Super alert: Shockwave",
            ["frFR"] = "Super alerte : Onde de choc",
        },
        explain = {
            ["enUS"] = "Enable this if you need an aggressive alert when Shockwave is cast.",
            ["frFR"] = "Cette option permet d'afficher une alerte agressive quand Onde de choc se produit.",
        },
    },
    version = 1,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Title"] = "Blackhorn Assault",
        ["SapperTrigger"] = "onto the deck",
        ["Sapper"] = "A sapper is there!",
        ["DrakesLeft"] = "Drakes left",
        ["Sunder"] = "You have %d Sunder Armor stacks!",
        ["Fire"] = "You're burning!",
    },
    ["frFR"] = {
        ["Title"] = "Assaut de Corne-Noire",
        ["SapperTrigger"] = "drake plonge",
        ["Sapper"] = "Un sapeur est là !",
        ["DrakesLeft"] = "Drakes restants",
        ["Sunder"] = "Vous avez %d fois Fracasser Armure !",
        ["Fire"] = "Vous brûlez !",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("Blackhorn", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    drakesDead = 0,

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(80);
        self:SetScoreBenchmarks(300, 480);

        self:DisableCombatCheck();
        self:DisableWipeCheck();
        self:SetTimeout(30);

        self:RegisterAdd(GORIONA_ID);
        self:IgnoreAdds();

        self:RegisterAnyYell(self:Localize("SapperTrigger"), "Sapper");
    end,

    OnBossFound = function(self)
        if self:IsHeroic() then
            self:DefineNearVictory(0.300, 0.90, 90);
        end
    end,

    OnEngaged = function(self)
        self:OnPhaseOne();
    end,

    OnCleared = function(self)
        Module:SetCooldown("BlackwingSpine", 60);
    end,

    OnUpdate = function(self)
        if UnitExists("boss1") then
            self:ZeroTimeout();
            self:OnEngaged();
        end
    end,

    OnDeath = function(self, guid, name)
        local id = self:GetMobID(guid);
        if id == 56855 or id == 56587 then
            self.drakesDead = self.drakesDead + 1;
        end
    end,

    -- ********
    -- Override
    -- ********

    BeginBerserk = function(self)
        if self:HasBerserk() then
            if self:GetPhase() == 2 then
                self.berserkStart = GetTime();
                self:SendViewEvent("OnTimerStart");
            end
            return true;
        end
        return false;
    end,

    GetDrakeCount = function(self)
        return max(0, 6 - self.drakesDead), 6;
    end,

    GetBoatHealth = function(self)
        if UnitExists("boss1") then
            return UnitHealth("boss1"), max(1, UnitHealthMax("boss1"));
      else
            return 0, 1;
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnPhaseOne = function(self)
        self:SetPhase(1);
        self:SetMainUnit(nil);
        self:StartMirrorValue(self:Localize("DrakesLeft"), self.GetDrakeCount, self.GetBoatHealth);

        if self:IsLFR() then
            self:InterruptTimer("Sapper");
        end
    end,

    OnPhaseTwo = function(self)
        if self:SetIncreasingPhase(2) then
            self:SetMainUnit(BLACKHORN_ID);
            self:EnableWipeCheck();
            self:BeginBerserk();
            self:StopMirrorValue();

            self:StartTimer("Shockwave");
            self:InterruptTimer("Onslaught");
            self:InterruptTimer("Sapper");
        end
    end,

    OnSunder = function(self, guid, stack)
        if self:IsMe(guid) then
            self:AlertThreshold("SunderThreshold", stack, self:Localize("Sunder"));
        end
    end,

    OnHarpoon = function(self)
        self:AlertIncomingSpell("dps_ranged/high", 108038);
    end,

    OnOnslaught = function(self)
        self:AlertIncomingSpell("all/high", 107588);
    end,

    OnShockwave = function(self)
        if self:GetSetting("ShockwaveAlert") then
            self:SuperAlert("", "", 2.5, true);
      else
            self:AlertIncomingSpell("all/high", 108046);
        end
    end,

    OnSapper = function(self)
        self:AlertCategorized("offensive/high", self:Localize("Sapper"));
    end,

    OnFlames = function(self, guid)
        if self:IsMe(guid) and self:CheckCooldown("Flames", 3) then
            self:AlertCategorized("all/high", self:Localize("Fire"));
        end
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(BLACKHORN_ID, true, false); -- ID, volatile, bypass
M:DeclareNPC(GORIONA_ID, true, false);
M:DeclareNPC(BOAT_ID, true, true);
M:DefineBerserk(250);
M:DefineFailCooldown(30);
M:DefineClearCooldown(600);

-- Events

M:RegisterEffectGainEvent(108040, "PhaseTwo");
M:RegisterEffectGainEvent(108043, "Sunder");
M:RegisterEffectGainEvent(108038, "Harpoon");
M:RegisterCastEvent(108046, "Shockwave");
M:RegisterCastEvent(107588, "Onslaught");
M:RegisterDamageEvent(108076, "Flames");

-- Timers

M:DefineRepeatingVoidTimer("Shockwave/108046", "all/critical", 23);
M:DefineInitialRepeatingVoidTimer("Onslaught/107588", "all/high", 35, 47.5);
M:DefineInitialRepeatingVoidTimer("Sapper/107752", "offensive/critical", 40, 70);

-- Actions

M:DeclareAction("PhaseTwo");
M:AddCallbackToAction("PhaseTwo", "OnPhaseTwo");

M:DeclareAction("Sunder");
M:AddCallbackToAction("Sunder", "OnSunder", "targetguid", "stack");

M:DeclareAction("Harpoon", 10);
M:AddCallbackToAction("Harpoon", "OnHarpoon");

M:DeclareAction("Shockwave");
M:AddTriggerTimerToAction("Shockwave", "Shockwave");
M:AddCallbackToAction("Shockwave", "OnShockwave");

M:DeclareAction("Onslaught");
M:AddTriggerTimerToAction("Onslaught", "Onslaught");
M:AddCallbackToAction("Onslaught", "OnOnslaught");

M:DeclareAction("Sapper");
M:AddTriggerTimerToAction("Sapper", "Sapper");
M:AddCallbackToAction("Sapper", "OnSapper");

M:DeclareAction("Flames");
M:AddCallbackToAction("Flames", "OnFlames", "targetguid");

-- Errors

M:DefineBossAggroError();
M:DefineAvoidDamageError(108076, 2); -- Twilight Flames
M:DefineAvoidDamageError(108046); -- Shockwave