﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- --------------------------------------------------------------------
-- **                              Locals                            **
-- --------------------------------------------------------------------

local DEATHWING_ID   = 53879;
local TENDRIL_ID     = 56341;
local ALT_TENDRIL_ID = 56575;

local PLASMA_AMOUNT = {
    ["25R"] = 200000,
    ["10N"] = 200000,
    ["25N"] = 300000,
    ["10H"] = 280000,
    ["25H"] = 420000,
};

local function plasmaFunc(self, uid)
    local current, max = self.module:GetPlasmaInfo(UnitGUID(uid));
    self:SetBarFill(current / max);
    self:SetBarColor(1, 0, 0);
    self:SetText(current);
end

-- --------------------------------------------------------------------
-- **                             Settings                           **
-- --------------------------------------------------------------------

local mySettings = {
    {
        id = "ShowPlasma",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Show Plasma counter",
            ["frFR"] = "Montrer le compteur de plasma",
        },
        explain = {
            ["enUS"] = "Display the amount of absorbed heals on players.\n\nThis option is ignored in LFR mode.",
            ["frFR"] = "Affiche le montant des soins absorbés sur les joueurs.\n\nCette option est ignorée en mode Outil Raids.",
        },
    },
    {
        id = "CinematicSkip",
        type = "BOOLEAN",
        default = true,
        label = {
            ["enUS"] = "Skip cinematic",
            ["frFR"] = "Passer la cinématique",
        },
        explain = {
            ["enUS"] = "Skip immediately the introduction cinematic.\n\nThis option is ignored for the first attempt.",
            ["frFR"] = "Passe immédiatement la cinématique d'introduction.\n\nCette option est ignorée lors du premier essai.",
        },
    },
    {
        id = "Grip",
        type = "BOOLEAN",
        action = "SYMBOL",
        default = false,
        exclusive = true,
        meta = { spell = 105490,
                 symbol = 1,
                 multiple = true },
    },
    version = 2,
};

-- --------------------------------------------------------------------
-- **                           Localization                         **
-- --------------------------------------------------------------------

local myLocalization = {
    ["enUS"] = {
        ["Title"]       = "Deathwing I",
        ["DragonSoul"]  = "Dragon Soul",
        ["Deathwing"]   = "Deathwing",
        ["FrozenSea"]   = "Above the Frozen Sea",
        ["PlasmaUI"]    = "Plasma",
        ["RollTrigger"] = "about to roll",
        ["Roll"]        = "Roll",
    },
    ["frFR"] = {
        ["Title"]       = "Aile de mort I",
        ["DragonSoul"]  = "Âme des dragons",
        ["Deathwing"]   = "Aile de mort",
        ["FrozenSea"]   = "Survol de la mer Gelée",
        ["PlasmaUI"]    = "Plasma",
        ["RollTrigger"] = "va faire",
        ["Roll"]        = "Tonneau",
    },
};

-- --------------------------------------------------------------------
-- **                            Definition                          **
-- --------------------------------------------------------------------

local M = Module:NewCataclysmRaid("BlackwingSpine", mySettings, myLocalization);

Module:InstallExtra(M, {
    -- **********
    -- Properties
    -- **********

    grip = { },
    plasma = { },
    plasmaMax = 1,
    mainTendril = false,
    _cinematicHook = false,

    -- ***************
    -- Custom handlers
    -- ***************

    OnEnable = function(self)
        self:SetImportance(85);
        self:SetScoreBenchmarks(360, 600);

        self:DisableCombatCheck();
        self:DisableWipeCheck();
        self:SetTimeout(30);

        self:RegisterRespawnableAdd(TENDRIL_ID);
		self:RegisterRespawnableAdd(ALT_TENDRIL_ID);
        self:IgnoreAdds();

        self:RegisterAnyYell(self:Localize("RollTrigger"), "Roll");

        self:CheckPlasmaDisplay();
        self:SetCinematicHook(self:GetSetting("CinematicSkip"));
    end,

    OnBossFound = function(self)
        self:SetMainUnit(nil); -- Mostly useless.
    end,

    OnDisable = function(self)
        self:SetCinematicHook(false);
    end,

    OnEngaged = function(self)
        self:EnableWipeCheck();
    end,

    OnCleared = function(self)
        Module:SetCooldown("BlackwingMadness", 75);
    end,

    OnStop = function(self)
        self:SetCinematicHook(false);
    end,

    OnUpdate = function(self)
        if UnitAffectingCombat("player") then
            self:OnEngaged();
        end
    end,

    OnDeath = function(self, guid, name)
        local id = self:GetMobID(guid);
        if id == TENDRIL_ID or id == ALT_TENDRIL_ID then
            self:OnSealArmorSuccess(guid);
        end
    end,

    OnHeal = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed)
        local abs = absorbed or 0;
        if abs > 0 then
            local current = self.plasma[targetGUID];
            if current then
                self.plasma[targetGUID] = max(0, current - abs);
            end
        end
    end,

    OnSettingUpdate = function(self, id, value)
        if id == "ShowPlasma" then
            self:CheckPlasmaDisplay();

    elseif id == "CinematicSkip" then
            self:SetCinematicHook(value);
        end
    end,

    -- *******
    -- Methods
    -- *******

    CheckPlasmaDisplay = function(self)
        if self:GetSetting("ShowPlasma") and not self:IsLFR() then
            self:StartUnitMonitor(self:Localize("PlasmaUI"), plasmaFunc);
      else
            self:StopUnitMonitor();
        end
    end,

    GetPlasmaInfo = function(self, guid)
        return self.plasma[guid] or 0, self.plasmaMax;
    end,

    SetCinematicHook = function(self, state)
        if not self._cinematicHook and state then
            if self:GetAttempt() > 1 then
                self._cinematicHook = MovieFrame:GetScript("OnEvent");
                MovieFrame:SetScript("OnEvent", function() GameMovieFinished() end);
            end

    elseif self._cinematicHook and not state then
            MovieFrame:SetScript("OnEvent", self._cinematicHook);
            self._cinematicHook = false;
        end
    end,

    -- ****************
    -- Special handlers
    -- ****************

    OnNuclearBlast = function(self)
        self:AlertIncomingSpell("all/critical", 105845);
    end,

    OnSealArmor = function(self, guid)
        local timer = 22.99;
        if self:IsLFR() then timer = 34.49; end

        if not self.mainTendril then
            self.mainTendril = guid;
            self:SetMainUnit(self:GetMobID(guid) or TENDRIL_ID);
            self:SendViewEvent("OnFloatingTimerStart", timer);
        end

        local EventList = self:GetWidget("EventList");
        if EventList then
            EventList:StartEvent("SealArmor:"..guid, nil, timer, 0, self.OnSealArmorFailure, self, guid);
        end
    end,

    OnSealArmorSuccess = function(self, guid)
        if self.mainTendril == guid then
            self.mainTendril = false;
            self:SetMainUnit(nil);
            self:SendViewEvent("OnFloatingTimerStop", 1.0);
        end

        local EventList = self:GetWidget("EventList");
        if EventList then
            EventList:StopEvent("SealArmor:"..guid);
        end
    end,

    OnSealArmorFailure = function(self, guid)
        if self.mainTendril == guid then
            self.mainTendril = false;
            self:SetMainUnit(nil);
            self:SendViewEvent("OnFloatingTimerStop");
        end
    end,

    OnGripTarget = function(self, guid)
        if self:IsMe(guid) then
            self:AlertSpell("all/high", 105490, guid);
        end
        tinsert(self.grip, guid);
    end,

    OnGrip = function(self)
        if self:GetSetting("Grip") then
            self:AnnounceUnitListSpell(self.grip, 105490, true, 8);
        end
        wipe(self.grip);
    end,

    OnPlasmaGain = function(self, guid)
        local amount = PLASMA_AMOUNT[ self:GetCode() ] or 0;

        self.plasmaMax = max(self.plasmaMax, amount);
        self.plasma[guid] = amount;
    end,

    OnPlasmaFade = function(self, guid)
        self.plasma[guid] = nil;
    end,
} );

-- --------------------------------------------------------------------
-- **                               Setup                            **
-- --------------------------------------------------------------------

M:DeclareBoss(DEATHWING_ID, true, true); -- ID, volatile, bypass
M:DeclareZone("Deathwing", false);
M:DeclareZone("FrozenSea", true);
M:DefineZoneBossFilter("DragonSoul", 6); -- 6 boss killed out of 8
M:DefineFailCooldown(30);
M:DefineClearCooldown(600);

-- Events

M:RegisterCastEvent(105845, "NuclearBlast");
M:RegisterCastEvent({105847, 105848}, "SealArmor");
M:RegisterEffectGainEvent({105490}, "GripTarget");
M:RegisterEffectGainEvent({105479, 109379}, "PlasmaGain");
M:RegisterEffectFadeEvent({105479, 109379}, "PlasmaFade");

-- Timers

M:DefineVoidTimer("Roll", "all/critical", 3);

-- Actions

M:DeclareAction("NuclearBlast", 8);
M:AddCallbackToAction("NuclearBlast", "OnNuclearBlast");

M:DeclareAction("SealArmor");
M:AddCallbackToAction("SealArmor", "OnSealArmor", "actorguid");

M:DeclareAction("GripTarget");
M:AddCallbackToAction("GripTarget", "OnGripTarget", "targetguid");
M:AddExternalActionToAction("GripTarget", "Grip");

M:DeclareAction("Grip", 1.5, 1.0);
M:AddCallbackToAction("Grip", "OnGrip");

M:DeclareAction("PlasmaGain");
M:AddCallbackToAction("PlasmaGain", "OnPlasmaGain", "targetguid");

M:DeclareAction("PlasmaFade");
M:AddCallbackToAction("PlasmaFade", "OnPlasmaFade", "targetguid");

M:DeclareAction("Roll", 3);
M:AddStartTimerToAction("Roll", "Roll");