﻿local Root = BossEncounter3;

local Module = Root:GetOrNewModule("Module");

local Generator = Module:GetOrNewModule("Generator");

local QuickModule = Root:GetOrNewModule("QuickModule");

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

local function sizeOf(t)
    if t then
        return #t;
  else
        return 0;
    end
end

local function OnQuickModuleEnable(self)
    local t = self._quickModuleInfo;
    Generator:Setup(self, t);
end

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

function Generator:Generate(name)
    local t = QuickModule:GetDescriptionTable(name);
    if t then
        self:Build(name, t);
    end
end

function Generator:GenerateAll() -- Used for debug purpose.
    local num = QuickModule:GetNumDescriptionTables();
    local i, t;
    for i=1, num do
        t = QuickModule:GetDescriptionTable(i);
        self:Build(t._name, t);
    end
end

function Generator:Build(name, t) -- t is a table describing the boss.
    local i;

    -- Build the core.
    local M = Module:New(name, t.game or "CATA", t.environment or "PARTY", t.settings, t.locale);

    -- Insert immediately extra methods and properties.
    Module:InstallExtra(M, t.extra or {});

    -- Inject a method that will be called at the appropriate time in the OnEnable handler.
    -- It will take tare of finishing setting up the module.
    M._quickModuleInfo = t;
    M.OnQuickModuleEnable = OnQuickModuleEnable;

    -- Declare module constant elements.
    if t.boss and t.boss > 0 then
        M:DeclareBoss(t.boss);
  else
        M:DeclareNoBoss();
    end
    if t.range and t.range > 0 then
        M:DefineMinRange(t.range);
    end
    if t.berserk and t.berserk > 0 then
        M:DefineBerserk(t.berserk);
    end

    -- Declare timers
    local timer, default, first;
    for i=1, sizeOf(t.timers) do
        timer = t.timers[i];
        default = timer.default or 0;
        first = timer.first or 0;

        if timer.id and timer.category then
            if default + first > 0 then
                local action = nil;
                if timer.callback then
                    if M[timer.callback] then
                        action = "T:"..timer.id;
                        M:DeclareAction(action);
                        M:AddCallbackToAction(action, timer.callback);
                  else
                        M:Debug("Timer '"..timer.id.."' uses a callback method that hasn't been defined.");
                    end
                end
                M:DefineTimerFull(timer.id, timer.category, default, first, timer.loop, first > 0, action, timer.finish);
                if timer.spell then
                    M:DefineTimerLabel(timer.id, timer.spell);
                end
          else
                M:Debug("Timer '"..timer.id.."' doesn't have either a 'first' or 'default' field.");
            end
      else
            M:Debug("'timers' entry #"..i.." is missing either 'id' or 'category' field.");
        end
    end

    -- Declare zones
    local zone;
    for i=1, sizeOf(t.zones) do
        zone = t.zones[i];

        if zone.id and zone.category and zone.color then
            if zone.duration and zone.radius and zone.type then
                M:DefineZone(zone.id, zone.category, zone.color, zone.radius, zone.duration, zone.multiple or false, zone.type, zone.x or 0, zone.y or 0);
          else
                M:Debug("Zone '"..zone.id.."' is missing 'duration', 'radius' or 'type' field.");
            end
      else
            M:Debug("'zones' entry #"..i.." is missing either 'id', 'category' or 'color' field.");
        end
    end

    -- Declare spells
    local count = 0;
    local spell;
    for i=1, sizeOf(t.spells) do
        spell = t.spells[i];

        if spell.id and spell.trigger then
            count = count + 1;

            -- Build the action responding to this.
            local action = "S:"..count.."-"..spell.trigger;
            M:DeclareAction(action, spell.cooldown or 0);
            if spell.onSelf then
                -- Self-only, do the self check.
                M:AddCheckSelfToAction(action);
            end
            if spell.zone then
                if M:HasZone(spell.zone) then
                    if spell.trigger == "effectfade" then
                        M:AddHideZoneToAction(action, spell.zone);
                  else
                        M:AddShowZoneToAction(action, spell.zone);
                    end
              else
                    M:Debug("'spells' entry #"..i.." tries to operate zone '"..spell.zone.."' but it hasn't been defined.");
                end
            end
            if spell.finishTimer then
                if M:HasTimer(spell.finishTimer) then
                    M:AddTriggerTimerToAction(action, spell.finishTimer);
              else
                    M:Debug("'spells' entry #"..i.." tries to trigger timer '"..spell.finishTimer.."' but it hasn't been defined.");
                end
            end
            if spell.startTimer then
                if M:HasTimer(spell.startTimer) then
                    M:AddStartTimerToAction(action, spell.startTimer);
              else
                    M:Debug("'spells' entry #"..i.." tries to start timer '"..spell.startTimer.."' but it hasn't been defined.");
                end
            end
            if spell.callback then
                if M[spell.callback] then
                    M:AddCallbackToAction(action, spell.callback, "targetguid", "stack");
              else
                    M:Debug("'spells' entry #"..i.." tries to use callback '"..spell.callback.."' but it hasn't been defined.");
                end
            end

            -- Register events triggering the action.
            if spell.trigger == "effectfade" then
                M:RegisterEffectFadeEvent(spell.id, action);

        elseif spell.trigger == "effectgain" then
                M:RegisterEffectGainEvent(spell.id, action);

        elseif spell.trigger == "damage" then
                M:RegisterDamageEvent(spell.id, action);

        elseif spell.trigger == "heal" then
                M:RegisterHealEvent(spell.id, action);

        elseif spell.trigger == "cast" then
                M:RegisterCastEvent(spell.id, action);
            end
      else
            M:Debug("'spells' entry #"..i.." is missing either 'id' or 'trigger' field.");
        end
    end

    -- Success!
    Module:Debug("Built successfully the QuickModule "..name.."!");
end

function Generator:Setup(M, t) -- M is the module being set up, t is a table describing the boss.
    local i;

    -- Now configure the module.
    if t.timeout then
        M:SetTimeout(t.timeout);
    end
    if t.safeEngage then
        M:EnableSafeEngage();
    end
    if t.ignoreStandardClear then
        M:DisableStandardClear();
    end
    if t.ignoreLeaveCombat then
        M:DisableCombatCheck();
    end
    if t.ignoreWipe then
        M:DisableWipeCheck();
    end

    -- Benchmark info.
    if t.greatTime and t.poorTime then
        M:SetScoreBenchmarks(t.greatTime, t.poorTime);
    end
    if t.timeCoeff then
        M:SetBenchmarksTimeCoeff(t.timeCoeff);
    end
    if t.ignoreResults or (t.ignoreDamageMeter and t.ignoreScore) then
        M:DisableEndSequence();
    end
    if t.ignoreScore or t.ignoreResults then
        M:DisableScore();
    end
    if t.ignoreDamageMeter or t.ignoreResults then
        M:DisableDamageMeter();
    end
    M:SetImportance(t.importance or 0);

    -- Declare adds or mooks.
    local numAdds = sizeOf(t.adds);
    local numMooks = sizeOf(t.mooks);
    if numAdds > 0 then
        -- There are adds.
        for i=1, numAdds do
            M:RegisterAdd(t.adds[i]);
        end

elseif numMooks > 0 then
        -- There are mooks.
        for i=1, numMooks do
            M:RegisterRespawnableAdd(t.mooks[i]);
        end
        M:IgnoreAdds();
        M:IgnoreAddsEngagement();
        M:SetRelevantDamagePending(false);
    end

    -- Set special yell triggers.
    local engageTrigger = M:Localize("EngageTrigger", true);
    local clearTrigger  = M:Localize("ClearTrigger", true);
    local failTrigger   = M:Localize("FailTrigger", true);
    if engageTrigger then
        M:RegisterAnyYell(engageTrigger, M.OnEngaged, M);
    end
    if clearTrigger then
        M:RegisterAnyYell(clearTrigger, M.OnCleared, M);
        M:SetClearAnimation("ALTERNATE");
    end
    if failTrigger then
        M:RegisterAnyYell(failTrigger, M.OnFailed, M, true);
    end

    -- Set yells.
    local yell, translation;
    for i=1, sizeOf(t.yells) do
        yell = t.yells[i];
        if yell.trigger and yell.callback and M[yell.callback] then
            translation = M:Localize(yell.trigger, true);
            if translation then
                M:RegisterAnyYell(translation, M[yell.callback], M);
          else
                M:Debug("Trigger '"..yell.trigger.."' used for yell #"..i.." is not defined in the locale table.");
            end
      else
            M:Debug("'yells' entry #"..i.." is missing either 'trigger' or a valid 'callback' field.");
        end
    end

    -- Set thresholds.
    local threshold, translation;
    for i=1, sizeOf(t.thresholds) do
        threshold = t.thresholds[i];
        if threshold.value and threshold.description then
            translation = M:Localize(threshold.description);
            if translation then
                if threshold.callback and M[threshold.callback] then
                    M:RegisterHealthThreshold(threshold.value, translation, M[threshold.callback], M);
              else
                    M:RegisterHealthThreshold(threshold.value, translation);
                end
          else
                M:Debug("Translation '"..threshold.description.."' used for threshold #"..i.." is not defined in the locale table.");
            end
      else
            M:Debug("'thresholds' entry #"..i.." is missing either 'value' or 'description' field.");
        end
    end

    -- Finished!
    M:Debug("Finished setup according to the QuickModule description table.");
end