local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Table = Root:GetOrNewModule("Table");

-- See events.lua, this is the section defining actions and commands.

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _actions = { },

    delayedActions = { },

    -- Methods
    -- *******

    -- ==============================================
    -- This method checks the existence of an action.
    -- ==============================================
    HasAction = function(self, name)
        return (self._actions[name] ~= nil);
    end,

    -- ==========================================
    -- This method accesses an action data table.
    -- ==========================================
    GetAction = function(self, name)
        local action = self._actions[name];
        if not action then
            self:Debug("Action "..name.." is not defined.");
        end
        return action;
    end,

    -- ===============================================================================================
    -- This method executes a previously defined action if it has no delay, or schedule its execution.
    -- ===============================================================================================
    ScheduleAction = function(self, name)
        local action = self:GetAction(name);
        if action then
            local delay = action.delay;
            if delay > 0 then
                local entry = Table:Alloc();
                entry.name = name;
                entry.timer = delay;
                tinsert(self.delayedActions, entry);
          else
                self:RunAction(name);
            end
        end
    end,

    -- ==========================================
    -- This updates the queue of delayed actions.
    -- ==========================================
    CheckActions = function(self, elapsed)
        local i, entry;
        for i=#self.delayedActions, 1, -1 do
            entry = self.delayedActions[i];
            entry.timer = max(0, entry.timer - elapsed);
            if entry.timer == 0 then
                tremove(self.delayedActions, i);
                self:RunAction(entry.name);
                Table:Recycle(entry);
            end
        end
    end,

    -- =================================================
    -- This method executes a previously defined action.
    -- =================================================
    RunAction = function(self, name)
        local action = self:GetAction(name);
        local cooldown = action.cooldown;

        -- First check for the action's cooldown.
        if cooldown <= 0 or self:CheckCooldown(action.name, cooldown) then
            -- Now execute each command.
            local i, command;
            for i=1, #action do
                command = action[i];
                if not self:RunCommand(command) then
                    break;
                end
            end
        end
    end,

    -- ==============================================================================
    -- The following methods allow you to define actions that will respond to events.
    -- ==============================================================================

    -- Merely declare the existence of an action (that does nothing).
    -- Cooldown specifies that the action can only be invoked once in a while (defaults to none: 0).
    -- Delay specifies how long the system will wait before really running the action (defaults to none: 0).
    -- After declaring the action, you fill it with commands which will get executed in sequential order (see below).

    DeclareAction = function(self, name, cooldown, delay)
        if self:HasAction(name) then
            self:Debug("Action "..name.." is already defined!");
      else
            self._actions[name] = {
                name = name,
                cooldown = cooldown or 0,
                delay = delay or 0,
            };
            return self._actions[name];
        end
    end,

    -- Set an action to call a callback when fired.
    -- Extra arguments (...) are the names of the event info fields you want to get as extra arguments in your callback.

    AddCallbackToAction = function(self, name, methodName, ...)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "callback",
                method = methodName,
                args = { },
            };
            local i, arg;
            for i=1, select('#', ...) do
                arg = select(i, ...);
                tinsert(newCommand.args, arg);
            end
            tinsert(action, newCommand);
        end
    end,

    -- Set an action to trigger a timer when fired.

    AddTriggerTimerToAction = function(self, name, timerName)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "triggertimer",
                timer = timerName,
            };
            tinsert(action, newCommand);
        end
    end,

    -- Set an action to start a timer when fired.

    AddStartTimerToAction = function(self, name, timerName)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "starttimer",
                timer = timerName,
            };
            tinsert(action, newCommand);
        end
    end,

    -- Set an action to hide a zone when fired.
    -- If the zone involves an unit, it will be queried in the context.

    AddHideZoneToAction = function(self, name, zoneName)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "hidezone",
                zone = zoneName,
            };
            tinsert(action, newCommand);
        end
    end,

    -- Set an action to show a zone when fired.
    -- If the zone involves an unit, it will be queried in the context.

    AddShowZoneToAction = function(self, name, zoneName)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "showzone",
                zone = zoneName,
            };
            tinsert(action, newCommand);
        end
    end,

    -- Set an action to schedule another action when fired.

    AddExternalActionToAction = function(self, name, actionName)
        local action = self:GetAction(name);
        if action then
            local newCommand = {
                type = "scheduleaction",
                action = actionName,
            };
            tinsert(action, newCommand);
        end
    end,

    -- Check the context within an action's execution is about you.
    -- This is only valid for actions that respond to an event.

    AddCheckSelfToAction = function(self, name)
        local action = self:GetAction(name);
        if action then
            tinsert(action, {type = "checkself"});
        end
    end,

    -- ==================================================
    -- This method executes a command (within an action).
    -- ==================================================
    RunCommand = function(self, command)
        local op = command.type;
        if op == "callback" then
            local method = self[command.method];
            if not method then
                self:Debug("Callback command failed: method "..command.method.." doesn't exist.");
          else
                local argList = Table:Alloc();
                local args = command.args;
                local i;
                for i=1, #args do
                    tinsert(argList, self:GetContextInfo(args[i]) or false);
                end
                method(self, unpack(argList));
                Table:Recycle(argList);
            end
            return true;

    elseif op == "triggertimer" then
            self:TriggerTimer(command.timer);
            return true;

    elseif op == "starttimer" then
            self:StartTimer(command.timer);
            return true;

    elseif op == "hidezone" then
            self:InterruptZone(command.zone, self:GetContextInfo("targetguid"));
            return true;

    elseif op == "showzone" then
            self:ShowZone(command.zone, self:GetContextInfo("targetguid"));
            return true;

    elseif op == "scheduleaction" then
            self:ScheduleAction(command.action);
            return true;

    elseif op == "checkself" then
            local myGUID = UnitGUID("player");
            return (self:GetContextInfo("targetguid") == myGUID or self:GetContextInfo("actorguid") == myGUID);
        end
    end,
} );