local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");

local RESPAWN_MIN_DELAY = 0.25; -- Time amount before a dead mob is flagged as respawnable.

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    ignoreAdds = false,
    ignoreAddsEngagement = false,

    adds = { },
    addsLookup = { },
    addsOutOfCombat = false,

    addReviveTimer = 0,
    addReviveList = false,
    addReviveDeath = { },

    -- Methods
    -- *******

    -- ============================================
    -- Determinate if adds should be killed to win.
    -- ============================================
    MustKillAdds = function(self)
        return (not self.ignoreAdds);
    end,

    -- =========================================
    -- Disable the need to kill the adds to win.
    -- =========================================
    IgnoreAdds = function(self)
        self.ignoreAdds = true;
    end,

    -- ============================================================
    -- Prevent the combat from being triggered by an add in combat.
    -- ============================================================
    IgnoreAddsEngagement = function(self)
        self.ignoreAddsEngagement = true;
    end,

    -- ===================
    -- Register a new add.
    -- ===================
    RegisterAdd = function(self, id)
        self:RegisterAddFull(id, false, true, nil, nil);
    end,

    -- ===============================
    -- Register a new respawnable add.
    -- ===============================
    RegisterRespawnableAdd = function(self, id)
        self:RegisterAddFull(id, true, true, nil, nil);
    end,

    -- ======================================================================================
    -- Register a new add in full mode with more options.
    -- Respawn should be true if another instance of the unit can replace the first on death.
    -- Respawn set to true is also fine to use with adds that can be revived.
    -- Selectable should be true for selectable adds.
    -- Class and role specify the class and role of the add, duh! Optionnal.
    -- ======================================================================================
    RegisterAddFull = function(self, id, canRespawn, selectable, class, role)
        tinsert(self.adds, {
            id = id,
            guid = nil,
            name = nil,
            respawn = canRespawn or false,
            dead = false,
            deathTime = 0,
        } );

        -- Put the add in the enemy list!
        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:AddUnit(id, selectable, class, role);
        end
    end,

    -- ====================================================================
    -- Remove an existing add. Do not use this method for respawnable adds.
    -- GUID can also be passed in "index" argument.
    -- ====================================================================
    RemoveAdd = function(self, index)
        if type(index) == "string" then
            local i;
            for i=#self.adds, 1, -1 do
                if self.adds[i].guid == index then
                    self:RemoveAdd(i);
                end
            end
      else
            local add = self.adds[index];
            if add then
                -- Shuffle with other entries in the add table.
                tremove(self.adds, index);

                local EnemyList = self:GetWidget("EnemyList");
                if EnemyList then
                    EnemyList:RemoveUnit(add.guid or add.id);
                end
                self:CheckStandardClear();
            end
        end
    end,

    -- ==============================
    -- Update currently watched adds.
    -- ==============================
    CheckAdds = function(self, elapsed)
        local outOfCombat = 0;
        local outOfCombatMax = 0;
        local i, add;
        for i=1, #self.adds do
            add = self.adds[i];
            if (not add.guid) or self:IsAddRespawnable(i) then -- Still unknown add or add that is set to be respawned.
                local guid, uid = Unit:GetGUIDFromMobID(add.id);
                if guid and uid and (not self:IsAdd(guid)) then
                    -- Got the mob!
                    add.name = UnitName(uid);
                    add.guid = guid;
                    self.addsLookup[guid] = true;

                    local EnemyList = self:GetWidget("EnemyList");
                    if EnemyList then
                        EnemyList:SetUnit(add.id, add.guid);
                    end

                    self:FireHandler("OnAddFound", add.guid, add.id, add.name);
                end
            end
            if add.guid then
                local uid = Unit:GetUID(add.guid);
                if uid then
                    self:ZeroTimeout();
                    if (not self.ignoreAddsEngagement) and self:CheckEngage(uid) then
                        self:OnEngaged();
                    end
                    if UnitIsDeadOrGhost(uid) and not add.dead then
                        -- The add has died.
                        self:SetAddDead(i);
                    end
                    outOfCombatMax = outOfCombatMax + 1;
                    if not UnitAffectingCombat(uid) and not add.dead then
                        outOfCombat = outOfCombat + 1;
                    end
                end
            end
        end
        self.addsOutOfCombat = (outOfCombatMax > 0 and outOfCombat == outOfCombatMax);
    end,

    -- =============================================================
    -- Check if dying entity <guid> is in one of the registered adds
    -- (ought to be called when a DEATH combat event is fired)
    -- =============================================================
    CheckAddDeath = function(self, guid)
        local id = Unit:GetMobID(guid);
        local i, add;
        for i=1, #self.adds do
            add = self.adds[i];
            if add.id == id then
                self:ZeroTimeout();
                self:SetAddDead(i);
            end
        end
    end,

    -- ===========================
    -- Flag the given add as dead.
    -- ===========================
    SetAddDead = function(self, index)
        local add = self.adds[index];
        if add then
            if not add.dead then
                -- Score the death.
                add.dead = true;
                add.deathTime = GetTime();
                if self:IsRevivableAdd(id) then
                    self.addReviveDeath[guid] = GetTime();
                    self:CheckReviveInterrupt();
                end
                self:FireHandler("OnAddDeath", add.guid, add.id, add.name);
                self:CheckStandardClear();
            end
        end
    end,

    -- =========================
    -- Return the number of adds
    -- =========================
    GetNumAdds = function(self)
        return #self.adds;
    end,

    -- ===============================
    -- Return the number of adds alive
    -- ===============================
    GetNumAddsAlive = function(self)
        local count = 0;
        local i;
        for i=1, #self.adds do
            if (not self.adds[i].dead) then
                count = count + 1;
            end
        end
        return count;
    end,

    -- ==============================
    -- Return the number of dead adds
    -- ==============================
    GetNumAddsDead = function(self)
        return self:GetNumAdds() - self:GetNumAddsAlive();
    end,

    -- ====================================================
    -- Returns info about an add. GUID and name returns can
    -- be nil if the engine has not found the add yet.
    -- ====================================================
    GetAddInfo = function(self, index)
        local add = self.adds[index];
        if add then
            return add.id, add.guid, add.name, add.dead, add.respawn;
        end
        return nil, nil, nil, false, false;
    end,

    -- ===========================================================================
    -- Return true if the specified add can be respawned (once it has been killed)
    -- ===========================================================================
    IsAddRespawnable = function(self, index)
        local add = self.adds[index];
        if add and add.dead and add.respawn then
            return GetTime() > (add.deathTime + RESPAWN_MIN_DELAY);
        end
        return false;
    end,

    -- ======================================================
    -- Get an add GUID from a mob ID. The result might be nil 
    -- if the add could not be found yet.
    -- ======================================================
    GetAddGUID = function(self, mobID)
        local i, id, guid;
        for i=1, self:GetNumAdds() do
            id, guid = self:GetAddInfo(i);
            if id == mobID then
                return guid; -- Can be nil if the engine has not found the add yet.
            end
        end
        return nil;
    end,

    -- ====================================================
    -- Check if unit of specified guid is a registered add.
    -- ====================================================
    IsAdd = function(self, guid)
        return self.addsLookup[guid] or false;
    end,

    -- ===============================================
    -- Check if all accessable adds are out of combat.
    -- ===============================================
    AreAddsOutOfCombat = function(self)
        return self.addsOutOfCombat;
    end,

    -- =============================================================================
    -- Register adds that will auto-revive themselves unless killed at the same time
    -- =============================================================================
    DefineRevivableAdds = function(self, list, reviveTimer)
        self.addReviveTimer = reviveTimer;
        self.addReviveList = list;
    end,

    -- ======================================================
    -- Determinate the time before an add auto-revives itself
    -- ======================================================
    GetReviveTimer = function(self, guid)
        local deathTime = self.addReviveDeath[guid];
        if not deathTime then
            return 0, 0;
      else
            return max(0, deathTime + self.addReviveTimer - GetTime()), self.addReviveTimer;
        end
    end,

    -- ===================================================
    -- Check if all adds belonging to the revive list have 
    -- been slain, thus interrupting the revive timer
    -- ===================================================
    CheckReviveInterrupt = function(self)
        -- If all revivable adds are flagged as dead, then interrupt all revival timers.
        local i, id, guid, name, dead;
        for i=1, self:GetNumAdds() do
            id, guid, name, dead = self:GetAddInfo(i);
            if self:IsRevivableAdd(id) and not dead then
                return; -- Still one add alive.
            end
        end
        wipe(self.addReviveDeath);
    end,

    -- ===============================================
    -- Specify if an add is flagged as a revivable add
    -- ===============================================
    IsRevivableAdd = function(self, id)
        if self.addReviveList then
            return self.addReviveList[id];
        end
        return false;
    end,
} );