local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");
local L = Root:GetOrNewModule("Localization");
local Environment = Root:GetOrNewModule("Environment");

local System = Root:GetOrNewModule("System");
local Stats = System:GetOrNewModule("Stats");

local sizeCustomizable = {
    ["LK"] = true,
    ["CATA"] = true,
};

local MAX_IMPORTANCE_ADJUSTMENT = 20; -- This is the max amount that will be substracted from importance value according to the number of times the boss has been killed.
local LFR_IMPORTANCE_ADJUSTMENT = -10; -- This is plainly added to the importance value if the difficulty is LFR. This is only applied for raid bosses.
local HEROIC_IMPORTANCE_ADJUSTMENT = 5; -- This is plainly added to the importance value if the difficulty is heroic. This is only applied for raid bosses.

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _name = "UNKNOWN",
    _game = "UNKNOWN", -- These will be filled externally.
    _environment = "UNKNOWN",

    slots = 1,
    difficulty = "NORMAL",
    startingPlayers = 0,
    customSlots = false,
    initialStateLocked = false,

    state = "UNKNOWN",
    test = false,
    complete = true,

    title = false,

    importance = 50,

    -- Methods
    -- *******

    -- ===============================
    -- Get the module's internal name.
    -- ===============================
    GetName = function(self)
        return self._name;
    end,

    -- ===============================================
    -- Get what game the module has been designed for.
    -- CATA, LK, TBC, WOW are possible values so far.
    -- ===============================================
    GetGame = function(self)
        return self._game;
    end,

    -- =============================================
    -- Get the environment this module should be in.
    -- Either SOLO, PARTY or RAID.
    -- =============================================
    GetEnvironment = function(self)
        return self._environment;
    end,

    -- ==============================================================
    -- Get the number of slots allowed in the environment by default.
    -- ==============================================================
    GetDefaultNumSlots = function(self)
        local environment = self:GetEnvironment();
        local game = self:GetGame();

        if environment == "PARTY" then
            return 5;

    elseif environment == "RAID" then
            if sizeCustomizable[game] then
                return select(2, Environment:GetInstanceFormat());
          else
                if game == "TBC" then
                    return 25;
              else
                    return 40;
                end
            end
      else
            return 1;
        end
    end,

    -- ===================================================
    -- Get the number of slots allowed in the environment.
    -- ===================================================
    GetNumSlots = function(self)
        if self.initialStateLocked then
            return max(self.slots, self.startingPlayers);
      else
            if self.customSlots then
                return self.customSlots;
          else
                return self:GetDefaultNumSlots();
            end
        end
    end,

    -- ==============================================================
    -- Explicitly set the number of slots allowed in the environment.
    -- ==============================================================
    SetMaxSlots = function(self, slots)
        if not self.customSlots then
            self.customSlots = slots;
        end
    end,

    -- ==========================================================
    -- Get the difficulty of the environment (NORMAL/HEROIC/LFR).
    -- ==========================================================
    GetDungeonDifficulty = function(self)
        if self.initialStateLocked then
            return self.difficulty;
      else
            return select(1, Environment:GetInstanceFormat());
        end
    end,

    -- ===================================================
    -- Shortcut for testing normal difficulty environment.
    -- ===================================================
    IsNormal = function(self)
        return (self:GetDungeonDifficulty() == "NORMAL");
    end,

    -- ===================================================
    -- Shortcut for testing heroic difficulty environment.
    -- ===================================================
    IsHeroic = function(self)
        return (self:GetDungeonDifficulty() == "HEROIC");
    end,

    -- ================================================
    -- Shortcut for testing LFR difficulty environment.
    -- ================================================
    IsLFR = function(self)
        return (self:GetDungeonDifficulty() == "LFR");
    end,

    -- ================================
    -- Get the code of the environment.
    -- ================================
    GetCode = function(self)
        return select(3, Environment:GetInstanceFormat());
    end,

    -- ======================================================================
    -- Get the number of players there was at the beginning of the encounter.
    -- ======================================================================
    GetStartingPlayers = function(self)
        if self.initialStateLocked then
            return self.startingPlayers;
      else
            return 0;
        end
    end,

    -- ================================================================
    -- Acquire and lock initial state infos once the module is engaged.
    -- (i.e: dynamic difficulty can no longer be changed in combat)
    -- ================================================================
    LockInitialState = function(self)
        if not self.initialStateLocked then
            self.slots = self:GetNumSlots();
            self.difficulty = self:GetDungeonDifficulty();
            self.startingPlayers = self:CountNumValidMembers();
            self.initialStateLocked = true;
        end
    end,

    -- ===================================================
    -- Get the module ID. An ID is unique for each module,
    -- for each difficulty mode and for each raid size.
    -- ===================================================
    GetID = function(self)
        local ID = self:GetName();
        local slots = self:GetNumSlots();
        local showSize = false;
        local showLFR = false;
        local showHeroic = false;
        local suffix = "";
        local displaySuffix = "";

        if sizeCustomizable[ self:GetGame() ] then
            suffix = suffix.."|"..slots;
            if self:GetEnvironment() == "RAID" then
                showSize = true;
            end
        end
        if self:IsLFR() then
            suffix = suffix.."|R";
            showLFR = true;
    elseif self:IsHeroic() then
            suffix = suffix.."|H";
            showHeroic = true;
        end

        if showLFR and showSize then
            displaySuffix = L:FormatLoc("Suffix:LFRSize", slots);

    elseif showHeroic and showSize then
            displaySuffix = L:FormatLoc("Suffix:HeroicSize", slots);

    elseif showHeroic then
            displaySuffix = L:Get("Suffix:Heroic");

    elseif showSize then
            displaySuffix = L:FormatLoc("Suffix:Size", slots);
        end

        return ID..suffix, displaySuffix;
    end,

    -- ============================
    -- Set the state of the module.
    -- ============================
    SetState = function(self, state)
        self.state = state;
    end,

    -- ============================
    -- Get the state of the module.
    -- ============================
    GetState = function(self)
        return self.state;
    end,

    -- =======================================================
    -- Check if the encounter is running (standby or engaged).
    -- =======================================================
    IsRunning = function(self)
        local state = self:GetState();
        return (state == "STANDBY" or state == "ENGAGED");
    end,

    -- =======================================================
    -- Check if the encounter is finished (cleared or failed).
    -- =======================================================
    IsFinished = function(self)
        local state = self:GetState();
        return (state == "CLEARED" or state == "FAILED");
    end,

    -- ==============================================================================
    -- Ask the module manager to disable us, scoring a wipe if the module is engaged.
    -- ==============================================================================
    Interrupt = function(self)
        if self:GetState() == "ENGAGED" then
            Stats:BossFailed(self);
            Module:SetCooldown(self:GetName(), self:GetFailCooldown());
        end
        self:StopMusic();
        self:Kill();
    end,

    -- =====================================
    -- Ask the module manager to disable us.
    -- =====================================
    Kill = function(self)
        return Module:Disable(self:GetName());
    end,

    -- ====================================
    -- Ask the module manager to enable us.
    -- ====================================
    Open = function(self)
        return Module:Enable(self:GetName());
    end,

    -- =============================================
    -- Test this module. This enables the test flag.
    -- =============================================
    Test = function(self, engageLater)
        if self:Open() then
            -- Special behavior for test mode.
            if self._bossID > 0 then
                self:OnBossFound("player");
            end
            if not engageLater then
                self.test = true;
                self:OnEngaged();
            end
            return true;
      else
            return false;
        end
    end,

    -- ==================
    -- Get the test flag.
    -- ==================
    IsTesting = function(self)
        return self.test;
    end,

    -- ========================
    -- Get the incomplete flag.
    -- ========================
    IsIncomplete = function(self)
        return (not self.complete);
    end,

    -- =============================
    -- Flag the fight as incomplete.
    -- =============================
    FlagAsIncomplete = function(self)
        self.complete = false;
    end,

    -- ===============================
    -- Get the title of the encounter.
    -- ===============================
    GetTitle = function(self)
        if self.title then
            return self.title;
      else
            return "?";
        end
    end,

    -- ======================================================
    -- Set the title of the encounter. Can only be done once.
    -- ======================================================
    SetTitle = function(self, title)
        if not self.title then
            self.title = title;
            self:SendViewEvent("OnTitleSet", title);

            -- Display title and attempt count.
            self:Message(self:GetTitle());
            self:ShowAttempt();
        end
    end,

    -- ===============================================
    -- Get the current attempt count of the encounter.
    -- ===============================================
    GetAttempt = function(self)
        local id = self:GetID();
        return select(2, Stats:GetInfo(id));
    end,

    -- =================================================
    -- Get the number of times the boss has been killed.
    -- =================================================
    GetKills = function(self)
        local id = self:GetID();
        return select(3, Stats:GetInfo(id));
    end,

    -- ====================================
    -- Set the importance of the encounter.
    -- ====================================
    SetImportance = function(self, importance)
        self.importance = importance;
    end,

    -- ============================================================
    -- Get the importance of the encounter. This number is adjusted
    -- downward as the number of times it has been killed increases
    -- and upward as the number of wipes increases.
    -- ============================================================
    GetImportance = function(self)
        local id = self:GetID();
        local _, attempt, kills = Stats:GetInfo(id)
        local value = self.importance - min(MAX_IMPORTANCE_ADJUSTMENT, kills * 2) + min(MAX_IMPORTANCE_ADJUSTMENT, attempt - 1);
        if self:IsLFR() and self:GetEnvironment() == "RAID" then value = value + LFR_IMPORTANCE_ADJUSTMENT; end
        if self:IsHeroic() and self:GetEnvironment() == "RAID" then value = value + HEROIC_IMPORTANCE_ADJUSTMENT; end
        return max(0, min(100, value));
    end,

    -- ==============================================================
    -- Get the base importance of the encounter (without adjustment).
    -- ==============================================================
    GetBaseImportance = function(self)
        return self.importance;
    end,
} );