﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");
local NPCInfo = Root:GetOrNewModule("NPCInfo");

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _bossID = 0,
    _bossCouncil = { },

    bossDefined = false,
    bossName = "?",
    bossGUID = "",
    bossHealth = 0,
    bossHealthMax = 0,
    bossDead = false,
    bossOutOfCombat = false,

    bossMultiDeath = false,

    healthThresholds = { },

    -- Methods
    -- *******

    -- ====================================================================
    -- Specify that a mob with the given mobID is affiliated with the
    -- boss module and should trigger it.
    -- ====================================================================
    DeclareNPC = function(self, mobID, volatile, bypass)
        NPCInfo:Register(mobID, self:GetName(), volatile, bypass);
    end,

    -- =====================================================
    -- Get the ID of the main boss, if it has been declared.
    -- Return nil if no main boss is declared.
    -- =====================================================
    GetBossID = function(self)
        if self._bossID > 0 then
            return self._bossID;
      else
            return nil;
        end
    end,

    -- ==========================================================
    -- Get the number of mobs in the council. This method is only
    -- available when you pass a table to DeclareBoss method.
    -- ==========================================================
    GetCouncilSize = function(self)
        return #self._bossCouncil;
    end,

    -- ================================
    -- Get the n-th mob of the council.
    -- ================================
    GetCouncilMob = function(self, index)
        return self._bossCouncil[index];
    end,

    -- ====================================================================
    -- Specify that a mob with the given mobID is the boss of the module.
    -- Will register the mob in the NPCInfo database so the module can be
    -- enabled when the mob gets seen by the raid.
    -- This method accepts a table as mobID argument to register a council.
    -- ====================================================================
    DeclareBoss = function(self, mobID, volatile, bypass)
        if type(mobID) == "table" then
            -- Multiple bosses declared.
            self._bossCouncil = mobID;
            self:DeclareNoBoss();
      else
            self._bossID = mobID;
        end
        self:DeclareNPC(mobID, volatile, bypass);
    end,

    -- =================================================================
    -- Specify that no mob can be considered as main boss of the module.
    -- You will probably have to specify adds to enable the module.
    -- =================================================================
    DeclareNoBoss = function(self)
        self._bossID = -1;
    end,

    -- ============================================================================================
    -- Specify that the boss is expected to die several times as part of encounter scenario.
    -- This makes the boss killable several times and invalidate the standard boss clear condition.
    -- Your module should overload OnBossDeath to specify the exact winning conditions.
    -- ============================================================================================
    EnableBossMultiDeath = function(self)
        self.bossMultiDeath = true;
    end,

    -- ========================================================
    -- Boss whose mobID has been defined in DeclareBoss will be
    -- searched and, once it is found, OnBossFound is called.
    -- Also add a row in the enemy list for the boss.
    -- If the combat consists of a council of bosses, this
    -- method will register them as adds.
    -- ========================================================
    SearchBoss = function(self)
        local id = self._bossID;
        if id > 0 then -- Boss has been declared as existing.
            self:SearchUnit(id, self.OnBossFound);
            local EnemyList = self:GetWidget("EnemyList");
            if EnemyList then
                EnemyList:AddUnit(id, true);
            end
            self:SetMainUnit(id);

    elseif id < 0 then -- Boss has been declared as not existing or it is a council.
            self:SetNoBoss();
            local i, mobID;
            for i=1, self:GetCouncilSize() do
                mobID = self:GetCouncilMob(i);
                self:RegisterAdd(mobID);
            end
        end
    end,

    -- ===============================
    -- Return if the main boss exists.
    -- ===============================
    HasBoss = function(self)
        return self.bossDefined;
    end,

    -- ============================
    -- Return the name of the boss.
    -- ============================
    GetBossName = function(self)
        if self:HasBoss() then
            return self.bossName;
      else
            return nil;
        end
    end,

    -- ============================
    -- Return the GUID of the boss.
    -- ============================
    GetBossGUID = function(self)
        if self:HasBoss() then
            return self.bossGUID;
      else
            return nil;
        end
    end,

    -- ======================================================
    -- Return the health of the boss (if boss is not visible,
    -- this will return the last health amount we saw).
    -- 3rd return is the boss health ratio.
    -- ======================================================
    GetBossHealth = function(self)
        return self.bossHealth, self.bossHealthMax, self.bossHealth / max(1, self.bossHealthMax);
    end,

    -- ==================================
    -- Return if the boss is dead or not.
    -- ==================================
    IsBossDead = function(self)
        return self.bossDead;
    end,

    -- ===========================================
    -- Return if the boss is out of combat or not.
    -- ===========================================
    IsBossOutOfCombat = function(self)
        return self.bossOutOfCombat;
    end,

    -- ===================================
    -- Set the passed UID as the boss unit
    -- ===================================
    SetBossUnit = function(self, uid)
        if self.bossDefined then -- Cannot define the boss twice.
            return;
        end

        self.bossDefined = true;
        self.bossName = UnitName(uid);
        self.bossGUID = UnitGUID(uid);
        self.bossDead = false;
        self.bossOutOfCombat = false;

        -- Set the title, if it isn't already set.
        self:SetTitle(self.bossName);

        -- Check if encounter is incomplete because we saw the boss already in combat.
        if UnitAffectingCombat(uid) then
            if self:MustSafelyEngage() then
                if self:IsUnitAffiliated(uid.."target") then
                    self:FlagAsIncomplete();
                end
          else
                self:FlagAsIncomplete();
            end
        end

        -- We have now discovered the GUID of the boss.
        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:SetUnit(self._bossID, self.bossGUID);
        end

        -- Do one iteration of boss update immediately, to set its health etc.
        self:CheckBoss(0);
    end,

    -- ====================================================================================
    -- Set the module as having no main boss unit. This will automatically flag it as dead.
    -- ====================================================================================
    SetNoBoss = function(self)
        self.bossDefined = false;
        self.bossDead = true;
        self.bossOutOfCombat = false;
    end,

    -- =======================================================
    -- Check if dying entity <guid> is the boss unit.
    -- (ought to be called when a DEATH combat event is fired)
    -- =======================================================
    CheckBossDeath = function(self, guid)
        if guid and self:GetBossGUID() == guid then
            self.bossHealth = 0;
            self.bossOutOfCombat = false;
            if not self.bossMultiDeath then
                self:FlagBossAsDead();
            end
            self:FireHandler("OnBossDeath", self.bossGUID, Unit:GetMobID(self.bossGUID), self.bossName);
        end
    end,

    -- ==================================================================
    -- Flag the boss as dead and check for winning conditions.
    -- Won't be called automatically if boss multi death flag is enabled.
    -- ==================================================================
    FlagBossAsDead = function(self)
        self.bossDead = true;
        self:CheckStandardClear();
    end,

    -- =====================================================================
    -- Check the main boss unit, including resetting timeout if it is found.
    -- =====================================================================
    CheckBoss = function(self, elapsed)
        if self:HasBoss() then
            local guid = self:GetBossGUID();
            local uid = Unit:GetUID(guid);

            if uid then
                self.bossHealth = UnitHealth(uid);
                self.bossHealthMax = UnitHealthMax(uid);

                -- We saw the boss, reset timeout.
                self:ZeroTimeout();

                -- Check engagement of the boss.
                if self:CheckEngage(uid) then
                    self:OnEngaged();
                end

                -- Check death of the boss.
                if (not self.bossDead) and (not self.bossMultiDeath) and UnitIsDeadOrGhost(uid) then
                    self:CheckBossDeath(guid);
                end

                -- Check combat state of the boss.
                self.bossOutOfCombat = (not UnitAffectingCombat(uid) and not self.bossDead);
            end
        end
    end,

    -- =============================================================================
    -- Set a custom health threshold delay. The delay is how far ahead the player is
    -- warned about an incoming health threshold. Its default value is 10%.
    -- =============================================================================
    SetHealthThresholdDelay = function(self, healthDelay)
        local EnemyList = self:GetWidget("EnemyList");
        if EnemyList then
            EnemyList:SetThresholdDelay(healthDelay);
        end
    end,

    -- =================================================================================
    -- Add a new callback to be fired when the boss gets below a given health threshold.
    -- =================================================================================
    RegisterHealthThreshold = function(self, threshold, label, callback, ...)
        if threshold then
            tinsert(self.healthThresholds, {
                threshold = threshold,
                callback = callback,
                label = label,
                args = {...},
            } );
            if label and self.bossGUID then
                local EnemyList = self:GetWidget("EnemyList");
                if EnemyList then
                    EnemyList:AddThreshold(self.bossGUID, label, threshold);
                end
            end
        end
    end,

    -- ====================================
    -- Fire the specified health threshold.
    -- ====================================
    TriggerHealthThreshold = function(self, index)
        local info = self.healthThresholds[index];
        if info then
            if info.callback then
                info.callback(unpack(info.args));
            end
            if info.label and self.bossGUID then
                local EnemyList = self:GetWidget("EnemyList");
                if EnemyList then
                    EnemyList:RemoveThreshold(self.bossGUID, info.label);
                end
            end
            tremove(self.healthThresholds, index);
        end
    end,

    -- ==================================================
    -- Evaluate each of the registered health thresholds.
    -- ==================================================
    CheckHealthThresholds = function(self)
        if self:GetState() == "ENGAGED" then
            local _, _, healthRatio = self:GetBossHealth();
            local i, info;
            for i=#self.healthThresholds, 1, -1 do
                info = self.healthThresholds[i];
                if healthRatio > 0.005 and healthRatio <= info.threshold then
                    self:TriggerHealthThreshold(i);
                end
            end
        end
    end,
} );