local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Authorization = Root:GetOrNewModule("Authorization");
local L = Root:GetOrNewModule("Localization");
local Social = Root:GetOrNewModule("Social");
local Unit = Root:GetOrNewModule("Unit");
local Network = Root:GetOrNewModule("Network");

local System = Root:GetOrNewModule("System");
local Symbol = System:GetOrNewModule("Symbol");

-- Categorized alerts may or may not appear according to the difficulty setting and role.
-- If a categorized alert doesn't match the role, it won't appear at all.
--
--                          *********************************************************************
--                          *       Low      *      Normal    *      High      *    Critical    *
-- **********************************************************************************************
-- *        Very easy       *      Shown     *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Easy          *    Invisible   *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *         Normal         *    Invisible   *    Invisible   *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Hard          *    Invisible   *    Invisible   *    Invisible   *      Shown     *
-- **********************************************************************************************
-- *        Very hard       *    Invisible   *    Invisible   *    Invisible   *    Invisible   *
-- **********************************************************************************************

local MAX_LIST_SIZE = 6; -- For announces and symbols.

Module:RegisterLib( {
    -- Methods
    -- *******

    -- ==============================================
    -- Send a general announce (uses raid warning).
    -- You must have the permission.
    -- It will try to display the message on party
    -- chat or say chat in other situations.
    -- ==============================================
    Announce = function(self, message, ignoreStyle)
        if not ignoreStyle then
            message = ">> "..message.." <<";
        end
        if IsInRaid() then
            if Authorization:Check("RAID_WARNING") then
                Social:Output("RAID_WARNING", message);
                return true;
            end
      else
            Social:Say(message);
            return true;
        end
        return false;
    end,

    -- =============================================================================
    -- Send an announce about a unit, optionnally putting a symbol on this unit at
    -- the same time and showing the symbol in the announce as well (if symbol > 0).
    -- The message must have a %s format tag (for the unit name).
    -- You must also have required authorizations of course.
    -- =============================================================================
    AnnounceUnit = function(self, message, unit, symbol, removeTimer)
        unit = Unit:ResolveUID(unit);
        if unit then
            local unitName = UnitName(unit);
            message = message:format(unitName);

            if symbol and symbol > 0 and Authorization:Check("SYMBOL") then
                self:PlaceSymbol(unit, symbol, removeTimer);
                message = "{rt"..symbol.."} "..message.." {rt"..symbol.."}";
            end
      else
            message = message:format("???");
        end
        return self:Announce(message);
    end,

    -- ===============================================================
    -- Send an announce about an unit being affected by a spell,
    -- optionnally putting a symbol on this unit at the same time
    -- and showing the symbol in the announce as well (if symbol > 0).
    -- You must also have required authorizations of course.
    -- ===============================================================
    AnnounceUnitSpell = function(self, spell, unit, symbol, removeTimer)
        unit = Unit:ResolveUID(unit);
        if unit then
            local spellName = GetSpellInfo(spell);
            local unitName = UnitName(unit);
            local message = self:FormatLoc("SpellAnnounce", spellName, unitName);

            if symbol and symbol > 0 and Authorization:Check("SYMBOL") then
                self:PlaceSymbol(unit, symbol, removeTimer);
                message = "{rt"..symbol.."} "..message.." {rt"..symbol.."}";
            end
            return self:Announce(message);
        end
        return false;
    end,

    -- =================================================
    -- Send a raid warning enumerating a list of people.
    -- Option: symbols can be put on each person.
    -- Once again, you must have the authorization(s).
    -- =================================================
    AnnounceUnitList = function(self, tbl, header, useSymbols, symbolRemoveTimer)
        -- First, check permissions.
        if (not Authorization:Check("RAID_WARNING") and header) or (useSymbols and not Authorization:Check("SYMBOL")) then
            return false;
        end
        if header then
            self:Announce(header);
        end
        local list = "";
        local count = 0;
        local i, uid, name;
        for i=1, #tbl do
            uid = Unit:GetUID(tbl[i]);
            if uid and count < MAX_LIST_SIZE then
                name = UnitName(uid) or "?";
                count = count + 1;
                if useSymbols then
                    name = string.format("{rt%d} %s", count, name);
                    self:PlaceSymbol(uid, count, symbolRemoveTimer);
                end
                if ( list == "" ) then
                    list = name;
              else
                    list = list .. " || " .. name;
                end
            end
        end
        return self:Announce(list, true);
    end,

    -- ============================================================================
    -- Same as AnnounceUnitList, but automatically sets the header to a spell name.
    -- ============================================================================
    AnnounceUnitListSpell = function(self, tbl, spell, useSymbols, symbolRemoveTimer)
        local spellName = GetSpellInfo(spell);
        if #tbl == 1 then
            return self:AnnounceUnitSpell(spell, tbl[1], useSymbols and 1, symbolRemoveTimer);
      else
            return self:AnnounceUnitList(tbl, spellName, useSymbols, symbolRemoveTimer);
        end
    end,

    -- ====================
    -- Write a say message.
    -- ====================
    Say = function(self, message)
        Social:Say(message, "SAY", false);
    end,

    -- ===========================
    -- Write a party/raid message.
    -- ===========================
    RaidMessage = function(self, message)
        Social:Say(message);
    end,

    -- ===========================================
    -- Send a custom BE3 net packet.
    -- Should be caught with OnNetCommand handler.
    -- Pass only strings and numbers.
    -- ===========================================
    SendPacket = function(self, ...)
        Network:Send("CUSTOM", "GROUP", ...);
    end,

    -- =============================================
    -- Same as SendPacket, but to a specific player.
    -- =============================================
    SendPacketToPlayer = function(self, target, ...)
        Network:Send("CUSTOM", target, ...);
    end,

    -- ================================================================
    -- Lock user symbols, so we won't erase them while putting symbols.
    -- ================================================================
    LockUserSymbols = function(self)
        Symbol:LockUserSymbols();
    end,

    -- ====================================
    -- Clear all previously placed symbols.
    -- Authorization required.
    -- ====================================
    ClearAllSymbols = function(self)
        Symbol:ClearAll();
    end,

    -- ===================================================================
    -- Remove a symbol from someone. Unit can be an UID, a GUID or a name.
    -- Authorization required.
    -- ===================================================================
    ClearSymbol = function(self, unit)
        return Symbol:ClearSymbol(unit);
    end,

    -- ================================================================
    -- Place a symbol on someone. Unit can be an UID, a GUID or a name.
    -- Authorization required.
    -- ================================================================
    PlaceSymbol = function(self, unit, symbol, removeTimer)
        return Symbol:PlaceSymbol(unit, symbol, removeTimer);
    end,

    -- ========================================
    -- Place a symbol on each entity of a list.
    -- Authorization required.
    -- ========================================
    PlaceSymbolList = function(self, tbl, removeTimer)
        local i;
        for i=1, min(MAX_LIST_SIZE, #tbl) do
            self:PlaceSymbol(tbl[i], i, removeTimer);
        end
    end,

    -- ==============================================
    -- Display a local message (not shown raid-wide).
    -- Duration is optionnal.
    -- ==============================================
    Message = function(self, message, duration)
        self:SendViewEvent("OnMessage", message, duration);
    end,

    -- ====================================================
    -- Display a local alert message (not shown raid-wide).
    -- Duration is optionnal.
    -- ====================================================
    Alert = function(self, message, duration)
        self:SendViewEvent("OnAlert", message, duration);
    end,

    -- ===========================================================
    -- Display an alert about an unit (must have a %s format tag).
    -- ===========================================================
    AlertUnit = function(self, message, unit, duration)
        unit = Unit:ResolveUID(unit);
        if unit then
            local unitName = UnitName(unit);
            message = message:format(unitName);
      else
            message = message:format("???");
        end
        self:Alert(message, duration);
    end,

    -- ==============================================================
    -- Check if an alert with the given category should be displayed.
    -- ==============================================================
    EvaluateAlertCategory = function(self, category)
        local total, importance, roleMod = self:GetCategoryValue(category);
        if roleMod >= 0 then
            total = total + self:GetDifficultyValue();
            return (total >= 2);
        end
        return false;
    end,

    -- =========================================================
    -- Display an alert if the local player has the correct role
    -- and a correct difficulty level.
    -- =========================================================
    AlertCategorized = function(self, category, message, duration)
        if self:EvaluateAlertCategory(category) then
            self:Alert(message, duration);
            return true;
      else
            return false;
        end
    end,

    -- ======================================================
    -- Display an alert about an unit if the local player has
    -- the correct role and a correct difficulty level.
    -- ======================================================
    AlertUnitCategorized = function(self, category, message, unit, duration)
        if self:EvaluateAlertCategory(category) then
            self:AlertUnit(message, unit, duration);
            return true;
      else
            return false;
        end
    end,

    -- ======================================================
    -- Look at a numerical setting and compare the value with
    -- the specified value. If the specified value is higher
    -- than the setting's value, then an alert is displayed.
    -- The alert must have a %d format token.
    -- ======================================================
    AlertThreshold = function(self, settingId, count, message)
        if count >= self:GetSetting(settingId) and self:CheckCooldown(settingId, 5) then
            self:Alert(message:format(count));
        end
    end,

    -- ==================================================
    -- Show a categorized alert about a spell on someone.
    -- ==================================================
    AlertSpell = function(self, category, spell, guid)
        local spellName = GetSpellInfo(spell);
        local uid = Unit:ResolveUID(guid);
        if spellName and uid then
            local message;
            if not self:IsMe(uid) then
                local name = UnitName(uid);
                message = self:FormatLoc("SpellOtherAlert", spellName, name);
          else
                message = self:FormatLoc("SpellSelfAlert", spellName);
            end
            self:AlertCategorized(category, message);
        end
    end,

    -- =================================================
    -- Show a categorized alert about an incoming spell.
    -- =================================================
    AlertIncomingSpell = function(self, category, spell)
        local spellName = GetSpellInfo(spell);
        if spellName then
            local message = self:FormatLoc("SpellIncomingAlert", spellName);
            self:AlertCategorized(category, message);
        end
    end,

    -- ===============================================================
    -- Attempt to display a super alert.
    -- What happens depends on actual settings of the user.
    -- Indeed, some players may have disabled intrusive alerts.
    -- Set focus to true if the alert needs attention from the player.
    -- ===============================================================
    SuperAlert = function(self, title, subTitle, duration, focus)
        self:SendViewEvent("OnSuperAlert", title, subTitle, duration, focus);
    end,

    -- ===================================================
    -- Tell the player to interrupt the goddamn mob!
    -- Will default to the boss if mobID is not specified.
    -- Alert won't fire if the boss is not targetted.
    -- Alert won't fire for healers.
    -- ===================================================
    InterruptMob = function(self, mobID, importance)
        local bossGUID = self:GetBossGUID();
        local bossID = self:GetMobID(bossGUID);
        mobID = mobID or bossID;
        if mobID and self:HasTarget(mobID) and self:HasInterrupt() then
            local category = "all/"..(importance or "high");
            if mobID == bossID or mobID == bossGUID then
                self:AlertCategorized(category, self:Localize("InterruptBoss"));
          else
                local mobName = UnitName("target");
                self:AlertCategorized(category, self:FormatLoc("InterruptUnit", mobName));
            end
        end
    end,

    -- =======================================
    -- Call this when a player becomes a bomb!
    -- =======================================
    Bomb = function(self, guid, importance)
        if self:IsMe(guid) then
            local category = "all/"..(importance or "high");
            self:AlertCategorized(category, self:Localize("BombSelf"));
      else
            local category = "all/"..(importance or "normal");
            self:AlertUnitCategorized(category, self:Localize("BombOther"), guid);
        end
    end,

    -- ===================================
    -- Display the current attempt number.
    -- ===================================
    ShowAttempt = function(self)
        self:Message(L:FormatLoc("Try", self:GetAttempt()));
    end,

    -- ======================
    -- Print a debug message.
    -- ======================
    Debug = function(self, message)
        Module:Debug("["..self:GetName().."] "..message);
    end,

    -- =======================================================
    -- Notify the user the boss module is not implemented yet.
    -- =======================================================
    NotImplemented = function(self)
        self:Message(L:Get("NotImplementedBoss1"));
        self:Message(L:Get("NotImplementedBoss2"));
    end,
} );