local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local System = Root:GetOrNewModule("System");
local Error = System:GetOrNewModule("Error");

local Unit = Root:GetOrNewModule("Unit");

-- Errors provide an interface to define possible errors done by players in the encounter.

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _hasErrors = false,
    _errors = { },

    -- Methods
    -- *******

    -- =============================================
    -- This method checks the existence of an error.
    -- =============================================
    HasError = function(self, name)
        return (self._errors[name] ~= nil);
    end,

    -- ============================================
    -- Start the error system, if we're allowed to.
    -- ============================================
    StartError = function(self)
        if self._hasErrors and self:GetSetting("EnableErrorTool") and self:GetEnvironment() == "RAID" then
            -- Upload definitions to the error system.
            Error:Clear();
            local id, info;
            for id, info in pairs(self._errors) do
                Error:Define(id, info.type, info.argument, info.tolerance, info.gracePeriod);
            end

            -- Then start it!
            Error:Start();
        end
    end,

    -- ======================
    -- Stop the error system.
    -- ======================
    StopError = function(self)
        Error:Stop();
    end,

    -- =========================================
    -- This method accesses an error data table.
    -- =========================================
    GetError = function(self, name)
        local error = self._errors[name];
        if not error then
            self:Debug("Error "..name.." is not defined.");
        end
        return error;
    end,

    -- ==================================================
    -- Flag the specified player as having done an error.
    -- ==================================================
    TriggerError = function(self, name, player)
        Error:Trigger(name, player);
    end,

    -- =====================================================
    -- Enable an error. All errors start enabled by default.
    -- =====================================================
    EnableError = function(self, name)
        Error:Enable(name);
    end,

    -- =================
    -- Disable an error.
    -- =================
    DisableError = function(self, name)
        Error:Disable(name);
    end,

    -- ==============================================
    -- Flag a player has been exempted from an error.
    -- ==============================================
    ExemptError = function(self, name, unit)
        local uid = Unit:ResolveUID(unit);
        if uid then
            local player = UnitName(uid);
            Error:Exempt(name, player, true);
        end
    end,

    -- =================================================
    -- The following methods allow you to define errors.
    -- =================================================

    -- The most general error definition method. It shouldn't be used by modules though.

    DefineError = function(self, name, type, argument, tolerance, gracePeriod)
        if self:HasError(name) then
            self:Debug("Error "..name.." is already defined!");
      else
            self._hasErrors = true;
            self._errors[name] = {
                type        = type,
                argument    = argument,
                tolerance   = tolerance,
                gracePeriod = gracePeriod,
            };
            return self._errors[name];
        end
    end,

    -- The following are the good methods to use.
    -- ******************************************
    -- tolerance corresponds to the times in a row the error can trigger without being validated.
    -- gracePeriod is the period during which this error cannot be validated.

    DefineAvoidEffectError = function(self, spell, tolerance, gracePeriod)
        if type(spell) == "table" then
            local id;
            for _, id in ipairs(spell) do
                self:DefineAvoidEffectError(id, tolerance, gracePeriod);
            end
      else
            self:DefineError("e:"..spell, "avoid-effect", spell, tolerance, gracePeriod);
        end
    end,

    DefineAvoidDamageError = function(self, spell, tolerance, gracePeriod)
        if type(spell) == "table" then
            local id;
            for _, id in ipairs(spell) do
                self:DefineAvoidDamageError(id, tolerance, gracePeriod);
            end
      else
            self:DefineError("d:"..spell, "avoid-damage", spell, tolerance, gracePeriod);
        end
    end,

    DefineAvoidBombError = function(self, spell, tolerance, gracePeriod)
        if type(spell) == "table" then
            local id;
            for _, id in ipairs(spell) do
                self:DefineAvoidBombError(id, tolerance, gracePeriod);
            end
      else
            self:DefineError("b:"..spell, "avoid-bomb", spell, tolerance, gracePeriod);
        end
    end,

    DefineAvoidInterruptError = function(self, spell, gracePeriod)
        if type(spell) == "table" then
            local id;
            for _, id in ipairs(spell) do
                self:DefineAvoidInterruptError(id, gracePeriod);
            end
      else
            self:DefineError("i:"..spell, "avoid-interrupt", spell, 0, gracePeriod);
        end
    end,

    DefineAggroError = function(self, mob, gracePeriod)
        if type(mob) == "table" then
            local id;
            for _, id in ipairs(mob) do
                self:DefineAggroError(id, gracePeriod);
            end
      else
            self:DefineError("a:"..mob, "aggro", mob, 0, gracePeriod or 3);
        end
    end,

    DefineBossAggroError = function(self, gracePeriod) -- This is a helper method that automatically defines aggro errors for the main boss(es).
        local bossID = self:GetBossID();
        if bossID then
            self:DefineAggroError(bossID, gracePeriod);
      else
            -- Check for council.
            local size = self:GetCouncilSize();
            local i, id;
            for i=1, size do
                id = self:GetCouncilMob(i);
                self:DefineAggroError(id, gracePeriod);
            end
        end
    end,

    DefineCustomError = function(self, name, tolerance, gracePeriod)
        self:DefineError(name, "custom", self:Localize(name), tolerance, gracePeriod);
    end,
} );