local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Table = Root:GetOrNewModule("Table");

-- Probably one of the most important libs to do the internal plumbering of the modules in a rather uniform way
-- as well as providing shortcuts for handling common events, such as "DO NOT STAY IN FIRE" alerts and such.
-- Also see actions.lua for the part that really "do" stuff.

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _castEvents = { },
    _damageEvents = { },
    _healEvents = { },
    _interruptEvents = { },
    _effectGainEvents = { },
    _effectFadeEvents = { },

    currentContext = { },

    effectHistory = { },
    activeEffects = { },

    -- Methods
    -- *******

    -- ============================
    -- Erase current context infos.
    -- ============================
    WipeContext = function(self)
        wipe(self.currentContext);
    end,

    -- ===========================
    -- Set a current context info.
    -- ===========================
    SetContextInfo = function(self, field, value)
        self.currentContext[field] = value;
    end,

    -- ===========================
    -- Get a current context info.
    -- ===========================
    GetContextInfo = function(self, field)
        return self.currentContext[field];
    end,

    -- =================================================
    -- Check if an effect is already active on a player.
    -- =================================================
    IsEffectPresent = function(self, guid, spellId)
        return self.activeEffects[guid..":"..spellId] or false;
    end,

    -- =====================================
    -- Register a gain or loss of an effect.
    -- =====================================
    RegisterEffectHistory = function(self, event, guid, spellId)
        local newEntry = Table:Alloc();
        newEntry.event = event;
        newEntry.guid = guid;
        newEntry.spell = spellId;
        newEntry.timestamp = GetTime();
        tinsert(self.effectHistory, newEntry);

        if event == "GAIN" then
            self.activeEffects[guid..":"..spellId] = true;

      elseif event == "FADE" then
            self.activeEffects[guid..":"..spellId] = false;
        end
    end,

    -- ================================================
    -- Erase the effect history and create a blank one.
    -- ================================================
    EraseEffectHistory = function(self)
        Table:Recycle(self.effectHistory, true);
        self.effectHistory = Table:Alloc();
        wipe(self.activeEffects);
    end,

    -- =============================================================================================================================
    -- The following methods allow you to watch for a spell cast through various means and set the event that'd be called if caught.
    -- =============================================================================================================================

    -- This version is specialized into simple casting messages (either casting instant spells or begin casting damage/debuff spells).

    RegisterCastEvent = function(self, spellId, action)
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterCastEvent(v, action);
            end
      else
            self._castEvents[spellId] = action;
        end
    end,

    -- This version is specialized into registering boss spells that get interrupted.

    RegisterInterruptEvent = function(self, spellId, action) 
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterInterruptEvent(v, action);
            end
      else
            self._interruptEvents[spellId] = action;
        end
    end,

    -- These versions are specialized into registering boss spells that damage or heal stuff.
    -- Filter specifies what the damage/heal must satisty in order to invoke the action. It can be "INSTANT", "PERIODIC" or "BOTH". Default to "INSTANT".

    RegisterDamageEvent = function(self, spellId, action, filter) 
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterDamageEvent(v, action, filter);
            end
      else
            self._damageEvents[spellId] = {action = action, filter = filter or "INSTANT"};
        end
    end,

    RegisterHealEvent = function(self, spellId, action, filter) 
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterHealEvent(v, action, filter);
            end
      else
            self._healEvents[spellId] = {action = action, filter = filter or "INSTANT"};
        end
    end,

    -- This version is specialized into registering boss effects (that may do a myriad of things) that get applied on units.
    -- Filter specifies what the stack count must satisty in order to invoke the action. It can be "FIRST", "NEXT" or "ALL". Default to "ALL".
    -- "FIRST" filter might fail under some specific circumstances, if you outrange the player and nothing appears on the combat log.

    RegisterEffectGainEvent = function(self, spellId, action, filter)
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterEffectGainEvent(v, action, filter);
            end
      else
            self._effectGainEvents[spellId] = {action = action, filter = filter or "ALL"};
        end
    end,

    -- This version is specialized into registering boss effects (that may do a myriad of things) that get removed from units.
    -- Filter specifies what the stack count must satisty in order to invoke the action. It can be "LAST", "PREV" or "ALL". Default to "ALL".

    RegisterEffectFadeEvent = function(self, spellId, action, filter)
        if type(spellId) == "table" then
            local v;
            for _, v in pairs(spellId) do
                self:RegisterEffectFadeEvent(v, action, filter);
            end
      else
            self._effectFadeEvents[spellId] = {action = action, filter = filter or "ALL"};
        end
    end,

    -- ==================================================
    -- Now the various handlers called from combat events
    -- ==================================================

    CheckCastEvent = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        local name = self._castEvents[spellId];
        if name then
            self:WipeContext();
            self:SetContextInfo("event", "CAST");
            self:SetContextInfo("actorguid", actorGUID);
            self:SetContextInfo("actorname", actorName);
            self:SetContextInfo("targetguid", targetGUID);
            self:SetContextInfo("targetname", targetName);
            self:SetContextInfo("spellid", spellId);
            self:SetContextInfo("spellname", spellName);
            self:ScheduleAction(name);
        end
    end,

    CheckInterruptEvent = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        local name = self._interruptEvents[spellId];
        if name then
            self:WipeContext();
            self:SetContextInfo("event", "INTERRUPT");
            self:SetContextInfo("actorguid", actorGUID);
            self:SetContextInfo("actorname", actorName);
            self:SetContextInfo("targetguid", targetGUID);
            self:SetContextInfo("targetname", targetName);
            self:SetContextInfo("spellid", spellId);
            self:SetContextInfo("spellname", spellName);
            self:ScheduleAction(name);
        end
    end,

    CheckDamageEvent = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, absorbed)
        local info = self._damageEvents[spellId];
        if info then
            local filter = info.filter;
            local name = info.action;
            if name and ((filter == "BOTH") or (filter == "INSTANT" and (not isPeriodic)) or (filter == "PERIODIC" and isPeriodic)) then
                self:WipeContext();
                self:SetContextInfo("event", "DAMAGE");
                self:SetContextInfo("actorguid", actorGUID);
                self:SetContextInfo("actorname", actorName);
                self:SetContextInfo("targetguid", targetGUID);
                self:SetContextInfo("targetname", targetName);
                self:SetContextInfo("spellid", spellId);
                self:SetContextInfo("spellname", spellName);
                self:SetContextInfo("isperiodic", isPeriodic);
                self:SetContextInfo("amount", amount);
                self:SetContextInfo("overkill", overkill);
                self:SetContextInfo("absorbed", absorbed);
                self:ScheduleAction(name);
            end
        end
    end,

    CheckHealEvent = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed)
        local info = self._healEvents[spellId];
        if info then
            local name = info.action;
            local filter = info.filter;
            local matchFilter = ((filter == "BOTH") or (filter == "INSTANT" and (not isPeriodic)) or (filter == "PERIODIC" and isPeriodic));
            if name and matchFilter then
                self:WipeContext();
                self:SetContextInfo("event", "HEAL");
                self:SetContextInfo("actorguid", actorGUID);
                self:SetContextInfo("actorname", actorName);
                self:SetContextInfo("targetguid", targetGUID);
                self:SetContextInfo("targetname", targetName);
                self:SetContextInfo("spellid", spellId);
                self:SetContextInfo("spellname", spellName);
                self:SetContextInfo("isperiodic", isPeriodic);
                self:SetContextInfo("amount", amount);
                self:SetContextInfo("overheal", overheal);
                self:SetContextInfo("absorbed", absorbed);
                self:ScheduleAction(name);
            end
        end
    end,

    CheckEffectEvent = function(self, event, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
        local info;
        local first = false;
        if event == "GAIN" then
            info = self._effectGainEvents[spellId];
            if not self:IsEffectPresent(targetGUID, spellId) then
                first = true;
                self:RegisterEffectHistory("GAIN", targetGUID, spellId);
            end
      else
            info = self._effectFadeEvents[spellId];
            if stackAmount <= 1 then
                self:RegisterEffectHistory("FADE", targetGUID, spellId);
            end
        end
        if info then
            local name = info.action;
            local filter = info.filter;
            local matchFilter;
            if event == "GAIN" then
                if first then
                    matchFilter = (filter == "FIRST" or filter == "ALL");
              else
                    matchFilter = (filter == "NEXT" or filter == "ALL");
                end
          else
                if stackAmount <= 1 then
                    matchFilter = (filter == "LAST" or filter == "ALL");
              else
                    matchFilter = (filter == "PREV" or filter == "ALL");
                end
            end
            if name and matchFilter then
                self:WipeContext();
                self:SetContextInfo("event", "EFFECT_"..event);
                self:SetContextInfo("actorguid", actorGUID);
                self:SetContextInfo("actorname", actorName);
                self:SetContextInfo("targetguid", targetGUID);
                self:SetContextInfo("targetname", targetName);
                self:SetContextInfo("spellid", spellId);
                self:SetContextInfo("spellname", spellName);
                self:SetContextInfo("stack", stackAmount);
                self:ScheduleAction(name);
            end
        end
    end,
} );