local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local System = Root:GetOrNewModule("System");
local Combat = System:GetOrNewModule("Combat");
local Stats = System:GetOrNewModule("Stats");

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    enableTime = 0,
    engageTime = 0,
    exitDelay = 0,
    clearDelay = 0,

    -- Methods
    -- *******

    -- ================================================
    -- Get the time amount the module has been enabled.
    -- ================================================
    GetEnableTime = function(self)
        if self.enableTime > 0 then
            return GetTime() - self.enableTime;
      else
            return 0;
        end
    end,

    -- ================================================
    -- Get the time amount the module has been engaged.
    -- ================================================
    GetEngageTime = function(self)
        if self.engageTime > 0 then
            return GetTime() - self.engageTime;
      else
            return 0;
        end
    end,

    -- =====================
    -- Enable combat handler
    -- =====================
    EnableCombatHandlers = function(self)
        Combat:RegisterCallback(self.OnCombatEvent, self);
    end,

    -- ======================
    -- Disable combat handler
    -- ======================
    DisableCombatHandlers = function(self)
        Combat:UnregisterCallback(self.OnCombatEvent, self);
    end,

    -- ==========================================================
    -- Fire an handler from the module, checking it exists first.
    -- ==========================================================
    FireHandler = function(self, handlerName, ...)
        local f = self[handlerName];
        if f then
            f(self, ...);
        end
    end,

    -- =============================================================
    -- Combat event handlers. Custom handlers may be added.
    -- They'll be called once the standard handler has done its job.
    -- =============================================================
    OnCombatEvent = {
        ["DEATH"] = function(self, guid, name)
            self:CheckAddDeath(guid);
            self:CreditPendingDamage(guid);
            self:CheckBossDeath(guid);

            self:FireHandler("OnDeath", guid, name);
        end,

        ["DAMAGE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed)
            if actorGUID and targetGUID then
                local damage = amount; -- - (overkill or 0) + (absorbed or 0);
                self:CreditDamage(actorGUID, targetGUID, damage);
            end

            if spellId then
                self:CheckDamageEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, absorbed);
                self:InterruptChainCastSuccessful(spellId);
            end

            self:FireHandler("OnDamage", actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed);
        end,

        ["HEAL"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed)
            if targetGUID and not self:IsIgnoringHealing() then
                local healing = amount - (overheal or 0);
                self:RollbackDamage(targetGUID, healing);
            end

            if spellId then
                self:CheckHealEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed);
                self:InterruptChainCastSuccessful(spellId);
            end

            self:FireHandler("OnHeal", actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed);
        end,

        ["KILL"] = function(self, actorGUID, actorName, targetGUID, targetName)
            if targetGUID and targetGUID == self:GetBossGUID() and actorGUID then
                self:SetCoupDeGraceDealer(actorGUID);
            end
        end,

        ["EFFECT_FADE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
            if spellId then
                self:CheckEffectEvent("FADE", actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount);
            end

            self:FireHandler("OnEffectFade", actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount);
        end,

        ["EFFECT_GAIN"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
            if spellId then
                self:CheckEffectEvent("GAIN", actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount);
                self:InterruptChainCastSuccessful(spellId);
            end

            self:FireHandler("OnEffectGain", actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount);
        end,

        ["CAST_START"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
            if spellId then
                self:CheckCastEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName);
                self:InterruptChainCastStarting(spellId, actorGUID);
            end

            self:FireHandler("OnCastStart", actorGUID, actorName, targetGUID, targetName, spellId, spellName);
        end,

        ["CAST_SUCCESS"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
            if spellId then
                self:CheckCastEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName);
            end

            self:FireHandler("OnCastSuccess", actorGUID, actorName, targetGUID, targetName, spellId, spellName);
        end,

        ["SUMMON"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
            if spellId then
                self:CheckCastEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName);
                self:InterruptChainCastSuccessful(spellId);
            end

            self:FireHandler("OnSummon", actorGUID, actorName, targetGUID, targetName, spellId, spellName);
        end,

        ["INTERRUPT"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
            if spellId then
                self:CheckInterruptEvent(actorGUID, actorName, targetGUID, targetName, spellId, spellName);
                self:InterruptChainCastInterrupted(spellId, actorGUID);
            end

            self:FireHandler("OnInterrupt", actorGUID, actorName, targetGUID, targetName, spellId, spellName);
        end,

        ["INSTAKILL"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
            if spellId then
                self:InterruptChainCastSuccessful(spellId);
            end

            self:FireHandler("OnInstakill", actorGUID, actorName, targetGUID, targetName, spellId, spellName);
        end,
    },

    -- =============================
    -- Fired when module is enabled.
    -- =============================
    OnEnable = function(self)
        self.enableTime = GetTime();
        self:SetState("STANDBY");
        self:SetDefaultTimeout();

        -- Prepare our components
        self:DeclareMySettings();
        self:ConnectGUI();
        self:EnableCombatHandlers();
        self:SelectLastMinuteMusic();

        -- Check for preset title
        self:CheckTitle();

        -- Search for the boss!
        self:SearchBoss();

        self:FireHandler("OnQuickModuleEnable");
        self:FireHandler("OnEnableCustom");

        -- Set the berserk timer, if it has been defined and the custom handler hasn't set it itself.
        self:CheckBerserk();

        -- Set the default end sequence if it hasn't been defined.
        self:SetDefaultEndSequence();

        -- Check if the radar should be displayed
        self:CheckRadar();

        -- Start the pre-battle music if it has been defined.
        self:PlayPreMusic();

        -- Enable the capture system.
        self:StartCapture();

        -- Run the equipment checker for the local player.
        self:CheckEquipment();
    end,

    -- ==============================
    -- Fired when module is disabled.
    -- ==============================
    OnDisable = function(self)
        self:FireHandler("OnDisableCustom"); -- Custom method is called first.

        -- Remove symbols the module has put.
        self:ClearAllSymbols();

        -- Restore overridden CVars.
        self:RestoreCVars();

        -- Release our components
        self:ClearMySettings();
        self:DisableCombatHandlers();
        self:DisconnectGUI();

        -- Stop the error system.
        self:StopError();

        -- Stop the capture system.
        self:StopCapture();
    end,

    -- =============================
    -- Fired when module is engaged.
    -- =============================
    OnEngaged = function(self)
        if self:GetState() ~= "STANDBY" then
            return; -- Cannot be engaged at this time (nor twice).
        end

        self.engageTime = GetTime();
        self:SetState("ENGAGED");
        self:ZeroTimeout();

        -- Take a snapshot of players taking part in the combat.
        self:LockInitialState();

        -- Start BGM of the fight, if it exists.
        self:PlayMusic();

        -- Begin berserk timer, if it has been set.
        if not self:BeginBerserk() then
            -- If there's no berserk timer, still start the regular timing.
            local poorTime = self:GetScoreBenchmarks();
            self:SendViewEvent("OnTimerStart", poorTime);
        end

        -- Start engage timers.
        self:StartEngageTimers();

        -- Take a snapshot of player-defined symbols.
        if not self:IsIncomplete() then
            self:LockUserSymbols();
        end

        -- Close options
        self:CloseOptions();

        -- Run the error system.
        self:StartError();

        self:FireHandler("OnEngagedCustom");
    end,

    -- ========================
    -- Fired when fight is won.
    -- ========================
    OnCleared = function(self)
        if self:IsFinished() then
            return; -- Cannot be finished twice!
        end

        self:SetState("CLEARED");
        self.exitDelay  = 8.00;
        self.clearDelay = 0.10; -- Give a little delay to figure out who scored the Coup de Grace.

        -- Validate the final stats of the fight.
        self:LockFinalState();

        -- Finalize GUI.
        self:FinalizeGUI();

        -- Register the success on the stats module.
        Stats:BossCleared(self);

        -- Block this boss module from being enabled again for some time.
        Module:SetCooldown(self:GetName(), self:GetClearCooldown());

        self:FireHandler("OnClearedCustom");
    end,

    -- ====================================================
    -- Fired when fight is lost or when the module expires.
    -- ====================================================
    OnFailed = function(self, isDefeat)
        if self:IsFinished() then
            return; -- Cannot be finished twice!
        end

        self:SetState("FAILED");
        self.exitDelay  = 5.00;
        self.clearDleay = 0.00;

        -- Validate the final stats of the fight.
        self:LockFinalState();

        -- Finalize GUI.
        self:FinalizeGUI();

        -- Stop music.
        self:StopMusic();

        -- Play end jingle.
        self:PlayFailJingle(isDefeat, self:GetEngageTime());

        if isDefeat then
            -- Happily register the failure as a wipe.
            Stats:BossFailed(self);

            self:Message(self:Localize("Failure"));

            -- Block this boss module from being enabled again for some time.
            Module:SetCooldown(self:GetName(), self:GetFailCooldown());
      else
            self:Message(self:Localize("Expiration"));
        end

        self:FireHandler("OnFailedCustom", isDefeat);
    end,

    -- ===================
    -- Fired periodically.
    -- ===================
    OnUpdate = function(self, elapsed)
        -- The unit search service.
        self:CheckSearches();

        -- The delayed actions system.
        self:CheckActions(elapsed);

        -- Check the main boss status.
        self:CheckBoss(elapsed);

        -- Now the adds handling portion.
        self:CheckAdds(elapsed);

        -- Check if we breached through one of the defined health thresholds.
        self:CheckHealthThresholds();

        -- Check interrupt time-out.
        self:CheckInterrupts();

        -- Check last minute trigger.
        self:CheckLastMinute();

        -- Check near victory trigger.
        self:CheckNearVictory();

        -- Now do the wipe check periodically.
        self:CheckWipe(elapsed);

        -- Now do the combat leave check.
        self:CheckCombat(elapsed);

        -- Check time-out value.
        self:CheckTimeout(elapsed);

        -- Check clear animation.
        if self.clearDelay > 0 then
            self.clearDelay = max(0, self.clearDelay - elapsed);
            if self.clearDelay == 0 then
                self:PlaySilence();
                self:SendViewEvent("OnBossCleared", self:GetClearAnimation(), self:HasCoupDeGrace());
            end
        end

        -- Check exit timer.
        if self:IsFinished() then
            local isCleared = (self:GetState() == "CLEARED");

            self.exitDelay = max(0, self.exitDelay - elapsed);
            if self.exitDelay == 0 and ((not UnitAffectingCombat("player")) or not isCleared) then
                if isCleared then
                    if self:EvaluateEndSequence() then
                        self:SendViewEvent("OnResult");
                    end
                end
                self:Kill(); -- Disable the module now.
            end
        end

        self:FireHandler("OnUpdateCustom", elapsed);
    end,

    -- ===============================
    -- Fired when boss has been found.
    -- ===============================
    OnBossFound = function(self, uid)
        self:FireHandler("OnBossFoundCustom", uid);

        -- Define this unit as the boss.
        self:SetBossUnit(uid);

        -- Set default score benchmark (in case it hasn't been previously defined explicitly).
        self:SetDefaultScoreBenchmarks();
    end,

    -- ==========================================
    -- Fired when boss has completed a spellcast.
    -- ==========================================
    OnBossCast = function(self, unit, spellId, spellName)
        self:FireHandler("OnBossCastCustom", unit, spellId, spellName);
    end,

    -- ================================================
    -- Default way of handling instance format changes.
    -- ================================================
    OnInstanceFormatChange = function(self, mode, players)
        self:CheckBerserk();

        self:FireHandler("OnInstanceFormatChangeCustom", mode, players);
    end,

    -- ==================================
    -- Default way of handling mob yells.
    -- ==================================
    OnMobYell = function(self, message, source, channel, target)
        self:CheckYell(message, source, target);
    end,

    -- ================================
    -- Default way of handling logouts.
    -- ================================
    OnLogout = function(self)
        self:RestoreCVars();
    end,
} );