﻿local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Table = Root:GetOrNewModule("Table");
local Unit = Root:GetOrNewModule("Unit");
local Social = Root:GetOrNewModule("Social");

local VALIDATION_DELAY = 0.75;

local INTERRUPT_STATE = { };

local function interruptFunc(self, uid)
    -- TODO
end

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _interruptSpells = { }, -- The list of spellIds that count as valid interrupts.
    _interruptImportance = "high",

    interruptChain = { },
    interruptCurrent = 0,
    interruptWarn = false,

    interruptCasting = { }, -- List all spells that must be interrupted that are being cast and the source mob.
    interruptMobs = { },    -- List all mobs casting spells that must be interrupted and spellId of the spell being cast.
    interruptPending = { }, -- List all spells that have theorically been interrupted but are pending the time validation.

    -- Methods
    -- *******

    -- API

    -- ==========================================================
    -- Set up the spells that will be monitored for interruption.
    -- ==========================================================
    DefineInterruptChain = function(self, spellIds, tolerance, gracePeriod, importance)
        if type(spellIds) == "number" then
            self._interruptSpells = {spellIds};
      else
            self._interruptSpells = spellIds;
        end
        if importance then
            self._interruptImportance = importance;
        end

        local i;
        for i=1, #self._interruptSpells do
            self:DefineError("interrupt"..i, "interrupt", self._interruptSpells[i], tolerance, gracePeriod);
        end
    end,

    -- ========================================================
    -- Set up a list of players assigned to interrupts.
    -- chainText is a list of players separated by a semicolon.
    -- SettingId is the name of the setting from which the
    -- list of players assigned to interrupt will be pulled.
    -- ========================================================
    StartInterruptChain = function(self, settingId)
        local chainText, decider = self:GetSetting(settingId);
        self:FillPlayerList(self.interruptChain, chainText);

        self.interruptCurrent = min(#self.interruptChain, 1);
        self.interruptWarn = self:IsMe(decider);

        if self.interruptCurrent > 0 then
--[[ TODO
            local UnitList = self:GetWidget("UnitList");
            if UnitList then
                UnitList:SetTitle(self:Localize("Interruption"));
                UnitList:SetPlayers(self.interruptChain);
                UnitList:SetUpdateFunc(interruptFunc);
            end
]]
            self:InterruptChainPlayerIsUp();
        end

        wipe(INTERRUPT_STATE);
    end,

    -- =================================
    -- Stop the current interrupt chain.
    -- =================================
    StopInterruptChain = function(self)
        self.interruptCurrent = 0;
--[[ TODO
        local UnitList = self:GetWidget("UnitList");
        if UnitList then
            UnitList:Clear();
        end
]]
    end,

    -- ========================================
    -- Check if the interrupt chain is running.
    -- ========================================
    IsInterruptChainRunning = function(self)
        return self.interruptCurrent > 0;
    end,

    -- Internal methods

    -- ====================================
    -- Get the size of the interrupt chain.
    -- ====================================
    GetInterruptChainSize = function(self)
        return #self.interruptChain;
    end,

    -- ====================================================
    -- Get the error name for the given spell to interrupt.
    -- ====================================================
    GetInterruptChainErrorName = function(self, spellId)
        local i;
        for i=1, #self._interruptSpells do
            if self._interruptSpells[i] == spellId then
                return "interrupt"..i;
            end
        end
        return nil;
    end,

    -- =================================================
    -- Get the index of a player in the interrupt chain.
    -- =================================================
    GetInterruptChainIndex = function(self, guid)
        local list = self.interruptChain;
        local i;
        for i=1, #list do
            if list[i] == guid then
                return i;
            end
        end
        return nil;
    end,

    -- =======================================================
    -- Get the GUID of the current unit assigned to interrupt.
    -- =======================================================
    GetInterruptChainCurrentGUID = function(self)
        return self.interruptChain[self.interruptCurrent];
    end,

    -- =======================================================
    -- Get the name of the current unit assigned to interrupt.
    -- =======================================================
    GetInterruptChainCurrentName = function(self)
        local guid = self:GetInterruptChainCurrentGUID();
        local uid = Unit:ResolveUID(guid);
        if uid then
            return UnitName(uid);
      else
            return UNKNOWN;
        end
    end,

    -- =====================================================================
    -- Mark a player in the interrupt chain as having succeeded and move on.
    -- =====================================================================
    InterruptChainSuccess = function(self, guid)
        local uid = Unit:GetUID(guid);
        if uid and self.interruptWarn then
            local name = UnitName(uid);
            Social:Output("GROUP", self:FormatLoc("InterruptChainSuccess", name));
        end

        INTERRUPT_STATE[guid] = 1;

        local index = self:GetInterruptChainIndex(guid);
        if index then
            self.interruptCurrent = math.fmod(index, self:GetInterruptChainSize()) + 1;
      else
            self.interruptCurrent = math.fmod(self.interruptCurrent, self:GetInterruptChainSize()) + 1;
        end
        self:InterruptChainPlayerIsUp();
    end,

    -- ============================================================================
    -- Mark the current player in the interrupt chain as having failed and move on.
    -- ============================================================================
    InterruptChainFailure = function(self, spellId)
        if self.interruptWarn then
            Social:Output("GROUP", self:Localize("InterruptChainFailed"));
        end

        local guid = self:GetInterruptChainCurrentGUID();
        INTERRUPT_STATE[guid] = -1;

        local errorName = self:GetInterruptChainErrorName(spellId);
        if errorName then
            local currentPlayer = self:GetInterruptChainCurrentName();
            self:TriggerError(errorName, currentPlayer);
        end
        self.interruptCurrent = math.fmod(self.interruptCurrent, self:GetInterruptChainSize()) + 1;
        self:InterruptChainPlayerIsUp();
    end,

    -- ===============================================================================
    -- Flag the current player in the interrupt chain as the next player to interrupt.
    -- ===============================================================================
    InterruptChainPlayerIsUp = function(self)
        if self.interruptWarn then
            Social:Output("GROUP", self:FormatLoc("InterruptChainNext", self:GetInterruptChainCurrentName()));
        end

        local guid = self:GetInterruptChainCurrentGUID();
        INTERRUPT_STATE[guid] = 0;
    end,

    -- ============================================================
    -- Check if the specified spell belongs to the interrupt chain.
    -- ============================================================
    IsInterruptChainSpell = function(self, spellId)
        local spells = self._interruptSpells;
        local i;
        for i=1, #spells do
            if spells[i] == spellId then
                return true;
            end
        end
        return false;
    end,

    -- Handlers

    -- ================================================================
    -- Called when a spell that is in the interrupt list is being cast.
    -- ================================================================
    InterruptChainCastStarting = function(self, spellId, guid)
        if self:IsInterruptChainRunning() and self:IsInterruptChainSpell(spellId) then
            self.interruptCasting[spellId] = guid;
            self.interruptMobs[guid] = spellId;

            if self:IsMe(self:GetInterruptChainCurrentGUID()) and self:HasTarget(guid) then
                local category = "all/"..self._interruptImportance;
                if guid == self:GetBossGUID() then
                    self:AlertCategorized(category, self:Localize("InterruptBoss"));
              else
                    local mobName = UnitName("target");
                    self:AlertCategorized(category, self:FormatLoc("InterruptUnit", mobName));
                end
            end
        end
    end,

    -- ====================================================================================
    -- Called when a spell that is in the interrupt list has been successfully interrupted.
    -- ====================================================================================
    InterruptChainCastInterrupted = function(self, spellId, guid)
        if self.interruptCasting[spellId] then
            local mob = self.interruptCasting[spellId];
            self.interruptMobs[mob] = nil;
            self.interruptCasting[spellId] = nil;

            local newEntry = Table:Alloc(); -- It's possible for the table to be never recycled
            newEntry.spell = spellId;       -- but in that case, the GC will take over.
            newEntry.guid = guid;
            newEntry.validation = GetTime() + VALIDATION_DELAY;
            tinsert(self.interruptPending, newEntry);
        end
    end,

    -- ==========================================================================
    -- Called when a spell that is in the interrupt list had a successful effect,
    -- such as damage done, healing done or effect applied.
    -- ==========================================================================
    InterruptChainCastSuccessful = function(self, spellId, guid)
        self:InterruptChainRollback(spellId);

        if self.interruptCasting[spellId] then
            local mob = self.interruptCasting[spellId];
            self.interruptMobs[mob] = nil;
            self.interruptCasting[spellId] = nil;

            self:InterruptChainFailure(spellId);
        end
    end,

    -- ====================================================================================
    -- Called when a spell that is supposed to be interrupted has passed through.
    -- Causes all instances of this spellId in the confirmation pipeline to be rolled back.
    -- ====================================================================================
    InterruptChainRollback = function(self, spellId)
        local i, entry;
        for i=#self.interruptPending, 1, -1 do
            entry = self.interruptPending[i];
            if entry.spell == spellId then
                self:InterruptChainFailure(spellId);
                Table:Recycle(entry);
                tremove(self.interruptPending, i);
            end
        end
    end,

    -- ============================
    -- Check interrupt validations.
    -- ============================
    CheckInterrupts = function(self)
        local i, entry;
        for i=#self.interruptPending, 1, -1 do
            entry = self.interruptPending[i];
            if GetTime() > entry.validation then
                self:InterruptChainSuccess(entry.guid);
                Table:Recycle(entry);
                tremove(self.interruptPending, i);
            end
        end
    end,
} );