local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");
local Audio = Root:GetOrNewModule("Audio");

local JADED_THRESHOLD = 11;
local IMPORTANT_JADED_THRESHOLD = 15;
local SWIFT_DEFEAT_THRESHOLD = 60;

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    music = false,
    preMusic = false,
    lastMinuteMusic = false,

    lastMinuteMusicPlayed = false,
    nearVictoryMusicPlayed = false,

    nearVictoryHealth = 0,
    nearVictoryRaidAlive = 0,
    nearVictoryTimeMargin = 0,
    nearVictoryCheck = false,

    -- Methods
    -- *******

    -- ================================================================
    -- Set the music BGM of the fight. Also can set a pre-battle music.
    -- ================================================================
    SetMusic = function(self, music, preMusic)
        self.music = music;
        if preMusic then
            self.preMusic = preMusic;
        end
    end,

    -- ==================================
    -- Change the music BGM of the fight.
    -- ==================================
    ChangeMusic = function(self, music, force)
        self:SetMusic(music);
        return self:PlayMusic(nil, force);
    end,

    -- ==============================================================
    -- Change music during the encounter. This method ensures special
    -- themes (near victory, last minute) will never be interrupted.
    -- If you don't provide name, the module BGM will be played.
    -- ==============================================================
    PlayMusic = function(self, name, force)
        if not name then
            name = self.music;
        end
        if not name then
            return Audio:PlayBattleMusic(self:GetImportance());
        end
        if name and (((not self.nearVictoryMusicPlayed) and (not self.lastMinuteMusicPlayed)) or force) then
            return Audio:PlayMusic(name);
        end
        return false;
    end,

    -- ===========================================
    -- Play the pre-battle music, if there is one.
    -- ===========================================
    PlayPreMusic = function(self)
        local attempt = self:GetAttempt();
        local environment = self:GetEnvironment();
        local baseImportance = self:GetBaseImportance();
        local importance = self:GetImportance();

        if attempt == 1 and environment == "RAID" and self:GetSetting("EnableJingle") then
            -- Yeaaah, first attempt of a raid boss!
            return Audio:PlayMusicOnce("FIRSTTRY");

    elseif self.preMusic then
            return Audio:PlayMusic(self.preMusic);

    elseif (attempt >= JADED_THRESHOLD and baseImportance < 80) or attempt >= IMPORTANT_JADED_THRESHOLD then
            -- Boss has been tried plenty of times. Fire the jaded theme!
            return Audio:PlayMusicType("PREBATTLE_EXHAUSTED") or Audio:PlayMusicType("PREBATTLE_NORMAL");

    elseif environment == "RAID" and importance > 70 then
            return Audio:PlayMusicType("PREBATTLE_NORMAL");
        end
        return false;
    end,

    -- =============================
    -- Play silence "special music".
    -- =============================
    PlaySilence = function(self)
        return Audio:PlayMusic("SILENCE");
    end,

    -- ===========================================
    -- Stop whatever music the module has started.
    -- ===========================================
    StopMusic = function(self)
        Audio:StopMusic();
    end,

    -- =================================================
    -- Return true once a special music theme is played.
    -- =================================================
    IsMusicLocked = function(self)
        return self.lastMinuteMusicPlayed or self.nearVictoryMusicPlayed;
    end,

    -- ===============================
    -- Select the "last minute" music.
    -- ===============================
    SelectLastMinuteMusic = function(self)
        self.lastMinuteMusic = Audio:SelectMusicType("HURRYUP") or false;
    end,

    -- =======================================================================
    -- Play the "last minute" music. It will always interrupt a regular music.
    -- =======================================================================
    PlayLastMinuteMusic = function(self)
        if not self:IsMusicLocked() and Audio:PlayMusicOnce(self.lastMinuteMusic) then
            self.lastMinuteMusicPlayed = true;
            return true;
        end
        return false;
    end,

    -- ===========================================================
    -- Check if conditions to trigger "last minute" theme are met.
    -- ===========================================================
    CheckLastMinute = function(self)
        if self:GetState() == "ENGAGED" then
            local duration = Audio:GetMusicDuration(self.lastMinuteMusic);
            local timeLeft = self:GetBerserkTimeLeft();
            if timeLeft < duration then
                self:PlayLastMinuteMusic();
            end
        end
    end,

    -- ========================================================================
    -- Play the "near victory" music. It will always interrupt a regular music.
    -- ========================================================================
    PlayNearVictoryMusic = function(self)
        if not self:IsMusicLocked() and Audio:PlayMusicType("NEARVICTORY") then
            self.nearVictoryMusicPlayed = true;
            return true;
        end
        return false;
    end,

    -- ============================================================
    -- Set the condition on boss health, time remaining and portion
    -- of raid alive to play "near victory" special theme.
    -- ============================================================
    DefineNearVictory = function(self, healthRatio, raidRatio, timeMargin)
        self.nearVictoryCheck = true;
        self.nearVictoryHealth = healthRatio;
        self.nearVictoryRaidAlive = raidRatio or 0;
        self.nearVictoryTimeMargin = timeMargin or 0;
    end,

    -- ============================================================
    -- Check if conditions to trigger "near Victory" theme are met.
    -- ============================================================
    CheckNearVictory = function(self)
        if self:GetState() == "ENGAGED" and self:HasBoss() and self.nearVictoryCheck then
            local _, _, healthRatio = self:GetBossHealth();
            local _, _, raidRatio = self:GetRaidAlive();
            local timeLeft = self:GetBerserkTimeLeft();
            if healthRatio > 0.005 and healthRatio < self.nearVictoryHealth and raidRatio >= self.nearVictoryRaidAlive and timeLeft >= self.nearVictoryTimeMargin then
                self:TriggerNearVictory();
            end
        end
    end,

    -- =================================
    -- Trigger the "near Victory" theme.
    -- =================================
    TriggerNearVictory = function(self)
        self.nearVictoryCheck = false;
        if self:PlayNearVictoryMusic() then
            local BossBar = self:GetWidget("BossBar");
            if BossBar then
                BossBar:Blink(true);
            end
        end
    end,

    -- =====================================================================
    -- Check if player is near to victory... so a defeat can be even bitter.
    -- =====================================================================
    IsNearVictory = function(self)
        return self.nearVictoryMusicPlayed;
    end,

    -- ===================================
    -- Jingle played after a boss failure.
    -- ===================================
    PlayFailJingle = function(self, isDefeat, combatTime)
        if not isDefeat then
            Audio:PlaySoundType("EXPIRATION");
    elseif self:IsNearVictory() then
            Audio:PlaySoundType("BIGFAIL");
    elseif combatTime < SWIFT_DEFEAT_THRESHOLD then
            Audio:PlaySoundType("SWIFTFAIL");
      else
            Audio:PlaySoundType("FAIL");
        end
    end,
} );