local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");
local Utility = Root:GetOrNewModule("Utility");

local System = Root:GetOrNewModule("System");
local Stats = System:GetOrNewModule("Stats");

local MANA_USER_MARGIN = 1.4;

local manaUserThreshold = {
     [1] =    70,
    [60] =  2000,
    [70] =  4000,
    [80] =  7000,
    [85] = 25000,
    [90] = 100000, -- (placeholder)
};

local expectedDps = {
     [1] = 10,
    [10] = 50,
    [40] = 200,
    [60] = 600,
    [70] = 1200,
    [80] = 4000,
    [85] = 7500,
	[90] = 50000, -- (placeholder)
};

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    hasBenchmarks = false,
    poorTime = 300,
    greatTime = 180,
    hasHeroicBenchmarks = false,
    poorTimeHeroic = 0,
    greatTimeHeroic = 0,
    timeCoeff = 1,

    finalPlayers = 0,
    finalManaRatio = 0,
    finalDuration = 0,
    finalStateLocked = false,

    voidDamage = false,
    ignoreHealing = false,
    strictRelevance = false,
    relevancePending = true,
    relevantTargets = { },

    damageDone = { },
    coupDeGrace = false,
    voidScore = false,
    clearAnim = false,
    endSequence = false,
    ignoreEndSequence = false,

    -- Methods
    -- *******

    -- ===============================================================================
    -- Set the values which are considered to be an average performance
    -- for the boss in terms of time needed to kill the boss.
    -- greatTime specifies how fast a great performance should be (100 pts).
    -- poorTime specifies how slow a crappy performance should be (0 pt).
    -- ===============================================================================
    SetScoreBenchmarks = function(self, greatTime, poorTime)
        self.poorTime = poorTime;
        self.greatTime = greatTime;
        self.hasBenchmarks = true;
    end,

    -- =======================================================
    -- Same as SetScoreBenchmarks, but for heroic mode. 
    -- If not specified, the same will be used as normal mode.
    -- =======================================================
    SetHeroicScoreBenchmarks = function(self, greatTime, poorTime)
        self.poorTimeHeroic = poorTime;
        self.greatTimeHeroic = greatTime;
        self.hasHeroicBenchmarks = true;
    end,

    -- =================================================================
    -- Some bosses need to have a multiplier for the time benchmark
    -- to account for the fact there are adds accompanying the boss etc.
    -- =================================================================
    SetBenchmarksTimeCoeff = function(self, coeff)
        self.timeCoeff = coeff;
    end,

    -- =======================
    -- Get poorTime, greatTime
    -- =======================
    GetScoreBenchmarks = function(self)
        if self.hasHeroicBenchmarks and self:IsHeroic() then
            return self.poorTimeHeroic * self.timeCoeff, self.greatTimeHeroic * self.timeCoeff;
      else
            return self.poorTime * self.timeCoeff, self.greatTime * self.timeCoeff;
        end
    end,

    -- ==============================================================
    -- Set default values for score benchmark. Overriding it is fine.
    -- ==============================================================
    SetDefaultScoreBenchmarks = function(self)
        if not self.hasBenchmarks then
            local difficultyCoeff;
            if self:IsHeroic() then
                difficultyCoeff = 1.25;
          else
                difficultyCoeff = 1.00;
            end

            local bossHealth, bossHealthMax = self:GetBossHealth();
            local level = UnitLevel("player");
            local dps = Utility:LinearInterpolation(expectedDps, level) * self:GetNumSlots() * 0.75;
            local time = bossHealthMax / (dps * difficultyCoeff);

            self.poorTime  = time * 3/2;
            self.greatTime = time * 2/3;
        end
    end,

    -- =============================================================
    -- Acquire and lock the final state at the end of the encounter:
    -- combat duration, number of players alive, raid mana ratio
    -- =============================================================
    LockFinalState = function(self)
        if not self.finalStateLocked then
            self.finalDuration = self:GetEngageTime();
            self.finalPlayers = self:GetRaidAlive();
            _, _, self.finalManaRatio = self:GetRaidMana();
            self.finalStateLocked = true;
        end
    end,

    -- =========================================================
    -- Read the final stats at the end of the encounter:
    -- combat duration, number of players alive, raid mana ratio
    -- =========================================================
    GetFinalStats = function(self)
        if self.finalStateLocked then
            return self.finalDuration, self.finalPlayers, self.finalManaRatio;
      else
            return 0, 0, 0;
        end
    end,

    -- ============================================================
    -- Get current players alive, starting players and ratio alive.
    -- ============================================================
    GetRaidAlive = function(self)
        local current = self:CountNumValidMembers();
        local start = max(self:GetStartingPlayers(), current);
        local ratio = current / max(start, 1);
        return current, start, ratio;
    end,

    -- =======================================================
    -- Get the max, current, ratio mana of the raid.
    -- Only those who have a big enough mana pool are counted.
    -- =======================================================
    GetRaidMana = function(self)
        Unit:EnumerateRaid();
        local currentMana, maxMana, ratio = 0, 0, 0;
        local curValue, maxValue, manaThreshold;
        local i, uid;
        for i=1, Unit:GetEnumerationSize() do
            uid = Unit:GetEnumerationUID(i);
            if UnitIsVisible(uid) and not UnitIsDeadOrGhost(uid) then
                curValue = UnitPower(uid, SPELL_POWER_MANA);
                maxValue = UnitPowerMax(uid, SPELL_POWER_MANA);
                if maxValue > 0 then
                    manaThreshold = Utility:LinearInterpolation(manaUserThreshold, UnitLevel(uid)) * MANA_USER_MARGIN;
                    if maxValue > manaThreshold then
                        currentMana = currentMana + curValue;
                        maxMana = maxMana + maxValue;
                    end
                end
            end
        end
        if maxMana > 0 then
            ratio = currentMana / maxMana;
        end
        return currentMana, maxMana, ratio;
    end,

    -- ================================================================================================
    -- Determinate if damage done to the given unit is relevant for the damage meter.
    -- It is now consolidated in a single string return.
    -- Possible returns:
    -- IGNORE - The damage will not be recorded at all.
    -- IRRELEVANT - The damage will be recorded but will automatically be considered as useless damage.
    -- PENDING - The damage will be recorded but will not be relevant until flagged so.
    -- RELEVANT - The damage is relevant.
    -- ================================================================================================
    IsRelevantTarget = function(self, guid)
        if not self.voidDamage and self:GetState() == "ENGAGED" then
            -- If the target is affiliated with us, ignore damage (friendly fire).
            local uid = Unit:GetUID(guid);
            if uid and self:IsUnitAffiliated(uid) then
                return "IGNORE";
            end

            -- Determinate if the unit is on the relevant targets list
            local id = Unit:GetMobID(guid);
            local specificRelevance = false;
            if self.relevantTargets[guid] or (id and self.relevantTargets[id]) then
                specificRelevance = true;
            end

            -- Check against explicit adds and boss units
            -- ******************************************
            if guid == self:GetBossGUID() then
                return "RELEVANT"; -- Damaging the main boss unit is generally relevant :)
            end

            if not specificRelevance and self:IsAdd(guid) then
                if self:MustKillAdds() then
                    -- This unit is a registered add. Damaging it should be relevant.
                    return "RELEVANT";
              else
                    -- However, if adds are optionnal (that means, killing them is not necessary to clear the boss), the score gain will be pending.
                    return "PENDING";
                end
            end

            -- Past this point, we check against mobs that are not adds nor boss
            -- *****************************************************************
            if self.strictRelevance then
                if specificRelevance then
                    if self.relevancePending then
                        return "PENDING";
                  else
                        return "RELEVANT";
                    end
              else
                    return "IRRELEVANT";
                end
          else
                if specificRelevance then
                    if self.relevancePending then
                        return "PENDING";
                  else
                        return "RELEVANT";
                    end
              else
                    return "PENDING";
                end
            end
        end
        return "IGNORE";
    end,

    -- ============================================================================
    -- Add a list of targets to the relevant targets list.
    -- You can pass GUIDs or mobIDs. GUIDs are of course more specific than mobIDs.
    -- ============================================================================
    AddRelevantTargets = function(self, ...)
        local nArgs = select('#', ...);
        local i, arg;
        for i=1, nArgs do
            arg = select(i, ...);
            self.relevantTargets[arg] = true;
        end
    end,

    -- ============================================================================
    -- Remove a list of targets to the relevant targets list.
    -- You can pass GUIDs or mobIDs. GUIDs are of course more specific than mobIDs.
    -- ============================================================================
    RemoveRelevantTargets = function(self, ...)
        local nArgs = select('#', ...);
        local i, arg;
        for i=1, nArgs do
            arg = select(i, ...);
            self.relevantTargets[arg] = nil;
        end
    end,

    -- ===============================================
    -- Specify that no damage must be counted excepted
    -- if it's against one of the defined target.
    -- ===============================================
    SetStrictRelevance = function(self, flag)
        self.strictRelevance = flag;
    end,

    -- ========================================================
    -- When false: specify that the targets flagged as relevant
    -- do not have to die to validate the damage done to them.
    -- ========================================================
    SetRelevantDamagePending = function(self, pending)
        self.relevancePending = pending;
    end,

    -- ====================================================
    -- Once called, damage data will no longer be recorded.
    -- ====================================================
    DisableDamageMeter = function(self)
        self.voidDamage = true;
    end,

    -- =====================================
    -- Check if the damage meter is enabled.
    -- =====================================
    IsDamageMeterEnabled = function(self)
        return (not self.voidDamage);
    end,

    -- ========================================
    -- Check if heals done on mobs are ignored.
    -- ========================================
    IsIgnoringHealing = function(self)
        return self.ignoreHealing;
    end,

    -- ===============================================
    -- Specify that heals done on the mobs will not be
    -- deducted from total damage done.
    -- ===============================================
    SetIgnoreHealing = function(self, flag)
        self.ignoreHealing = flag;
    end,

    -- =====================================================
    -- Get the table listing damage done on a unit.
    -- Possible damage types: RELEVANT, IRRELEVANT, PENDING.
    -- =====================================================
    GetTargetDamageTable = function(self, target, damageType)
        local t = self.damageDone;
        if not t[target] then
            t[target] = { };
        end
        t = t[target];
        if not t[damageType] then
            t[damageType] = { };
        end
        return t[damageType];
    end,

    -- ==============================================================================
    -- Determinate if pending damage done to a dying unit should be credited to their
    -- respective damage dealers. Generally, this is true excepted in some cases.
    -- ==============================================================================
    MayCreditPendingDamage = function(self, guid)
        local result = self:IsRelevantTarget(guid);
        return (result == "RELEVANT" or result == "PENDING");
    end,

    -- ============================================================================
    -- Credit damage to a damage dealer, either as relevant, pending or irrelevant.
    -- ============================================================================
    CreditDamage = function(self, source, target, amount)
        local result = self:IsRelevantTarget(target);
        if result ~= "IGNORE" then
            local targetTable = self:GetTargetDamageTable(target, result);
            targetTable[source] = (targetTable[source] or 0) + amount;
        end
    end,

    -- ====================================
    -- Remove some damage done to a target.
    -- ====================================
    RollbackDamage = function(self, target, amount)
        local pendingTable    = self:GetTargetDamageTable(target, "PENDING");
        local relevantTable   = self:GetTargetDamageTable(target, "RELEVANT");
        local irrelevantTable = self:GetTargetDamageTable(target, "IRRELEVANT");

        local source, damage;
        local total = 0;

        for source, damage in pairs(pendingTable)    do total = total + damage; end
        for source, damage in pairs(relevantTable)   do total = total + damage; end
        for source, damage in pairs(irrelevantTable) do total = total + damage; end

        if ( total <= 0 ) then return; end

        for source, damage in pairs(pendingTable) do
            pendingTable[source] = pendingTable[source] * ((total - amount) / total);
        end
        for source, damage in pairs(relevantTable) do
            relevantTable[source] = relevantTable[source] * ((total - amount) / total);
        end
        for source, damage in pairs(irrelevantTable) do
            irrelevantTable[source] = irrelevantTable[source] * ((total - amount) / total);
        end
    end,

    -- =========================================================================
    -- Credit all pending damage on a target to their respective damage dealers.
    -- The magic word ADDS credit all pending damage on defined adds.
    -- =========================================================================
    CreditPendingDamage = function(self, target)
        if target == "ADDS" then
            local i, guid;
            for i=1, self:GetNumAdds() do
                guid = select(2, self:GetAddInfo(i));
                if guid then
                    self:CreditPendingDamage(guid);
                end
            end
      else
            local pendingTable  = self:GetTargetDamageTable(target, "PENDING");
            local relevantTable = self:GetTargetDamageTable(target, "RELEVANT");

            local source, damage;
            for source, damage in pairs(pendingTable) do
                relevantTable[source] = (relevantTable[source] or 0) + damage;
                pendingTable[source] = 0;
            end
        end
    end,

    -- =============================
    -- Access the damage data table.
    -- =============================
    GetDamageData = function(self)
        return self.damageDone;
    end,

    -- ===============================================
    -- Set the person who has dealt the Coup de Grace.
    -- ===============================================
    SetCoupDeGraceDealer = function(self, guid)
        self.coupDeGrace = guid;
    end,

    -- ===============================================
    -- Get the person who has dealt the Coup de Grace.
    -- ===============================================
    GetCoupDeGraceDealer = function(self)
        return self.coupDeGrace;
    end,

    -- ========================================================
    -- Check if we are the one to have dealt the Coup de Grace.
    -- ========================================================
    HasCoupDeGrace = function(self)
        local guid = self:GetCoupDeGraceDealer();
        return (guid and guid == UnitGUID("player"));
    end,

    -- =================================================
    -- Once called, no score will be shown when cleared.
    -- =================================================
    DisableScore = function(self)
        self.voidScore = true;
    end,

    -- ==============================
    -- Check if the score is enabled.
    -- ==============================
    IsScoreEnabled = function(self)
        return (not self.voidScore);
    end,

    -- ============================================================================
    -- Set the clear anim displayed when the fight is won. Can only be called once.
    -- ============================================================================
    SetClearAnimation = function(self, anim)
        if not self.clearAnim then
            self.clearAnim = anim;
        end
    end,

    -- ===================================================
    -- Get the clear anim displayed when the fight is won.
    -- ===================================================
    GetClearAnimation = function(self)
        return self.clearAnim or "NORMAL";
    end,

    -- ========================================================================================
    -- Determinate if the end sequence should be displayed. This function should be called only
    -- after the OnCleared handler has been invoked or it will not return a correct value.
    -- ========================================================================================
    EvaluateEndSequence = function(self)
        if not self:IsEndSequenceEnabled() then
            return false;

       elseif self:GetSetting("MaskEndSequence") and not self:GetEnvironment() == "SOLO" then
            local name, attempts, attemptScore, speedScore, techniqueScore, clearTime, previousTime, previousScore, numKilled, sequence, dps = Stats:GetCurrentInfo();
            if numKilled >= 3 and previousScore then
                local newScore = (attemptScore or 0) + (speedScore or 0) + (techniqueScore or 0);
                if newScore <= previousScore then
                    return false;
                end
            end
      else
            return true;
        end
    end,

    -- ===============================================================================
    -- Set the end sequence to be used when the fight is won. Can only be called once.
    -- ===============================================================================
    SetEndSequence = function(self, sequence)
        if not self.endSequence then
            self.endSequence = sequence;
        end
    end,

    -- ===================================================
    -- Set the default end sequence when the fight is won.
    -- ===================================================
    SetDefaultEndSequence = function(self)
        local environment = self:GetEnvironment();
        if environment == "SOLO" or (environment == "PARTY" and self:GetBaseImportance() < 50) then
            self:SetEndSequence("SMALL");
      else
            self:SetEndSequence("FULL");
        end
    end,

    -- ======================================================
    -- Get the end sequence to be used when the fight is won.
    -- ======================================================
    GetEndSequence = function(self)
        return self.endSequence;
    end,

    -- ================================================
    -- Disable the end sequence upon winning the fight.
    -- ================================================
    DisableEndSequence = function(self)
        self.ignoreEndSequence = true;
    end,

    -- =====================================================================
    -- Check if the end sequence is enabled (but not necessarily displayed).
    -- =====================================================================
    IsEndSequenceEnabled = function(self)
        return self.endSequence and (not self.ignoreEndSequence);
    end,
} );