local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

-- Timers provide a rather standard interface to start and manage timers, without removing the possibility of doing things yourself for unusual boss modules.

-- Timers use a category string to figure out which roles should see the timer and for which difficulty setting.
-- Refer to category.lua for exact mechanics.
--
-- According to the difficulty setting and final category value, the timer is displayed as such:
--
--                          *********************************************************************
--                          *        0       *        1       *        2       *        3       *
-- **********************************************************************************************
-- *        Very easy       *      Shown     *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Easy          *     Hidden     *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *         Normal         *    Invisible   *     Hidden     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Hard          *    Invisible   *    Invisible   *     Hidden     *      Shown     *
-- **********************************************************************************************
-- *        Very hard       *    Invisible   *    Invisible   *    Invisible   *     Hidden     *
-- **********************************************************************************************
--
-- Special notes:
-- * "ooc" category string should be used for timers which have no impact on combat. These are always displayed, in a normal fashion.
-- * "invisible" category string should be used for timers which won't be visible in the GUI, ever.

local INVISIBLE_THRESHOLD = 0;
local HIDDEN_THRESHOLD = 1;
local ALERT_THRESHOLD = 2;
local CRITICAL_THRESHOLD = 3;

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _timers = { },

    scheduledTimers = { },

    -- Methods
    -- *******

    -- ============================================
    -- This method checks the existence of a timer.
    -- ============================================
    HasTimer = function(self, name)
        return self._timers[name] ~= nil;
    end,

    -- ======================================
    -- This method checks a timer is running.
    -- ======================================
    IsTimerRunning = function(self, name)
        return self.scheduledTimers[name] or false;
    end,

    -- ========================================
    -- This method accesses a timer data table.
    -- ========================================
    GetTimer = function(self, name)
        local timer = self._timers[name];
        if not timer then
            self:Debug("Timer "..name.." is not defined.");
        end
        return timer;
    end,

    -- =============================================
    -- Schedule timers with an initial trigger time.
    -- =============================================
    StartEngageTimers = function(self)
        if self:IsIncomplete() then return; end

        local timer;
        for _, timer in pairs(self._timers) do
            if timer.startAtEngage then
                self:StartTimer(timer.name, timer.firstTimer);
            end
        end
    end,

    -- =================
    -- Schedule a timer.
    -- =================
    StartTimer = function(self, name, timer)
        local info = self:GetTimer(name);
        if info and not self.scheduledTimers[name] then
            local timeAmount = timer or info.timer;
            local display = info.display;
            local icon = info.icon;

            -- Figure out display mode according to category.
            local invisible = false;
            local hidden = false;
            local alert = 0;
            local value, importanceValue = self:GetFilterValue(info.category);
            if value <= INVISIBLE_THRESHOLD then
                invisible = true;
        elseif value <= HIDDEN_THRESHOLD then
                hidden = true;
        elseif importanceValue >= CRITICAL_THRESHOLD then
                alert = 2;
        elseif importanceValue >= ALERT_THRESHOLD then
                alert = 1;
            end

            -- Send to GUI.
            local EventList = self:GetWidget("EventList");
            if EventList then
                EventList:StartEvent(name, (not invisible) and display, timeAmount, alert, self.OnTimerExpire, self, name);
                if icon then
                    EventList:SetIcon(name, icon);
                end
                if hidden then
                    EventList:HideEventTimer(name);
                end
            end

            self.scheduledTimers[name] = true;
        end
    end,

    -- ====================================================================
    -- Trigger a timer, running its action and removing the timer from GUI.
    -- ====================================================================
    TriggerTimer = function(self, name)
        self:InterruptTimer(name);

        local info = self:GetTimer(name);
        if info then
            if info.autoRepeat then
                self:StartTimer(name);
            end

            local action = info.action;
            if action then
                self:ScheduleAction(action);
            end
        end
    end,

    -- =====================
    -- Interrupt all timers.
    -- =====================
    InterruptAllTimers = function(self)
        local name, value;
        for name, value in pairs(self.scheduledTimers) do
            if value then
                self:InterruptTimer(name);
            end
        end
    end,

    -- ========================================
    -- Interrupt a timer, removing it from GUI.
    -- ========================================
    InterruptTimer = function(self, name)
        if self.scheduledTimers[name] then
            local EventList = self:GetWidget("EventList");
            if EventList then
                EventList:StopEvent(name);
            end
            self.scheduledTimers[name] = false;
        end
    end,

    -- =========================================================
    -- Callback fired from the GUI upon completion of the timer.
    -- =========================================================
    OnTimerExpire = function(self, name)
        local info = self:GetTimer(name);
        if info then
            if info.triggerOnExpire then
                self:TriggerTimer(name);
          else
                self:InterruptTimer(name);
            end
        end
    end,

    -- =================================================
    -- The following methods allow you to define timers.
    -- =================================================

    -- The most general timer definition method.
    -- *****************************************
    -- category is a string that is parsed to determinate if timer should be displayed on the GUI (according to role and difficulty setting).
    -- timer is the average time amount between 2 activations of the timer.
    -- firstTimer is the average time amount until the first activation of the timer.
    -- autoRepeat specifies that the timer will be automatically repeated upon being triggered.
    -- startAtEngage specifies if the timer will be started as soon as the boss is engaged, at the condition that the fight is not flagged as incomplete.
    -- action is the name of the action called once timer is triggered.
    -- triggerOnExpire, when true, means the action will be triggered when the timer expires, instead of requiring a manual "triggertimer" action.

    DefineTimerFull = function(self, name, category, timer, firstTimer, autoRepeat, startAtEngage, action, triggerOnExpire)
        local label;
        name, label = strsplit("/", name, 2);

        if self:HasTimer(name) then
            self:Debug("Timer "..name.." is already defined!");
      else
            self._timers[name] = {
                name            = name,
                display         = self:Localize(name) or name,
                icon            = nil,
                category        = category,
                timer           = timer or firstTimer,
                firstTimer      = firstTimer or timer,
                autoRepeat      = autoRepeat or false,
                startAtEngage   = startAtEngage or false,
                action          = action or nil,
                triggerOnExpire = triggerOnExpire or false,
            };
            if label then
                self:DefineTimerLabel(name, tonumber(label) or label);
            end
            return self._timers[name];
        end
    end,

    -- This sets specifically the label to be displayed for a given timer.
    -- If label is a number, then it will be interpreted as a spell ID and the label will be set to the spell's name.

    DefineTimerLabel = function(self, name, label)
        local info = self:GetTimer(name);
        if info then
            if type(label) == "number" then
                -- Spell ID
                local display, _, icon = GetSpellInfo(label);
                info.display = display or label;
                info.icon = icon;
          else
                -- Direct text
                info.display = label;
            end
        end
    end,

    -- These versions (plus others below) define a timer that is automatically repeated or not once triggered.

    DefineTimer = function(self, name, category, timer, firstTimer, action, triggerOnExpire)
        self:DefineTimerFull(name, category, timer, firstTimer, false, false, action, triggerOnExpire);
    end,

    DefineRepeatingTimer = function(self, name, category, timer, firstTimer, action, triggerOnExpire)
        self:DefineTimerFull(name, category, timer, firstTimer, true, false, action, triggerOnExpire);
    end,

    -- This version defines a timer that is automatically started upon boss engagement.

    DefineInitialTimer = function(self, name, category, timer, firstTimer, action, triggerOnExpire)
        self:DefineTimerFull(name, category, timer, firstTimer, false, true, action, triggerOnExpire);
    end,

    DefineInitialRepeatingTimer = function(self, name, category, timer, firstTimer, action, triggerOnExpire)
        self:DefineTimerFull(name, category, timer, firstTimer, true, true, action, triggerOnExpire);
    end,

    -- Defines a timer that triggers no action.
    -- It can be used simply for a graphical display of periodic events that do not need any particular action.

    DefineVoidTimer = function(self, name, category, timer, firstTimer)
        self:DefineTimer(name, category, timer, firstTimer);
    end,

    DefineRepeatingVoidTimer = function(self, name, category, timer, firstTimer)
        self:DefineRepeatingTimer(name, category, timer, firstTimer, nil, false);
    end,

    -- This version additionnally is automatically started upon boss engagement.

    DefineInitialVoidTimer = function(self, name, category, timer, firstTimer)
        self:DefineInitialTimer(name, category, timer, firstTimer, nil, false);
    end,

    DefineInitialRepeatingVoidTimer = function(self, name, category, timer, firstTimer)
        self:DefineInitialRepeatingTimer(name, category, timer, firstTimer, nil, false);
    end,
} );