local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");

local System = Root:GetOrNewModule("System");
local Distance = System:GetOrNewModule("Distance");
local Capture = System:GetOrNewModule("Capture");
local EquipChecker = System:GetOrNewModule("EquipChecker");

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    cooldowns = { },
    oldCVars = { },

    -- Methods
    -- *******

    -- ===============================================================================
    -- Determinate the mob ID from a GUID. This method is merely a convenience mirror.
    -- ===============================================================================
    GetMobID = function(self, guid)
        return Unit:GetMobID(guid);
    end,

    -- =================================================================================
    -- Determinate a mob GUID from a mob ID. This method is merely a convenience mirror.
    -- =================================================================================
    GetMobGUID = function(self, id)
        return Unit:GetGUIDFromMobID(id, false);
    end,

    -- ====================================================
    -- Determinate if the given unit is affiliated with us.
    -- ====================================================
    IsUnitAffiliated = function(self, unit)
        return Unit:IsPartner(unit);
    end,

    -- =======================================
    -- Determinate if the given unit is alive.
    -- =======================================
    IsUnitAlive = function(self, unit)
        unit = Unit:ResolveUID(unit);
        return unit and UnitExists(unit) and not UnitIsDeadOrGhost(unit);
    end,

    -- ================================================================
    -- Count the number of visible and alive party and raid members.
    -- You are accounted for in this method.
    -- ================================================================
    CountNumValidMembers = function(self, excludeOutOfCombat)
        Unit:EnumerateRaid();
        local count = 0;
        local i, uid;
        for i=1, Unit:GetEnumerationSize() do
            uid = Unit:GetEnumerationUID(i);
            if UnitIsVisible(uid) and not UnitIsDeadOrGhost(uid) then
                if UnitAffectingCombat(uid) or not excludeOutOfCombat then
                    count = count + 1;
                end
            end
        end
        return count;
    end,

    -- ==============================================================================================
    -- Allows one to reduce the number of calls that can be done to a function over a period of time.
    -- Duration is in seconds. Return true if the cooldown is ready, return false if not.
    -- ==============================================================================================
    CheckCooldown = function(self, identifier, duration)
        local t = self.cooldowns;
        local last = t[identifier];
        if (not last) or (GetTime() > (last + duration)) then
            t[identifier] = GetTime();
            return true;
        end
        return false;
    end,

    -- =================================================================================================
    -- Try to determinate the distance (in yards) between two arbitrary units as accurately as possible. 
    -- This is only possible if the units are friendly or else the approximation will be rough and 
    -- must involve the player. GUIDs/names/UIDs can all be passed to this function.
    -- Nil will be returned if no result can be provided. Unit2 will default to "player".
    -- =================================================================================================
    GetDistance = function(self, unit1, unit2)
        unit2 = Unit:ResolveUID(unit2 or "player");
        unit1 = Unit:ResolveUID(unit1);
        return Distance:GetUnitDistance(unit1, unit2);
    end,

    -- ====================================================
    -- Get the GUID, name and UID of the target of an unit.
    -- The unit is the boss by default.
    -- ====================================================
    GetTargetInfo = function(self, otherUnit)
        local unit = Unit:ResolveUID(otherUnit or self:GetBossGUID());
        if unit then
            local targetUID, targetGUID, targetName;
            targetUID = unit.."target";
            if UnitExists(targetUID) then
                targetGUID = UnitGUID(targetUID);
                targetName = UnitName(targetUID);
                return targetGUID, targetName, targetUID;
            end
        end
        return nil, nil, nil;
    end,

    -- =========================================================================
    -- Get the GUID, name and UID of the unit who is currently tanking the unit.
    -- The unit is the boss by default.
    -- =========================================================================
    GetTankUnit = function(self, otherUnit)
        return Unit:GetTankOf(otherUnit or self:GetBossGUID());
    end,

    -- =====================================
    -- Determinate if the specified unit has
    -- the aggro from the boss.
    -- =====================================
    HasBossAggro = function(self, unit)
        unit = Unit:ResolveUID(unit);
        if unit then
            local bossGUID = self:GetBossGUID();
            if bossGUID then
                local bossUID = Unit:GetUID(bossGUID);
                if bossUID then
                    local isTanking = UnitDetailedThreatSituation(unit, bossUID);
                    return isTanking;
                end
            end
        end
        return false;
    end,

    -- ==========================================
    -- Return true if "unit" is the local player.
    -- ==========================================
    IsMe = function(self, unit)
        unit = Unit:ResolveUID(unit);
        if unit and UnitIsUnit(unit, "player") then
            return true;
      else
            return false;
        end
    end,

    -- =====================================================================
    -- Return timeLeft, stackCount on a buff or debuff of an unit.
    -- Return nil, nil if the buff or debuff is not present.
    -- =====================================================================
    GetEffectInfo = function(self, unit, spellId)
        unit = Unit:ResolveUID(unit);
        if unit then
            local startTime, duration, stackCount = Unit:SearchEffect(unit, true, spellId, false);
            if not startTime then
                startTime, duration, stackCount = Unit:SearchEffect(unit, false, spellId, false);
            end
            if startTime then
                local timeLeft;
                if startTime > 0 and duration > 0 then
                    timeLeft = max(startTime + duration - GetTime(), 0);
              else
                    timeLeft = math.huge; -- Infinite duration effect.
                end
                return timeLeft, stackCount;
            end
        end
        return nil, nil;
    end,

    -- ===============================================================
    -- Return true if the local player has the specified mob targeted.
    -- ===============================================================
    HasTarget = function(self, mobID)
        local guid = UnitGUID("target");
        if guid then
            local id = Unit:GetMobID(guid);
            if (id and id == mobID) or (guid == mobID) then
                return true;
            end
        end
        return false;
    end,

    -- ========================================
    -- Start the capture system, if allowed to.
    -- ========================================
    StartCapture = function(self)
        if self:GetSetting("EnableCapture") then
            Capture:Start(self);
        end
    end,

    -- ========================
    -- Stop the capture system.
    -- ========================
    StopCapture = function(self)
        Capture:Stop();
    end,

    -- ===========================
    -- Enumerate all raid members.
    -- ===========================
    EnumerateRaid = function(self)
        Unit:EnumerateRaid();
    end,

    -- ===================================================
    -- Enumerate all raid members with the specified role.
    -- ===================================================
    EnumerateRaidByRole = function(self, role)
        Unit:EnumerateRaidByRole(role);
    end,

    -- ================================
    -- Get the size of the enumeration.
    -- ================================
    GetEnumerationSize = function(self)
        return Unit:GetEnumerationSize();
    end,

    -- ===============================
    -- Get one UID in the enumeration.
    -- ===============================
    GetEnumerationUID = function(self, index)
        return Unit:GetEnumerationUID(index);
    end,

    -- =================================================================================
    -- Fill a list of player GUID from a list of player names, separated by a semicolon.
    -- =================================================================================
    FillPlayerList = function(self, t, nameList)
        wipe(t);
        local names = { strsplit(";", nameList) };
        local i, name, uid;
        for i=1, #names do
            name = strtrim(names[i]);
            uid = Unit:ResolveUID(name);
            if uid and Unit:IsPartner(uid) then
                tinsert(t, UnitGUID(uid));
            end
        end
    end,

    -- =============================================================================================
    -- Take a table that is interpreted as an area list and see in which area the specified unit is.
    -- Area list takes id-table pairs, tables have two elements: X and Y coordinates.
    -- Return the id of the first area matching or nil if none matches.
    -- The unit examined is the player by default.
    -- =============================================================================================
    GetArea = function(self, areaList, unit)
        unit = Unit:ResolveUID(unit or "player");
        if unit then
            local px, py = Distance:GetUnitPosition(unit);
            if px and py then
                local radius = areaList["radius"] or 0.05;
                local id, coords;
                for id, coords in pairs(areaList) do
                    if id ~= "radius" then
                        if Distance:GetMapDistance(px, py, coords[1], coords[2]) < radius then
                            return id;
                        end
                    end
                end
            end
        end
        return nil;
    end,

    -- =============================================================
    -- Check if specified unit is casting the specified spell.
    -- The exact spellID doesn't have to match, only the spell name.
    -- Return timeLeft, castTime if casting the spell.
    -- Return nil, nil if not casting the spell.
    -- The unit is the boss by default.
    -- =============================================================
    GetCastingInfo = function(self, spellId, unit)
        unit = Unit:ResolveUID(unit or self:GetBossGUID());
        local spellName = GetSpellInfo(spellId);
        if unit and spellName then
            local name, _, _, _, startTime, endTime = UnitCastingInfo(unit);
            if name == spellName then
                local timeLeft = max(0, endTime / 1000 - GetTime());
                local castTime = (endTime - startTime) / 1000;
                return min(castTime, timeLeft), castTime;
            end
        end
        return nil, nil;
    end,

    -- ==============================================================
    -- Check the equipment of the local player. The check is delayed.
    -- ==============================================================
    CheckEquipment = function(self)
        if not self:HasAction("_CheckEquipment") then
            self:DeclareAction("_CheckEquipment", 0, 8);
            self:AddCallbackToAction("_CheckEquipment", "CheckEquipmentExecute");
        end
        self:ScheduleAction("_CheckEquipment");
    end,

    -- ================================================
    -- Execute the equipment check of the local player.
    -- ================================================
    CheckEquipmentExecute = function(self)
        if self:GetImportance() > 0 and not UnitAffectingCombat("player") then
            if EquipChecker:Audit("player") < 0 then
                self:SendViewEvent("OnEquipCheckResult", EquipChecker:GetText());
            end
        end
    end,

    -- ===================================
    -- Get the role of the specified unit.
    -- ===================================
    GetUnitRole = function(self, unit)
        return Unit:GetRole(unit);
    end,

    -- ================================================
    -- Override a CVar temporarily, restoring it later.
    -- ================================================
    OverrideCVar = function(self, name, value)
        local val = GetCVar(name);
        val = tonumber(val) or val or value;

        self.oldCVars[name] = val;
        SetCVar(name, value);
    end,

    -- ============================================
    -- Restore all CVars that have been overridden.
    -- ============================================
    RestoreCVars = function(self)
        local name, value;
        for name, value in pairs(self.oldCVars) do
            SetCVar(name, value);
        end
        wipe(self.oldCVars);
    end,
} );