local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local ZoneInfo = Root:GetOrNewModule("ZoneInfo");

local L = Root:GetOrNewModule("Localization");
local Audio = Root:GetOrNewModule("Audio");

local WIPE_CHECK_INTERVAL = 3;
local OUT_OF_COMBAT_DELAY = 3;

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    safeEngage = false,

    checkStandardClear = true,

    checkWipe = true,
    nextWipeCheck = WIPE_CHECK_INTERVAL,

    checkCombat = true,
    outOfCombatTimer = 0,

    phase = 0,

    _failCooldown = 20,
    _clearCooldown = 30,

    _enableZones = { },
    _zoneInstanceName = "",
    _zoneNumBossFilter = false,

    -- Methods
    -- *******

    -- ===================================================
    -- Return true if bosses and adds should be considered
    -- engaged only when they target someone of our raid.
    -- ===================================================
    MustSafelyEngage = function(self)
        return self.safeEngage;
    end,

    -- ==================================
    -- Toggle on the safe engage feature.
    -- ==================================
    EnableSafeEngage = function(self)
        self.safeEngage = true;
    end,

    -- ====================================================
    -- Check if the boss module should be engaged according
    -- to the combat state of the given unit.
    -- ====================================================
    CheckEngage = function(self, uid)
        if (not self:MustSafelyEngage()) or self:IsUnitAffiliated(uid.."target") then
            return UnitAffectingCombat(uid);
        end
        return false;
    end,

    -- ==============================================
    -- Disable the standard clear for this boss.
    -- The module will have to call OnCleared itself.
    -- ==============================================
    DisableStandardClear = function(self)
        self.checkStandardClear = false;
    end,

    -- ====================================================================
    -- Basic check to see if boss and adds are dead to clear the encounter.
    -- ====================================================================
    CheckStandardClear = function(self)
        if self:IsBossDead() and self.checkStandardClear then
            if self:GetNumAddsAlive() == 0 or not self:MustKillAdds() then
                self:OnCleared();
            end
        end
    end,

    -- ===============================
    -- Disable the wipe check feature.
    -- ===============================
    DisableWipeCheck = function(self)
        self.checkWipe = false;
    end,

    -- ==============================
    -- Enable the wipe check feature.
    -- ==============================
    EnableWipeCheck = function(self)
        self.checkWipe = true;
    end,

    -- ================
    -- Check for wipes.
    -- ================
    CheckWipe = function(self, elapsed)
        if self:GetState() == "ENGAGED" and not self:IsTesting() and self.checkWipe then
            self.nextWipeCheck = self.nextWipeCheck - elapsed;
            if self.nextWipeCheck <= 0 then
                self.nextWipeCheck = WIPE_CHECK_INTERVAL;

                if self:CountNumValidMembers(true) == 0 then
                    self:OnFailed(true);
                end
            end
        end
    end,

    -- =========================================
    -- Disable the boss-or-adds-in-combat check.
    -- =========================================
    DisableCombatCheck = function(self)
        self.checkCombat = false;
    end,

    -- ========================================
    -- Enable the boss-or-adds-in-combat check.
    -- ========================================
    EnableCombatCheck = function(self)
        self.checkCombat = true;
    end,

    -- ===========================================
    -- Check for boss or adds leaving combat mode.
    -- ===========================================
    CheckCombat = function(self, elapsed)
        if self:GetState() == "ENGAGED" and not self:IsTesting() and self.checkCombat then
            if self:IsBossOutOfCombat() and self:AreAddsOutOfCombat() then
                self.outOfCombatTimer = self.outOfCombatTimer + elapsed;
                if self.outOfCombatTimer >= OUT_OF_COMBAT_DELAY then
                    self:OnFailed(true);
                end
          else
                self.outOfCombatTimer = 0;
            end
        end
    end,

    -- ====================================
    -- Get the current phase of the combat.
    -- ====================================
    GetPhase = function(self)
        return self.phase;
    end,

    -- =====================================================================
    -- Set the current phase of the combat. Return true if set successfully.
    -- =====================================================================
    SetPhase = function(self, phase)
        if self.phase ~= phase then
            self.phase = phase;
            if phase > 0 then
                self:Message(L:Get("Phase").." "..phase);
                Audio:PlaySound("PHASECHANGE");
            end
            return true;
        end
        return false;
    end,

    -- =============================================================
    -- Similar to set phase, but only allows phase counter to go up.
    -- =============================================================
    SetIncreasingPhase = function(self, phase)
        if self:GetPhase() < phase then
            return self:SetPhase(phase);
      else
            return false;
        end
    end,

    -- ================================================================
    -- Set the cooldown that is triggered on the module upon a failure.
    -- ================================================================
    DefineFailCooldown = function(self, duration)
        self._failCooldown = duration;
    end,

    -- ================================================================
    -- Get the cooldown that is triggered on the module upon a failure.
    -- ================================================================
    GetFailCooldown = function(self)
        return self._failCooldown;
    end,

    -- ================================================================
    -- Set the cooldown that is triggered on the module upon a success.
    -- ================================================================
    DefineClearCooldown = function(self, duration)
        self._clearCooldown = duration;
    end,

    -- ================================================================
    -- Get the cooldown that is triggered on the module upon a success.
    -- ================================================================
    GetClearCooldown = function(self)
        return self._clearCooldown;
    end,

    -- ===================================================================
    -- Specify that the zone listed in localization table at the specified
    -- key will cause the boss module to be enabled and maintained alive
    -- as long as the player stays in any of the declared zone.
    -- ===================================================================
    DeclareZone = function(self, key, requireBossFilter)
        local zoneName = strlower(self:Localize(key));
        self._enableZones[zoneName] = true;
        ZoneInfo:Register(zoneName, self:GetName(), requireBossFilter);
    end,

    -- ==============================================================
    -- Return true if the provided zone is one of the declared zones.
    -- "zone" should be lowercase.
    -- ==============================================================
    IsDeclaredZone = function(self, zone)
        return self._enableZones[zone] ~= nil;
    end,

    -- ====================================================================
    -- Specify that the declared zones will cause the boss module to be
    -- enabled if and only if you have killed the specified number of boss.
    -- ====================================================================
    DefineZoneBossFilter = function(self, key, numBoss)
        self._zoneInstanceName = self:Localize(key);
        self._zoneNumBossFilter = numBoss;
    end,

    -- ===============================================================================
    -- Return true if the player's raid lockout matches the required raid lockout.
    -- Will always return false in LFR (i.e: find a workaround or compromise for LFR).
    -- ===============================================================================
    IsMatchingBossFilter = function(self)
        if not self._zoneNumBossFilter then
            return true;

    elseif not self:IsLFR() then
            local i, instanceName, defeatedBosses;
            for i=1, GetNumSavedInstances() do
                instanceName, _, _, _, _, _, _, _, _, _, _, defeatedBosses = GetSavedInstanceInfo(i);
                if instanceName == self._zoneInstanceName and defeatedBosses == self._zoneNumBossFilter then
                    return true;
                end
            end
        end
        return false;
    end,
} );