local Root = BossEncounter3;
local Module = Root:GetOrNewModule("Module");

local Unit = Root:GetOrNewModule("Unit");

local System = Root:GetOrNewModule("System");
local Distance = System:GetOrNewModule("Distance");

-- Zones provide an interface to show and manage zones on the radar.

-- Zones use a category string to figure out which roles should see the zone and for which difficulty setting.
-- Refer to category.lua for exact mechanics.
--
-- According to the difficulty setting and final category value, the zone is displayed as such:
--
--                          *********************************************************************
--                          *        0       *        1       *        2       *        3       *
-- **********************************************************************************************
-- *        Very easy       *      Shown     *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Easy          *    Invisible   *      Shown     *      Shown     *      Shown     *
-- **********************************************************************************************
-- *         Normal         *    Invisible   *    Invisible   *      Shown     *      Shown     *
-- **********************************************************************************************
-- *          Hard          *    Invisible   *    Invisible   *    Invisible   *      Shown     *
-- **********************************************************************************************
-- *        Very hard       *    Invisible   *    Invisible   *    Invisible   *    Invisible   *
-- **********************************************************************************************
--
-- Special notes:
-- * As you can see, there's not much variation with the zone system behavior according to the difficulty setting (i.e: the zone is either shown or invisible).
-- * That being said, difficulty is also taken in account in the radar to determinate which of its features should be on or off.

local INVISIBLE_THRESHOLD = 1;

Module:RegisterLib( {
    -- Properties [those starting with _ won't be reset between 2 sessions]
    -- **********

    _hasZones = false,
    _zones = { },

    _minRange = false,
    minRange = false,

    -- Methods
    -- *******

    -- ==================================
    -- Define the min range of the radar.
    -- ==================================
    DefineMinRange = function(self, value)
        self._minRange = value;
    end,

    -- ==================================
    -- Change the min range on the radar.
    -- ==================================
    ChangeMinRange = function(self, value)
        self.minRange = value;
        self:CheckRadar();
    end,

    -- =====================================
    -- Check if the radar should be enabled.
    -- =====================================
    HasRadar = function(self)
        if self.minRange or self._hasZones then
            return true;
      else
            return false;
        end
    end,

    -- =================================
    -- Enable the radar if it should be.
    -- =================================
    CheckRadar = function(self)
        if not self.minRange then
            self.minRange = self._minRange;
        end
        if self:HasRadar() then
            self:SendViewEvent("OnRadarRequest", self.minRange);
        end
    end,

    -- ===========================================
    -- This method checks the existence of a zone.
    -- ===========================================
    HasZone = function(self, name)
        return (self._zones[name] ~= nil);
    end,

    -- =======================================
    -- This method accesses a zone data table.
    -- =======================================
    GetZone = function(self, name)
        local zone = self._zones[name];
        if not zone then
            self:Debug("Zone "..name.." is not defined.");
        end
        return zone;
    end,

    -- ===================================================================
    -- Display a zone on the radar. Several display flavors are available.
    -- ===================================================================

    -- Duration is optional. If not specified, the default duration will be used.

    ShowZone = function(self, name, unit, duration) -- This is the simple, automatic version. Unit should be nil for "fixed" zones.
        local info = self:GetZone(name);
        if info then
            if info.type == "fixed" then
                self:ShowZoneFixed(name, info.x, info.y, duration);

        elseif info.type == "aimed" then
                self:ShowZoneAimed(name, unit, duration);

        elseif info.type == "anchored" then
                self:ShowZoneAnchored(name, unit, duration);
            end
        end
    end,

    ShowZoneAnchored = function(self, name, unit, duration)
        local info = self:GetZone(name);
        if info then
            if not info.allowMultiple then
                self:InterruptZone(name);
          else
                self:InterruptZone(name, unit);
            end
            if self:GetFilterValue(info.category) > INVISIBLE_THRESHOLD then
                self:SendViewEvent("OnRadarZoneAnchored", name, info.color, info.radius, unit, duration or info.duration);
            end
        end
    end,

    ShowZoneFixed = function(self, name, x, y, duration)
        local info = self:GetZone(name);
        if info then
            if not info.allowMultiple then
                self:InterruptZone(name);
            end
            if self:GetFilterValue(info.category) > INVISIBLE_THRESHOLD then
                self:SendViewEvent("OnRadarZoneCircular", name, info.color, info.radius, x, y, duration or info.duration);
            end
        end
    end,

    ShowZoneAimed = function(self, name, unit, duration)
        local x, y = Distance:GetUnitPosition(unit);
        if x and y then
            self:ShowZoneFixed(name, x, y, duration);
        end
    end,

    -- ====================
    -- Interrupt all zones.
    -- ====================
    InterruptAllZones = function(self)
        self:SendViewEvent("OnRadarZoneFlush");
    end,

    -- ====================================================================================
    -- Interrupt a zone. Unit is an optional argument allowing you to distinguish
    -- between several instances of anchored zone, in the case you want to remove only one.
    -- ====================================================================================
    InterruptZone = function(self, name, unit)
        self:SendViewEvent("OnRadarZoneInterrupted", name, unit);
    end,

    -- ==================================
    -- Modify the size of an active zone.
    -- ==================================
    ModifyZoneSize = function(self, name, add, factor)
        self:SendViewEvent("OnRadarZoneModSize", name, add, factor);
    end,

    -- ===============================
    -- Set the size of an active zone.
    -- ===============================
    SetZoneSize = function(self, name, size)
        self:SendViewEvent("OnRadarZoneSetSize", name, size);
    end,

    -- ================================================
    -- The following methods allow you to define zones.
    -- ================================================

    -- The most general zone definition method.
    -- ****************************************
    -- category is a string that is parsed to determinate if zone should be displayed on the GUI (according to role and difficulty setting).
    -- color is the meaning of the zone. Can be "green", "yellow", "red".
    -- radius is the radius in yards of the zone.
    -- allowMultiple specifies if existing zones should be removed when a new one is placed.
    -- type specifies how the zone is displayed. Can be "fixed" (spawns at fixed location), "aimed" (spawns where an unit is, doesn't move), "anchored" (follows an unit).
    -- x and y are the spawn coordinates for fixed zones.

    DefineZone = function(self, name, category, color, radius, duration, allowMultiple, type, x, y)
        if self:HasZone(name) then
            self:Debug("Zone "..name.." is already defined!");
      else
            self._hasZones = true;
            self._zones[name] = {
                name          = name,
                category      = category,
                color         = color or "yellow",
                radius        = radius or 5,
                duration      = duration or 5,
                allowMultiple = allowMultiple or false,
                type          = type,
                x             = x,
                y             = y,
            };
            return self._zones[name];
        end
    end,

    DefineFixedZone = function(self, name, category, color, radius, duration, x, y)
        self:DefineZone(name, category, color, radius, duration, false, "fixed", x, y);
    end,

    DefineAimedZone = function(self, name, category, color, radius, duration, allowMultiple)
        self:DefineZone(name, category, color, radius, duration, allowMultiple, "aimed", 0, 0);
    end,

    DefineAnchoredZone = function(self, name, category, color, radius, duration, allowMultiple)
        self:DefineZone(name, category, color, radius, duration, allowMultiple, "anchored", 0, 0);
    end,
} );