﻿-- --------------------------------------------------------------------
-- ////////////////////////////////////////////////////////////////////
-- --                          MODULE PART                           --
-- \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
-- --------------------------------------------------------------------

local Root = BossEncounter3;

local Module = Root:NewSubCore("Module");

local Generator = Module:GetOrNewModule("Generator");

local Table = Root:GetOrNewModule("Table");
local Utility = Root:GetOrNewModule("Utility");
local L = Root:GetOrNewModule("Localization");

Module.printPrefix = "|cff00ff00BE3 Module|r - ";

local activeModule = nil;
local repository = { };

local libs = { };

local cooldowns = { ["GENERAL"] = math.huge }; -- The view is expected to remove the general cooldown once it is ready.

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function GetFromRepository(moduleName)
    local module = repository[moduleName];
    if not module then
        -- Module is not in the repository. Try to generate it if it's a QuickModule.
        Generator:Generate(moduleName);
        module = repository[moduleName]; -- Try checking the repository again.
        if not module then
            -- The generation gave us nothing.
            Module:Debug("Module "..moduleName.." doesn't exist and we're unable generate it either.");
        end
    end
    return module;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> lib: the lib table.                                           *
-- ********************************************************************
-- * Register a new library.                                          *
-- ********************************************************************
function Module:RegisterLib(lib)
    local methods = { };
    local properties = { };
    local k, v;
    for k, v in pairs(lib) do
        if type(v) == "function" then
            methods[k] = v;
      else
            properties[k] = v;
        end
    end
    local newLib = {
        methods = methods,
        properties = properties,
    };
    tinsert(libs, newLib);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the internal name of the module.                        *
-- * >> game: one of the following: WOW, TBC, LK, CATA.               *
-- * >> environment: SOLO, PARTY, RAID.                               *
-- * >> settings: the table describing the possible settings of the   *
-- *              module. Can be nil.                                 *
-- * >> localization: the table containing localization data of the   *
-- *                  new module. Can be nil.                         *
-- ********************************************************************
-- * Create a new module object.                                      *
-- ********************************************************************
function Module:New(name, game, environment, settings, localization)
    if repository[name] or self[name] then
        error("Invalid module name: "..name, 0);
    end

    local t = { -- Provide the persistent core properties.
        _name = name,
        _game = game or "UNKNOWN",
        _environment = environment or "UNKNOWN",
        _settings = settings or {},
        _localization = localization or {},
        _defaultProperties = {},
    };
    self:LocalizeSettings(t._settings);
    self:InstallLibs(t);

    repository[name] = t;
    return t; -- Past this point, the module has standard methods and properties.
end

-- ********************************************************************
-- * Same as New, but directly includes the Cataclysm game argument.  *
-- ********************************************************************
function Module:NewCataclysm(name, environment, settings, localization)
    return self:New(name, "CATA", environment, settings, localization);
end

-- ********************************************************************
-- * Same as NewCataclysm, but includes RAID environment.             *
-- ********************************************************************
function Module:NewCataclysmRaid(name, settings, localization)
    return self:New(name, "CATA", "RAID", settings, localization);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the module object.                                    *
-- ********************************************************************
-- * Install all shared libraries into a fresh module.                *
-- ********************************************************************
function Module:InstallLibs(module)
    local i, lib, k, v;
    for i=1, #libs do
        lib = libs[i];
        for k, v in pairs(lib.methods) do
            module[k] = v;
        end
    end
    self:ResetLibs(module);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the module object.                                    *
-- ********************************************************************
-- * Put all library properties back to their default value.          *
-- ********************************************************************
function Module:ResetLibs(module)
    local i, lib, k, v;
    for i=1, #libs do
        lib = libs[i];
        for k, v in pairs(lib.properties) do
            if k:sub(1, 1) ~= "_" or module[k] == nil then -- persistent (names with '_') will never be overwritten.
                if type(v) == "table" then
                    module[k] = Table:Copy(v); -- Copy the table itself and not the reference!
              else
                    module[k] = v; -- Simple value.
                end
            end
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the module object.                                    *
-- * >> extra: extra table, containing properties and methods.        *
-- ********************************************************************
-- * Install extra members into a fresh module.                       *
-- * An extra property may not overwrite existing properties.         *
-- * Methods can be overwritten.                                      *
-- * Handlers will be renamed to Custom to keep the original intact.  *
-- ********************************************************************
function Module:InstallExtra(module, extra)
    local k, v, isMethod, alreadyExists;
    for k, v in pairs(extra) do
        isMethod = (type(v) == "function");
        alreadyExists = (module[k] ~= nil);
        if k:sub(1, 2) == "On" and isMethod and alreadyExists then
            -- Redirect OnX to OnXCustom if there is a conflict so that base handlers are not overwritten.
            -- Base handlers are programmed to issue a call to the custom version if it exists after their own.
            k = k.."Custom";
        end
        if (not isMethod) and alreadyExists then
            error("Tried to overload "..k.." member, which is not a method! Module: "..module._name, 0);
      else
            module[k] = v;
        end
        if (not isMethod) and k:sub(1, 1) ~= "_" then
            module._defaultProperties[k] = v;
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the module object.                                    *
-- ********************************************************************
-- * Reset extra properties of a module, as long as they are not      *
-- * flagged as persistent (persistent properties start with '_').    *
-- ********************************************************************
function Module:ResetExtra(module)
    local k, v;
    for k, v in pairs(module._defaultProperties) do
        if type(v) == "table" then
            module[k] = Table:Copy(v); -- Copy the table itself and not the reference!
      else
            module[k] = v; -- Simple value.
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> moduleName: the name of the module.                           *
-- * >> checkBossFilter: if true, the boss filter of the module will  *
-- *                     be checked before enabling the module.       *
-- ********************************************************************
-- * Set one of modules as the active module, allowing events to      *
-- * reach this module and firing its OnEnable handler.               *
-- * Return true on success.                                          *
-- ********************************************************************
function Module:Enable(moduleName, checkBossFilter)
    local module = GetFromRepository(moduleName);
    if module then
        if not activeModule and (not self:HasCooldown(moduleName)) and (not self:HasCooldown("GENERAL")) then
            if (not checkBossFilter) or module:IsMatchingBossFilter() then
                activeModule = module;

                -- Move the module to a section where it will be able to receive events.
                self[moduleName] = module;
                repository[moduleName] = nil;
                tinsert(self.moduleList, moduleName);

                self:ResetLibs(module);
                self:ResetExtra(module);
                module:OnEnable();
                return true;
            end
        end
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> moduleName: the name of the module.                           *
-- ********************************************************************
-- * Disable the specified module, putting it back in the repository  *
-- * and firing its OnDisable handler.                                *
-- * This only works for the active module since there's only         *
-- * one active module at a time.                                     *
-- * Return true on success.                                          *
-- ********************************************************************
function Module:Disable(moduleName)
    if activeModule and moduleName == activeModule:GetName() then
        local module = self[moduleName];

        -- Move the module back to its repository.
        self[moduleName] = nil;
        repository[moduleName] = module;
        local i;
        for i=#self.moduleList, 1, -1 do
            if self.moduleList[i] == moduleName then
                tremove(self.moduleList, i);
            end
        end

        activeModule = nil;
        module:OnDisable();

        self:SetCooldown("GENERAL", 2);
        return true;
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> moduleName: the name of the module.                           *
-- * >> engageLater: if true, the engage won't be engaged at once.    *
-- ********************************************************************
-- * Enable the specified module in test mode.                        *
-- ********************************************************************
function Module:Test(moduleName, engageLater)
    local module = GetFromRepository(moduleName);
    if module then
        self:SetCooldown(moduleName, 0);
        self:SetCooldown("GENERAL", 0);
        module:Test(engageLater);
    end
end

-- ********************************************************************
-- * Return true if there is currently an active module.              *
-- ********************************************************************
function Module:HasActive()
    return (activeModule ~= nil);
end

-- ********************************************************************
-- * Get the active module if there's one.                            *
-- ********************************************************************
function Module:GetActive()
    return activeModule;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> moduleName: the name of the module.                           *
-- * >> duration: duration of the cooldown, in seconds.               *
-- ********************************************************************
-- * Prevent the module from being enabled for a while.               *
-- ********************************************************************
function Module:SetCooldown(moduleName, duration)
    cooldowns[moduleName] = GetTime() + duration;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> moduleName: the name of the module.                           *
-- ********************************************************************
-- * Check if the specified module has a cooldown enabled.            *
-- ********************************************************************
function Module:HasCooldown(moduleName)
    return GetTime() < (cooldowns[moduleName] or 0);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> settings: the settings table that will be localized.          *
-- ********************************************************************
-- * Check if the specified module has a cooldown enabled.            *
-- ********************************************************************
function Module:LocalizeSettings(settings)
    local i, setting, meta;
    for i=1, #settings do
        setting = settings[i];
        meta = setting.meta;
        if meta and (not setting.label) then
            if meta.spell then
                local spellLink = GetSpellLink(meta.spell) or "?";
                if meta.threshold then
                    setting.label = L:FormatLoc("SettingThreshold", spellLink);
                    setting.explain = L:FormatLoc("SettingThresholdExplain", spellLink);

            elseif meta.silent and meta.symbol then
                    local symbolLink = Utility:GetSymbolLink(meta.symbol);
                    setting.label = L:FormatLoc("SettingSpellSymbol", spellLink);
                    if meta.multiple then
                        setting.explain = L:FormatLoc("SettingSpellSymbolMultipleExplain", spellLink, symbolLink);
                  else
                        setting.explain = L:FormatLoc("SettingSpellSymbolSingleExplain", symbolLink, spellLink);
                    end
              else
                    setting.label = L:FormatLoc("SettingSpell", spellLink);
                    if meta.symbol then
                        local symbolLink = Utility:GetSymbolLink(meta.symbol);
                        if meta.multiple then
                            setting.explain = L:FormatLoc("SettingSpellMultipleSymbolExplain", spellLink, symbolLink);
                      else
                            setting.explain = L:FormatLoc("SettingSpellSingleSymbolExplain", spellLink, symbolLink);
                        end
                  else
                        if meta.multiple then
                            setting.explain = L:FormatLoc("SettingSpellMultipleExplain", spellLink);
                      else
                            setting.explain = L:FormatLoc("SettingSpellSingleExplain", spellLink);
                        end
                    end
                end
          else
                -- Nothing.
            end
        end
        if meta and meta.heroic then
            if setting.label then
                setting.label = L:ReadTable(setting.label)..L:Get("SettingHeroicOnly");
            end
            if setting.explain then
                setting.explain = L:ReadTable(setting.explain)..L:Get("SettingHeroicOnlyExplain");
            end
        end
    end
end