﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Save = Root:GetOrNewModule("Save");

local Combat = System:GetOrNewModule("Combat");

local Capture = System:GetOrNewModule("Capture");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Provides ways to memorize the abilities of bosses.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

Capture.module = nil;
Capture.current = { };

-- --------------------------------------------------------------------
-- **                           Methods                              **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Check if the capture module is running.                          *
-- ********************************************************************
function Capture:IsRunning()
    return self.module ~= nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the boss module for which the capture is operated.    *
-- ********************************************************************
-- * Start capture system for a module.                               *
-- ********************************************************************
function Capture:Start(module)
    if not self:IsRunning() then
        self.module = module;
        Combat:RegisterCallback(self.OnCombatEvent, self);
    end
end

-- ********************************************************************
-- * Stop capture system for a module.                                *
-- ********************************************************************
function Capture:Stop()
    if self:IsRunning() then
        -- Insert the capture data!
        if self.module:GetEngageTime() > 30 and #self.current > 0 then
            local DB = self.DB;
            local id = strlower(self.module:GetID());
            if not DB[id] then DB[id] = { }; end
            tinsert(DB[id], self.current);
        end

        self.module = nil;
        self.current = { };
        Combat:UnregisterCallback(self.OnCombatEvent, self);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> type: the type of the event.                                  *
-- *          Can be "cast", "effect", "death", "damage", "heal".     *
-- * >> id: the ID of the spell, if applicable.                       *
-- * >> target: GUID of the target, if applicable.                    *
-- * >> solvable: true if the parser should try to determinate        *
-- *              the interval between those events.                  *
-- * >> multiple: true if the spell can cause multiple targets to be  *
-- *              hit at the same time.                               *
-- ********************************************************************
-- * Register an event in the current capture.                        *
-- ********************************************************************
function Capture:Register(type, id, target, solvable, multiple)
    local timestamp = self.module:GetEngageTime();

    tinsert(self.current, {timestamp = timestamp, type = type, id = id, target = target, solvable = solvable, multiple = multiple});
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: GUID of the checked unit.                               *
-- ********************************************************************
-- * Check if specified unit should have its abilities logged.        *
-- ********************************************************************
function Capture:IsValidUnit(guid)
    if self:IsRunning() then
        local M = self.module;
        local strictMatching = (M:IsAdd(guid) or M:GetBossGUID() == guid);
        local approxMatching = (M:GetEngageTime() > 0 and not M:IsUnitAffiliated(guid));
        return strictMatching or approxMatching;
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the encounter.                                  *
-- ********************************************************************
-- * Get the number of samples available for a given encounter.       *
-- ********************************************************************
function Capture:GetNumSamples(id)
    local DB = self.DB;
    if DB and DB[id] then
        return #DB[id];
  else
        return 0;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the encounter.                                  *
-- * >> sample: which sample. 1 is the first.                         *
-- ********************************************************************
-- * Display capture data for a specified encounter.                  *
-- ********************************************************************
function Capture:Show(id, sample)
    if sample >= 1 and sample <= self:GetNumSamples(id) then
        System:Print("---[ "..id:gsub("|", "||"):upper().." sample "..sample.." ]---");
        local info = self.DB[id][sample];
        local i, entry;
        for i=1, #info do
            entry = info[i];
            System:Print(("[%.2f] %s (%s) -> %s"):format(entry.timestamp, entry.type, tostring(entry.id or "nil"), tostring(entry.target or "nil")));
        end
        return true;
  else
        return false;
    end
end

-- ********************************************************************
-- * Erase all capture data.                                          *
-- ********************************************************************
function Capture:Clear()
    self.DB = { };
    Save:Set("system", "capture", self.DB);
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

Capture.OnBootComplete = function(self)
    -- Grab the DB table.
    self.DB = Save:Get("system", "capture") or { };
    Save:Set("system", "capture", self.DB);
end


Capture.OnCombatEvent = {
    ["DEATH"] = function(self, guid, name)
        if self:IsValidUnit(guid) then
            self:Register("death", nil, guid, false, false);
        end
    end,

    ["DAMAGE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed)
        if self:IsValidUnit(actorGUID) then
            self:Register("damage", spellId, targetGUID, true, true);
        end
    end,

    ["HEAL"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, school, absorbed)
        if self:IsValidUnit(actorGUID) then
            self:Register("heal", spellId, targetGUID, true, true);
        end
    end,

    ["EFFECT_GAIN"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
        if self:IsValidUnit(actorGUID) then
            self:Register("effect", spellId, targetGUID, true, true);
        end
    end,

    ["CAST_START"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        if self:IsValidUnit(actorGUID) then
            self:Register("cast", spellId, targetGUID, true, false);
        end
    end,

    ["CAST_SUCCESS"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        if self:IsValidUnit(actorGUID) then
            self:Register("cast", spellId, targetGUID, true, false);
        end
    end,

    ["SUMMON"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        if self:IsValidUnit(actorGUID) then
            self:Register("cast", spellId, targetGUID, true, false);
        end
    end,
};