﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Combat = System:GetOrNewModule("Combat");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Allows to register combat callbacks.

-- --------------------------------------------------------------------
-- **                               Data                             **
-- --------------------------------------------------------------------

local callbacks = { };

local SCHOOL_PHYSICAL = 0x01;

-- ===========================================================
-- Possible events (in uppercase) with given additionnal data:
-- ===========================================================
-- DAMAGE: actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed
-- MISS: actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, missType, amountMissed, school
-- HEAL: actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overheal, absorbed
-- ENERGIZE: actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, powerType
-- DRAIN: actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, powerType, gainedAmount
-- INTERRUPT: actorGUID, actorName, targetGUID, targetName, spellId, spellName, interruptedSpellId, interruptedSpellName
-- INSTAKILL: actorGUID, actorName, targetGUID, targetName, spellId, spellName

-- KILL: actorGUID, actorName, targetGUID, targetName
-- DEATH: guid, name

-- CAST_START: actorGUID, actorName, targetGUID, targetName, spellId, spellName
-- CAST_SUCCESS: actorGUID, actorName, targetGUID, targetName, spellId, spellName
-- SUMMON: actorGUID, actorName, targetGUID, targetName, spellId, spellName

-- EFFECT_GAIN: actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount
-- EFFECT_FADE: actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> table: a table containing functions whose indexes are the     *
-- *           names of all events we want to process.                *
-- * >> obj: the object the table belongs to. May be nil.             *
-- ********************************************************************
-- * Adds a combat callback table to the callback list.               *
-- ********************************************************************
function Combat:RegisterCallback(table, obj)
    tinsert(callbacks, {obj = obj, vector = table});
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> table: a table containing functions whose indexes are the     *
-- *           names of all events we want to process.                *
-- * >> obj: the object the table belongs to. May be nil.             *
-- ********************************************************************
-- * Remove a combat callback from the callback list.                 *
-- ********************************************************************
function Combat:UnregisterCallback(table, obj)
    local i;
    for i=#callbacks, 1, -1 do
        if callbacks[i].vector == table and callbacks[i].obj == obj then
            tremove(callbacks, i);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> event: the combat event type.                                 *
-- * >> ...: the additionnal infos related to the event.              *
-- ********************************************************************
-- * Notifies all currently registered callbacks of a combat event.   *
-- ********************************************************************
function Combat:FireCallbacks(event, ...)
    local i, callback, handler, obj;
    for i=1, #callbacks do
        callback = callbacks[i];
        handler = callback.vector[event];
        if handler then
            obj = callback.obj;
            if obj then
                handler(obj, ...);
          else
                handler(...);
            end
        end
    end
end

-- --------------------------------------------------------------------
-- **                        Combat handlers                         **
-- --------------------------------------------------------------------

function Combat:OnRawCombatEvent(...)
    local timestamp, event, hideCaster, sGUID, sName, _, _, tGUID, tName, _, _ = ...
    local idx = 12;

    local spellId, spellName, spellSchool;
    local extraSpellId, extraSpellName;
    local amount, overkill, school, resisted, blocked, absorbed, overhealing;
    local missType, isOffHand, amountMissed;
    local extraAmount;
    local powerType;

    -- Note: full absorb events are treated as normal 0 damage events.

    if event == "PARTY_KILL" then
        self:FireCallbacks("KILL", sGUID, sName, tGUID, tName);

elseif event == "UNIT_DIED" or event == "UNIT_DESTROYED" then
        self:FireCallbacks("DEATH", tGUID, tName);

elseif event == "SWING_DAMAGE" then
        amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
        self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, nil, nil, false, amount, overkill, school, absorbed);

elseif event == "SWING_MISSED" then
        missType, isOffHand, amountMissed = select(idx, ...);
        if missType == "ABSORB" then 
            self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, nil, nil, false, 0, 0, SCHOOL_PHYSICAL, amountMissed);
      else
            self:FireCallbacks("MISS", sGUID, sName, tGUID, tName, nil, nil, false, missType, amountMissed, SCHOOL_PHYSICAL);
        end

elseif event == "RANGE_DAMAGE" then 
        spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
        self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, overkill, school, absorbed);

elseif event == "RANGE_MISSED" then 
        spellId, spellName, spellSchool, missType, isOffHand, amountMissed = select(idx, ...);
        if missType == "ABSORB" then 
            self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, 0, 0, spellSchool, amountMissed);
      else
            self:FireCallbacks("MISS", sGUID, sName, tGUID, tName, spellId, spellName, false, missType, amountMissed, spellSchool);
        end

elseif event == "DAMAGE_SHIELD" then
        spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
        self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, overkill, school, absorbed);

elseif event == "DAMAGE_SHIELD_MISSED" then 
        spellId, spellName, spellSchool, missType, isOffHand, amountMissed = select(idx, ...);
        if missType == "ABSORB" then 
            self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, 0, 0, spellSchool, amountMissed);
      else
            self:FireCallbacks("MISS", sGUID, sName, tGUID, tName, spellId, spellName, false, missType, amountMissed, spellSchool);
        end

elseif event == "ENVIRONMENTAL_DAMAGE" then
        _, amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
        self:FireCallbacks("DAMAGE", nil, nil, tGUID, tName, nil, nil, false, amount, overkill, school, absorbed);

elseif event == "DAMAGE_SPLIT" then
	spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
        self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, overkill, school, absorbed);

elseif strsub(event, 1, 6) == "SPELL_" then
        spellId, spellName, spellSchool = select(idx, ...);
        idx = idx + 3;

        if event == "SPELL_DAMAGE" or event == "SPELL_BUILDING_DAMAGE" then
            amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
            self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, overkill, school, absorbed);

    elseif event == "SPELL_MISSED" then 
            missType, isOffHand, amountMissed = select(idx, ...);
            if missType == "ABSORB" then 
                self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, false, 0, 0, spellSchool, amountMissed);
          else
                self:FireCallbacks("MISS", sGUID, sName, tGUID, tName, spellId, spellName, false, missType, amountMissed, spellSchool);
            end

    elseif event == "SPELL_HEAL" or event == "SPELL_BUILDING_HEAL" then 
            amount, overhealing, absorbed = select(idx, ...);
            self:FireCallbacks("HEAL", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, overhealing, absorbed);

    elseif event == "SPELL_ENERGIZE" then 
            amount, powerType = select(idx, ...);
            self:FireCallbacks("ENERGIZE", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, powerType);

    elseif event == "SPELL_DRAIN" or event == "SPELL_LEECH" then
            amount, powerType, extraAmount = select(idx, ...);
            self:FireCallbacks("DRAIN", sGUID, sName, tGUID, tName, spellId, spellName, false, amount, powerType, extraAmount);

    elseif event == "SPELL_PERIODIC_MISSED" then
            missType, isOffHand, amountMissed = select(idx, ...);
            if missType == "ABSORB" then 
                self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, true, 0, 0, spellSchool, amountMissed);
          else
                self:FireCallbacks("MISS", sGUID, sName, tGUID, tName, spellId, spellName, true, missType, amountMissed, spellSchool);
            end

    elseif event == "SPELL_PERIODIC_DAMAGE" then
            amount, overkill, school, resisted, blocked, absorbed = select(idx, ...);
            self:FireCallbacks("DAMAGE", sGUID, sName, tGUID, tName, spellId, spellName, true, amount, overkill, school, absorbed);

    elseif event == "SPELL_PERIODIC_HEAL" then
            amount, overhealing, absorbed = select(idx, ...);
            self:FireCallbacks("HEAL", sGUID, sName, tGUID, tName, spellId, spellName, true, amount, overhealing, absorbed);

    elseif event == "SPELL_PERIODIC_ENERGIZE" then 
            amount, powerType = select(idx, ...);
            self:FireCallbacks("ENERGIZE", sGUID, sName, tGUID, tName, spellId, spellName, true, amount, powerType);

    elseif event == "SPELL_PERIODIC_DRAIN" or event == "SPELL_PERIODIC_LEECH" then
            amount, powerType, extraAmount = select(idx, ...);
            self:FireCallbacks("DRAIN", sGUID, sName, tGUID, tName, spellId, spellName, true, amount, powerType, extraAmount);

    elseif event == "SPELL_INTERRUPT" then
            extraSpellId, extraSpellName = select(idx, ...);
            self:FireCallbacks("INTERRUPT", sGUID, sName, tGUID, tName, extraSpellId, extraSpellName, spellId, spellName);

    elseif event == "SPELL_INSTAKILL" then
            self:FireCallbacks("INSTAKILL", sGUID, sName, tGUID, tName, spellId, spellName);

    elseif event == "SPELL_AURA_APPLIED" then
            self:FireCallbacks("EFFECT_GAIN", sGUID, sName, tGUID, tName, spellId, spellName, 1);

    elseif event == "SPELL_AURA_APPLIED_DOSE" then
            _, amount = select(idx, ...);
            self:FireCallbacks("EFFECT_GAIN", sGUID, sName, tGUID, tName, spellId, spellName, amount);

    elseif event == "SPELL_AURA_REMOVED" then
            self:FireCallbacks("EFFECT_FADE", sGUID, sName, tGUID, tName, spellId, spellName, 1);

    elseif event == "SPELL_AURA_REMOVED_DOSE" then
            _, amount = select(idx, ...);
            self:FireCallbacks("EFFECT_FADE", sGUID, sName, tGUID, tName, spellId, spellName, amount);

    elseif event == "SPELL_CAST_START" then
            self:FireCallbacks("CAST_START", sGUID, sName, tGUID, tName, spellId, spellName);

    elseif event == "SPELL_CAST_SUCCESS" then
            self:FireCallbacks("CAST_SUCCESS", sGUID, sName, tGUID, tName, spellId, spellName);

    elseif event == "SPELL_SUMMON" then
            self:FireCallbacks("SUMMON", sGUID, sName, tGUID, tName, spellId, spellName);
        end
    end
end