﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Unit = Root:GetOrNewModule("Unit");
local Save = Root:GetOrNewModule("Save");
local Utility = Root:GetOrNewModule("Utility");

local Distance = System:GetOrNewModule("Distance");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Service that provides distance services while in combat.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

Distance.currentID = "";
Distance.meterScale = 0;

-- Vars for the scale determination algorithm

Distance.moving = false;
Distance.speed = 0;
Distance.startX = 0;
Distance.startY = 0;
Distance.lastX = 0;
Distance.lastY = 0;
Distance.direction = 0;
Distance.facing = 0;
Distance.startTime = 0;

local DEVIATION_TOLERANCE    = 1.8; -- In degrees.
local SPEED_MOD_TOLERANCE    = 0.01; -- Ratio.
local FACING_TOLERANCE       =  18; -- In degrees.
local DURATION_REQUIRED      = 1.5;
local MINIMAL_MAP_MOVEMENT   = 0.00001;
local DETERMINATION_COOLDOWN = 3;

local CORRECTION_FACTOR = 0.93; -- Correct the error of the measure.

-- --------------------------------------------------------------------
-- **                               Hooks                            **
-- --------------------------------------------------------------------

local function hook_SetMapToCurrentZone(self)
    Distance:Refresh();
end

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x, y: coordinates to normalize.                               *
-- ********************************************************************
-- * Normalize map coordinates so that Y axis has the same "weight"   *
-- * as X axis. This is required before any meaningful manipulation.  *
-- ********************************************************************
function Distance:NormalizePosition(x, y)
    return x, y / 1.5;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x1, y1: first point coordinates.                              *
-- * >> x2, y2: second point coordinates.                             *
-- ********************************************************************
-- * Get the 2D distance between 2 map points.                        *
-- * Distance is adjusted so X has the same weight as Y.              *
-- ********************************************************************
function Distance:GetMapDistance(x1, y1, x2, y2)
    return (((x2 - x1)/1.0)^2 + ((y2 - y1)/1.5)^2)^0.5;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x1, y1: first point coordinates.                              *
-- * >> x2, y2: second point coordinates.                             *
-- ********************************************************************
-- * Get the 2D distance between 2 points.                            *
-- ********************************************************************
function Distance:GetDistance(x1, y1, x2, y2)
    return ((x2 - x1)^2 + (y2 - y1)^2)^0.5;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> x1, y1: first point coordinates.                              *
-- * >> x2, y2: second point coordinates.                             *
-- ********************************************************************
-- * Get the angle from the origin of the trigonometric circle        *
-- * and the line joining 2 points.                                   *
-- ********************************************************************
function Distance:Atanfull(x1, y1, x2, y2)
    x2 = x2 - x1;
    y2 = y2 - y1;
    if x2 == 0 then
        if y2 > 0 then
            return 90;
    elseif y2 < 0 then
            return -90;
      else
            return 0;
        end
elseif x2 < 0 then
        return atan(y2 / x2) + 180;
  else
        return atan(y2 / x2);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> a1, a2: the angles, in degrees.                               *
-- ********************************************************************
-- * Get the distance between two angles, the shortest being used.    *
-- ********************************************************************
function Utility:GetAngleDistance(a1, a2)
    local difference = a2 - a1;
    while difference < -180 do
        difference = difference + 360;
    end
    while difference > 180 do
        difference = difference - 360;
    end
    return difference;
end

-- ********************************************************************
-- * Get the current meter scale of the current map scale.            *
-- ********************************************************************
function Distance:GetMeterScale()
    return self.meterScale;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: GUID, name or UID of the unit queried.                  *
-- ********************************************************************
-- * Get the map position of the unit.                                *
-- * Return nil, nil if not available.                                *
-- ********************************************************************
function Distance:GetUnitPosition(unit)
    unit = Unit:ResolveUID(unit);
    if unit then
        local x, y = GetPlayerMapPosition(unit);
        if x > 0 and y > 0 then
            return x, y;
        end
    end
    return nil, nil;
end

-- ********************************************************************
-- * Refresh the distance system to make sure current map scale       *
-- * is loaded.                                                       *
-- ********************************************************************
function Distance:Refresh()
    local level = max(GetCurrentMapDungeonLevel(), 1);
    local ID = strupper(GetMapInfo() or "")..":"..level;
    if ID ~= self.currentID then
        self.currentID = ID;
        self.meterScale = (self.DB[self.currentID] or 0) * CORRECTION_FACTOR;

        if self.meterScale > 0 then
            System:DebugFmt("Loaded scale for zone [%s]: %d", self.currentID, self.meterScale);
      else
            System:DebugFmt("Unknown scale for zone [%s], will try to determinate.", self.currentID);
        end

        self.moving = false; -- Interrupt any movement determination process.
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit, otherUnit: the two units that are checked.              *
-- ********************************************************************
-- * Determinate the distance between two units.                      *
-- * A good estimation will be found for two friendly units.          *
-- * A very rough one will be returned if you want the distance       *
-- * to an enemy unit. This estimation is generally sufficient to     *
-- * determinate if you are in melee range or not though.             *
-- ********************************************************************
function Distance:GetUnitDistance(unit, otherUnit)
    otherUnit = otherUnit or "player";

    local x1, y1 = self:GetUnitPosition(unit);
    local x2, y2 = self:GetUnitPosition(otherUnit);

    if x1 and y1 and x2 and y2 and self.meterScale > 0 then
        return floor(self.meterScale * self:GetMapDistance(x1, y1, x2, y2) * 10 + 0.5) / 10; -- 1-digit floating precision.
  else
        -- Fall back to the standard, approximate way, always involving the local player.
        if unit == "player" then
            return self:GetApproxDistanceFrom(otherUnit);
    elseif otherUnit == "player" then
            return self:GetApproxDistanceFrom(unit);
      else
            return 50000;
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: the remote unit to check the distance to.               *
-- ********************************************************************
-- * Approximate the distance between you and a remote unit.          *
-- * Being very inaccurate, it should be used only as a way to        *
-- * discriminate melee from ranged positions.                        *
-- ********************************************************************
function Distance:GetApproxDistanceFrom(unit)
    local distance = 50000;

    if UnitIsUnit("player", unit) then
        distance = 0;

elseif UnitIsVisible(unit) then
        if CheckInteractDistance(unit, 3) then
            distance = 10;

    elseif CheckInteractDistance(unit, 4) then
            distance = 28;
        end
    end

    return distance;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: the remote unit to check the distance to.               *
-- ********************************************************************
-- * Check if you are roughly in melee range from specified unit.     *
-- ********************************************************************
function Distance:InMeleeOf(unit)
    return self:GetUnitDistance("player", unit) <= 10;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

Distance.OnBootComplete = function(self)
    -- Grab the DB table.
    self.DB = Save:Get("system", "zoneScale") or { };
    Save:Set("system", "zoneScale", self.DB);

    -- FIXME: temporary hack to make sure everyone has the same scale for Yor'sahj
    self.DB["DRAGONSOUL:3"] = 428;

    hooksecurefunc("SetMapToCurrentZone", hook_SetMapToCurrentZone);

    SetMapToCurrentZone();
end

Distance.OnZoneChange = function(self)
    if self.DB then
        SetMapToCurrentZone();
    end
end

Distance.OnMapUpdate = function(self)
    if self.DB then
        self:Refresh();
    end
end

Distance.OnUpdate = function(self, elapsed)
    if self.meterScale == 0 and self.currentID ~= "" then
        local x, y = self:GetUnitPosition("player");
        local speed = GetUnitSpeed("player");

        if self.moving then
            if x and y then
                local angle = self:Atanfull(self.startX, -self.startY, x, -y);
                local deviation = 0;
                local facingDeviation = 0;
                local totalTime = GetTime() - self.startTime;

                if totalTime > (DURATION_REQUIRED/2) then
                    if not self.direction then
                        self.direction = angle;
                    end
                    deviation = abs(Utility:GetAngleDistance(angle, self.direction));
                    facingDeviation = abs(Utility:GetAngleDistance(self.facing, self.direction));
                end

                if IsFlying() or IsFalling() or IsSwimming() then
                    self.moving = false;
                    System:Debug("Is falling/flying/swimming, aborting.");

            elseif self:GetMapDistance(self.lastX, self.lastY, x, y) < MINIMAL_MAP_MOVEMENT then
                    self.moving = false;
                    System:Debug("No movement on map, aborting.");

            elseif speed ~= 0 and abs(speed - self.speed) > SPEED_MOD_TOLERANCE then
                    self.moving = false;
                    System:Debug("Movement speed has changed too much, aborting.");

            elseif deviation > DEVIATION_TOLERANCE then
                    self.moving = false;
                    System:Debug("Movement direction has deviated too much, aborting.");

            elseif facingDeviation > FACING_TOLERANCE then
                    self.moving = false;
                    System:Debug("Movement direction is out of touch with player's facing, aborting.");

            elseif speed == 0 then
                    self.moving = false;
                    if totalTime > DURATION_REQUIRED then
                        local distance = totalTime * self.speed;
                        local mapDistance = self:GetMapDistance(self.startX, self.startY, x, y);
                        local scale = floor(distance / mapDistance + 0.5);

                        System:DebugFmt("Finished successfully movement approximation. Real distance: %.2f yds, map distance: %.4f, scale: %d", distance, mapDistance, scale);
                        System:DebugFmt("Set scale %d for zone [%s].", scale, self.currentID);

                        self.DB[self.currentID] = scale;
                        self.currentID = "";
                        self:Refresh();
                  else
                        System:Debug("Movement duration is too short.");
                    end
                end

                self.lastX, self.lastY = x, y;
           else
                System:Debug("Lost map coordinates.");
                self.moving = false;
            end

    elseif speed > 0 and x and y and (GetTime() - self.startTime) > DETERMINATION_COOLDOWN then
            self.startX, self.startY = x, y;
            self.lastX, self.lastY = x, y;
            self.speed = speed;
            self.direction = nil;
            self.facing = math.fmod(GetPlayerFacing() * 180 / math.pi + 90, 360);
            self.startTime = GetTime();
            self.moving = true;

            System:DebugFmt("Figuring out current zone scale. Facing: %d°, speed: %d yds/sec", self.facing, self.speed);
        end
    end
end