﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");

local EquipChecker = System:GetOrNewModule("EquipChecker");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Service that determinates if there are fishy (pun intended) items equipped on a player.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

EquipChecker.names = { };
EquipChecker.status = { };
EquipChecker.try = { };
EquipChecker.next = { };
EquipChecker.current = 0;

EquipChecker.emptySlots = { };
EquipChecker.badItems = { };
EquipChecker.text = "";

local relevantSlots = {"HEADSLOT", "NECKSLOT", "SHOULDERSLOT", "BACKSLOT", "CHESTSLOT", "WRISTSLOT", "HANDSSLOT", "WAISTSLOT", "LEGSSLOT", "FEETSLOT", "FINGER0SLOT", "FINGER1SLOT", "TRINKET0SLOT", "TRINKET1SLOT", "MAINHANDSLOT", "SECONDARYHANDSLOT"};

local WAIT_TIME = 1.50; -- Minimal wait time before the next new try on the unit.

local MAX_TRY = 10; -- Number of inspections on an unit, if some items are missing.

-- Possible status:
-- [1] = passed
-- [0] = unknown
-- [-1] = failed

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

local function AvgItemLevel(unit)
    local delta = UnitLevel("player") - UnitLevel(unit);
    if abs(delta) <= 5 then
        return GetAverageItemLevel() * (1 - 0.03 * delta);
  else
        return 0; -- Disable the equipment checker against lower level players.
    end
end

local function DoAudit(self, unit)
    wipe(self.emptySlots);
    wipe(self.badItems);
    self.text = "";

    local threshold = AvgItemLevel(unit) * 0.80;
    if threshold >= 50 then -- Don't bother for low level players.
        local isLocalPlayer = UnitIsUnit("player", unit);
        local playerName = UnitName(unit);
        local i;
        for i=1, #relevantSlots do
            local slot = relevantSlots[i];
            local slotId, textureName = GetInventorySlotInfo(slot);
            if slotId then
                local itemLink = GetInventoryItemLink(unit, slotId);
                if itemLink then
                    local _, _, itemRarity, itemLevel = GetItemInfo(itemLink);
                    if itemRarity ~= 7 and itemLevel then
                        if itemLevel < threshold then
                            tinsert(self.badItems, itemLink);
                        end
                    end
              else
                    if slot ~= "SECONDARYHANDSLOT" then -- TODO: refine for 1H
                        tinsert(self.emptySlots, slot);
                    end
                end
            end
        end
        if #self.emptySlots > 0 then
            if isLocalPlayer then
                self.text = L:Get("EquipChecker-SelfEmpty");
          else
                self.text = L:FormatLoc("EquipChecker-OtherEmpty", playerName, #self.emptySlots);
            end
    elseif #self.badItems > 0 then
            if isLocalPlayer then
                self.text = L:FormatLoc("EquipChecker-SelfBad", table.concat(self.badItems, ", "));
          else
                self.text = L:FormatLoc("EquipChecker-OtherBad", playerName, table.concat(self.badItems, ", "));
            end
        end
    end

    return (#self.emptySlots == 0 and #self.badItems == 0);
end

local function Abort(self, index)
    self.status[index] = 0;
    self.try[index] = 0;
    self.next[index] = 0;
    System:Print(L:FormatLoc("EquipChecker-Warning", self.names[index]));
end

local function Incomplete(self, index)
    self.status[index] = 0;
    self.try[index] = 0;
    self.next[index] = 0;
    System:Print(L:FormatLoc("EquipChecker-Incomplete", self.names[index]));
end

local function Validate(self, index)
    self.status[index] = 1;
    self.try[index] = 0;
    self.next[index] = 0;
    System:Print(self.names[index].." : OK");
end

local function Bad(self, index)
    self.status[index] = -1;
    self.try[index] = 0;
    self.next[index] = 0;
    System:Print(self:GetText());
end

local function Continue(self, index)
    self.try[index] = self.try[index] - 1;
    self.next[index] = GetTime() + WAIT_TIME;
end

local function TryAudit(self, current)
    local tryLeft = self.try[current];
    if tryLeft > 0 then
        local name = self.names[current];
        local unit = Unit:ResolveUID(name);
        if unit and CheckInteractDistance(unit, 1) and CanInspect(unit) then
            if InspectFrame then
                HideUIPanel(InspectFrame); -- TODO: should be uncommented once Blizzard writes code a little more careful for guild inspect.
                InspectFrame.unit = "player"; -- TODO: temporary workaround hack.
            end
            NotifyInspect(unit);

            DoAudit(self, unit);

            if #self.emptySlots > 0 then
                if tryLeft > 1 then
                    Continue(self, current);
              else
                    Incomplete(self, current);
                end
          else
                if #self.badItems > 0 then
                    Bad(self, current);
              else
                    Validate(self, current);
                end
            end
            return;
        end
    end
    -- If we reach this point, inspect failed.
    Abort(self, current);
end

local function PickNext(self)
    local index = nil;
    local soonest = nil;
    local i, next;
    for i=1, #self.next do
        if self.try[i] > 0 then
            next = self.next[i];
            if (not soonest) or (next < soonest) then
                index = i;
                soonest = next;
            end
        end
    end
    return index;
end

local function Finish(self)
    self.current = 0;

    local feedback = "";
    local i;
    for i=1, #self.names do
        if self.status[i] < 0 then
            -- Bad audit!
            if #feedback > 0 then
                feedback = feedback .. ", ";
            end
            feedback = feedback .. self.names[i];
        end
    end

    System:Print(L:Get("EquipChecker-End"));
    if #feedback > 0 then
        System:InvokeHandler("OnEquipCheckResult", L:FormatLoc("EquipChecker-OtherSummary", feedback));
    end
end

-- ********************************************************************
-- * Return true if you may perform an equipment check (audit) now.   *
-- ********************************************************************
function EquipChecker:CanAudit()
    return self.current == 0;
end

-- ********************************************************************
-- * Get the result text of the last audit.                           *
-- ********************************************************************
function EquipChecker:GetText()
    return self.text;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: the player unit to audit.                               *
-- * This method only allows "player".                                *
-- * See AuditAll to inspect other players' equipment.                *
-- ********************************************************************
-- * Check if the specified unit has the correct equipment.           *
-- * Return  1 if everything is alright.                              *
-- * Return  0 if we can't audit the unit (unknown situation).        *
-- * Return -1 if the audit is bad.                                   *
-- ********************************************************************
function EquipChecker:Audit(unit)
    unit = Unit:ResolveUID(unit);
    if unit and self:CanAudit() and UnitIsUnit(unit, "player") then
        if DoAudit(self, unit) then
            return 1;
      else
            return -1;
        end
  else
        return 0;
    end
end

-- ********************************************************************
-- * This will check the equipment of all raid members.               *
-- * It is asynchronous and results won't be available immediately.   *
-- * This method will print text to the console.                      *
-- * OnEquipCheckResult handler will be fired if something was found  *
-- * by the equip checker, passing the text as a parameter.           *
-- ********************************************************************
function EquipChecker:AuditAll()
--[[
    if self:CanAudit() then
        wipe(self.status);
        wipe(self.names);
        wipe(self.try);
        wipe(self.next);

        local now = GetTime();
        local i, baseUID;
        if IsInRaid() then
            baseUID = "raid";
      else
            baseUID = "party";
        end
        for i=1, GetNumGroupMembers() do
            if not UnitIsUnit("player", baseUID..i) then
                local name = UnitName(baseUID..i);
                tinsert(self.status, 0);
                tinsert(self.names, name);
                tinsert(self.try, MAX_TRY);
                tinsert(self.next, now + i * 0.05);
            end
        end

        if #self.status > 0 then
            System:Print(L:Get("EquipChecker-Begin"));
            self.current = 1;
      else
            System:Print(L:Get("EquipChecker-NoGroup"));
        end
    end
]]
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function EquipChecker:OnUpdate(elapsed)
    local current = self.current;
    if current > 0 and GetTime() > self.next[current] then
        TryAudit(self, current);

        local next = PickNext(self);
        if next then
            self.current = next;
      else
            Finish(self);
        end
    end
end