﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Table = Root:GetOrNewModule("Table");
local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");
local Social = Root:GetOrNewModule("Social");
local Utility = Root:GetOrNewModule("Utility");
local NPCName = Root:GetOrNewModule("NPCName");
local Sort = Root:GetOrNewModule("Sort");

local Combat = System:GetOrNewModule("Combat");

local Error = System:GetOrNewModule("Error");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Provides ways to memorize and detect errors of players during
-- the execution of a boss encounter.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

Error.running = false;
Error.begin = -1;
Error.final = -1;
Error.players = 0;
Error.current = { };
Error.definitions = { };
Error.order = { };

Error.spell = { };
Error.aggro = { };
Error.nextMobScan = 0;

local TOLERANCE_RESET = 5;

local SCAN_RATE = 3;

local MAX_OCCURENCE_STRING_LENGTH = 80;

-- DEBUG
--[[
Error.order = { "test" };
Error.definitions = { ["test"] = { type = "avoid-effect", spell = 11111, text = "toto", enabled = true } };
Error.current = { ["test"] = {
    ["Wayatt"]  = { occurences = {0, 1, 2}       },
    ["Cradot"]  = { occurences = {0, 1, 2, 3}    },
    ["Meaty"]   = { occurences = { }             },
    ["Lamelie"] = { occurences = {0}             },
    ["Khally"]  = { occurences = {0, 1, 2, 3, 4} },
} };
]]

-- Fun stuff
local function getReference(totalCounter)
    local count, unit, explain = 0, nil, nil;
--[[
    local a = totalCounter["Wayatt"] or 0;
    local b = totalCounter["Cradot"] or 0;
    if a > 0 and b > 0 then
        if a > b then
            count, unit, explain = b, "Watot", "Le Watot est une unité définie comme le rapport du nombre d'erreurs du joueur sur le nombre d'erreurs de Wayatt et Cradot mdrrr.";
      else
            count, unit, explain = a, "Crayatt", "Le Crayatt est une unité définie comme le rapport du nombre d'erreurs du joueur sur le nombre d'erreurs de Cradot et Wayatt.";
        end
elseif a > 0 then
        count, unit, explain = a, "Wayatt", "Le Wayatt est une unité définie comme le rapport du nombre d'erreurs du joueur sur le nombre d'erreurs de Wayatt mdrrr.";
elseif b > 0 then
        count, unit, explain = b, "Cradot", "Le Cradot est une unité définie comme le rapport du nombre d'erreurs du joueur sur le nombre d'erreurs de Cradot.";
    end
]]
    return count, unit, explain;
end


-- --------------------------------------------------------------------
-- **                           Methods                              **
-- --------------------------------------------------------------------

function Error:AccessPlayerData(id, player)
    local playerList = self.current[id];
    local playerData = playerList[player];
    if not playerData then
        playerData = Table:Alloc();
        playerData.occurences = { };
        playerData.triggerCount = 0;
        playerData.lastTrigger = 0;
        playerData.exempted = false;
        playerData.lethal = false;
        playerList[player] = playerData;
    end
    return playerData;
end

-- ********************************************************************
-- * Check if the error module is running.                            *
-- ********************************************************************
function Error:IsRunning()
    return self.running;
end

-- ********************************************************************
-- * Start the error system.                                          *
-- ********************************************************************
function Error:Start()
    if not self:IsRunning() then
        self.running = true;
        self.begin = GetTime();
        Unit:EnumerateRaid();
        self.players = Unit:GetEnumerationSize();
        Combat:RegisterCallback(self.OnCombatEvent, self);
    end
end

-- ********************************************************************
-- * Stop the error system.                                           *
-- ********************************************************************
function Error:Stop()
    if self:IsRunning() then
        self.running = false;
        self.final = GetTime();
        Combat:UnregisterCallback(self.OnCombatEvent, self);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- *                                                                  *
-- * >> type: the type of error.                                      *
-- *          Can be: "avoid-effect", "avoid-interrupt",              *
-- *                  "avoid-damage", "avoid-bomb",                   *
-- *                  "aggro", "interrupt", "custom".                 *
-- *                                                                  *
-- * >> argument: depends on the type of the error.                   *
-- * For "avoid-X": the spellID of the spell.                         *
-- * For "interrupt": the spellID of the spell to interrupt.          *
-- * For "aggro": the mobID of the creature. Won't trigger for tanks. *
-- * For "custom": the string explaining the error.                   *
-- *                                                                  *
-- * Note: avoid-bomb is a more restrictive version of avoid-damage.  *
-- *       avoid-bomb error will not trigger for the player that has  *
-- *       the spell debuff, but will trigger for other players.      *
-- *                                                                  *
-- * >> tolerance: how many times in a row the error may trigger      *
-- *               without being validated.                           *
-- *                                                                  *
-- * >> gracePeriod: period during which this error is discarded.     *
-- ********************************************************************
-- * Define an error.                                                 *
-- ********************************************************************
function Error:Define(id, type, argument, tolerance, gracePeriod)
    self:Erase(id);

    -- Add the definition
    local t = Table:Alloc();
    t.type = type;
    t.tolerance = tolerance or 0;
    t.gracePeriod = gracePeriod or 0;
    t.enabled = true;

    if type == "custom" then
        t.text = argument
  else
        t.text = L:Get("Error:"..type);
        if type == "aggro" then
            t.mob = argument;
            self:AddToAggroMonitor(t.mob, id);
      else
            t.spell = argument;
            self:AddToSpellMonitor(t.spell, type, id);
        end
    end
    self.definitions[id] = t;
    tinsert(self.order, id);

    -- Add the table to register the errors
    t = Table:Alloc();
    self.current[id] = t;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- ********************************************************************
-- * Erase the definition of an error.                                *
-- ********************************************************************
function Error:Erase(id)
    local info = self.definitions[id];
    if info then
        if info.type == "aggro" then
            self:RemoveFromAggroMonitor(info.mob);
      else
            self:RemoveFromSpellMonitor(info.spell);
        end
        self.definitions[id] = Table:Recycle(info);
        self.current[id] = Table:Recycle(self.current[id], true);

        local list = self.order;
        local i;
        for i=#list, 1, -1 do
            if list[i] == id then
                tremove(list, i);
            end
        end
    end
end

-- ********************************************************************
-- * Erase all definition of errors.                                  *
-- ********************************************************************
function Error:Clear()
    local id;
    for id in pairs(self.definitions) do
        self:Erase(id);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- * >> player: the player name who has done the error.               *
-- * >> lethal: true if the error is lethal.                          *
-- ********************************************************************
-- * Trigger an error in the current error list.                      *
-- * Player must be a valid player name for non-custom errors.        *
-- * Anything can be specified for custom errors.                     *
-- ********************************************************************
function Error:Trigger(id, player, lethal)
    local info = self.definitions[id];
    local timeElapsed = GetTime() - self.begin;
    local uid = Unit:GetUID(player);

    if info and info.enabled and timeElapsed >= info.gracePeriod and ((uid and UnitIsPlayer(uid)) or (info.type == "custom")) then
        local playerData = self:AccessPlayerData(id, player);
        if not playerData.exempted then
            if (GetTime() - playerData.lastTrigger) > TOLERANCE_RESET then
                playerData.triggerCount = 0;
            end
            playerData.triggerCount = playerData.triggerCount + 1;
            playerData.lastTrigger = GetTime();
            if lethal then playerData.lethal = true; end

            if playerData.triggerCount > info.tolerance or lethal then
                -- You're doing it wrong mate!
                tinsert(playerData.occurences, timeElapsed);
            end
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- * >> player: the name of the player who gets the exemption.        *
-- * >> state: the exemption state.                                   *
-- ********************************************************************
-- * Mark a player as being exempted from the specified error.        *
-- ********************************************************************
function Error:Exempt(id, player, state)
    local playerData = self:AccessPlayerData(id, player);
    playerData.exempted = state;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- ********************************************************************
-- * Allow an error to be triggered and validated.                    *
-- ********************************************************************
function Error:Enable(id)
    local info = self.definitions[id];
    if info then
        info.enabled = true;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the error.                                      *
-- ********************************************************************
-- * Disallow an error to be triggered or validated.                  *
-- ********************************************************************
function Error:Disable(id)
    local info = self.definitions[id];
    if info then
        info.enabled = false;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mobID: the mob ID to watch aggro of.                          *
-- * >> errorID: the error that is triggered upon gaining aggro.      *
-- ********************************************************************
-- * Register a mob for which we'll monitor the aggro.                *
-- ********************************************************************
function Error:AddToAggroMonitor(mobID, errorID)
    local t = Table:Alloc();
    t.id = mobID;
    t.guid = nil;
    t.current = nil;
    t.error = errorID;
    tinsert(self.aggro, t);
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> mobID: the mob ID that is no longer observed.                 *
-- ********************************************************************
-- * Unregister a mob from the aggro monitor.                         *
-- ********************************************************************
function Error:RemoveFromAggroMonitor(mobID)
    local list = self.aggro;
    local i;
    for i=#list, 1, -1 do
        if list[i].id == mobID then
            tremove(list, i);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> spellID: the spell ID that is being monitored.                *
-- * >> mode: the mode of detection of the spell.                     *
-- * >> errorID: the error that is triggered upon gaining aggro.      *
-- ********************************************************************
-- * Register a spell to pay attention to.                            *
-- ********************************************************************
function Error:AddToSpellMonitor(spellID, mode, errorID)
    local t = Table:Alloc();
    t.mode = mode;
    t.error = errorID;
    self.spell[spellID] = t;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> spellID: the spell ID that is no longer observed.             *
-- ********************************************************************
-- * Register a spell to pay attention to.                            *
-- ********************************************************************
function Error:RemoveFromSpellMonitor(spellID)
    local t = self.spell[spellID];
    if t then
        self.spell[spellID] = Table:Recycle(t);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> destination: where the output is sent.                        *
-- ********************************************************************
-- * Display the list of errors to the chat.                          *
-- * Destination is one of valid values of "where" of Output method.  *
-- ********************************************************************
function Error:PrintToChat(destination)
    Social:Output(destination, L:Get("ErrorTitle"));

    local totalErrors = 0;
    local totalCounter = Table:Alloc();
    local ids = self.order;
    local i, id, errorDef, playerList, player, playerData, occurences;
    for i=1, #ids do
        id = ids[i];
        errorDef = self.definitions[id];

        playerList = self.current[id];
        for player, playerData in pairs(playerList) do
            occurences = playerData.occurences;

            totalErrors = totalErrors + #occurences;
            totalCounter[player] = (totalCounter[player] or 0) + #occurences;

            if #occurences > 0 then
                local token;
                if errorDef.mob then
                    token = "["..(NPCName:Get(errorDef.mob) or "?").."]";
            elseif errorDef.spell then
                    token = GetSpellLink(errorDef.spell) or "[?]";
              else
                    token = "";
                end
                if playerData.lethal then
                    token = token.." "..L:Get("ErrorFatalTag");
                end

                local timeString = "";
                local j, timeElapsed;
                for j=1, #occurences do
                    timeElapsed = occurences[j];
                    if #timeString < MAX_OCCURENCE_STRING_LENGTH then
                        if #timeString > 0 then
                            timeString = timeString..", ";
                        end
                        timeString = timeString..Utility:FormatCountdown("%M:%S", timeElapsed);
                  else
                        break;
                    end
                end

                Social:Output(destination, ("►%s %s %s (x%d) → %s"):format(player, errorDef.text, token, #occurences, timeString));
            end
        end
    end

    if not self.running then
        local totalMinutes = (self.final - self.begin + 0.00001) / 60;
        local numPlayers = max(1, self.players);
        Social:Output(destination, L:FormatLoc("ErrorTotal", totalErrors, totalErrors / (totalMinutes * numPlayers)));

        local referenceCounter, unit, explain = getReference(totalCounter);
        if referenceCounter > 0 then
            local list = { };
            for player, count in pairs(totalCounter) do
                tinsert(list, {player = player, count = count / referenceCounter});
            end
            Sort:NumericField(list, "count", true);
            Social:Output(destination, ("----- Récapitulatif en %s -----"):format(unit));
            Social:Output(destination, explain);
            for i=1, #list do
                Social:Output(destination, ("× %s : %.2f %s"):format(list[i].player, list[i].count, unit));
            end
        end
    end

    Table:Recycle(totalCounter);
end

-- ********************************************************************
-- * Check if there is one person assigned to tanking that is dead.   *
-- ********************************************************************
function Error:IsOneTankDead()
    local result = false;
    local i, uid;
    Unit:EnumerateRaid();
    for i=1, Unit:GetEnumerationSize() do
        uid = Unit:GetEnumerationUID(i);
        if Unit:GetRole(uid) == "TANK" and UnitIsDeadOrGhost(uid) then
            result = true;
            break;
        end
    end
    return result;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

Error.OnCombatEvent = {
    ["DAMAGE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, isPeriodic, amount, overkill, school, absorbed)
        local info = self.spell[spellId];
        if info and (info.mode == "avoid-damage" or info.mode == "avoid-bomb") then
            self:Trigger(info.error, targetName, overkill > 0);
        end
    end,

    ["EFFECT_GAIN"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
        local info = self.spell[spellId];
        if info then
            if info.mode == "avoid-effect" then
                self:Trigger(info.error, targetName);

        elseif info.mode == "avoid-bomb" then
                self:Exempt(info.error, targetName, true);
            end
        end
    end,

    ["EFFECT_FADE"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName, stackAmount)
        local info = self.spell[spellId];
        if info and info.mode == "avoid-bomb" then
            self:Exempt(info.error, targetName, false);
        end
    end,

    ["INTERRUPT"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        local info = self.spell[spellId];
        if info and info.mode == "avoid-interrupt" then
            self:Trigger(info.error, targetName);
        end
    end,
};

function Error:OnUpdate(elapsed)
    local list = self.aggro;
    local now = GetTime();
    local i, entry;
    if now > self.nextMobScan then
        self.nextMobScan = now + SCAN_RATE;
        for i=1, #list do
            entry = list[i];
            if not entry.guid then
                entry.guid = Unit:GetGUIDFromMobID(entry.id, false);
            end
        end
    end
    if self:IsRunning() then
        for i=1, #list do
            entry = list[i];
            if entry.guid then
                local _, tankName = Unit:GetTankOf(entry.guid);
                if tankName and tankName ~= entry.current then
                    entry.current = tankName;
                    self:OnAggro(tankName, entry.error);
                end
            end
        end
    end
end

function Error:OnAggro(player, error)
    if not self:IsOneTankDead() and Unit:GetRole(player) ~= "TANK" then
        self:Trigger(error, player);
    end
end