﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Network = Root:GetOrNewModule("Network");
local Sort = Root:GetOrNewModule("Sort");
local Version = Root:GetOrNewModule("Version");
local Authorization = Root:GetOrNewModule("Authorization");

local Settings = System:GetOrNewModule("Settings");

local SettingPriority = System:GetOrNewModule("SettingPriority");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Determinates who has priority on a given exclusive setting.
-- An exclusive setting state is determinated by only 1 people
-- in the raid for the whole raid.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

local settingRankScore = {
    ["NORMAL"] = 0,
    ["ASSIST"] = 10,
    ["LEADER"] = 50,
    ["ADMIN"] = 99,
};

SettingPriority.list = { };
SettingPriority.lastBroadcast = nil;
SettingPriority.nextBroadcast = nil;

local BROADCAST_DELAY = 1.00; -- Min interval between broadcasts.

local function serializeValue(v)
    if v == true then
        return "1";
elseif v == false or v == nil then
        return "0";
  else
        return tostring(v);
    end
end

local function deserializeValue(v)
    if v == "1" then
        return true;
elseif v == "0" then
        return false;
  else
        return tonumber(v) or tostring(v);
    end
end

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Reset priority list. Should be called on a refresh.              *
-- ********************************************************************
function SettingPriority:Reset()
    wipe(self.list);
    self:Broadcast();
    self:Query();
end

-- ********************************************************************
-- * Ask the raid to send the state of their exclusive settings.      *
-- * This will allow us to maintain the priority list.                *
-- ********************************************************************
function SettingPriority:Query()
    Network:Send("SETTING_QUERY", "GROUP");
end

-- ********************************************************************
-- * Tell the raid of the state of our exclusive settings.            *
-- * This will allow them to maintain their priority list.            *
-- ********************************************************************
function SettingPriority:Broadcast()
    if self.lastBroadcast and (GetTime() - self.lastBroadcast) < BROADCAST_DELAY then
        self.nextBroadcast = self.lastBroadcast + BROADCAST_DELAY;
  else
        -- Past this point, we execute the broadcasting.
        self.lastBroadcast = GetTime();

        local i, id, value;
        for i=1, Settings:GetNum() do
            id = Settings:Get(i);
            if Settings:IsExclusive(id) then
                -- This setting is exclusive. Broadcast it.
                value = Settings:ReadReal(id);
                Network:Send("SETTING_NOTIFY", "GROUP", id, serializeValue(value), Version:GetFullInteger());
            end
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> name: the name of the player.                                 *
-- * >> id: the ID of the setting that changed.                       *
-- * >> value: its new value.                                         *
-- * >> version: version of the player (full integer).                *
-- ********************************************************************
-- * Acknowledge a change in a setting's value for a given player.    *
-- * Will also make sure this player is in our priority list.         *
-- ********************************************************************
function SettingPriority:Change(name, id, value, version)
    if Settings:IsExclusive(id) and value and version then
        local found = false;
        local i, t;
        for i=1, #self.list do
            t = self.list[i];
            if t.name == name then
                found = true;
                break;
            end
        end
        if not found then
            t = { name = name, version = version, score = 0 };
            tinsert(self.list, t);
        end
        t[id] = value;
    end
end

-- ********************************************************************
-- * Recalculate the list order, based on the rank and the            *
-- * version each member of the raid has.                             *
-- ********************************************************************
function SettingPriority:Arrange()
    local i, t, auth;
    for i=1, #self.list do
        t = self.list[i];
        if Network:IsAdmin(t.name) then
            auth = "ADMIN";
      else
            auth = Authorization:GetRank(t.name);
        end
        t.score = 100 * t.version + settingRankScore[auth];
    end
    Sort:Name(self.list);
    Sort:NumericField(self.list, "score", true); -- Highest score people first.
end

-- ********************************************************************
-- * SettingPriority:Resolve(id)                                      *
-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the setting that is checked.                    *
-- ********************************************************************
-- * Determinate who has the final word over a setting and its value. *
-- * Return decider, value.                                           *
-- ********************************************************************
function SettingPriority:Resolve(id)
    if Settings:IsExclusive(id) then
        self:Arrange();
        local i, t;
        for i=1, #self.list do
            t = self.list[i];
            if t[id] ~= nil and Settings:HasPermission(id, t.name) then
                break; -- Gotcha!
          else
                t = nil;
            end
        end
        if t then
            return t.name, t[id];
        end
    end
    -- Well, past this point we have either an incorrect setting or incomplete list.
    local me = UnitName("player");
    return me, Settings:ReadReal(id);
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

SettingPriority.OnBootComplete = function(self)
    -- Register net commands.
    Network:Define("SETTING_QUERY", "S", "NORMAL", false);
    Network:Define("SETTING_NOTIFY", "s", "NORMAL", true);
end

SettingPriority.OnNetCommand = function(self, command, sender, ...)
    if command == "SETTING_NOTIFY" then
        local id, value, version = ...
        self:Change(sender, id, deserializeValue(value), version);

elseif command == "SETTING_QUERY" then
        self:Broadcast();
    end
end

SettingPriority.OnUpdate = function(self, elapsed)
    if self.nextBroadcast and (GetTime() > self.nextBroadcast) then
        self.nextBroadcast = nil;
        self:Broadcast();
    end
end

SettingPriority.Debug = function(self)
    System:Debug("-- SETTING PRIORITY LIST:");
    local i, t, k, v;
    for i=1, #self.list do
        t = self.list[i];
        System:DebugFmt("%d. %s (ver %d) = %d", i, t.name, t.version, t.score);
        for k, v in pairs(t) do
            if k ~= "name" and k ~= "version" then
                System:Debug(k.."="..tostring(v or false));
            end
        end
    end
end

SettingPriority.OnSettingUpdate = function(self, id, value)
    self:Broadcast();
end

SettingPriority.OnSettingsRefresh = function(self)
    self:Reset();
end