﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Save = Root:GetOrNewModule("Save");
local Social = Root:GetOrNewModule("Social");
local Sort = Root:GetOrNewModule("Sort");
local Unit = Root:GetOrNewModule("Unit");
local L = Root:GetOrNewModule("Localization");

local Combat = System:GetOrNewModule("Combat");

local Stats = System:GetOrNewModule("Stats");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Provides ways to memorize the stats of the various boss encounters
-- and to pull the data when needed.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

local currentInfo = {
    name = nil,
    attempts = 0,
    attempt = 0,
    speed = 0,
    technique = 0,
    clearTime = nil,
    previousTime = nil,
    previousScore = nil,
    numKilled = 0,
    sequence = "NORMAL",
    dps = { },
};

-- FOR DEBUG
--[[
local currentInfo = {
    name = "Test",
    attempts = 3,
    attempt = 100,
    speed = 100,
    technique = 105,
    clearTime = 10,
    previousTime = 5,
    previousScore = 150,
    numKilled = 0,
    sequence = "FULL",
    dps = {
        coupDeGrace = 2,
    },
};
do
    -- Generate a test sample
    local i, dps, sum, maximum;
    sum, maximum = 0, 0;
    for i=1, 20 do
        dps = 15000 / ((i-1)/8 + 1);
        sum = sum + dps;
        tinsert(currentInfo.dps, {guid = i, name = "Guy "..i, dps = dps, totalDps = dps, pet = 0, class = "DRUID", special = nil});
        maximum = math.max(maximum, dps);
    end
    currentInfo.dps.average = sum / #currentInfo.dps;
    currentInfo.dps.total = sum;
    currentInfo.dps.maximum = maximum;
end
]]

local tryCountScore = {
    [1] = 120,
    [2] = 100,
    [3] =  75,
    [4] =  50,
    [5] =  25,
};

local EMPTY_TABLE = { };

-- --------------------------------------------------------------------
-- **                          Constants                             **
-- --------------------------------------------------------------------

-- After this amount of seconds has passed since the last boss attempt,
-- the attempt counter for this boss will be reset.

local ATTEMPT_COUNT_EXPIRE_TIMER = 21600;

-- --------------------------------------------------------------------
-- **                           Methods                              **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the identifier of the boss fight.                         *
-- * This ID depends on the boss module itself.                       *
-- ********************************************************************
-- * Get history data about a given boss encounter.                   *
-- * These informations are character-specific.                       *
-- * Return:                                                          *
-- *    name, numAttempts, numKilled, bestScore, bestTime             *
-- *                                                                  *
-- * name: the name of the encounter, usually the main boss name.     *
-- * numAttempts: the current attempt number. 1 if it is the first    *
-- * time the boss is fought. The attempt number will reset after     *
-- * enough time has passed since last attempt.                       *
-- * numKilled: the number of times the boss has been killed.         *
-- * bestScore, bestTime: best score/time data for this boss.         *
-- * bestScore and bestTime will be nil if the boss was never killed. *
-- ********************************************************************
function Stats:GetInfo(id)
    local t = self:Access(id, false);
    if t then
        local numAttempts = self:GetAttemptCount(t) + 1;
        return t.name, numAttempts, t.kills, t.score, t.combatTime;
  else
        return UNKNOWN, 1, 0, nil, nil;
    end
end

-- ********************************************************************
-- * Get infos about the boss encounter that has just been cleared.   *
-- * It will provide comparison data with the previous record for     *
-- * that encounter.                                                  *
-- * Return:                                                          *
-- *    name, attempts, attemptScore, speedScore, techniqueScore,     *
-- * clearTime, previousTime, previousScore, numKilled, sequence, dps *
-- *                                                                  *
-- * name: the name of the encounter, usually the main boss name.     *
-- * attempts: the attempts used to defeat the encounter.             *
-- * attempt, score, technique: the various score criterias.          *
-- * clearTime, previousTime: the time the encounter lasted.          *
-- * previousScore: the previous best score.                          *
-- * numKilled: the number of times the boss has been killed.         *
-- * sequence: the clear sequence that should be played.              *
-- * dps: the table containing DPS data for members of the raid.      *
-- *                                                                  *
-- * Note that clearTime will be nil if the fight was incomplete.     *
-- * previousTime and previousScore will be nil if no record exists.  *
-- ********************************************************************
function Stats:GetCurrentInfo()
    local t = currentInfo;
    return t.name, t.attempts, t.attempt, t.speed, t.technique, t.clearTime, t.previousTime, t.previousScore, t.numKilled, t.sequence, t.dps;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the boss module that announces proudly that the boss  *
-- *            has been killed.                                      *
-- ********************************************************************
-- * Acknowledges that the player has killed a boss.                  *
-- * This method will directly query data from the boss module.       *
-- * Once this method is called, the GetCurrentInfo method will be    *
-- * available to get the score of the boss that has been killed.     *
-- ********************************************************************
function Stats:BossCleared(module)
    local M = module;
    local id, displaySuffix = M:GetID();

    local t = self:Access(id, true);
    if ( not t ) then return; end

    -- PHASE 1 - Prepare data for processing

    -- Load previous history for this boss, if any.
    local _, numAttempts, numKilled, prevScore, prevTime = self:GetInfo(id);

    local battleName = M:GetTitle()..displaySuffix;
    local clearSequence = M:GetEndSequence();
    local battleDuration, remainingPlayers, manaRatio = M:GetFinalStats();
    local poorTime, greatTime = M:GetScoreBenchmarks();
    local startingPlayers = M:GetStartingPlayers();
    local maxPlayers = M:GetNumSlots();
    local surviveRatio = remainingPlayers / startingPlayers;
    local incomplete = M:IsIncomplete();

    -- PHASE 2 - Calculate each criteria and the final score.

    local attempt, speed, technique;
    local totalPoints;

    -- Attempt score calculation:
    -- * 120 points on the first try.
    -- * 100 points on the second try.
    -- *  75 points on 3rd.
    -- *  50 points on 4th.
    -- *  25 points on 5th.
    -- *   0 point  on 6th+.

    attempt = tryCountScore[numAttempts] or 0;

    -- Speed calculation:
    -- * 50 points if the player has cleared the boss in the expected time limit.
    -- * The result is between 0~120 points.

    if incomplete or (not poorTime) or (not greatTime) then
        speed = 0;
  else
        if battleDuration >= poorTime then
            speed = 0;
      else
            speed = (poorTime - battleDuration) / (poorTime - greatTime) * 100;
            speed = min(120, max(0, speed));
            speed = floor(speed + 0.5);
        end
    end

    -- Technique calculation:
    -- * Up to 75 pts based on the % of people that survived.
    -- * Up to 25 pts based on the % of mana remaining to the raid.
    -- * Bonus granted for each player slot unused below the maximal number of players (24 players out of 25 => +5 pts, 9 players out of 10 => +12.5 pts).

    local survivalPoints = (exp(10/3 * (surviveRatio-1)) - 0.05 * (1-surviveRatio)) * 75;
    local manaPoints     = min(1, manaRatio * 1.33) * 25; -- If the raid has up to 3/4 mana left, we still consider it as full mana.
    local bonusPoints    = (maxPlayers - startingPlayers) / maxPlayers * 125;

    technique = survivalPoints + manaPoints + bonusPoints;
    technique = min(120, max(0, technique));
    technique = floor(technique + 0.5);

    -- Final score calculation:
    -- * Simply the sum of all three criterias.
    -- * It is between 0~360.

    totalPoints = attempt + speed + technique;

    -- PHASE 3 - Update data for this boss encounter.

    if not prevScore or prevScore < totalPoints then
        t.score = totalPoints;
    end
    if not incomplete and ((not prevTime) or (prevTime > battleDuration)) then
        t.combatTime = battleDuration;
    end
    t.name = battleName;
    t.kills = numKilled + 1;
    t.attempts = 0;
    t.lastTry = 0;
    t.lastClear = time();

    -- PHASE 4 - Set current info table, if the sequence hasn't been skipped.

    local scoreEnabled = M:IsScoreEnabled();
    local damageMeterEnabled = M:IsDamageMeterEnabled();
    local sequenceEnabled = M:IsEndSequenceEnabled();
    local enabled = (scoreEnabled or damageMeterEnabled) and sequenceEnabled;

    if enabled then
        currentInfo.name = battleName;
        currentInfo.attempts = numAttempts;
        currentInfo.previousTime = prevTime;
        currentInfo.previousScore = prevScore;
        currentInfo.numKilled = numKilled + 1;
        currentInfo.sequence = clearSequence;

        -- Build clear time
        if (not incomplete) then
            currentInfo.clearTime = battleDuration;
      else
            currentInfo.clearTime = nil;
            Root:Print(L:Get("IncompleteFightExplain"));
        end

        -- Build score
        if scoreEnabled then
            currentInfo.attempt = attempt;
            currentInfo.speed = speed;
            currentInfo.technique = technique;
      else
            currentInfo.attempt = nil;
            currentInfo.speed = nil;
            currentInfo.technique = nil;
        end

        -- Build damage table
        if damageMeterEnabled then
            currentInfo.dps = self:CreateDpsTable(M:GetDamageData(), battleDuration, M:GetCoupDeGraceDealer());
      else
            currentInfo.dps = nil;
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> module: the boss module that announces shamefully that the    *
-- *            boss has been PHAILED.                                *
-- ********************************************************************
-- * Acknowledges that the player pathetically wiped on a boss.       *
-- * Be careful not to call this when it is not a wipe.               *
-- ********************************************************************
function Stats:BossFailed(module)
    local M = module;
    local id = M:GetID();

    local t = self:Access(id, true);
    if ( not t ) then return; end

    t.attempts = self:GetAttemptCount(t) + 1;
    t.lastTry = time();
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the boss fight whose data is accessed.          *
-- * >> writeMode: set this to create a table if no data currently    *
-- *               exists for this boss fight.                        *
-- ********************************************************************
-- * Access the data table of a boss fight.                           *
-- * Will return the data table or nil.                               *
-- ********************************************************************
function Stats:Access(id, writeMode)
    if ( not self.DB ) then return nil; end

    local t = self.DB[id];
    if ( not t ) and ( writeMode ) then
        t = {
            name = UNKNOWN, -- Populate the new entry with default values.
            attempts = 0,
            kills = 0,
            score = nil,
            combatTime = nil,
            lastClear = 0,
            lastTry = 0,
        };
        self.DB[id] = t;
    end
    return t or nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> dataTable: a boss fight data table.                           *
-- ********************************************************************
-- * Get the real attempt number, taking in account the expire timer. *
-- ********************************************************************
function Stats:GetAttemptCount(dataTable)
    local elapsed = time() - dataTable.lastTry;
    if ( dataTable.lastTry == 0 or elapsed > ATTEMPT_COUNT_EXPIRE_TIMER ) then
        return 0;
  else
        return dataTable.attempts;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> damageTable: the damage done table of the module.             *
-- * >> battleDuration: the duration of the battle.                   *
-- * >> coupDeGrace: who landed the Coup de Grace.                    *
-- ********************************************************************
-- * Generate a DPS table from a raw damage done table.               *
-- ********************************************************************
function Stats:CreateDpsTable(damageTable, battleDuration, coupDeGrace)
    local sourceTable = { };      -- Table containing relevant damage per source.
    local sourceTotalTable = { }; -- Table containing total damage per source.
    local sourcePetTable = { };   -- Table listing the pet contribution per source.
    local dpsTable = { };         -- Final output as DPS

    if battleDuration > 0 then
        local target, targetTable, source, damage;
        for target, targetTable in pairs(damageTable) do
            -- Note: total damage is relevant + pending + irrelevant damage.
            for source, damage in pairs(targetTable["RELEVANT"] or EMPTY_TABLE) do
                sourceTable[source] = (sourceTable[source] or 0) + damage;
                sourceTotalTable[source] = (sourceTotalTable[source] or 0) + damage;
            end
            for source, damage in pairs(targetTable["PENDING"] or EMPTY_TABLE) do
                sourceTotalTable[source] = (sourceTotalTable[source] or 0) + damage;
            end
            for source, damage in pairs(targetTable["IRRELEVANT"] or EMPTY_TABLE) do
                sourceTotalTable[source] = (sourceTotalTable[source] or 0) + damage;
            end
        end

        -- We have now the relevant & total damage by each source.
        -- Determinate now which sources are part of the raid.
        -- Handle pets, etc.
        local isPartner, unitType, name, parameter; -- parameter is "class" if the unit is a player, "master" if it is a pet and has no meaning otherwise.
        for source, damage in pairs(sourceTable) do
            isPartner, unitType, name, parameter = Unit:GetAffiliation(source);
            if isPartner then
                if unitType == "PET" and parameter then
                    sourcePetTable[parameter] = (sourcePetTable[parameter] or 0) + damage;

            elseif unitType == "PLAYER" then
                    tinsert(dpsTable, {guid = source, name = name or "?", class = parameter or nil});
                end
            end
        end

        -- Finally, compile the final DPS done after taking pets in account.
        local info, thisDps, thisTotalDps;
        local sum, total, maximum = 0, 0, 0;
        for _, info in ipairs(dpsTable) do
            -- The useful dps from the pet is computed beforehand.
            if sourcePetTable[info.guid] then
                info.pet = sourcePetTable[info.guid] / battleDuration;
          else
                info.pet = 0;
            end
            thisDps = sourceTable[info.guid] / battleDuration + info.pet;
            thisTotalDps = sourceTotalTable[info.guid] / battleDuration + info.pet;
            info.dps = thisDps;
            info.totalDps = thisTotalDps;
            sum = sum + thisDps;
            total = total + 1;
            maximum = max(maximum, thisDps);
        end

        -- Remove false damage dealers (that is to say, healers in most cases...).
        local threshold = sum / (3 * max(1, total));
        local i;
        for i=#dpsTable, 1, -1 do
            if dpsTable[i].dps < threshold then
                tremove(dpsTable, i);
            end
        end

        -- Recompute stats after false damage dealers have been removed.
        sum, total = 0, 0;
        for _, info in ipairs(dpsTable) do
            sum = sum + info.dps;
            total = total + 1;
        end
        if total > 0 then
            dpsTable.average = sum / total;
      else
            dpsTable.average = 0;
        end
        dpsTable.maximum = maximum;
        dpsTable.total = sum;

        -- Sort the table
        Sort:NumericField(dpsTable, "dps", true);
  else
        dpsTable.average = 0;
        dpsTable.maximum = 0;
        dpsTable.total = 0;
    end

    dpsTable.coupDeGrace = coupDeGrace;

    return dpsTable;
end

-- ********************************************************************
-- * Stats:PrintDpsToChat(dpsTable, destination)                      *
-- ********************************************************************
-- * Arguments:                                                       *
-- * >> dpsTable: the DPS table being printed.                        *
-- * >> destination: where the table should be printed.               *
-- ********************************************************************
-- * Print a DPS table on the chat.                                   *
-- * Destination is one of valid values of "where" of Output method.  *
-- ********************************************************************
function Stats:PrintDpsToChat(dpsTable, destination)
    Social:Output(destination, L:Get("DamageMeter-Title"));

    local i, usefulDps, totalDps, petDps, pct, wastedDps, relevancePct, petString;
    for i=1, #dpsTable do
        usefulDps = dpsTable[i].dps;
        totalDps  = dpsTable[i].totalDps;
        petDps    = dpsTable[i].pet;

        pct = usefulDps * 100 / dpsTable.total;
        wastedDps = max(0, totalDps - usefulDps);
        relevancePct = usefulDps * 100 / totalDps;

        if petDps > 0 then
            petString = L:FormatLoc("DamageMeter-Pet", petDps);
      else
            petString = "";
        end

        Social:Output(destination, L:FormatLoc("DamageMeter-Entry", i, dpsTable[i].name, usefulDps, petString, pct, wastedDps, relevancePct + 0.5));
    end

    Social:Output(destination, L:FormatLoc("DamageMeter-Average", dpsTable.average));
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

Stats.OnBootComplete = function(self)
    local guid = UnitGUID("player");

    -- Grab the DB table.
    self.DB = Save:Get("system", "stats:"..guid) or { };
    Save:Set("system", "stats:"..guid, self.DB);

    -- Listen to SUMMON events to figure out the owners of pets.
    Combat:RegisterCallback(self.OnCombatEvent, self);
end


Stats.OnCombatEvent = {
    ["SUMMON"] = function(self, actorGUID, actorName, targetGUID, targetName, spellId, spellName)
        if actorGUID and targetGUID and targetName then
            Unit:SetPetAffiliation(targetGUID, targetName, actorGUID);
        end
    end,
};