﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local Unit = Root:GetOrNewModule("Unit");
local Table = Root:GetOrNewModule("Table");
local Authorization = Root:GetOrNewModule("Authorization");

local Symbol = System:GetOrNewModule("Symbol");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Service that provides smart symbol services in combat.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

local userSymbols = { };
local modSymbols = { };

Symbol.nextExpireCheck = 0;

local EXPIRE_CHECK_RATE = 0.3;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Check we can manipulate symbols.                                 *
-- ********************************************************************
function Symbol:HasPermission()
    return Authorization:Check("SYMBOL");
end

-- ********************************************************************
-- * This method will make sure symbols that are currently on players *
-- * will be memorized as persistent symbols.                         *
-- ********************************************************************
function Symbol:LockUserSymbols()
    wipe(userSymbols);

    local i, guid, name, uid, symbol;
    for i=1, Unit:GetNumUID() do
        guid, name, uid = Unit:GetUID(i);
        if uid then
            symbol = GetRaidTargetIndex(uid) or 0;
            if symbol > 0 then
                userSymbols[guid] = symbol;
            end
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit queried.                           *
-- ********************************************************************
-- * Get the symbol users have defined for specified unit before      *
-- * starting the fight.                                              *
-- ********************************************************************
function Symbol:GetUserSymbol(guid)
    return userSymbols[guid] or 0;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit queried.                           *
-- ********************************************************************
-- * Get the symbol the mod has defined for specified unit.           *
-- ********************************************************************
function Symbol:GetModSymbol(guid)
    local info = modSymbols[guid];
    if info and GetTime() < info.expire then
         return info.symbol;
    end
    return 0;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit set.                               *
-- * >> symbol: the symbol set.                                       *
-- * >> duration: display duration of the symbol.                     *
-- ********************************************************************
-- * Set the mod-defined symbol for specified unit.                   *
-- ********************************************************************
function Symbol:SetModSymbol(guid, symbol, duration)
    local newInfo = Table:Alloc();
    newInfo.symbol = symbol;
    newInfo.expire = GetTime() + duration;
    modSymbols[guid] = newInfo;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit affected.                          *
-- ********************************************************************
-- * Flag the symbol of an unit has expired.                          *
-- ********************************************************************
function Symbol:RemoveModSymbol(guid)
    local info = modSymbols[guid];
    if info then
        info.expire = -1;
        self.nextExpireCheck = 0;
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: name/guid/uid of the unit that gets the symbol.         *
-- * >> symbol: the symbol index.                                     *
-- * >> timer: duration of the symbol display (not required).         *
-- ********************************************************************
-- * This method will place a symbol on a unit.                       *
-- * A display timer can be specified.                                *
-- ********************************************************************
function Symbol:PlaceSymbol(unit, symbol, timer)
    local unit = Unit:ResolveUID(unit);
    if unit then
        local guid = UnitGUID(unit);
        if self:ExecuteSymbol(unit, symbol) then
            self:SetModSymbol(guid, symbol, timer or math.huge);
        end
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: name/guid/uid of the unit which gets cleared.           *
-- ********************************************************************
-- * This method will clear symbol from an unit.                      *
-- ********************************************************************
function Symbol:ClearSymbol(unit)
    local unit = Unit:ResolveUID(unit);
    if unit then
        local guid = UnitGUID(unit);
        self:RemoveModSymbol(guid);
    end
end

-- ********************************************************************
-- * Clear all symbols that we have defined.                          *
-- ********************************************************************
function Symbol:ClearAll()
    local guid;
    for guid in pairs(modSymbols) do
        self:RemoveModSymbol(guid);
    end
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> unit: name/guid/uid of the unit that gets the symbol.         *
-- * >> symbol: the symbol index.                                     *
-- ********************************************************************
-- * Try to place a symbol on the unit.                               *
-- ********************************************************************
function Symbol:ExecuteSymbol(unit, symbol)
    if symbol and self:HasPermission() then
        local uid = Unit:ResolveUID(unit);
        if uid then
            if GetRaidTargetIndex(uid) ~= symbol then
                SetRaidTarget(uid, symbol);
            end
            return true;
        end
    end
    return false;
end

-- ********************************************************************
-- * Cancel expired symbols and place back user symbols.              *
-- ********************************************************************
function Symbol:CheckExpired()
    local now = GetTime();
    local guid, info;
    for guid, info in pairs(modSymbols) do
        if now > info.expire then
            self:ExecuteSymbol(guid, self:GetUserSymbol(guid));
            modSymbols[guid] = Table:Recycle(modSymbols[guid]);
        end
    end
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

Symbol.OnUpdate = function(self, elapsed)
    if GetTime() > self.nextExpireCheck then
        self.nextExpireCheck = GetTime() + EXPIRE_CHECK_RATE;
        self:CheckExpired();
    end
end
