﻿-- --------------------------------------------------------------------
-- ////////////////////////////////////////////////////////////////////
-- --                          SYSTEM PART                           --
-- \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
-- --------------------------------------------------------------------

local Root = BossEncounter3;

local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");

local NPCInfo = Root:GetOrNewModule("NPCInfo");
local ZoneInfo = Root:GetOrNewModule("ZoneInfo");

local Module = Root:GetOrNewModule("Module");

local System = Root:NewSubCore("System");

System.printPrefix = "|cff00ff00BE3 System|r - ";

System.nextZoneCheck = 0;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> guid: the GUID of the unit checked.                           *
-- ********************************************************************
-- * Return the number of raid / party members that target the unit.  *
-- * Return percentage, count, outOf.                                 *
-- ********************************************************************
function System:GetNumSelections(guid)
    if guid then
        -- Get the count
        Unit:EnumerateRaid();
        local outOf = Unit:GetEnumerationSize();
        local count = 0;
        local i, uid, targetUID;
        for i=1, outOf do
            uid = Unit:GetEnumerationUID(i);
            if UnitIsUnit("player", uid) then
                targetUID = "target";
          else
                targetUID = uid.."target";
            end
            if UnitGUID(targetUID) == guid then
                count = count + 1;
            end
        end
        return count / outOf, count, outOf;
    end
    return 0, 0, 0;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> uid: the UID of the unit checked.                             *
-- ********************************************************************
-- * Determinate if the given unit is a valid boss.                   *
-- * Return its boss module name if it is as well as volatile flag.   *
-- * Return nil if not.                                               *
-- ********************************************************************
function System:IsValidBoss(uid)
    if uid then
        if UnitExists(uid) and UnitName(uid) ~= UNKNOWN then
            if not UnitIsDeadOrGhost(uid) then
                local guid = UnitGUID(uid);
                local id = Unit:GetMobID(guid);
                local module, volatile, bypass = NPCInfo:Get(id);
                if module then
                    if bypass or UnitCanAttack("player", uid) then
                        return module, volatile;
                    end
                end
            end
        end
    end
    return nil, nil;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> id: the ID of the NPC.                                        *
-- ********************************************************************
-- * Test a boss module of the specified NPC.                         *
-- ********************************************************************
function System:Test(id)
    local module = NPCInfo:Get(id);
    if module then
        Module:Test(module);
    end
end

-- ********************************************************************
-- * Stop the current active module.                                  *
-- * If it's already engaged, it will also be flagged as a wipe.      *
-- ********************************************************************
function System:StopActiveModule()
    if Module:HasActive() then
        Module:GetActive():Interrupt();
  else
        Root:Print(L:Get("NoActiveModule"));
    end
end

-- --------------------------------------------------------------------
-- **                           Handlers                             **
-- --------------------------------------------------------------------

System.OnTargetChange = function(self)
    if not Module:HasActive() then
        local moduleName, volatile = self:IsValidBoss("target");
        if moduleName then
            -- We have selected an unit which is referenced to as a boss and we have the module that shall handle it.
            Module:Enable(moduleName);
        end
    end
end

System.OnUnitListUpdate = function(self)
    if not Module:HasActive() then
        -- Browse the unit list.
        local i, guid, name, uid;
        local moduleName, volatile;
        for i=1, Unit:GetNumUID() do
            guid, name, uid = Unit:GetUID(i);
            moduleName, volatile = self:IsValidBoss(uid);
            if moduleName then
                -- Phat guy is a boss. Trigger him at once, or when 20% of the raid has selected him.
                if volatile or self:GetNumSelections(guid) >= 0.20 then

                    if Module:Enable(moduleName) then
                        return;
                    end
                end
            end
        end
    end
end

System.OnZoneChange = function(self)
    self.nextZoneCheck = 0;
end

System.OnUpdate = function(self, elapsed)
    local now = GetTime();
    if now > self.nextZoneCheck then
        self.nextZoneCheck = now + 5;

        local zone = strlower(GetSubZoneText());
        local activeModule = Module:GetActive();
        if not activeModule then
            local modules = ZoneInfo:Get(zone);
            local i;
            for i=1, #modules do
                if Module:Enable(modules[i].name, modules[i].hasFilter) then
                    return;
                end
            end
      else
            if activeModule:IsDeclaredZone(zone) then
                activeModule:ZeroTimeout();
            end
        end
    end
end