﻿local Root = BossEncounter3;
local System = Root:GetOrNewModule("System");

local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");
local Network = Root:GetOrNewModule("Network");
local Version = Root:GetOrNewModule("Version");

local VersionQuery = System:GetOrNewModule("VersionQuery");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Service allowing you to query version of other players.

-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

local VERSION_COOLDOWN = 2.5;
local MULTIPLE_QUERY_TIME = 4.0;
local SINGLE_QUERY_TIME = 2.0;

VersionQuery.running = false;
VersionQuery.timeOut = nil;

VersionQuery.answerCooldown = { };
VersionQuery.queryCooldown = nil;

VersionQuery.answers = { };
VersionQuery.names = { };

-- Possible answers:
-- [false] = not answered yet / no answer
-- [-1] = disconnected
-- [0] = unknown
-- [1, 2, 3...] = version 0.0.1, 0.0.2, 0.0.3...

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

-- ********************************************************************
-- * Return true if you may perform a version query now.              *
-- ********************************************************************
function VersionQuery:CanQuery()
    if not self.running then
        if not self.queryCooldown or GetTime() > self.queryCooldown then
            return true;
        end
    end
    return false;
end

-- ********************************************************************
-- * Arguments:                                                       *
-- * >> recipient: a player name or a player GUID or GUILD or GROUP.  *
-- ********************************************************************
-- * Query the version of BE2.                                        *
-- ********************************************************************
function VersionQuery:Query(recipient)
    if not self:CanQuery() then return false; end

    recipient = Network:GetRecipient(recipient);
    if not recipient then return false; end

    wipe(self.answers);
    wipe(self.names);

    if recipient == "GROUP" then
        -- Group version check.
        local i, baseUID;
        if IsInRaid() then
            baseUID = "raid";
      else
            baseUID = "party";
        end
        for i=1, GetNumGroupMembers() do
            if not UnitIsUnit("player", baseUID..i) then
                local name = UnitName(baseUID..i);
                if UnitIsConnected(baseUID..i) then
                    tinsert(self.answers, false);
              else
                    tinsert(self.answers, -1);
                end
                tinsert(self.names, name);
            end
        end

elseif recipient == "GUILD" then
        -- Guild version check.
        local num = GetNumGuildMembers(false);
        local name, online;
        local i;
        for i=1, num do
            name, _, _, _, _, _, _, _, online = GetGuildRosterInfo(i);
            if online and name ~= UnitName("player") then
                tinsert(self.answers, false);
                tinsert(self.names, name);
            end
        end
  else
        -- Individual.
        tinsert(self.answers, false);
        tinsert(self.names, recipient);
    end

    if #self.names == 1 then
        self.timeOut = GetTime() + SINGLE_QUERY_TIME;
  else
        self.timeOut = GetTime() + MULTIPLE_QUERY_TIME;
    end
    self.running = true;
    self.queryCooldown = GetTime() + VERSION_COOLDOWN;

    Network:Send("VERSION_QUERY", recipient);

    return true;
end

-- ********************************************************************
-- * Show the version query results. Should be called at the timeout. *
-- ********************************************************************
function VersionQuery:PrintResults()
    System:Print(L:FormatLoc("Version-Results", #self.names));
    System:Print(L:FormatLoc("Version-Self", Version:GetString()));

    local i, name, result;
    for i=1, #self.names do
        name = self.names[i];
        result = self.answers[i];

        if not result then
            result = L:Get("Version-NoAnswer");
    elseif result == 0 then
            result = L:Get("Version-Unknown");
    elseif result == -1 then
            result = L:Get("Version-Disconnected");
      else
            result = L:FormatLoc("Version-Answer", Version:GetStringFromInt(result or 0));
        end

        System:PrintFmt("%d. %s: %s", i, name, result);
    end

    self.running = false;
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

VersionQuery.OnBootComplete = function(self)
    -- Register net commands.
    Network:Define("VERSION_QUERY", "V", "NORMAL", false);
    Network:Define("VERSION_ANSWER", "v", "NORMAL", false);
end

VersionQuery.OnNetCommand = function(self, command, sender, ...)
    if command == "VERSION_QUERY" then
        if GetTime() > (self.answerCooldown[sender] or 0) then
            self.answerCooldown[sender] = GetTime() + VERSION_COOLDOWN;

            -- Okay, not a spam, can reply.
            Network:Send("VERSION_ANSWER", sender, Version:GetFullInteger());
        end

elseif command == "VERSION_ANSWER" then
        if self.running then
            local allDone = true;
            local i;
            for i=1, #self.names do
                if self.names[i] == sender then
                    self.answers[i] = tonumber(select(1, ...)) or 0;
                end
                if not self.answers[i] then
                    allDone = false;
                end
            end
            if allDone then
                self:PrintResults();
            end
        end
    end
end

VersionQuery.OnUpdate = function(self)
    if self.running then
        if GetTime() > self.timeOut then
            -- Time up.
            self:PrintResults();
        end
    end
end