﻿local Root = BossEncounter3;

local GUI = Root:GetOrNewModule("GUI");
local Module = Root:GetOrNewModule("Module");

local L = Root:GetOrNewModule("Localization");
local Unit = Root:GetOrNewModule("Unit");

local System = Root:GetOrNewModule("System");
local Stats = System:GetOrNewModule("Stats");
local Settings = System:GetOrNewModule("Settings");

local View = System:GetOrNewModule("View");

-- --------------------------------------------------------------------
-- **                           Description                          **
-- --------------------------------------------------------------------

-- Defines the GUI available to modules and an usage interface.

-- --------------------------------------------------------------------
-- **                            Properties                          **
-- --------------------------------------------------------------------

View.enabled = false;
View.testing = false;
View.healRange = false;

-- --------------------------------------------------------------------
-- **                             Methods                            **
-- --------------------------------------------------------------------

function View:GetEventList()
    return self.EventList;
end

function View:GetEnemyList()
    return self.EnemyList;
end

function View:GetUnitList()
    return self.UnitList;
end

function View:GetBossBar()
    return self.BossBar;
end

function View:GetRadar()
    return self.Radar;
end

-- Prepare elements for usage.
function View:Enable()
    if self.enabled then return; end
    if self.testing then self:Disable(); end
    self.enabled = true;

    self.StatusWindow:ScrollIn();
    self.StatusWindow:DisableOptions(true);
    self.MainTimer:SetMode(1); -- Main timer starts in watch mode.
    self.EventList:Clear();
    self.EnemyList:Clear();
    self.EnemyList:SetExtraAlgorithm(nil);
    self.EnemyList:SetMainUnit(nil);
    self.EnemyList:SetThresholdDelay(nil);
    self.UnitList:Clear();
    self.UnitList:SetUpdateFunc(nil);
    self.NormalMessage:Clear();
    self.AlertMessage:Clear();
    self.SuperAlert:Clear();
    self.MirrorBar:Clear();

    -- Animations start hidden.
    self.ClearedAnim:Stop(true);
    self.ResultAnim:Stop();

    -- Setup the wrench icon.
    if Settings:GetNumSpecific() > 0 then
        self.StatusWindow:AllowOptions();
    end
end

-- Remove all visible elements.
function View:Disable()
    if not self.enabled and not self.testing then return; end
    self.enabled = false;
    self.testing = false;

    self.StatusWindow:ScrollOut();
    self.MainTimer:Stop();
    self.Radar:ScrollOut();
    self.Radar:SetMinRange(0);
    self.Radar:Clear();
    self.EventList:Clear();
    self.EnemyList:Clear();
    self.UnitList:Clear();
    self.BossBar:ScrollOut();
    self.NormalMessage:Clear();
    self.AlertMessage:Clear();
    self.SuperAlert:Clear();
    self.FloatingTimer:Stop();
    self.MirrorBar:Clear();
    self.Popup:Close();

    -- Make some animations disappear. Some others remain.
    self.ClearedAnim:Stop();

    self:DisableHealRange();
end

-- Replay the last victory sequence, if available and allowed.
function View:ReplayLastSequence()
    local name = Stats:GetCurrentInfo();
    if name then
        if not self.enabled then
            self.ResultAnim:Play();
        end
  else
        Root:Print(L:Get("NoSequenceReplay"));
    end
end

-- Ask the view to open the options frame of the mod.
function View:OpenOptions()
    if not self.MusicManager:IsShown() then
        self.SettingFrame:Open(not self.enabled); -- Open module options only if the GUI is enabled by a module elsewise, we use global options.
    end
end

-- Ask the view to open the music manager frame of the mod.
function View:OpenMusicManager()
    if not self.SettingFrame:IsShown() then
        self.MusicManager:Open();
    end
end

-- Ask the view to close the options frame of the mod.
function View:CloseOptions()
    self.SettingFrame:Close();
    self.MusicManager:Close(); -- also close the music manager.

    self:DisableHealRange();
end

-- Return true if messages (both normal and alerts) cannot be displayed for now. This is true when a super alert is running.
function View:IsMessageLocked()
    return self.SuperAlert:IsShown();
end

-- Display the size slider frame to change the size of an object.
function View:StartSizeChange(target, ref)
    self.SizeSlider:Open(target, ref);
end

-- Toggle the display of the radar.
function View:ToggleRadar()
    if self.Radar:IsShown() and (not self.Radar:ScrollingOut()) and (not self.enabled) then
        self.Radar:ScrollOut();

elseif not self.Radar:IsShown() then
        self:OnRadarRequest();
    end
end

-- Toggle the healer range check through the radar.
function View:CheckHealRange()
    if not self.healRange then
        self:EnableHealRange();
  else
        self:DisableHealRange();
    end
end

-- Enable the healer range check.
function View:EnableHealRange()
    if not self.healRange then
        Unit:EnumerateRaid();
        local count = 0;
        local i, unit;
        for i=1, Unit:GetEnumerationSize() do
            unit = Unit:GetEnumerationUID(i);
            if Unit:GetRole(unit) == "HEALER" then
                count = count + 1;
                self:OnRadarZoneAnchored("_heal_", "green", 40, unit, 900);
            end
        end
        if count > 0 then
            self.healRange = true;
            System:Print(L:Get("HealRangeExplain"));

            self:OnRadarRequest();
            self.Radar:SetVisibility(100);
      else
            System:Print(L:Get("HealRangeInvalid"));
        end
    end
end

-- Disable the healer range check.
function View:DisableHealRange()
    if self.healRange then
        self.healRange = false;

        self.Radar:RemoveZone("_heal_");
        self.Radar:SetVisibility(nil);
    end
end

-- --------------------------------------------------------------------
-- **                            Handlers                            **
-- --------------------------------------------------------------------

function View:OnBootComplete()
    -- Regular widgets creation.
    self.StatusWindow = GUI:Request("StatusWindow", "BossEncounter3_StatusWindow");
    self.StatusWindow:SetDefaultPosition("TOP", "TOP", 0, -16);
    self.MainTimer = self.StatusWindow:GetImportantTimer();
    self.Radar = GUI:Request("Radar", "BossEncounter3_Radar");
    self.Radar:SetDefaultPosition("CENTER", "RIGHT", -384, -192);
    self.BossBar = GUI:Request("BossBar", "BossEncounter3_BossBar");
    self.BossBar:SetDefaultPosition("BOTTOM", "BOTTOM", 0, 216);
    self.EventList = GUI:Request("EventList", "BossEncounter3_EventList");
    self.EventList:SetDefaultPosition("TOP", "RIGHT", -256, 64);
    self.EnemyList = GUI:Request("EnemyList", "BossEncounter3_EnemyList");
    self.EnemyList:SetDefaultPosition("TOP", "LEFT", 288, 64);
    self.EnemyList:ConnectBossBar(self.BossBar);
    self.UnitList = GUI:Request("UnitList", "BossEncounter3_UnitList");
    self.UnitList:SetDefaultPosition("TOP", "RIGHT", -256, 212);
    self.NormalMessage = GUI:Request("NormalMessage", "BossEncounter3_NormalMessage");
    self.NormalMessage:SetDefaultPosition("CENTER", "CENTER", 0, 64);
    self.AlertMessage = GUI:Request("AlertMessage", "BossEncounter3_AlertMessage");
    self.AlertMessage:SetDefaultPosition("CENTER", "CENTER", 0, 0);
    self.SuperAlert = GUI:Request("SuperAlert", "BossEncounter3_SuperAlert");
    self.SuperAlert:SetDefaultPosition("CENTER", "CENTER", 0, 0);
    self.SuperAlert:SetParent(WorldFrame);
    self.FloatingTimer = GUI:Request("FloatingTimer", "BossEncounter3_FloatingTimer");
    self.FloatingTimer:SetDefaultPosition("CENTER", "CENTER", 0, 200);
    self.MirrorBar = GUI:Request("MirrorBar", "BossEncounter3_MirrorBar");
    self.Popup = GUI:Request("Popup", "BossEncounter3_Popup");

    -- Animations creation. Animations have fixed positions.
    self.ClearedAnim = GUI:RequestChild("ClearedAnim", "BossEncounter3_ClearedAnim", WorldFrame);
    self.ClearedAnim:SetPoint("CENTER", WorldFrame, "CENTER", 0, 0);
    self.ClearedAnim:SetFrameStrata("DIALOG");
    self.ResultAnim = GUI:RequestChild("ResultAnim", "BossEncounter3_ResultAnim", WorldFrame);
    self.ResultAnim:SetAllPoints(WorldFrame);
    self.ResultAnim:SetFrameStrata("DIALOG");

    -- Configuration widgets creation.
    self.SettingFrame = GUI:Request("SettingFrame", "BossEncounter3_SettingFrame");
    self.SettingFrame:SetDefaultPosition("CENTER", "CENTER", 0, 0);
    self.MusicManager = GUI:Request("MusicManager", "BossEncounter3_MusicManager");
    self.MusicManager:SetDefaultPosition("CENTER", "CENTER", 0, 0);
    self.SizeSlider = GUI:Request("SizeSlider", "BossEncounter3_SizeSlider");

    -- Tell the module manager we're ready by removing the global cooldown.
    Module:SetCooldown("GENERAL", 0);
end

-- These are events sent by the module. These are very simple so the View can directly respond to them.

function View:OnTitleSet(title)
    self.StatusWindow:SetTitle(title);
end

function View:OnBerserkSet(timer)
    self.MainTimer:SetMode(-1);
    self.MainTimer:SetValue(timer);
end

function View:OnTimerStart(max)
    self.MainTimer:Start(max);
    self.StatusWindow:DisableOptions();
end

function View:OnTimerSuspend()
    self.MainTimer:Stop();
end

function View:OnTimerResume(timer)
    self.MainTimer:SetValue(timer);
    self.MainTimer:Resume();
end

function View:OnBossCleared(animation, coupDeGrace)
    self.ClearedAnim:Play(animation, coupDeGrace);
end

function View:OnResult()
    self.ResultAnim:Play();
end

function View:OnMessage(message, duration)
    if not self:IsMessageLocked() then
        self.NormalMessage:Add(message, duration);
    end
end

function View:OnAlert(message, duration)
    if not self:IsMessageLocked() then
        self.AlertMessage:Display(message, duration);
    end
end

function View:OnSuperAlert(title, subTitle, duration, focus)
    if not self:IsMessageLocked() then
        self.SuperAlert:Display(title, subTitle, duration, focus);

        -- Wipe messages with less priority.
        self.NormalMessage:Clear();
        self.AlertMessage:Clear();
        self.Popup:Close();
    end
end

function View:OnRadarRequest(minRange)
    self.Radar:ScrollIn();

    if minRange then
        self.Radar:SetMinRange(minRange);
    end
end

function View:OnRadarZoneFlush()
    self.Radar:Clear();
end

function View:OnRadarZoneAnchored(name, color, radius, unit, duration) -- Circular zone moving with an unit.
    self.Radar:AddAnchoredZone(name, color, radius, unit, duration);
end

function View:OnRadarZoneCircular(name, color, radius, x, y, duration) -- (x; y) are the coordinates on the map.
    self.Radar:AddFixedZone(name, color, radius, x, y, duration);
end

function View:OnRadarZoneModSize(name, add, factor)
    self.Radar:ModifyZoneSize(name, add, factor);
end

function View:OnRadarZoneSetSize(name, radius)
    self.Radar:SetZoneSize(name, radius);
end

function View:OnRadarZoneInterrupted(name, unit)
    self.Radar:RemoveZone(name, unit);
end

function View:OnMirrorValueStart(label, object, method, methodAlt)
    self.MirrorBar:Start(label, object, method, methodAlt);
end

function View:OnMirrorValueStop()
    self.MirrorBar:Stop();
end

function View:OnFloatingTimerStart(timer)
    if timer then
        self.FloatingTimer:SetMode(-1);
        self.FloatingTimer:SetValue(timer);
  else
        self.FloatingTimer:SetMode(1);
    end
    self.FloatingTimer:Start();
end

function View:OnFloatingTimerStop(removeDelay)
    self.FloatingTimer:Stop(removeDelay);
end

function View:OnPopupRequest(type, title, text)
    self.Popup:Open(type, title, text);
end

function View:OnEquipCheckResult(text)
    self:OnPopupRequest("WARNING", L:Get("EquipChecker-Title"), text);
end

function View:Test()
    if self.testing then
        self:Disable();
        System:Print(L:Get("TestDisabled"));

elseif not self.enabled then
        self.testing = true;
        System:Print(L:Get("TestEnabled"));

        -- Test message
        self:OnMessage(L:Get("TestMessage"), 30);

        -- Test alert
        self:OnAlert(L:Get("TestAlert"), 30);

        -- Test status box
        self.StatusWindow:ScrollIn();
        self.StatusWindow:DisableOptions(true);
        self.StatusWindow:SetTitle("Test");
        self.MainTimer:SetMode(1);
        self.MainTimer:Start(60);

        -- Test events
        self.EventList:StartEvent("Test1", "Test 1", 120);
        self.EventList:StartEvent("Test2", "Test 2",  60);

        -- Test enemy
        self.EnemyList:AddUnit(UnitGUID("player"), true);
        self.EnemyList:SetExtraAlgorithm("mana");
        self.EnemyList:SetMainUnit(UnitGUID("player"));

        -- Test units
        self.UnitList:SetTitle("Test");
        self.UnitList:SetPlayers("ALL");
        self.UnitList:SetUpdateFunc(nil);

        -- Test radar
        self:OnRadarRequest(10);
        self:OnRadarZoneAnchored("test", "red", 20, "player", 10);
        self:DisableHealRange();
    end
end
